package de.uni_potsdam.hpi.isg.efes.cli;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintStream;
import java.sql.SQLException;
import java.util.Map.Entry;
import java.util.Properties;

import org.apache.commons.cli.BasicParser;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.xml.sax.SAXException;

import de.uni_potsdam.hpi.isg.efes.csg.datafit.DataFitReport;
import de.uni_potsdam.hpi.isg.efes.csg.effort.EffortEstimate;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.BasicNodeMapping;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.DataSet;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.IntegrationScenario;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.adapter.SpicyConfigurationScenarioBuilder;
import de.uni_potsdam.hpi.isg.efes.pipeline.Pipeline;
import de.uni_potsdam.hpi.isg.efes.pipeline.configuration.Configurations;
import de.uni_potsdam.hpi.isg.efes.pipeline.configuration.model.Configuration;
import de.uni_potsdam.hpi.isg.efes.spicy.SpicyConfiguration;
import de.uni_potsdam.hpi.isg.efes.spicy.SpicyParser;
import de.uni_potsdam.hpi.isg.efes.spicy.extensions.TableMappingConfiguration;

public class Main {

	private static Logger logger = LogManager.getLogger("efes-cli");

	public static void main(String[] args) throws FileNotFoundException, SAXException, IOException, SQLException {
		Parameters parameters = new Parameters(args);
		if (!parameters.isReady()) {
			System.exit(1);
		}

		IntegrationScenario integrationScenario = createIntegrationScenario(parameters);
		Pipeline pipeline = createPipeline(parameters);
		runPipeline(parameters, integrationScenario, pipeline);
	}

	private static IntegrationScenario createIntegrationScenario(Parameters parameters) throws SAXException,
			IOException, FileNotFoundException, SQLException {

		SpicyParser spicyParser = new SpicyParser();
		String[] spicyFilePaths = parameters.getSpicyFiles();
		SpicyConfigurationScenarioBuilder builder = new SpicyConfigurationScenarioBuilder();
		for (String spicyFilePath : spicyFilePaths) {
			SpicyConfiguration spicyConfiguration = spicyParser.parse(new FileInputStream(spicyFilePath));
			builder.add(spicyConfiguration);
		}

		String[] tableMappingFilePaths = parameters.getTableMappingFiles();
		for (String tableMappingFilePath : tableMappingFilePaths) {
			TableMappingConfiguration tableMappingConfiguration = TableMappingConfiguration.load(new File(tableMappingFilePath));
			builder.add(tableMappingConfiguration);
		}

		IntegrationScenario integrationScenario = builder.build();
		if (parameters.createIdentityMapping()) {
			DataSet targetDataSet = integrationScenario.getTargetDataSet();
			for (DataSet sourceDataSet : integrationScenario.getSourceDataSets()) {
				BasicNodeMapping oldMapping = integrationScenario.getMappings().get(sourceDataSet);
				if (!oldMapping.getNodeCorrespondences().isEmpty()) {
					logger.warn("Not creating an identity mapping for because there are already correspondences.");
					continue;
				}
				BasicNodeMapping mapping = BasicNodeMapping.createIdentityMapping(sourceDataSet, targetDataSet);
				integrationScenario.setMapping(sourceDataSet, mapping);
			}
		}
		return integrationScenario;
	}

	private static Pipeline createPipeline(Parameters parameters) {
		Configuration configuration;
		if (parameters.hasPipelineConfiguration()) {
			File configFile = new File(parameters.getPipelineConfiguration());
			configuration = Configurations.loadConfiguration(configFile);
		} else {
			configuration = Configurations.createExampleConfiguration();
		}
		Pipeline pipeline = configuration.createPipeline();
		return pipeline;
	}

	private static void runPipeline(Parameters parameters, IntegrationScenario integrationScenario, Pipeline pipeline) {
		String unit = parameters.getUnit();

		PrintStream output = System.out;
		if (parameters.hasOutputFile()) {
			String filePath = parameters.getOutputFile();
			try {
				output = new PrintStream(filePath);
			} catch (Exception e) {
				logger.catching(e);
				logger.error("Cannot write to '{}', using standard out instead.", filePath);
			}
		}

		pipeline.process(integrationScenario, unit);
		DataFitReport dataFitReport = pipeline.getDataFitReport();
		dataFitReport.print(output);
		EffortEstimate estimate = pipeline.getEffortEstimate();
		output.println();
		estimate.print(output);
		output.println();
		estimate.printCsv(output);
		output.println();

		if (output != System.out) {
			output.close();
		}
	}

	private static class Parameters {

		private Options options;
		private CommandLine commandLine;

		public Parameters(String[] args) {
			initialize();
			if (args == null || args.length == 0) {
				printUsage();
			} else {
				parse(args);
			}
		}

		private void printUsage() {
			HelpFormatter helpFormatter = new HelpFormatter();
			helpFormatter.printHelp("efes <options...>", options);
		}

		@SuppressWarnings("static-access")
		private void initialize() {
			this.options = new Options();

			Option spicy = OptionBuilder
					.withArgName("spicy files")
					.withLongOpt("spicy")
					.hasArg()
					.isRequired()
					.withDescription(
							"a semicolon-separated list of ++Spicy configuration files "
									+ "that describes the integration scenario").create("s");
			options.addOption(spicy);

			Option tableMapping = OptionBuilder.withArgName("table mapping").withLongOpt("table").hasArg()
					.withDescription("a semicolon-separated list of table mapping files (optional)").create("t");
			options.addOption(tableMapping);

			Option identityMapping = OptionBuilder.withArgName("identity mapping").withLongOpt("identity-mapping")
					.withDescription("instead of the provided mapping generate an identity mapping").create();
			options.addOption(identityMapping);

			Option unit = OptionBuilder.withArgName("unit").withLongOpt("unit").hasArg().isRequired()
					.withDescription("the unit to be estimated").create("u");
			options.addOption(unit);

			Option configuration = OptionBuilder.withArgName("configuration file").withLongOpt("pipeline").hasArg()
					.withDescription("a configuraton file for the efes pipeline (optional)").create("p");
			options.addOption(configuration);

			Option output = OptionBuilder.withArgName("output file").withLongOpt("output").hasArg()
					.withDescription("output file to write the results to (optional)").create("o");
			options.addOption(output);

			Option property = OptionBuilder.withArgName("property=value").hasArgs(2).withValueSeparator()
					.withDescription("Java system properties").create("D");
			options.addOption(property);

		}

		private void parse(String[] args) {
			CommandLineParser parser = new BasicParser();
			try {
				this.commandLine = parser.parse(options, args);
			} catch (ParseException exp) {
				System.out.println("Could not start efes: " + exp.getMessage());
				System.out.println("Run efes without parameters for usage.");
			}

			Properties javaProperties = this.commandLine.getOptionProperties("D");
			for (Entry<Object, Object> entry : javaProperties.entrySet()) {
				System.setProperty((String) entry.getKey(), (String) entry.getValue());
			}
		}

		public boolean isReady() {
			return this.commandLine != null;
		}

		public String[] getSpicyFiles() {
			String optionValue = this.commandLine.getOptionValue("s");
			if (optionValue == null) {
				return new String[0];
			}
			return optionValue.split(";");
		}

		public String[] getTableMappingFiles() {
			String optionValue = this.commandLine.getOptionValue("t");
			if (optionValue == null) {
				return new String[0];
			}
			return optionValue.split(";");
		}

		public boolean hasPipelineConfiguration() {
			return this.commandLine.hasOption("p");
		}

		public String getPipelineConfiguration() {
			return this.commandLine.getOptionValue("p");
		}

		public String getUnit() {
			return this.commandLine.getOptionValue("u");
		}

		public boolean createIdentityMapping() {
			return this.commandLine.hasOption("identity-mapping");
		}

		public boolean hasOutputFile() {
			return this.commandLine.hasOption("o");
		}

		public String getOutputFile() {
			return this.commandLine.getOptionValue("o");
		}

	}
}
