package de.uni_potsdam.hpi.isg.efes.csg.datafit.mapping;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.DataFitModel;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.DataFitReport;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.NodeCorrespondence;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.NodeMapping;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.RelationshipCorrespondence;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.RelationshipMapping;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.RelationshipMatcher;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.IdentityNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.DataSet;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.IntegrationScenario;
import de.uni_potsdam.hpi.isg.efes.csg.util.RelationalSchemaToCSGConverter;

public class MappingInspector implements DataFitModel {

	private static final Logger logger = LogManager.getLogger(MappingInspector.class.getSimpleName());

	private IntegrationScenario integrationScenario;

	@Override
	public void setIntegrationScenario(IntegrationScenario scenario) {
		this.integrationScenario = scenario;
	}

	@Override
	public void evaluateScenarioTo(DataFitReport report) {
		logger.info("Running mapping inspector.");

		ReportSection reportSection = new ReportSection();

		BasicSchemaGraph targetSchema = this.integrationScenario.getTargetDataSet().getSchema();

		for (DataSet sourceDataSet : this.integrationScenario.getSourceDataSets()) {
			BasicSchemaGraph sourceSchema = sourceDataSet.getSchema();
			NodeMapping nodeMapping = this.integrationScenario.getMappings().get(sourceDataSet);
			RelationshipMatcher relationshipMatcher = new RelationshipMatcher(sourceSchema, targetSchema, nodeMapping);
			RelationshipMapping relationshipMapping = relationshipMatcher.run();

			inspect(sourceSchema, targetSchema, nodeMapping, relationshipMapping, reportSection);
		}

		report.add(reportSection);
		logger.info("Mapping inspector completed.");
	}


	private void inspect(BasicSchemaGraph sourceSchema, BasicSchemaGraph targetSchema, NodeMapping nodeMapping,
			RelationshipMapping mapping, ReportSection reportSection) {
		Collection<Node> allTargetSchemaNode = targetSchema.getAtomicNodes().values();

		for (Node targetSchemaNode : allTargetSchemaNode) {
			if (targetSchemaNode.isSimpleIdentityNode()) {
				IdentityNode tableNode = (IdentityNode) targetSchemaNode;

				// See if the target node is mapped.
				Collection<NodeCorrespondence> targetNodeCorrespondences = mapping
						.getCorrespondencesForTargetNode(tableNode);
				boolean isTableMapped = !targetNodeCorrespondences.isEmpty();
				if (!isTableMapped) {
					continue;
				}

				TableMapping tableMapping = new TableMapping(tableNode);

				// It is a table node, search attributes.
				List<Relationship> attributeRelationships = findAttributeRelationships(tableNode, targetSchema);

				// Evaluate each attribute.
				for (Relationship attributeRelationship : attributeRelationships) {
					evaluateAttribute(attributeRelationship, targetSchema, mapping, tableMapping, sourceSchema);
				}

				Set<String> tableNames = new HashSet<>();
				for (Relationship relationship : tableMapping.getMappedRelationships()) {
					extractTableNames(relationship, tableNames);
				}
				logger.trace("Collected tables for {}: {}", tableNode, tableNames);
				tableMapping.setNumInvolvedSourceTables(tableNames.size());

				reportSection.add(tableMapping);
			}
		}

	}

	private void evaluateAttribute(Relationship attributeRelationship, BasicSchemaGraph targetSchema,
			RelationshipMapping mapping, TableMapping tableMapping, BasicSchemaGraph sourceSchema) {

		Node attributeNode = attributeRelationship.getEndNode();
		if (attributeNode.isMetaData()) {
			if (isForeignKeyNode(attributeNode, targetSchema)) {
				tableMapping.incNumMetaDataForeignKeyAttributes();

			} else if (isPrimaryKey(attributeRelationship, targetSchema)) {
				tableMapping.setHasMetaDataPrimaryKey(true);

			} else {
				tableMapping.incNumOtherMetaData();
			}

		} else {
			Collection<RelationshipCorrespondence> correspondences = mapping.getCorrespondencesForTargetRelationship(attributeRelationship);
			for (RelationshipCorrespondence correspondence : correspondences) {
				tableMapping.addMappedRelationship(correspondence.getSourceRelationship());
			}
//			Relationship translatedRelationship = mapping.translateTargetRelationship(attributeRelationship);
//			if (translatedRelationship != null) {
//				tableMapping.addMappedRelationship(translatedRelationship);
//			}

		}
	}

	private List<Relationship> findAttributeRelationships(IdentityNode tableNode, BasicSchemaGraph targetSchema) {
		List<Relationship> attributeRelationships = new ArrayList<>();
		Collection<Relationship> relationships = targetSchema.getAtomicRelationships(tableNode);
		for (Relationship relationship : relationships) {
			if (relationship.isAtomic() && relationship.isExplicit() && relationship.getEndNode().isSimpleValueNode()) {
				attributeRelationships.add(relationship);
			}
		}
		return attributeRelationships;
	}

	private boolean isPrimaryKey(Relationship attributeRelationship, BasicSchemaGraph targetSchema) {
		NaturalNumberSet cardinality = targetSchema.getCardinality(attributeRelationship);
		NaturalNumberSet inverseCardinality = targetSchema.getCardinality(attributeRelationship.getInverse());

		return cardinality.equals(NaturalNumberSet.ONE) && inverseCardinality.equals(NaturalNumberSet.ONE);
	}

	/**
	 * Checks if the given node has an equality relationship with another node with cardinality 1.
	 */
	private boolean isForeignKeyNode(Node attributeNode, BasicSchemaGraph targetSchema) {
		Collection<Relationship> metaDataRelationships = targetSchema.getAtomicRelationships(attributeNode);
		for (Relationship metaDataRelationship : metaDataRelationships) {
			if (metaDataRelationship.isAtomic()
					&& metaDataRelationship.getContentType() == Relationship.ContentType.EQUALITY) {
				NaturalNumberSet cardinality = targetSchema.getCardinality(metaDataRelationship);
				if (cardinality.equals(NaturalNumberSet.ONE)) {
					return true;
				}
			}
		}
		return false;
	}

	private void extractTableNames(Relationship relationship, Set<String> tableNames) {
		Collection<Relationship> atomicRelationships = relationship.getAtomicRelationships();
		for (Relationship atomicRelationship : atomicRelationships) {
			extractTableName(atomicRelationship.getStartNode(), tableNames);
			extractTableName(atomicRelationship.getEndNode(), tableNames);
		}
	}

	private void extractTableName(Node node, Set<String> tableNames) {
		if (node == null) {
			return;
		}

		if (node.getOperationType() != Node.Operation.NONE) {
			logger.warn("Expected only atomic nodes, found {}.", node);
			return;
		}

		String tableName = RelationalSchemaToCSGConverter.extractTableName(node);
		tableNames.add(tableName);
	}

	public static class ReportSection implements DataFitReport.Section {

		private Collection<TableMapping> tableMappings = new ArrayList<>();

		public void add(TableMapping tableMapping) {
			this.tableMappings.add(tableMapping);
		}

		public Collection<TableMapping> getTableMappings() {
			return tableMappings;
		}

		@Override
		public void print(PrintStream stream) {
			stream.println("Mapping complexity report");
			stream.println("=========================");
			stream.format(" %d mappings:\n", this.tableMappings.size());
			for (TableMapping mapping : this.tableMappings) {
				stream.format(" * %s\n", mapping);
			}
		}

		@Override
		public void printCsv(PrintStream stream) {
			stream.println("target table,mapped attributes,metadata PKs,metadata FKs,source tables");
			for (TableMapping mapping : this.tableMappings) {
				stream.format("%s,%d,%d,%d,%d\n", mapping.tableNode, mapping.mappedRelationships.size(),
						mapping.hasMetaDataPrimaryKey ? 1 : 0, mapping.numMetaDataForeignKeyAttributes,
						mapping.numInvolvedSourceTables);
			}
		}
	}

	/**
	 * 
	 * @author Sebastian Kruse
	 */
	public static class TableMapping {

		private IdentityNode tableNode;
		private int numMetaDataForeignKeyAttributes, numOtherMetaData;
		private boolean hasMetaDataPrimaryKey;
		private Collection<Relationship> mappedRelationships;
		private int numInvolvedSourceTables;

		public TableMapping(IdentityNode tableNode) {
			super();
			this.tableNode = tableNode;
			this.mappedRelationships = new ArrayList<>();
		}

		public IdentityNode getTableNode() {
			return tableNode;
		}

		public int getNumMetaDataForeignKeyAttributes() {
			return numMetaDataForeignKeyAttributes;
		}

		public void incNumMetaDataForeignKeyAttributes() {
			this.numMetaDataForeignKeyAttributes++;
		}

		public int getNumOtherMetaData() {
			return numOtherMetaData;
		}

		public void incNumOtherMetaData() {
			this.numOtherMetaData++;
		}

		public boolean hasMetaDataPrimaryKey() {
			return hasMetaDataPrimaryKey;
		}

		public void setHasMetaDataPrimaryKey(boolean hasMetaDataPrimaryKey) {
			this.hasMetaDataPrimaryKey = hasMetaDataPrimaryKey;
		}

		public Collection<Relationship> getMappedRelationships() {
			return mappedRelationships;
		}

		public void addMappedRelationship(Relationship relationship) {
			this.mappedRelationships.add(relationship);
		}

		public int getNumInvolvedSourceTables() {
			return numInvolvedSourceTables;
		}

		public void setNumInvolvedSourceTables(int numInvolvedSourceTables) {
			this.numInvolvedSourceTables = numInvolvedSourceTables;
		}

		@Override
		public String toString() {
			return "TableMapping [tableNode=" + tableNode + ", numMetaDataForeignKeyAttributes="
					+ numMetaDataForeignKeyAttributes + ", numOtherMetaData=" + numOtherMetaData
					+ ", hasMetaDataPrimaryKey=" + hasMetaDataPrimaryKey + ", mappedRelationships="
					+ mappedRelationships + ", numInvolvedSourceTables=" + numInvolvedSourceTables + "]";
		}

	}

}
