package de.uni_potsdam.hpi.isg.efes.csg.datafit.structure;

import it.unimi.dsi.fastutil.ints.IntIterator;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.DataFitModel;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.DataFitReport;
import de.uni_potsdam.hpi.isg.efes.csg.instance.VirtualSchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.integration.IntegratedMapping;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.BasicNodeMapping;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.BasicRelationshipMapping;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.MappedDataStoreAdapter;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.NodeMapping;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.RelationshipCorrespondence;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.RelationshipMapping;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.RelationshipMatcher;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.CompositeRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.DataSet;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.IntegrationScenario;
import de.uni_potsdam.hpi.isg.utils.util.CollectionUtil;

public class StructureFitInspector implements DataFitModel {

	private static Logger logger = LogManager.getLogger(StructureFitInspector.class.getSimpleName());

	private IntegrationScenario integrationScenario;

	private boolean useSourceSchema = false;

	public StructureFitInspector() {
		super();
	}

	public void setIntegrationScenario(IntegrationScenario integrationScenario) {
		this.integrationScenario = integrationScenario;
	}

	@Override
	public void evaluateScenarioTo(DataFitReport report) {

		List<DataSet> sourceDataSets = this.integrationScenario.getSourceDataSets();
		DataSet targetDataSet = this.integrationScenario.getTargetDataSet();
		Map<DataSet, BasicNodeMapping> mappings = this.integrationScenario.getMappings();

		BasicRelationshipMapping identityMapping = new BasicRelationshipMapping(
				BasicNodeMapping.createIdentityMapping(this.integrationScenario.getTargetDataSet()));
		BasicSchemaGraph trimmedTargetSchema = trimSchema(targetDataSet.getSchema(), identityMapping);

		Map<DataSet, RelationshipMapping> relationshipMappings = new HashMap<DataSet, RelationshipMapping>();

		if (sourceDataSets.isEmpty()) throw new IllegalStateException("No source data sets.");

		for (DataSet sourceDataSet : sourceDataSets) {
			BasicSchemaGraph sourceSchema = sourceDataSet.getSchema();
			BasicNodeMapping mapping = mappings.get(sourceDataSet);

			NodeMapping trimmedNodeMapping = trimNodeMapping(mapping, trimmedTargetSchema);

			// Run relationship matcher on trimmed node mapping.
			logger.debug("Running relationship matcher on {} and {}.", sourceDataSet, targetDataSet);
			RelationshipMatcher relationshipMatcher = new RelationshipMatcher(sourceSchema, trimmedTargetSchema,
					trimmedNodeMapping);
			RelationshipMapping relationshipMapping = relationshipMatcher.run();
			relationshipMappings.put(sourceDataSet, relationshipMapping);

			for (Entry<Relationship, List<RelationshipCorrespondence>> c : relationshipMapping
					.getRelationshipCorrespondences().entrySet()) {
				logger.trace(" {} -> {}", c.getKey(), c.getValue());
			}
		}

		Collection<RelationshipMapping> mappingsToUse = new ArrayList<>();
		if (useSourceSchema) {
			mappingsToUse.add(identityMapping);
		}
		mappingsToUse.addAll(relationshipMappings.values());
		RelationshipMapping mapping;
		if (mappingsToUse.size() > 1) {
			mapping = new IntegratedMapping(mappingsToUse);
		} else {
			mapping = CollectionUtil.getAny(mappingsToUse);
		}

		ReportSection reportSection = inspect(mapping, trimmedTargetSchema);
		report.add(reportSection);
	}

	private BasicSchemaGraph trimSchema(BasicSchemaGraph schema, BasicRelationshipMapping mapping) {
		BasicSchemaGraph trimmedSchema = new BasicSchemaGraph(schema.getName());

		// Copy all atomic, non-meta-data nodes.
		for (Node node : schema.getAtomicNodes().values()) {
			if (isAtomicAndNotMetaData(node)) {
				trimmedSchema.add(node);
			}
		}

		// Copy all atomic, non-meta-data relationships.
		for (Relationship relationship : schema.getAllRelationships()) {
			if (relationship.isAtomic() && isAtomicAndNotMetaData(relationship.getStartNode())
					&& isAtomicAndNotMetaData(relationship.getEndNode())) {
				trimmedSchema.add(relationship);
				NaturalNumberSet cardinality = schema.getCardinality(relationship);
				trimmedSchema.addConstraint(relationship, cardinality);
				mapping.addCorrespondence(new RelationshipCorrespondence(relationship, relationship));
			}
		}

		// "Condense" relationships that involve meta-data nodes.
		for (Node startNode : schema.getAtomicNodes().values()) {

			// We must start at an atomic, non-meta-data node.
			if (!isAtomicAndNotMetaData(startNode)) {
				continue;
			}

			// Inspect all relationships.
			for (Relationship relationship : schema.getAtomicRelationships(startNode)) {

				// The next node must be an atomic meta-data node.
				Node nextNode = relationship.getEndNode();
				if (!relationship.isAtomic() || !isAtomicAndMetaData(nextNode)) {
					continue;
				}

				// Now we must extend the node until we reach a
				// non-meta-data node.
				logger.trace("Need to pursue {}.", relationship);
				List<Relationship> pursuedRelationships = new ArrayList<>();
				pursuedRelationships.add(relationship);
				List<Relationship> results = new ArrayList<>();
				extendToNonMetaDataNode(pursuedRelationships, schema, results);
				logger.trace("{} results", results.size());

				// Pick the relationship with the least complexity.
				int minComplexity = -1;
				int maxComplexity = -1;
				Relationship minComplexityRelationship = null;
				for (Relationship result : results) {
					int complexity = result.getComplexity();
					Node endNode = result.getEndNode();
					if (startNode.equals(endNode)) {
						logger.debug("Detected meta-data loop that will be neglected: {}", result);
						continue;
					} else if (minComplexityRelationship == null || minComplexity > complexity) {
						minComplexity = complexity;
						minComplexityRelationship = result;
					}
					maxComplexity = Math.max(maxComplexity, complexity);
				}
				if (minComplexityRelationship == null) {
					logger.trace("Could not extend {} to a non-meta-data relationship.");
				} else {
					logger.trace("Detected non-meta-data relationship {} (from {}).", minComplexityRelationship,
							results);
					logger.trace("Min complexity: {}, max complexity: {}", minComplexity, maxComplexity);
					Node endNode = minComplexityRelationship.getEndNode();

					trimmedSchema.addExplicitRelationships(startNode, endNode);

					NaturalNumberSet cardinality = schema.getCardinality(minComplexityRelationship);
					trimmedSchema.addConstraint(startNode, endNode, cardinality);
					logger.debug("Adding {}->{} (C={}).", startNode, endNode, cardinality);

					cardinality = schema.getCardinality(minComplexityRelationship.getInverse());
					trimmedSchema.addConstraint(endNode, startNode, cardinality);
					logger.debug("Adding {}->{} (C={}).", endNode, startNode, cardinality);

					Relationship addedRelationship = trimmedSchema.getAtomicRelationship(startNode, endNode);
					mapping.addCorrespondence(new RelationshipCorrespondence(minComplexityRelationship,
							addedRelationship));
					mapping.addCorrespondence(new RelationshipCorrespondence(minComplexityRelationship.getInverse(),
							addedRelationship.getInverse()));

				}
			}
		}

		return trimmedSchema;
	}

	private void extendToNonMetaDataNode(List<Relationship> pursuedRelationships, BasicSchemaGraph schema,
			List<Relationship> results) {
		Relationship previousRelationship = pursuedRelationships.get(pursuedRelationships.size() - 1);
		Node previousNode = previousRelationship.getStartNode();
		Node startNode = previousRelationship.getEndNode();
		Collection<Relationship> nextRelationships = schema.getAtomicRelationships(startNode);

		for (Relationship nextRelationship : nextRelationships) {
			Node endNode = nextRelationship.getEndNode();
			if (pursuedRelationships.contains(nextRelationship) || endNode.equals(previousNode)) {
				continue;
			}

			pursuedRelationships.add(nextRelationship);
			if (isAtomicAndNotMetaData(endNode)) {
				CompositeRelationship compositeRelationship = new CompositeRelationship(pursuedRelationships);
				results.add(compositeRelationship);
			} else if (isAtomicAndMetaData(endNode)) {
				extendToNonMetaDataNode(pursuedRelationships, schema, results);
			}
			pursuedRelationships.remove(pursuedRelationships.size() - 1);
		}

	}

	/**
	 * Creates a trimmed copy of the given mapping, that only has target nodes that occur in the given trimmed schema.
	 * 
	 * @param mapping
	 *            is the mapping that shall be trimmed
	 * @param trimmedSchema
	 *            is the schema that contains the nodes to be retained
	 * @return the trimmed mapping of nodes
	 */
	private NodeMapping trimNodeMapping(BasicNodeMapping mapping, BasicSchemaGraph trimmedSchema) {
		BasicNodeMapping trimmedNodeMapping = new BasicNodeMapping(mapping);
		Set<Node> trimmedNodes = new HashSet<>(mapping.getNodeCorrespondences().keySet());
		trimmedNodes.removeAll(trimmedSchema.getAtomicNodes().values());
		for (Node trimmedNode : trimmedNodes) {
			trimmedNodeMapping.removeCorrespondencesForTargetNode(trimmedNode);
		}
		return trimmedNodeMapping;
	}

	private boolean isAtomicAndNotMetaData(Node node) {
		return node.getOperationType() == Node.Operation.NONE && !node.isMetaData() && !isUnmappedForeignKeyNode(node);
	}

	private boolean isAtomicAndMetaData(Node node) {
		return node.getOperationType() == Node.Operation.NONE && node.isMetaData();
	}

	private boolean isUnmappedForeignKeyNode(Node node) {
		for (BasicNodeMapping mapping : this.integrationScenario.getMappings().values()) {
			if (!mapping.getCorrespondencesForTargetNode(node).isEmpty()) {
				return false;
			}
		}

		BasicSchemaGraph targetSchema = this.integrationScenario.getTargetDataSet().getSchema();
		Collection<Relationship> relationships = targetSchema.getAtomicRelationships(node);
		for (Relationship relationship : relationships) {
			if (!relationship.isExplicit()
					&& targetSchema.getConstraints(relationship).toConstraint().getCardinality()
							.isSubsetOf(NaturalNumberSet.ONE)) {
				return true;
			}
		}
		return false;
	}

	private ReportSection inspect(RelationshipMapping mapping, BasicSchemaGraph targetSchema) {

		VirtualSchemaGraphInstance targetInstance = new VirtualSchemaGraphInstance(targetSchema);
		if (mapping.isBackedByInstance()) targetInstance
				.acquireDataStoreAdapter(new MappedDataStoreAdapter.MappedDataStore(mapping));
		ReportSection reportSection = inspectTargetRelationships(mapping, targetInstance);

		return reportSection;
	}

	private ReportSection inspectTargetRelationships(RelationshipMapping mapping,
			VirtualSchemaGraphInstance targetInstance) {
		Set<Relationship> handledRelationships = new HashSet<>();
		ReportSection reportSection = new ReportSection();

		for (Relationship targetRelationship : targetInstance.getSchema().getAllRelationships()) {
			if (!handledRelationships.add(targetRelationship)) {
				logger.trace("Looks like {} is handled twice.", targetRelationship);
				continue;
			}

			NaturalNumberSet sourceCardinality = mapping.estimateSourceCardinality(targetRelationship);
			BasicSchemaGraph targetSchema = targetInstance.getSchema();
			Constraint targetConstraint = targetSchema.getConstraints(targetRelationship).toConstraint();
			NaturalNumberSet targetCardinality = targetConstraint.getCardinality();
			targetInstance.setActualCardinality(targetRelationship, sourceCardinality);

			if (!sourceCardinality.isSubsetOf(targetCardinality)) {
				logger.debug("Schema-level violation of {}: {}", targetConstraint, sourceCardinality);
				int violatingElements = -1;
				if (mapping.isBackedByInstance()) {
					NaturalNumberBag cardinalityBag = mapping.getRealSourceCardinality(targetRelationship);
					logger.debug("Looking onto data revealed cardinalities: {}", cardinalityBag);
					violatingElements = 0;
					for (IntIterator i = cardinalityBag.getCountedCardinalities().iterator(); i.hasNext();) {
						int cardinality = i.nextInt();
						if (!targetCardinality.hasElement(cardinality)) {
							violatingElements += cardinalityBag.getCountFor(cardinality);
						}
					}
					logger.debug("Counted {} violations", violatingElements);
					targetInstance.setActualCardinality(targetRelationship, cardinalityBag.toInterval());
				}
				Relationship sourceRelationship = mapping.translateTargetRelationship(targetRelationship);
				reportSection.addStructureViolation(new Violation(targetConstraint, sourceCardinality,
						violatingElements, sourceRelationship));
			} else {
				logger.debug("No violation of: {} is subset of {} ", targetConstraint, sourceCardinality);
			}

			reportSection.setSchemaGraphInstance(targetInstance);

		}

		return reportSection;
	}

	public void setUseSourceData(boolean useSourceData) {
		this.useSourceSchema = useSourceData;
	}

	public static class ReportSection implements DataFitReport.Section {

		private List<Violation> structureViolations = new ArrayList<>();
		private VirtualSchemaGraphInstance sgInstance;

		void addStructureViolation(Violation violation) {
			this.structureViolations.add(violation);
		}

		public List<Violation> getStructureViolations() {
			return this.structureViolations;
		}

		void setSchemaGraphInstance(VirtualSchemaGraphInstance sgInstance) {
			this.sgInstance = sgInstance;
		}

		public VirtualSchemaGraphInstance getSchemaGraphInstance() {
			return sgInstance;
		}

		@Override
		public void print(PrintStream stream) {
			stream.println("Structure fit report");
			stream.println("====================");
			stream.format(" %d structure violations:\n", this.structureViolations.size());
			for (Violation violation : this.structureViolations) {
				stream.format(" * %s\n", violation);
			}
		}

		@Override
		public void printCsv(PrintStream stream) {
			// TODO
		}

	}

	public static class Violation {

		private Constraint constraint;
		private NaturalNumberSet actualCardinality;
		private int numViolatingElements;
		private Relationship sourceRelationship;

		public Violation(Constraint constraint, NaturalNumberSet actualCardinality, int numViolatingElements,
				Relationship sourceRelationship) {
			super();
			this.constraint = constraint;
			this.actualCardinality = actualCardinality;
			this.numViolatingElements = numViolatingElements;
			this.sourceRelationship = sourceRelationship;
		}

		public NaturalNumberSet getActualCardinality() {
			return actualCardinality;
		}

		public Constraint getConstraint() {
			return constraint;
		}

		public Relationship getSourceRelationship() {
			return sourceRelationship;
		}

		@Override
		public String toString() {
			return String.format("%3d violations of %s: %s<=%s.", numViolatingElements, constraint, actualCardinality,
					sourceRelationship);
		}

	}

}
