package de.uni_potsdam.hpi.isg.efes.csg.datafit.values;

import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;

import java.io.PrintStream;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.datafit.DataFitModel;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.DataFitReport;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.conclusion.ConclusionModel;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.data_access.ColumnIterator;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.data_access.ColumnIterator.ValueRetriever;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.data_access.String2IntValueRetriever;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.ValueFitMetric;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.ValueFitMetric.SourceColumnStatistic;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.ValueFitMetric.SourceOperator;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.ValueFitMetric.TargetColumnStatistic;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.ValueFitMetric.TargetOperator;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.BasicNodeMapping;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.NodeCorrespondence;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Datatype;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Datatype.GeneralType;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.ValueNode;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.DataSet;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.IntegrationScenario;
import de.uni_potsdam.hpi.isg.efes.csg.util.RelationalSchemaGraphUtil;
import de.uni_potsdam.hpi.isg.utils.db.Database;
import de.uni_potsdam.hpi.isg.utils.util.CollectionUtil;

public class ValueCharacteristicsInspector implements DataFitModel {

	private static final Logger logger = LogManager.getLogger(ValueCharacteristicsInspector.class.getSimpleName());

	private IntegrationScenario integrationScenario;

	private List<ValueFitMetric> registeredMetrics = new ArrayList<ValueFitMetric>();

	private ConclusionModel conclusionModel;

	@Override
	public void setIntegrationScenario(IntegrationScenario scenario) {
		this.integrationScenario = scenario;
	}

	public void register(ValueFitMetric metric) {
		if (this.registeredMetrics.contains(metric)) {
			logger.warn("Prevented {} from being registered a second time.", metric);
			return;
		}
		this.registeredMetrics.add(metric);
	}

	public void setConclusionModel(ConclusionModel conclusionModel) {
		this.conclusionModel = conclusionModel;
	}

	@Override
	public void evaluateScenarioTo(DataFitReport report) {
		List<DataSet> sourceDataSets = this.integrationScenario.getSourceDataSets();
		DataSet targetDataSet = this.integrationScenario.getTargetDataSet();
		Map<DataSet, BasicNodeMapping> mappings = this.integrationScenario.getMappings();

		if (sourceDataSets.isEmpty()) throw new IllegalStateException("No source data sets found.");

		ReportSection reportSection = new ReportSection();
		for (DataSet sourceDataSet : sourceDataSets) {
			BasicNodeMapping mapping = mappings.get(sourceDataSet);
			profileValues(sourceDataSet, mapping, targetDataSet, reportSection);

		}

		drawConclusions(reportSection);

		report.add(reportSection);
	}

	private ReportSection profileValues(DataSet sourceDataSet, BasicNodeMapping mapping, DataSet targetDataSet,
			ReportSection reportSection) {
		if (sourceDataSet.getDataStore() == null || sourceDataSet.getType() != DataSet.Type.RELATIONAL) {
			logger.error("Cannot perform value characteristics inspection on source {}.", sourceDataSet);
			return reportSection;
		}

		if (targetDataSet.getDataStore() == null || targetDataSet.getType() != DataSet.Type.RELATIONAL) {
			logger.error("Cannot perform value characteristics inspection on target {}.", targetDataSet);
			return reportSection;
		}

		for (Entry<Node, List<NodeCorrespondence>> entry : mapping.getNodeCorrespondences().entrySet()) {
			Node targetNode = entry.getKey();
			List<NodeCorrespondence> nodeCorrespondences = entry.getValue();

			// Only consider value node correspondences.
			if (!(targetNode instanceof ValueNode) || nodeCorrespondences.isEmpty()) continue;

			ValueNode targetValueNode = (ValueNode) targetNode;

			CorrespondenceLoop: for (NodeCorrespondence nodeCorrespondence : nodeCorrespondences) {
				Node sourceNode = nodeCorrespondence.getSourceNode();
				if (!sourceNode.isSimpleValueNode() || sourceNode.isMetaData()) continue;

				// Assumption of cascading FKs: leave out correspondence if it mirrors a PK correspondence.
				List<Node> sourceNodeReferences = RelationalSchemaGraphUtil.findReferencedAttributeNodes(
						sourceDataSet.getSchema(), sourceNode);
				if (RelationalSchemaGraphUtil.isForeignKeyNode(targetNode, targetDataSet.getSchema())) {
					Node targetPk = CollectionUtil.getAny(RelationalSchemaGraphUtil.findReferencedAttributeNodes(
							targetDataSet.getSchema(), targetNode));
					for (NodeCorrespondence pkCorrespondence : mapping.getCorrespondencesForTargetNode(targetPk)) {
						Node sourcePk = pkCorrespondence.getSourceNode();
						if (sourceNode.equals(sourcePk) || sourceNodeReferences.contains(sourcePk)) {
							logger.info("Skipping FK correspondence: {}.", nodeCorrespondence);
							continue CorrespondenceLoop;
						}
					}
				}

				ValueNode sourceValueNode = (ValueNode) sourceNode;
				Datatype targetDataType = targetValueNode.getDatatype();
				Datatype sourceDataType = sourceValueNode.getDatatype();

				logger.info("Need to compare {} ({}) and {} ({}).", targetNode, targetDataType, sourceNode,
						sourceDataType);

				ColumnIterator.ValueRetriever sourceValueRetriever = findValueRetriever(sourceDataType, targetDataType);
				if (sourceValueRetriever == null) {
					logger.warn("No suitable value retriever for conversion of {} to {}.", sourceDataType,
							targetDataType);
					continue;
				}

				List<ValueFitMetric> applicableMetrics = findApplicableMetrics(targetDataType);
				if (applicableMetrics == null || applicableMetrics.isEmpty()) {
					logger.warn("No metrics found for {}.", targetDataType);
					continue;
				}

				Database<?> sourceDb = (Database<?>) sourceDataSet.getDataStore();
				Map<ValueFitMetric, ValueFitMetric.SourceColumnStatistic> sourceMetricResults = gatherSourceResults(
						applicableMetrics, sourceValueNode, sourceDb, sourceValueRetriever);

				Database<?> targetDb = (Database<?>) targetDataSet.getDataStore();
				Map<ValueFitMetric, ValueFitMetric.TargetColumnStatistic> targetMetricResults = gatherTargetResults(
						applicableMetrics, targetValueNode, targetDb);

				for (Entry<ValueFitMetric, ValueFitMetric.SourceColumnStatistic> sourceMetricResultEntry : sourceMetricResults
						.entrySet()) {

					ValueFitMetric similarityMetric = sourceMetricResultEntry.getKey();
					ValueFitMetric.SourceColumnStatistic sourceResult = sourceMetricResultEntry.getValue();
					reportSection.addSourceColumnStatistic(sourceValueNode, sourceResult);

					ValueFitMetric.TargetColumnStatistic targetResult = targetMetricResults.get(similarityMetric);
					if (targetMetricResults.isEmpty()) {
						logger.warn("Could not find a target metric result for {}.", similarityMetric);
						continue;
					}
					reportSection.addTargetColumnStatistic(targetValueNode, targetResult);

					ValueFitMetric.Result result = targetResult.compare(sourceResult);
					reportSection.addMetricResult(nodeCorrespondence, result);

					logger.debug("Comparison of {} between {} and {}:", similarityMetric, sourceValueNode,
							targetValueNode);
					logger.debug(" {} (relevance={}): {} <-> {}", result.getFitValue(), result.getRelevance(),
							sourceResult, targetResult);
				}
			}
		}

		return reportSection;
	}

	private ValueRetriever findValueRetriever(Datatype sourceDataType, Datatype targetDataType) {

		GeneralType sourceGeneralType = sourceDataType.getGeneralType();
		GeneralType targetGeneralType = targetDataType.getGeneralType();
		if (sourceGeneralType == targetGeneralType) {
			return ColumnIterator.DefaultValueRetriever.INSTANCE;
		}
		logger.debug("Different data types. Source is {} and target is {}.", sourceDataType, targetDataType);

		if (targetDataType.canBeConvertedFrom(sourceGeneralType)) {
			return ColumnIterator.DefaultValueRetriever.INSTANCE;
		}

		logger.debug("Incompatible data types. Source is {} and target is {}.", sourceDataType, targetDataType);
		if (targetGeneralType == Datatype.GeneralType.INTEGER) {
			return String2IntValueRetriever.INSTANCE;
		}

		return null;
	}

	private List<ValueFitMetric> findApplicableMetrics(Datatype targetDataType) {
		List<ValueFitMetric> applicableMetrics = new ArrayList<>();
		for (ValueFitMetric metric : this.registeredMetrics) {
			if (metric.isApplicableFor(targetDataType.getGeneralType())) {
				applicableMetrics.add(metric);
			}
		}
		return applicableMetrics;
	}

	private Map<ValueFitMetric, ValueFitMetric.SourceColumnStatistic> gatherSourceResults(
			List<ValueFitMetric> applicableMetrics, ValueNode valueNode, Database<?> db,
			ValueRetriever sourceValueRetriever) {
		Map<ValueFitMetric, ValueFitMetric.SourceOperator> metricOperators = new HashMap<>();
		for (ValueFitMetric metric : applicableMetrics) {
			metricOperators.put(metric, metric.createSourceOperator());
		}

		Map<ValueFitMetric, ValueFitMetric.SourceColumnStatistic> metricResults = new HashMap<>();
		try (ColumnIterator iterator = new ColumnIterator(valueNode, db)) {
			iterator.setValueRetriever(sourceValueRetriever);

			while (iterator.findNextValue()) {
				for (ValueFitMetric.SourceOperator operator : metricOperators.values()) {
					operator.update(iterator);
				}
			}

			for (Entry<ValueFitMetric, ValueFitMetric.SourceOperator> operatorEntry : metricOperators.entrySet()) {
				ValueFitMetric metric = operatorEntry.getKey();
				SourceOperator operator = operatorEntry.getValue();
				SourceColumnStatistic result = operator.concludeSourceResult();
				metricResults.put(metric, result);
			}

		} catch (SQLException e) {
			logger.catching(e);
		}
		return metricResults;
	}

	private Map<ValueFitMetric, ValueFitMetric.TargetColumnStatistic> gatherTargetResults(
			List<ValueFitMetric> applicableMetrics, ValueNode valueNode, Database<?> db) {
		Map<ValueFitMetric, ValueFitMetric.TargetOperator> metricOperators = new HashMap<>();
		for (ValueFitMetric metric : applicableMetrics) {
			metricOperators.put(metric, metric.createTargetOperator());
		}

		Map<ValueFitMetric, ValueFitMetric.TargetColumnStatistic> metricResults = new HashMap<>();
		try (ColumnIterator iterator = new ColumnIterator(valueNode, db)) {

			while (iterator.findNextValue()) {
				for (ValueFitMetric.TargetOperator operator : metricOperators.values()) {
					operator.update(iterator);
				}
			}

			for (Entry<ValueFitMetric, ValueFitMetric.TargetOperator> operatorEntry : metricOperators.entrySet()) {
				ValueFitMetric metric = operatorEntry.getKey();
				TargetOperator operator = operatorEntry.getValue();
				TargetColumnStatistic result = operator.concludeTargetResult();
				metricResults.put(metric, result);
			}

		} catch (SQLException e) {
			logger.catching(e);
		}
		return metricResults;
	}

	private void drawConclusions(ReportSection reportSection) {
		this.conclusionModel.initialize(this.integrationScenario, reportSection);
		this.conclusionModel.drawConclusions();
	}

	public static class ReportSection implements DataFitReport.Section {

		private Map<ValueNode, List<ValueFitMetric.SourceColumnStatistic>> sourceColumnStatisticsPerNode = new HashMap<>();
		private Map<ValueNode, List<ValueFitMetric.TargetColumnStatistic>> targetColumnStatisticsPerNode = new HashMap<>();
		private Map<NodeCorrespondence, List<ValueFitMetric.Result>> metricResultsPerCorrespondence = new HashMap<>();
		private List<Issue> issues = new ArrayList<>();

		void addSourceColumnStatistic(ValueNode sourceValueNode, SourceColumnStatistic sourceResult) {
			addMapEntry(sourceValueNode, sourceResult, this.sourceColumnStatisticsPerNode);
		}

		void addTargetColumnStatistic(ValueNode targetValueNode, TargetColumnStatistic targetResult) {
			addMapEntry(targetValueNode, targetResult, this.targetColumnStatisticsPerNode);
		}

		void addMetricResult(NodeCorrespondence correspondence, ValueFitMetric.Result result) {
			addMapEntry(correspondence, result, this.metricResultsPerCorrespondence);
		}

		public void addIssue(Issue issue) {
			logger.debug("Adding issue to report section: {}", issue);
			this.issues.add(issue);
		}

		private <S, T> void addMapEntry(S key, T statistic, Map<S, List<T>> statisticsPerNode) {
			List<T> statistics = statisticsPerNode.get(key);
			if (statistics == null) {
				statistics = new ArrayList<>();
				statisticsPerNode.put(key, statistics);
			}
			statistics.add(statistic);
		}

		public ValueFitMetric.SourceColumnStatistic getSourceColumnStatistic(ValueNode sourceNode,
				Class<? extends ValueFitMetric> metricClass) {
			return getEntry(sourceNode, metricClass, this.sourceColumnStatisticsPerNode);
		}

		public ValueFitMetric.TargetColumnStatistic getTargetColumnStatistic(ValueNode targetNode,
				Class<? extends ValueFitMetric> metricClass) {
			return getEntry(targetNode, metricClass, this.targetColumnStatisticsPerNode);
		}

		public ValueFitMetric.Result getResult(NodeCorrespondence correspondence,
				Class<? extends ValueFitMetric> metricClass) {
			return getEntry(correspondence, metricClass, this.metricResultsPerCorrespondence);
		}

		private <S, T extends ValueFitMetric.MetricAssociate> T getEntry(S key,
				Class<? extends ValueFitMetric> metricClass, Map<S, List<T>> store) {
			List<T> entries = store.get(key);
			if (entries == null) return null;

			for (T entry : entries) {
				if (entry.getMetricClass().equals(metricClass)) return entry;
			}

			return null;
		}

		public Map<NodeCorrespondence, List<ValueFitMetric.Result>> getMetricResults() {
			return Collections.unmodifiableMap(metricResultsPerCorrespondence);
		}

		public List<Issue> getIssues() {
			return Collections.unmodifiableList(this.issues);
		}

		@Override
		public void print(PrintStream stream) {
			stream.println("Value fit report");
			stream.println("================");
			stream.format(" %d issues:\n", this.issues.size());
			for (Issue issue : this.issues) {
				stream.format(" * %s\n", issue);
			}
		}

		@Override
		public void printCsv(PrintStream printStream) {
			// TODO Auto-generated method stub

		}
	}

	public static class Issue {

		public static final int NO_COMPLEXITY = 0;
		public static final int LOW_COMPLEXITY = 1;
		public static final int MEDIUM_COMPLEXITY = 2;
		public static final int HIGH_COMPLEXITY = 3;

		private Object2IntMap<String> parameters = new Object2IntOpenHashMap<>();

		public void setParameter(String key, int value) {
			this.parameters.put(key, value);
		}

		public Object2IntMap<String> getParameters() {
			return this.parameters;
		}

	}

}
