package de.uni_potsdam.hpi.isg.efes.csg.datafit.values.conclusion;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.ValueCharacteristicsInspector.Issue;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.ValueCharacteristicsInspector.ReportSection;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.AverageStringLengthMetric;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.CharacterHistogramMetric;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.ConstancyMetric;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.ExponentialHistogramMetric;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.FillStatusMetric;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.MeanValueMetric;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.RangeOverlapMetric;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.TextPatternMetric;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.TopKValuesMetric;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.ValueFitMetric;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics.ValueFitMetric.Result;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.BasicNodeMapping;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.NodeCorrespondence;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.ValueNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.DataSet;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.IntegrationScenario;

public class SequentialConclusionModel implements ConclusionModel {

	private double mediumComplexityBoundary = 0.5;
	private double lowComplexityBoundary = 0.8;
	private double noComplexityBoundary = 0.9;

	private double criticalFillStatusDiff = 0.3;
	private double minDomainConstancy = 0.7;
	private int minDomainEvidence = 5;

	private static final Logger logger = LogManager
			.getLogger(SequentialConclusionModel.class.getSimpleName());
	private IntegrationScenario integrationScenario;
	private ReportSection reportSection;

	@Override
	public void initialize(IntegrationScenario scenario,
			ReportSection reportSection) {
		this.integrationScenario = scenario;
		this.reportSection = reportSection;
	}

	@Override
	public void drawConclusions() {
		DataSet targetDataSet = this.integrationScenario.getTargetDataSet();
		BasicSchemaGraph targetSchema = targetDataSet.getSchema();

		List<DataSet> sourceDataSets = this.integrationScenario
				.getSourceDataSets();
		Map<DataSet, BasicNodeMapping> mappings = this.integrationScenario
				.getMappings();

		if (sourceDataSets.isEmpty())
			throw new IllegalStateException();

		for (DataSet sourceDataSet : sourceDataSets) {
			BasicNodeMapping mapping = mappings.get(sourceDataSet);

			for (Node targetNode : targetSchema.getAtomicNodes().values()) {
				if (!targetNode.isSimpleValueNode() || targetNode.isMetaData()) {
					continue;
				}
				ValueNode targetValueNode = (ValueNode) targetNode;
				logger.trace("Inspection target node {}.", targetValueNode);

				Collection<NodeCorrespondence> correspondences = mapping
						.getCorrespondencesForTargetNode(targetNode);
				if (correspondences.isEmpty()) {
					logger.trace(
							"Skipping inspection of {}, though, fill status might be an issue.",
							targetNode);
					continue;
				}

				for (NodeCorrespondence correspondence : correspondences) {
					ValueNode sourceValueNode = (ValueNode) correspondence
							.getSourceNode();
					handleOneToOneCorrespondence(sourceValueNode,
							targetValueNode, correspondence);
				}
			}
		}
	}

	private void handleOneToOneCorrespondence(ValueNode sourceValueNode,
			ValueNode targetValueNode, NodeCorrespondence correspondence) {

		logger.debug("Inspection of correspondence {} -> {}.", sourceValueNode,
				targetValueNode);

		FillStatusMetric.ColumnStatistic targetFillStatistic = (FillStatusMetric.ColumnStatistic) reportSection
				.getTargetColumnStatistic(targetValueNode,
						FillStatusMetric.class);
		if (targetFillStatistic == null) {
			logger.debug("Skipping {}, no fill statistic for it.", targetValueNode);
			return;
		}
		double targetFillStatus = targetFillStatistic.getFillStatus();

		FillStatusMetric.ColumnStatistic sourceFillStatistic = (FillStatusMetric.ColumnStatistic) reportSection
				.getSourceColumnStatistic(sourceValueNode,
						FillStatusMetric.class);
		double sourceFillStatus = sourceFillStatistic.getFillStatus();

		// Step 1: Check how many values are required by the target and provided
		// by the source (if there is no not null constraint).
		if (!hasNotNullConstraint(targetValueNode)) {
			logger.trace("Source fill status: {}, target fill status: {}",
					sourceFillStatus, targetFillStatus);
			if (targetFillStatus * targetFillStatus - sourceFillStatus
					* sourceFillStatus > criticalFillStatusDiff) {
				double desiredSourceFillStatus = Math
						.sqrt((targetFillStatus * targetFillStatus)
								- criticalFillStatusDiff);
				int neededValues = (int) Math
						.round((desiredSourceFillStatus - sourceFillStatus)
								* sourceFillStatistic.getColumnSize());
				if (neededValues > 0) {
					logger.debug("Too few elements: {}.", correspondence);

					Issue issue = new FewElementsIssue(correspondence,
							neededValues);
					this.reportSection.addIssue(issue);
				}
			}
		}

		// Step 2: Check if there are values that cannot be compared and are
		// fundamentally different.
		int numInconvertibleValues = sourceFillStatistic
				.getNumInconvertibleValues();
		int numDistinctIncovertibleValues = sourceFillStatistic
				.getNumDistinctIncovertibleValues();
		if (numInconvertibleValues > 0) {
			logger.debug("Incompatible source values: {}.", correspondence);
			Issue convertValuesIssue = new DifferentRepresentationsIssue(
					correspondence, numInconvertibleValues,
					numDistinctIncovertibleValues, Issue.HIGH_COMPLEXITY, true);
			this.reportSection.addIssue(convertValuesIssue);
		}

		// Step 3: If there are still values that can be compared, do more
		// specific checks.
		if (sourceFillStatistic.getNumConvertibleValues() == 0) {
			logger.trace("No comparable values for {}.", correspondence);
			return;
		}

		switch (targetValueNode.getDatatype().getGeneralType()) {
		case STRING:
			// compare strings
			compareStrings(sourceValueNode, targetValueNode, correspondence);
			break;

		case INTEGER:
			compareIntegers(sourceValueNode, targetValueNode, correspondence);
			break;

		default:
			logger.warn("Cannot draw specific conclusions for {}.",
					targetValueNode.getDatatype());
		}
	}

	private boolean hasNotNullConstraint(ValueNode targetValueNode) {
		BasicSchemaGraph targetSchema = this.integrationScenario
				.getTargetDataSet().getSchema();
		for (Relationship relationship : targetSchema
				.getAtomicRelationships(targetValueNode)) {
			if (relationship.isExplicit()
					&& relationship.getEndNode().isSimpleIdentityNode()) {
				NaturalNumberSet cardinality = targetSchema
						.getCardinality(relationship.getInverse());
				if (cardinality.equals(NaturalNumberSet.ONE)) {
					return true;
				}
			}
		}
		return false;
	}

	private void compareStrings(ValueNode sourceValueNode,
			ValueNode targetValueNode, NodeCorrespondence correspondence) {

		// Find out if the source value describes a domain.
		if (isDomainRestricted(targetValueNode, true)) {
			logger.trace("Assume {} to be restricted by a domain.",
					targetValueNode);

			// Check if the source value also is a domain.
			if (isNotDomainRestricted(sourceValueNode, false)) {
				addTooSpecificIssue(correspondence);
			} else {
				ValueFitMetric.Result topKValuesMetricResult = this.reportSection
						.getResult(correspondence, TopKValuesMetric.class);
				addUncriticalDifferentRepresentationsIssue(correspondence,
						topKValuesMetricResult);
			}

		} else if (isNotDomainRestricted(targetValueNode, true)) {
			logger.trace("Assume {} not to be restricted by a domain.",
					targetValueNode);

			// Check if the source is domain-restricted.
			if (isDomainRestricted(sourceValueNode, false)) {
				addTooGeneralIssue(correspondence);

			} else {
				ValueFitMetric.Result formattingResult = reportSection
						.getResult(correspondence, TextPatternMetric.class);
				ValueFitMetric.Result characterHistogramResult = reportSection
						.getResult(correspondence,
								CharacterHistogramMetric.class);
				ValueFitMetric.Result averageStringLengthResult = reportSection
						.getResult(correspondence,
								AverageStringLengthMetric.class);
				addUncriticalDifferentRepresentationsIssue(correspondence,
						formattingResult, characterHistogramResult,
						averageStringLengthResult);
			}
		} else {
			logger.trace(
					"No assumption, whether {} is restricted by a domain.",
					targetValueNode);

			ValueFitMetric.Result formattingResult = reportSection.getResult(
					correspondence, TextPatternMetric.class);
			ValueFitMetric.Result characterHistogramResult = reportSection
					.getResult(correspondence, CharacterHistogramMetric.class);
			ValueFitMetric.Result averageStringLengthResult = reportSection
					.getResult(correspondence, AverageStringLengthMetric.class);
			addUncriticalDifferentRepresentationsIssue(correspondence,
					formattingResult, characterHistogramResult,
					averageStringLengthResult);
		}
	}

	private void compareIntegers(ValueNode sourceValueNode,
			ValueNode targetValueNode, NodeCorrespondence correspondence) {

		// Find out if the source value describes a domain.
		if (isDomainRestricted(targetValueNode, true)) {
			logger.trace("Assume {} to be restricted by a domain.",
					targetValueNode);

			// Check if the source value also is a domain.
			if (isNotDomainRestricted(sourceValueNode, false)) {
				addTooSpecificIssue(correspondence);
			} else {
				ValueFitMetric.Result topKValuesMetricResult = this.reportSection
						.getResult(correspondence, TopKValuesMetric.class);
				addUncriticalDifferentRepresentationsIssue(correspondence,
						topKValuesMetricResult);
			}

		} else if (isNotDomainRestricted(targetValueNode, true)) {
			logger.trace("Assume {} not to be restricted by a domain.",
					targetValueNode);

			// Check if the source is domain-restricted.
			if (isDomainRestricted(sourceValueNode, false)) {
				addTooGeneralIssue(correspondence);

			} else {
				ValueFitMetric.Result rangeOverlapResult = reportSection
						.getResult(correspondence, RangeOverlapMetric.class);
				ValueFitMetric.Result histogramResult = reportSection
						.getResult(correspondence,
								ExponentialHistogramMetric.class);
				ValueFitMetric.Result meanResult = reportSection.getResult(
						correspondence, MeanValueMetric.class);
				addUncriticalDifferentRepresentationsIssue(correspondence,
						rangeOverlapResult, histogramResult, meanResult);
			}
		} else {
			logger.trace(
					"No assumption, whether {} is restricted by a domain.",
					targetValueNode);

			ValueFitMetric.Result rangeOverlapResult = reportSection.getResult(
					correspondence, RangeOverlapMetric.class);
			addUncriticalDifferentRepresentationsIssue(correspondence,
					rangeOverlapResult);
		}
	}

	/**
	 * Estimates whether the column represented by the given node is restricted
	 * by a domain. For a positive outcome, there must be enough values in the
	 * column and the constancy metric must have a certain value.
	 * 
	 * @param node
	 *            represents a column of the underlying database
	 * @return whether the column is domain-restricted
	 */
	private boolean isDomainRestricted(ValueNode node, boolean isTargetNode) {
		ConstancyMetric.ColumnStatistic constancy = getConstancy(node,
				isTargetNode);
		FillStatusMetric.ColumnStatistic fillStatus = getFillStatus(node,
				isTargetNode);

		return constancy.getConstancy() >= minDomainConstancy
				&& fillStatus.getNumConvertibleValues() >= minDomainEvidence;
	}

	/**
	 * Estimates whether the column represented by the given node is not
	 * restricted by a domain. For a positive outcome, there must be enough
	 * values in the column and the constancy metric must have a certain value.
	 * 
	 * @param node
	 *            represents a column of the underlying database
	 * @return whether the column is not domain-restricted
	 */
	private boolean isNotDomainRestricted(ValueNode node, boolean isTargetNode) {
		ConstancyMetric.ColumnStatistic constancy = getConstancy(node,
				isTargetNode);
		FillStatusMetric.ColumnStatistic fillStatus = getFillStatus(node,
				isTargetNode);

		return constancy.getConstancy() < minDomainConstancy
				&& fillStatus.getNumConvertibleValues() >= minDomainEvidence;
	}

	private ConstancyMetric.ColumnStatistic getConstancy(ValueNode node,
			boolean isTargetNode) {
		if (!isTargetNode) {
			return (ConstancyMetric.ColumnStatistic) reportSection
					.getSourceColumnStatistic(node, ConstancyMetric.class);
		} else {
			return (ConstancyMetric.ColumnStatistic) reportSection
					.getTargetColumnStatistic(node, ConstancyMetric.class);
		}
	}

	private FillStatusMetric.ColumnStatistic getFillStatus(ValueNode node,
			boolean isTargetNode) {
		if (!isTargetNode) {
			return (FillStatusMetric.ColumnStatistic) reportSection
					.getSourceColumnStatistic(node, FillStatusMetric.class);
		} else {
			return (FillStatusMetric.ColumnStatistic) reportSection
					.getTargetColumnStatistic(node, FillStatusMetric.class);
		}
	}

	private void addTooSpecificIssue(NodeCorrespondence correspondence) {
		logger.debug("Too specific: {}.", correspondence);

		FillStatusMetric.ColumnStatistic fillStatus = (FillStatusMetric.ColumnStatistic) this.reportSection
				.getSourceColumnStatistic(
						(ValueNode) correspondence.getSourceNode(),
						FillStatusMetric.class);

		Issue generalizeIssue = new TooSpecificIssue(correspondence,
				fillStatus.getNumConvertibleValues(),
				fillStatus.getNumDistinctConvertibleValues());
		this.reportSection.addIssue(generalizeIssue);
	}

	private void addTooGeneralIssue(NodeCorrespondence correspondence) {
		FillStatusMetric.ColumnStatistic fillStatus = (FillStatusMetric.ColumnStatistic) this.reportSection
				.getSourceColumnStatistic(
						(ValueNode) correspondence.getSourceNode(),
						FillStatusMetric.class);
		ValueFitMetric.Result constancyResult = this.reportSection.getResult(
				correspondence, ConstancyMetric.class);

		int complexity = toComplexity(constancyResult);
		if (complexity != Issue.NO_COMPLEXITY) {
			Issue generalizeIssue = new TooGeneralIssue(correspondence,
					fillStatus.getNumConvertibleValues(),
					fillStatus.getNumDistinctConvertibleValues(), complexity);
			this.reportSection.addIssue(generalizeIssue);
		}
	}

	private void addUncriticalDifferentRepresentationsIssue(
			NodeCorrespondence correspondence,
			ValueFitMetric.Result... complexityDrivers) {
		FillStatusMetric.ColumnStatistic fillStatus = (FillStatusMetric.ColumnStatistic) this.reportSection
				.getSourceColumnStatistic(
						(ValueNode) correspondence.getSourceNode(),
						FillStatusMetric.class);

		int complexity = toComplexity(complexityDrivers);
		if (complexity != Issue.NO_COMPLEXITY) {
			logger.info("Uncritical different representations: {}.",
					correspondence);

			DifferentRepresentationsIssue convertValuesIssue = new DifferentRepresentationsIssue(
					correspondence, fillStatus.getNumConvertibleValues(),
					fillStatus.getNumDistinctConvertibleValues(), complexity,
					false);
			convertValuesIssue.setAdditionalInformation(String.format(
					"%.3f <- %s",
					calculateRelevanceWeightedMean(complexityDrivers),
					Arrays.toString(complexityDrivers)));
			this.reportSection.addIssue(convertValuesIssue);
		}
	}

	@SuppressWarnings("unused")
	private double calculateHarmonicMean(Result[] results) {
		int numResults = 0;
		double denominator = 0;
		for (Result result : results) {
			if (result.isValid()) {
				double fitValue = result.getFitValue();
				if (fitValue == 0) {
					return 0;
				}
				numResults++;
				denominator += 1 / fitValue;
			}
		}
		return numResults / denominator;
	}

	private double calculateRelevanceWeightedMean(Result[] results) {
		double nominator = 0;
		double denominator = 0;
		for (Result result : results) {
			if (result.isValid()) {
				nominator += result.getFitValue() * result.getRelevance();
				denominator += result.getRelevance();
			}
		}
		return denominator == 0 ? 1 : nominator / denominator;
	}

	private int toComplexity(Result... results) {
		if (results.length == 0) {
			throw new IllegalArgumentException();
		}

		// Calculate harmonic mean of results' fit values.
		// double mean = calculateHarmonicMean(results);
		double mean = calculateRelevanceWeightedMean(results);
		logger.info("{} is the mean of {}", mean, Arrays.toString(results));
		if (mean >= noComplexityBoundary) {
			return Issue.NO_COMPLEXITY;
		} else if (mean >= lowComplexityBoundary) {
			return Issue.LOW_COMPLEXITY;
		} else if (mean >= mediumComplexityBoundary) {
			return Issue.MEDIUM_COMPLEXITY;
		}
		return Issue.HIGH_COMPLEXITY;
	}

	public void setMediumComplexityBoundary(double mediumComplexityBoundary) {
		this.mediumComplexityBoundary = mediumComplexityBoundary;
	}

	public void setLowComplexityBoundary(double lowComplexityBoundary) {
		this.lowComplexityBoundary = lowComplexityBoundary;
	}

	public void setNoComplexityBoundary(double noComplexityBoundary) {
		this.noComplexityBoundary = noComplexityBoundary;
	}

	public void setCriticalFillStatusDiff(double criticalFillStatusDiff) {
		this.criticalFillStatusDiff = criticalFillStatusDiff;
	}

	public void setMinDomainConstancy(double minDomainConstancy) {
		this.minDomainConstancy = minDomainConstancy;
	}

	public void setMinDomainEvidence(int minDomainEvidence) {
		this.minDomainEvidence = minDomainEvidence;
	}

	public void setIntegrationScenario(IntegrationScenario integrationScenario) {
		this.integrationScenario = integrationScenario;
	}

	public void setReportSection(ReportSection reportSection) {
		this.reportSection = reportSection;
	}

}
