package de.uni_potsdam.hpi.isg.efes.csg.datafit.values.data_access;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.nodes.ValueNode;
import de.uni_potsdam.hpi.isg.efes.csg.util.RelationalSchemaToCSGConverter;
import de.uni_potsdam.hpi.isg.utils.db.Database;

public class ColumnIterator implements AutoCloseable {

	private static final Logger logger = LogManager.getLogger(ColumnIterator.class.getSimpleName());

	private static final String ITERATION_SQL_QUERY = "SELECT %1$s AS value, COUNT(*) AS count "
			+ "FROM %2$s GROUP BY %1$s";

	private Statement statement;
	private ResultSet resultSet;
	private boolean needToCloseDb;
	private ValueRetriever valueRetriever = ColumnIterator.DefaultValueRetriever.INSTANCE;

	private Database<?> db;

	public ColumnIterator(ValueNode valueNode, Database<?> db) throws SQLException {
		this(RelationalSchemaToCSGConverter.extractTableName(valueNode), RelationalSchemaToCSGConverter
				.extractAttributeNames(valueNode).get(0), db);
	}

	public ColumnIterator(String tableName, String columnName, Database<?> db) throws SQLException {
		this.db = db;
		if (this.needToCloseDb = !db.isOpen()) {
			db.open();
		}
		this.statement = db.getConnection().createStatement();
		String sqlQuery = String.format(ITERATION_SQL_QUERY, columnName, tableName);
		this.resultSet = this.statement.executeQuery(sqlQuery);
	}

	public boolean findNextValue() throws SQLException {
		return this.resultSet.next();
	}

	public boolean wasNull() throws SQLException {
		return this.resultSet.wasNull();
	}

	/**
	 * Tells if the value retrieved via the latest called <code>getValueAsXXX()</code> was valid.
	 */
	public boolean wasValid() {
		return this.valueRetriever.couldRetrieveValidValue();
	}
	
	/**
	 * Tells if the registered value retriever can cast the current column value to its target datatype.
	 */
	public boolean isValid() throws SQLException {
		return this.valueRetriever.isValueValid(this.resultSet);
	}

	public String getValueAsString() throws SQLException {
		return this.valueRetriever.getString(this.resultSet);
	}

	public int getValueAsInt() throws SQLException {
		return this.valueRetriever.getInt(this.resultSet);
	}

	public int getValueCount() throws SQLException {
		return this.resultSet.getInt(2);
	}

	public void setValueRetriever(ValueRetriever valueRetriever) {
		if (valueRetriever == null) {
			throw new IllegalArgumentException();
		}

		this.valueRetriever = valueRetriever;
	}

	public void close() {
		if (this.needToCloseDb) {
			try {
				db.close();
			} catch (SQLException e) {
				logger.catching(e);
			}
		}
		if (this.statement != null) {
			try {
				this.statement.close();
			} catch (SQLException e) {
				logger.catching(e);
			} finally {
				this.statement = null;
			}
		}
	}

	@Override
	public String toString() {
		return "ColumnIterator [valueRetriever=" + valueRetriever + ", db=" + db + "]";
	}

	public static interface ValueRetriever {

		int getInt(ResultSet resultSet) throws SQLException;

		String getString(ResultSet resultSet) throws SQLException;

		boolean couldRetrieveValidValue();

		/**
		 * In comparison to {@link #couldRetrieveValidValue()}, this method
		 * checks the validity of the underlying value without having to
		 * retrieved it beforehand.
		 * @throws SQLException 
		 */
		boolean isValueValid(ResultSet resultSet) throws SQLException;

	}

	public static class DefaultValueRetriever implements ValueRetriever {

		public static DefaultValueRetriever INSTANCE = new DefaultValueRetriever();

		private boolean couldRetrieveValidValue;

		protected void setRetrieveSuccess(boolean isSuccess) {
			this.couldRetrieveValidValue = isSuccess;
		};

		@Override
		public boolean couldRetrieveValidValue() {
			return couldRetrieveValidValue;
		}

		@Override
		public boolean isValueValid(ResultSet resultSet) throws SQLException {
			// No casting, so no problems.
			return true;
		}
		
		@Override
		public int getInt(ResultSet resultSet) throws SQLException {
			setRetrieveSuccess(true);
			return resultSet.getInt(1);
		}

		@Override
		public String getString(ResultSet resultSet) throws SQLException {
			setRetrieveSuccess(true);
			return resultSet.getString(1);
		}

		@Override
		public String toString() {
			return "DefaultValueRetriever";
		}

	}

}
