package de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import de.uni_potsdam.hpi.isg.efes.csg.nodes.Datatype;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Datatype.GeneralType;

public abstract class AbstractValueFitMetric implements ValueFitMetric {

	protected Set<Datatype.GeneralType> supportedTypes = new HashSet<>();

	public AbstractValueFitMetric(GeneralType... supportedTypes) {
		super();
		this.supportedTypes = new HashSet<>(Arrays.asList(supportedTypes));
	}

	@Override
	public boolean isApplicableFor(GeneralType type) {
		return this.supportedTypes.contains(type);
	}

	@Override
	public String toString() {
		return getClass().getSimpleName();
	}

	@Override
	public int hashCode() {
		return getClass().hashCode() + 1;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		return this.getClass().equals(obj.getClass());
	}

	public static class DefaultResult implements ValueFitMetric.Result {

		private double fitValue;
		private double relevance;
		private Class<? extends ValueFitMetric> metricClass;
		private boolean isValid;

		public DefaultResult(double fitValue, double relevance,
				Class<? extends ValueFitMetric> metricClass) {
			super();

			if (Double.isNaN(fitValue)) {
				this.isValid = false;
				this.relevance = 0d;
			} else {
				if (!(fitValue >= 0 && fitValue <= 1)) {
					// Allow minor deviations.
					if (fitValue >= 0.0001 && fitValue <= 1.0001) {
						fitValue = Math.min(1, Math.max(0, fitValue));
					} else {
						throw new IllegalArgumentException(String.format(
								"Illegal fit value %f for %s.", fitValue,
								metricClass.getSimpleName()));
					}
				}

				if (metricClass == null) {
					throw new IllegalArgumentException();
				}

				this.fitValue = fitValue;
				this.relevance = relevance;
				this.isValid = true;
			}
			this.metricClass = metricClass;
		}

		public double getFitValue() {
			return fitValue;
		}

		@Override
		public Class<? extends ValueFitMetric> getMetricClass() {
			return this.metricClass;
		}

		@Override
		public boolean isValid() {
			return this.isValid;
		}

		@Override
		public double getRelevance() {
			return this.relevance;
		}

		@Override
		public String toString() {
			return String.format(
					"DefaultResult [fit=%.3f, relevance=%.3f, metricClass=%s]",
					this.fitValue, this.relevance,
					this.metricClass.getSimpleName());
		}

	}

	protected static double correctFitValue(double fitValue) {
		return Math.min(Math.max(fitValue, 0), 1);
	}

	public abstract static class DefaultColumnStatistic implements
			ColumnStatistic {

		private Class<? extends ValueFitMetric> metricClass;

		public DefaultColumnStatistic(
				Class<? extends ValueFitMetric> metricClass) {
			super();
			this.metricClass = metricClass;
		}

		@Override
		public Class<? extends ValueFitMetric> getMetricClass() {
			return metricClass;
		}

	}

}
