package de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics;

import java.sql.SQLException;

import org.apache.commons.math3.distribution.NormalDistribution;
import org.apache.logging.log4j.LogManager;

import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.data_access.ColumnIterator;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Datatype;

public class AverageStringLengthMetric extends AbstractSymmetricValueFitMetric {

	public AverageStringLengthMetric() {
		super(Datatype.GeneralType.STRING);
	}

	@Override
	protected Operator createOperator() {
		return new Operator();
	}

	public static class Operator extends SymmetricOperator {

		private long cummulatedLength, numValues, squareLength;

		@Override
		public void update(ColumnIterator columnIterator) throws SQLException {
			String value = columnIterator.getValueAsString();
			int count = columnIterator.getValueCount();

			if (value != null && !value.isEmpty()) {
				cummulatedLength += value.length() * count;
				squareLength += value.length() * value.length() * count;
				numValues += count;
			}

		}

		@Override
		public ColumnStatistic concludeResult() {
			double averageLength;
			double variance;
			double relevance = 0;
			if (this.numValues == 0) {
				averageLength = Double.NaN;
				variance = Double.NaN;
				relevance = 0;
			} else {
				averageLength = cummulatedLength / (double) numValues;
				variance = squareLength / (double) numValues - averageLength * averageLength;
				if (averageLength == 0) {
					relevance = 0;
				} else {
					double stDev = Math.sqrt(variance);
					relevance = Math.sqrt(correctFitValue(1 - (stDev / averageLength)));
					LogManager.getRootLogger().debug("Average: {}, Standard Deviation: {}, Relevance: {}.",
							averageLength, stDev, relevance);
				}
			}
			return new ColumnStatistic(averageLength, variance, relevance);
		}

	}

	public static class ColumnStatistic extends SymmetricColumnStatistic {

		private double averageLength;
		private double variance;

		public ColumnStatistic(double averageLength, double variance, double relevance) {
			super(AverageStringLengthMetric.class, relevance);
			this.averageLength = averageLength;
			this.variance = variance;
		}

		@Override
		public double calculateFitValue(ValueFitMetric.SourceColumnStatistic otherResult) {
			if (!(otherResult instanceof ColumnStatistic)) {
				throw new IllegalArgumentException("Cannot compare average string length result with " + otherResult
						+ ".");
			}
			ColumnStatistic other = (ColumnStatistic) otherResult;
			if (this.variance > 0) {
				NormalDistribution distribution = new NormalDistribution(this.averageLength, Math.sqrt(this.variance));
				return distribution.density(other.averageLength) / distribution.density(this.averageLength);
			} else {
				return other.averageLength == this.averageLength ? 1 : 0;
			}
		}

		@Override
		public String toString() {
			return "Result [averageLength=" + averageLength + ", stDev=" + Math.sqrt(variance) + "]";
		}

	}

}
