package de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics;

import it.unimi.dsi.fastutil.chars.Char2DoubleMap;
import it.unimi.dsi.fastutil.chars.Char2DoubleRBTreeMap;
import it.unimi.dsi.fastutil.chars.Char2IntMap;
import it.unimi.dsi.fastutil.chars.Char2IntOpenHashMap;

import java.sql.SQLException;

import org.apache.logging.log4j.LogManager;

import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.data_access.ColumnIterator;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Datatype;

public class CharacterHistogramMetric extends AbstractSymmetricValueFitMetric {

	public CharacterHistogramMetric() {
		super(Datatype.GeneralType.STRING);
	}

	@Override
	protected Operator createOperator() {
		return new Operator();
	}
	
	public static class Operator extends SymmetricOperator {
		
		private Char2IntMap characterCounts = new Char2IntOpenHashMap();
		private int overallCount = 0;

		@Override
		public void update(ColumnIterator columnIterator) throws SQLException {
			String value = columnIterator.getValueAsString();
			int count = columnIterator.getValueCount();

			if (value != null && !value.isEmpty() && columnIterator.wasValid()) {
				overallCount += value.length() * count;
				for (int i = 0; i < value.length(); i++) {
					char c = value.charAt(i);
					int currentCount = characterCounts.get(c);
					characterCounts.put(c, currentCount + count);
				}
			}

		}

		@Override
		public ColumnStatistic concludeResult() {
			Char2DoubleMap characterFrequencies = new Char2DoubleRBTreeMap();
			double squareSum = 0;
			for (Char2IntMap.Entry entry : characterCounts.char2IntEntrySet()) {
				char c = entry.getCharKey();
				int count = entry.getIntValue();
				
				double frequency = count / (double) overallCount;
				characterFrequencies.put(c, frequency);
				
				squareSum += frequency * frequency;
			}
			// The less characters there are, the more relevant is the character histogram.
//			double relevance = characterFrequencies.isEmpty() ? 0 : Math.exp(1 - characterFrequencies.size());
			double relevance = 0;
			if (!characterFrequencies.isEmpty()) {
				relevance = Math.sqrt(squareSum / characterFrequencies.size());
			}
			// Boost relevance and normalize by a set of 25 characters.
			relevance = correctFitValue(relevance * 25d);
			LogManager.getRootLogger().debug("Characters: {}, relevance: {}.", characterFrequencies.size(), relevance);
			return new ColumnStatistic(characterFrequencies, relevance);
		}

	}

	public static class ColumnStatistic extends SymmetricColumnStatistic {

		private Char2DoubleMap characterFrequencies;

		public ColumnStatistic(Char2DoubleMap characterFrequencies, double relevance) {
			super(CharacterHistogramMetric.class, relevance);
			this.characterFrequencies = characterFrequencies;
		}

		@Override
		public double calculateFitValue(ValueFitMetric.SourceColumnStatistic otherResult) {
			if (!(otherResult instanceof ColumnStatistic)) {
				throw new IllegalArgumentException("Cannot compare character frequencies with " + otherResult + ".");
			}
			ColumnStatistic other = (ColumnStatistic) otherResult;
			if (this.characterFrequencies.isEmpty() || other.characterFrequencies.isEmpty()) {
				return 1;
			}
			
			// Soft intersection.
			double intersect = 0;
			for (Char2DoubleMap.Entry entry : other.characterFrequencies.char2DoubleEntrySet()) {
				char c = entry.getCharKey();
				double otherFrequency = entry.getDoubleValue();
				double thisFrequency = this.characterFrequencies.get(c);
				if (otherFrequency <= Math.sqrt(thisFrequency)) {
					intersect += otherFrequency;
				} else {
					intersect += Math.min(thisFrequency, otherFrequency);
				}
			}
			return correctFitValue(intersect);
		}

		@Override
		public String toString() {
			return "ColumnStatistic [characterFrequencies=" + characterFrequencies + "]";
		}

//		@Override
//		public String toString() {
//			return "Result [#characters=" + characterFrequencies.size() + "]";
//		}

		
		
	}

}
