package de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics;

import it.unimi.dsi.fastutil.chars.CharOpenHashSet;
import it.unimi.dsi.fastutil.chars.CharSet;
import it.unimi.dsi.fastutil.doubles.DoubleIterator;
import it.unimi.dsi.fastutil.ints.Int2ObjectMap;
import it.unimi.dsi.fastutil.objects.Object2DoubleMap;
import it.unimi.dsi.fastutil.objects.Object2DoubleOpenHashMap;
import it.unimi.dsi.fastutil.objects.Object2IntMap;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import uk.ac.shef.wit.simmetrics.similaritymetrics.Levenshtein;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.data_access.ColumnIterator;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Datatype;
import de.uni_potsdam.hpi.isg.utils.util.Counter;

public class TextPatternMetric extends AbstractSymmetricValueFitMetric {

	private static final Logger logger = LogManager.getLogger(TextPatternMetric.class.getSimpleName());

	public TextPatternMetric() {
		super(Datatype.GeneralType.STRING);
	}

	@Override
	protected Operator createOperator() {
		return new Operator();
	}

	public static class Operator extends SymmetricOperator {

		private Counter<TextPattern> patternCounter = new Counter<>();
		private int overallCount = 0;

		@Override
		public void update(ColumnIterator columnIterator) throws SQLException {
			String value = columnIterator.getValueAsString();
			int count = columnIterator.getValueCount();

			logger.trace("Value: {}, count: {}, column iterator: {}", value, count, columnIterator);

			if (value != null && columnIterator.wasValid() && !value.isEmpty()) {
				TextPattern pattern = TextPattern.parse(value);
				pattern.generalize();
				this.patternCounter.count(pattern, count);
				this.overallCount += count;
			}

		}

		@Override
		public ColumnStatistic concludeResult() {
			Int2ObjectMap<List<TextPattern>> reverse = this.patternCounter.reverse();
			logger.trace("Collected patterns:");
			for (Int2ObjectMap.Entry<List<TextPattern>> entry : reverse.int2ObjectEntrySet()) {
				logger.debug("\t{}\t{}", entry.getIntKey(), entry.getValue());
			}

			Object2DoubleMap<TextPattern> patternFrequencies = new Object2DoubleOpenHashMap<>();
			patternFrequencies.defaultReturnValue(0);

			// Relevance: Calculate the constancy of patterns as relevance.
			double relevance = 0;
			for (Object2IntMap.Entry<TextPattern> entry : this.patternCounter.getCounts().object2IntEntrySet()) {
				TextPattern pattern = entry.getKey();
				int count = entry.getIntValue();

				double frequency = count / (double) overallCount;
				patternFrequencies.put(pattern, frequency);

				relevance += count * Math.log(count);
			}
			relevance = this.overallCount > 1 ? relevance / (this.overallCount * Math.log(this.overallCount))
					: this.overallCount;

			logger.debug("Calculated pattern frequencies (relevance={}): {}", relevance, patternFrequencies);
			return new ColumnStatistic(patternFrequencies, relevance);
		}

	}

	public static class ColumnStatistic extends SymmetricColumnStatistic {

		private Object2DoubleMap<TextPattern> patternFrequencies;

		public ColumnStatistic(Object2DoubleMap<TextPattern> patternFrequencies, double relevance) {
			super(TextPatternMetric.class, relevance);
			this.patternFrequencies = patternFrequencies;
		}

		@Override
		protected double calculateFitValue(SourceColumnStatistic otherResult) {
			return calculateOverlap(otherResult);
		}

		@SuppressWarnings("unused")
		private double findMaxFrequency() {
			double maxFrequency = 0d;
			for (DoubleIterator i = this.patternFrequencies.values().iterator(); i.hasNext();) {
				maxFrequency = Math.max(maxFrequency, i.nextDouble());
			}
			return maxFrequency;
		}

		public double calculateOverlap(ValueFitMetric.SourceColumnStatistic otherResult) {
			if (!(otherResult instanceof ColumnStatistic)) {
				throw new IllegalArgumentException("Cannot compare character frequencies with " + otherResult + ".");
			}
			ColumnStatistic other = (ColumnStatistic) otherResult;
			if (this.patternFrequencies.isEmpty() && other.patternFrequencies.isEmpty()) {
				return 1;
			}

			double intersect = 0;
			Levenshtein levenshtein = new Levenshtein();
			for (Object2DoubleMap.Entry<TextPatternMetric.TextPattern> entry : other.patternFrequencies
					.object2DoubleEntrySet()) {
				TextPatternMetric.TextPattern pattern = entry.getKey();
				String patternString = pattern.buildStringRepresentation();
				double otherFrequency = entry.getDoubleValue();
				double intersectShare = 0;
				for (Object2DoubleMap.Entry<TextPattern> thisPatternEntry : this.patternFrequencies.object2DoubleEntrySet()) {
					String thisPatternString = thisPatternEntry.getKey().buildStringRepresentation();
					double intersectShareCandidate = levenshtein.getSimilarity(patternString, thisPatternString);
					double thisFrequency = thisPatternEntry.getDoubleValue();
					if (otherFrequency <= 5 * thisFrequency) {
						intersectShareCandidate *= otherFrequency;
					} else {
						intersectShareCandidate *= Math.min(thisFrequency, otherFrequency);
					}
					intersectShare = Math.max(intersectShare, intersectShareCandidate);
				}
				intersect += intersectShare;
			}
			return correctFitValue(intersect);
		}

		/**
		 * Calculates the best matching pattern from all source patterns.
		 * Therefore quadratic in runtime.
		 * 
		 * @param otherResult
		 * @return
		 */
		public double calculateBestMatch(ValueFitMetric.SourceColumnStatistic otherResult) {

			if (!(otherResult instanceof ColumnStatistic)) {
				throw new IllegalArgumentException("Cannot compare character frequencies with " + otherResult + ".");
			}

			ColumnStatistic source = (ColumnStatistic) otherResult;
			if (this.patternFrequencies.isEmpty() && source.patternFrequencies.isEmpty()) {
				return 1;
			}

			Levenshtein levenshtein = new Levenshtein();

			double bestMatchSum = 0;
			for (Object2DoubleMap.Entry<TextPattern> srcEntry : source.patternFrequencies.object2DoubleEntrySet()) {

				TextPattern srcPattern = srcEntry.getKey();
				String srcRepresentation = srcPattern.buildStringRepresentation();
				double srcFrequency = srcEntry.getDoubleValue();

				double bestMatch = 0;
				for (Object2DoubleMap.Entry<TextPattern> thisEntry : this.patternFrequencies.object2DoubleEntrySet()) {
					double thisFrequency = thisEntry.getDoubleValue();
					double frequencyScore = Math.min(1, 2 * thisFrequency / srcFrequency);

					TextPattern thisPattern = thisEntry.getKey();
					String thisRepresentation = thisPattern.buildStringRepresentation();
					double similarity = levenshtein.getSimilarity(srcRepresentation, thisRepresentation);

					double score = similarity * frequencyScore;
					bestMatch = Math.max(bestMatch, score);
				}

				bestMatchSum += bestMatch * srcFrequency;
			}

			return correctFitValue(bestMatchSum);
		}

		@Override
		public String toString() {
			return "Result [#patterns=" + patternFrequencies.size() + "]";
		}

	}

	/**
	 * A text pattern describes a set of strings by a sequence of {@link Token}
	 * objects.
	 * 
	 * @author Sebastian Kruse
	 */
	public static class TextPattern {

		private static final CharSet separators = new CharOpenHashSet(",;()/[]{}".toCharArray());

		private static final TextPattern EMPTY_PATTERN = new TextPattern(Collections.<Token> emptyList());
		private List<Token> tokens;

		public TextPattern(List<Token> tokens) {
			this.tokens = tokens;
		}

		public void generalize() {
			Iterator<Token> iterator = tokens.iterator();
			Token prevToken = null;
			while (iterator.hasNext()) {
				Token token = iterator.next();
				if (prevToken != null && prevToken.getType().equals(token.getType())) {
					prevToken.setRepeated(true);
					iterator.remove();
				}
				prevToken = token;
			}
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((tokens == null) ? 0 : tokens.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			TextPattern other = (TextPattern) obj;
			if (tokens == null) {
				if (other.tokens != null)
					return false;
			} else if (!tokens.equals(other.tokens))
				return false;
			return true;
		}

		public String buildStringRepresentation() {
			StringBuilder sb = new StringBuilder();
			for (Token token : this.tokens) {
				token.appendStringRepresentation(sb);
			}
			return sb.toString();
		}

		@Override
		public String toString() {
			// return "TextPattern [tokens=" + tokens + "]";
			return buildStringRepresentation();
		}

		public static TextPattern parse(String text) {
			if (text == null)
				return EMPTY_PATTERN;

			List<Token> tokens = new ArrayList<>(3);

			String[] mainTokens = text.split("\\s+");
			for (String mainToken : mainTokens) {

				List<String> subtokens = furtherSplit(mainToken);
				for (String subtoken : subtokens) {

					tokens.add(Token.forString(subtoken));
				}
			}

			return new TextPattern(tokens);
		}

		private static List<String> furtherSplit(String mainToken) {
			List<String> tokens = new ArrayList<>(1);
			int startPos = 0;
			for (int pos = 0; pos < mainToken.length(); pos++) {
				char c = mainToken.charAt(pos);
				boolean isSeparator = separators.contains(c);

				if (isSeparator) {
					if (startPos < pos) {
						tokens.add(mainToken.substring(startPos, pos));
					}
					tokens.add(mainToken.substring(pos, pos + 1));
					startPos = pos + 1;
				}

			}
			if (startPos < mainToken.length()) {
				tokens.add(mainToken.substring(startPos));
			}

			return tokens;
		}

	}

	/**
	 * A token represents a part of a {@link TextPattern}. It describes a class
	 * of substrings in a string, e.g., words, number, and separators.
	 * 
	 * @author Sebastian Kruse
	 */
	public static class Token {

		private static final String UNKNOWN = "unknown";

		private static final String MIXED = "mixed";

		private static final String NUMBER = "number";

		private static final String WORD = "word";

		private String type;

		private boolean isRepeated = false;

		private Token(String type) {
			super();
			this.type = type;
		}

		public Object getType() {
			return this.type;
		}

		public boolean isRepeated() {
			return isRepeated;
		}

		public void setRepeated(boolean isRepeated) {
			this.isRepeated = isRepeated;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + (isRepeated ? 1231 : 1237);
			result = prime * result + ((type == null) ? 0 : type.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			Token other = (Token) obj;
			if (isRepeated != other.isRepeated)
				return false;
			if (type == null) {
				if (other.type != null)
					return false;
			} else if (!type.equals(other.type))
				return false;
			return true;
		}

		public void appendStringRepresentation(StringBuilder sb) {
			if (this.type.equals(WORD)) {
				sb.append("a");
			} else if (this.type.equals(NUMBER)) {
				sb.append("0");
			} else if (this.type.equals(MIXED)) {
				sb.append("x");
			} else if (this.type.equals(UNKNOWN)) {
				sb.append("?");
			} else {
				sb.append(this.type);
			}

			if (this.isRepeated) {
				sb.append("r");
			}
		}

		@Override
		public String toString() {
			return "Token [type=" + type + "]" + (isRepeated ? "*" : "");
		}

		public static Token forString(String token) {
			boolean canBeNumber = true;
			boolean canBeText = true;

			for (int pos = 0; pos < token.length(); pos++) {
				char c = token.charAt(pos);
				boolean isSeparator = TextPattern.separators.contains(c);
				boolean isLetter = Character.isLetter(c);
				boolean isDigit = Character.isDigit(c);
				if (isSeparator) {
					return new Token(String.valueOf(c));
				}

				canBeNumber &= isDigit;
				canBeText &= isLetter;

				if (!canBeNumber && !canBeText) {
					return new Token(MIXED);
				}
			}

			if (canBeNumber)
				return new Token(NUMBER);

			if (canBeText)
				return new Token(WORD);

			logger.error("Did not expect control flow to pass this point.");
			return new Token(UNKNOWN);
		}

	}

}
