package de.uni_potsdam.hpi.isg.efes.csg.datafit.values.metrics;

import it.unimi.dsi.fastutil.objects.Object2DoubleMap;
import it.unimi.dsi.fastutil.objects.Object2DoubleOpenHashMap;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import de.uni_potsdam.hpi.isg.efes.csg.datafit.values.data_access.ColumnIterator;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Datatype.GeneralType;
import de.uni_potsdam.hpi.isg.utils.util.Strings;

public class TopKValuesMetric extends AbstractSymmetricValueFitMetric {

	private static final int DEFAULT_K = 10;

	private int k;

	public TopKValuesMetric() {
		this(DEFAULT_K);
	}

	public TopKValuesMetric(int k) {
		super();
		this.k = k;
	}

	@Override
	public boolean isApplicableFor(GeneralType type) {
		return true;
	}

	@Override
	protected Operator createOperator() {
		return new Operator(k);
	}

	public static class Operator extends SymmetricOperator {

		private int k;

		private List<ValueCount> valueCounts = new ArrayList<>();

		private int overallCount = 0;

		public Operator(int k) {
			super();
			this.k = k;
		}

		@Override
		public void update(ColumnIterator columnIterator) throws SQLException {
			String value = columnIterator.getValueAsString();
			int count = columnIterator.getValueCount();

			if (value != null && !value.isEmpty() && columnIterator.wasValid()) {
				overallCount += count;

				int index;
				for (index = valueCounts.size() - 1; index >= 0; index--) {
					ValueCount oldValueCount = this.valueCounts.get(index);
					if (oldValueCount.getCount() >= count) {
						break;
					}
				}
				index++;
				if (index < this.k) {
					while (this.valueCounts.size() >= this.k) {
						this.valueCounts.remove(this.valueCounts.size() - 1);
					}
					this.valueCounts.add(index, new ValueCount(value, count));
				}
			}
		}

		@Override
		public ColumnStatistic concludeResult() {
			Object2DoubleMap<String> valueFrequencies = new Object2DoubleOpenHashMap<>();
			double overallFrequencies = 0;
			for (ValueCount valueCount : this.valueCounts) {
				double frequency = valueCount.getCount() / (double) this.overallCount;
				valueFrequencies.put(valueCount.getValue(), frequency);
				valueFrequencies.defaultReturnValue(0d);
				overallFrequencies += overallFrequencies;
			}
			return new ColumnStatistic(valueFrequencies, overallFrequencies);
		}

	}

	public static class ColumnStatistic extends SymmetricColumnStatistic {

		private Object2DoubleMap<String> topWordFrequencies;

		public ColumnStatistic(Object2DoubleMap<String> wordFrequencies, double relevance) {
			super(TopKValuesMetric.class, relevance);
			this.topWordFrequencies = wordFrequencies;
		}

		@Override
		public double calculateFitValue(ValueFitMetric.SourceColumnStatistic otherResult) {
			if (!(otherResult instanceof ColumnStatistic)) {
				throw new IllegalArgumentException("Cannot compare character frequencies with " + otherResult + ".");
			}
			ColumnStatistic other = (ColumnStatistic) otherResult;
			if (this.topWordFrequencies.isEmpty() && other.topWordFrequencies.isEmpty()) {
				return 1;
			}

			double intersect = 0;
			for (Object2DoubleMap.Entry<String> entry : this.topWordFrequencies.object2DoubleEntrySet()) {
				String str = entry.getKey();
				double thisFrequency = entry.getDoubleValue();
				double otherFrequency = other.topWordFrequencies.getDouble(str);
				intersect += Math.min(thisFrequency, otherFrequency);
			}
			return intersect;
		}

		@Override
		public String toString() {
			return "Result [topValues=" + Strings.limit(this.topWordFrequencies.toString(), 100) + "]";
		}

	}

	public static class ValueCount {

		private String value;
		private int count;

		public ValueCount(String value, int count) {
			super();
			this.value = value;
			this.count = count;
		}

		public String getValue() {
			return value;
		}

		public int getCount() {
			return count;
		}

		@Override
		public String toString() {
			return "[" + count + "x " + "" + value + "]";
		}

	}

}
