package de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist;

import it.unimi.dsi.fastutil.objects.Object2IntMap;

import javax.el.ExpressionFactory;
import javax.el.ValueExpression;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.odysseus.el.util.SimpleContext;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskList.Task;

public abstract class TaskEstimationFunction {
	
	private static final Logger logger = LogManager.getLogger(TaskEstimationFunction.class.getSimpleName());

	private String estimatedTaskName;
	
	private String key;

	private String unit;

	public static TaskEstimationFunction createLinearFunction(String taskName, String unit, double linearFactor,
			double offset) {
		return new LinearTaskEstimationFunction(taskName, unit, linearFactor, offset);
	}

	public static TaskEstimationFunction createConstantFunction(String taskName, String unit, double estimate) {
		return new LinearTaskEstimationFunction(taskName, unit, 0, estimate);
	}
	
	public static TaskEstimationFunction createJuelExpressionFunction(String taskName, String unit, String juelExpression) {
		return new JuelTaskEstimationFunction(taskName, unit, juelExpression);
	}

	protected TaskEstimationFunction(String estimatedTaskName, String unit) {
		super();
		this.estimatedTaskName = estimatedTaskName;
		this.unit = unit;
	}

	public boolean matches(Task task) {
		if (this.key == null) {
			this.key = simplify(this.estimatedTaskName);
		}
		String simplifiedTaskName = simplify(task.getName());
		return simplifiedTaskName.equals(this.key);
	}

	private String simplify(String taskName) {
		StringBuilder sb = new StringBuilder();
		for (int index = 0; index < taskName.length(); index++) {
			char ch = taskName.charAt(index);
			if (Character.isLetterOrDigit(ch)) {
				sb.append(Character.toLowerCase(ch));
			}
		}
		return sb.toString();
	}

	public double estimate(Task task) {
		return calculateEstimate(task.getParameters());
	}

	public String getUnit() {
		return unit;
	}

	@Override
	public String toString() {
		return "TaskEstimationFunction [estimatedTaskName=" + estimatedTaskName + ", unit=" + unit + "]";
	}

	protected abstract double calculateEstimate(Object2IntMap<String> parameters);

	private static class LinearTaskEstimationFunction extends TaskEstimationFunction {

		private double linearFactor, offset;

		public LinearTaskEstimationFunction(String estimatedTaskName, String unit, double linearFactor, double offset) {
			super(estimatedTaskName, unit);
			this.linearFactor = linearFactor;
			this.offset = offset;
		}

		protected double calculateEstimate(int timesToDo) {
			return linearFactor * timesToDo + offset;
		}
		
		@Override
		protected double calculateEstimate(Object2IntMap<String> parameters) {
			int repetitions = parameters.getInt("repetitions");
			return calculateEstimate(repetitions);
		}

	}

	private static class JuelTaskEstimationFunction extends TaskEstimationFunction {

		private static final ExpressionFactory EXPRESSION_FACTORY = ExpressionFactory.newInstance();
		private SimpleContext context;
		private ValueExpression functionExpression;

		protected JuelTaskEstimationFunction(String estimatedTaskName, String unit, String juelExpression) {
			super(estimatedTaskName, unit);
			context = new SimpleContext();
			functionExpression = EXPRESSION_FACTORY.createValueExpression(context, juelExpression, double.class);
		}

		protected double calculateEstimate(Object2IntMap<String> parameters) {
			for (Object2IntMap.Entry<String> parameter : parameters.object2IntEntrySet()) {
				String key = parameter.getKey();
				Integer value = parameter.getValue();
				String keyExpression = String.format("#{%s}", key);
				logger.trace("Logging '{}' -> '{}'", keyExpression, value);
				EXPRESSION_FACTORY.createValueExpression(this.context, keyExpression, int.class).setValue(this.context, value);
			}
			double result = (double) this.functionExpression.getValue(this.context);
			return result;
		}

	}

}
