package de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist;

import it.unimi.dsi.fastutil.objects.Object2DoubleMap;
import it.unimi.dsi.fastutil.objects.Object2DoubleOpenHashMap;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.datafit.DataFitReport;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.DataFitReport.Section;
import de.uni_potsdam.hpi.isg.efes.csg.effort.EffortEstimate;
import de.uni_potsdam.hpi.isg.efes.csg.effort.EffortModel;
import de.uni_potsdam.hpi.isg.efes.csg.effort.format.EffortFormatter;
import de.uni_potsdam.hpi.isg.efes.csg.effort.format.Formatters;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskList.Task;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.IntegrationScenario;

public class TaskListModel implements EffortModel {

	private static Logger logger = LogManager.getLogger(TaskListModel.class.getSimpleName());

	private IntegrationScenario integrationScenario;
	private Map<Class<? extends Section>, Submodel> submodels;
	private Collection<TaskEstimationFunction> taskEstimationFunctions = new HashSet<>();
	private Profile profile;

	public TaskListModel(Profile profile) {
		this.profile = profile;
		this.submodels = new HashMap<>();
	}

	public void register(Submodel submodel) {
		for (Class<? extends Section> sectionClass : submodel.getSupportedReportSectionClasses()) {
			register(sectionClass, submodel);
		}
	}

	public void register(Class<? extends Section> sectionClass, Submodel submodel) {
		if (!submodel.getSupportedReportSectionClasses().contains(sectionClass)) {
			throw new IllegalArgumentException();
		}

		if (this.submodels.containsKey(sectionClass)) {
			logger.warn("There is already a submodel registered for {}. "
					+ "{} will not be registered for this section class.", sectionClass, submodel);
			return;
		}

		submodel.setIntegrationScenario(this.integrationScenario);
		submodel.setProfile(this.profile);
		this.submodels.put(sectionClass, submodel);
	}

	public void register(TaskEstimationFunction taskEstimationFunction) {
		if (!this.taskEstimationFunctions.add(taskEstimationFunction)) {
			logger.warn("Could not add task estimation function {}.", taskEstimationFunction);
		}
	}

	@Override
	public Estimate evaluate(DataFitReport report, String unit) {
		Estimate result = new Estimate(unit);

		for (Entry<Class<? extends Section>, Section> entry : report.getSections().entrySet()) {
			Class<? extends Section> sectionClass = entry.getKey();
			Section section = entry.getValue();

			Submodel handlingSubmodel = this.submodels.get(sectionClass);
			if (handlingSubmodel == null) {
				logger.warn("Cannot handle report sections of class {}", sectionClass);
				continue;
			}

			logger.debug("Running task list submodel {}.", handlingSubmodel.getClass().getSimpleName());
			TaskList taskList = handlingSubmodel.createTaskList(section, result);
			logger.debug("Tasks for {}:", sectionClass.getSimpleName());
			for (Task task : taskList.getTasks()) {
				logger.debug(task);
				estimate(task, unit);
				result.add(task);
			}
		}

		return result;
	}

	private void estimate(Task task, String unit) {
		for (TaskEstimationFunction function : this.taskEstimationFunctions) {
			if (unit.equals(function.getUnit()) && function.matches(task)) {
				double estimate = function.estimate(task);
				task.setEstimationValue(estimate);
				return;
			}
		}

		logger.error("Could not find an estimation function for {}.", task);
	}

	@Override
	public void setIntegrationScenario(IntegrationScenario integrationScenario) {
		this.integrationScenario = integrationScenario;
		for (Submodel submodel : this.submodels.values()) {
			submodel.setIntegrationScenario(integrationScenario);
		}
	}

	public static interface Submodel {

		Collection<Class<? extends Section>> getSupportedReportSectionClasses();

		void setIntegrationScenario(IntegrationScenario integrationScenario);

		void setProfile(Profile profile);

		TaskList createTaskList(DataFitReport.Section section, Estimate estimate);

	}

	public static class Estimate implements EffortEstimate {

		private double estimatedValue = 0;

		private String estimateUnit;

		private List<Task> tasks = new ArrayList<>();

		private List<String> comments = new ArrayList<>();

		public Estimate(String estimateUnit) {
			super();
			this.estimateUnit = estimateUnit;
		}

		@Override
		public void addComment(String comment) {
			this.comments.add(comment);
		}

		@Override
		public double getEstimatedValue() {
			return estimatedValue;
		}

		@Override
		public String getEstimateUnit() {
			return estimateUnit;
		}

		@Override
		public void print(PrintStream stream) {
			EffortFormatter formatter = Formatters.getFormatter(getEstimateUnit());

			stream.println("Task list effort estimate");
			stream.println("=========================");
			stream.println(" Overall estimation");
			stream.println(" ------------------");
			stream.println(" " + formatter.formatEffort(estimatedValue, estimateUnit));
			stream.println();
			stream.println(" Categories");
			stream.println(" ----------");
			Object2DoubleMap<String> taskGroups = sumEffortsByCategory();
			for (Object2DoubleMap.Entry<String> entry : taskGroups.object2DoubleEntrySet()) {
				String category = entry.getKey();
				double estimate = entry.getDoubleValue();
				stream.println(" " + category);
				stream.println("   " + formatter.formatEffort(estimate, estimateUnit));
			}
			stream.println();
			stream.println(" Tasks");
			stream.println(" -----");
			int number = 1;
			for (Task task : orderTasksByEffort()) {
				stream.format(" %3d. %s (%s)\n", number, task.getName(), task.getNotes());
				stream.format("        Parameters:  %s\n", task.getParameters());
				stream.format("        Subestimate: %s\n", formatter.formatEffort(task, this.estimateUnit));
				number++;
			}
			if (!comments.isEmpty()) {
				stream.println();
				stream.println(" Comments");
				stream.println(" --------");
				for (String comment : comments) {
					stream.format("  * %s\n", comment);
				}
			}
		}
		
		@Override
		public void printCsv(PrintStream stream) {

			// Scan through all tasks and index parameters.
			List<String> parameters = new ArrayList<>();
			for (Task task : this.tasks) {
				for (String parameter : task.getParameters().keySet()) {
					if (!parameters.contains(parameter)) {
						parameters.add(parameter);
					}
				}
			}
			
			Collections.sort(parameters);
			stream.print("name,category,estimate");
			for (String parameter : parameters) {
				stream.format(",%s", parameter);
			}
			stream.println(",notes");
			
			for (Task task : tasks) {
				stream.format(Locale.ENGLISH, "%s,%s,%.3f", task.getName(), task.getCategory(), task.getEstimationValue());
				for (String parameter : parameters) {
					if (task.getParameters().containsKey(parameter)) {
						int value = task.getParameters().getInt(parameter);
						stream.format(",%d", value);
					} else {
						stream.print(",");
					}
				}
				stream.format(",%s\n", task.getNotes());
			}
			
		}

		private List<Task> orderTasksByEffort() {
			List<Task> tasks = new ArrayList<>(this.tasks);
			Collections.sort(tasks, new Comparator<Task>() {
				@Override
				public int compare(Task t1, Task t2) {
					return Double.compare(t2.getEstimationValue(), t1.getEstimationValue());
				}
			});
			return tasks;
		}

		private Object2DoubleMap<String> sumEffortsByCategory() {
			Object2DoubleMap<String> result = new Object2DoubleOpenHashMap<>();
			result.defaultReturnValue(0);
			for (Task task : this.tasks) {
				if (task.getEstimationValue() != Task.NO_ESTIMATE) {
					double subtotal = result.getDouble(task.getCategory());
					subtotal += task.getEstimationValue();
					result.put(task.getCategory(), subtotal);
				}
			}
			return result;
		}

		private void add(Task task) {
			this.tasks.add(task);
			if (!task.hasEstimationValue()) {
				logger.warn("Task {} does not have an estimation value.", task);
			} else {
				this.estimatedValue += task.getEstimationValue();
			}
		}

	}

	public static enum Profile {
		HIGH_QUALITY, LOW_EFFORT
	}

}
