package de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.repair;

import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;
import it.unimi.dsi.fastutil.objects.ObjectIterator;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.StatisticBagOperator;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskList.Task;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskListModel.Profile;
import de.uni_potsdam.hpi.isg.efes.csg.instance.SchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.CompositeRelationship;

public abstract class AbstractSchemaGraphOperationDescriptor<SGO extends AbstractSchemaGraphOperationDescriptor<SGO>.AbstractSchemaGraphOperation>
		implements SchemaGraphOperation.Descriptor<SGO> {

	private static Logger logger = LogManager.getLogger(AbstractSchemaGraphOperationDescriptor.class.getSimpleName());

	@Override
	public boolean isOperationApplicableTo(Constraint violatedConstraint, SchemaGraphInstance sgInstance) {

		NaturalNumberSet actualCardinality = sgInstance.getActualCardinality(violatedConstraint.getRelationship());
		if (!checkCardinalityViolation(violatedConstraint.getCardinality(), actualCardinality)) {
			return false;
		}

		if (!checkGraphPattern(violatedConstraint.getRelationship(), sgInstance)) {
			return false;
		}

		return true;
	}

	protected boolean checkGraphPattern(Relationship relationship, SchemaGraphInstance schemaGraphInstance) {
		List<Relationship> relationships;
		if (relationship instanceof CompositeRelationship) {
			CompositeRelationship compositeRelationship = (CompositeRelationship) relationship;
			relationships = compositeRelationship.getComposedRelationships();
		} else {
			relationships = Collections.singletonList(relationship);
		}

		int patternIndex = 0;
		if (!checkGraphPatternComponent(relationships.get(0).getStartNode(), patternIndex)) {
			return false;
		}

		patternIndex++;
		for (Relationship partialRelationship : relationships) {
			if (!checkGraphPatternComponent(partialRelationship, patternIndex)
					|| !checkGraphPatternComponent(partialRelationship.getEndNode(), patternIndex + 1)) {
				return false;
			}
		}

		return true;
	}

	private boolean checkGraphPatternComponent(Object graphElement, int patternIndex) {
		Class<?>[] graphPattern = getGraphPattern();
		if (patternIndex >= graphPattern.length) {
			return false;
		}

		Class<?> patternElement = graphPattern[patternIndex];
		return patternElement.isInstance(graphElement);
	}

	protected boolean checkCardinalityViolation(NaturalNumberSet prescribedCardinality,
			NaturalNumberSet actualCardinality) {
		NaturalNumberSet targetCardinality = getTargetCardinality();
		return prescribedCardinality.isSubsetOf(targetCardinality) && !actualCardinality.isSubsetOf(targetCardinality);
	}

	abstract protected Class<?>[] getGraphPattern();

	@Override
	public boolean matchesProfile(Profile profile) {
		return getProfiles().contains(profile);
	}

	abstract List<Profile> getProfiles();

	@Override
	public SGO applyTo(Constraint constraint, RepairSchemaGraphInstance sgInstance) {
		SGO operation = createOperation(constraint, sgInstance);
		operation.searchSideEffectRelationships(sgInstance);
		sgInstance.register(operation);
		return operation;
	}

	abstract protected SGO createOperation(Constraint constraint, RepairSchemaGraphInstance sgInstance);

	abstract protected NaturalNumberSet getTargetCardinality();

	public abstract class AbstractSchemaGraphOperation implements SchemaGraphOperation {

		private static final int UNDEFINED = -1;
		protected static final int INVERSE = 0;
		protected static final int START_TABLE_TO_ATTRIBUTE = 1;
		protected static final int END_TABLE_TO_ATTRIBUTE = 2;
		protected static final int START_ATTRIBUTE_TO_ATTRIBUTE = 3;
		protected static final int END_ATTRIBUTE_TO_ATTRIBUTE = 4;
		protected static final int START_ATTRIBUTE_TO_TABLE = 5;
		protected static final int END_ATTRIBUTE_TO_TABLE = 6;
		protected static final int ATTRIBUTE_TO_START_TABLE = 7;
		protected static final int ATTRIBUTE_TO_END_TABLE = 8;
		protected static final int ATTRIBUTE_TO_START_ATTRIBUTE = 9;
		protected static final int ATTRIBUTE_TO_END_ATTRIBUTE = 10;
		protected static final int TABLE_TO_START_ATTRIBUTE = 11;
		protected static final int TABLE_TO_END_ATTRIBUTE = 12;
		protected static final int START_TABLE_TO_TABLE = 12;
		protected static final int TABLE_TO_START_TABLE = 13;
		protected static final int END_TABLE_TO_TABLE = 14;
		protected static final int TABLE_TO_END_TABLE = 15;

		private List<SchemaGraphOperation> predecessors = new ArrayList<>();
		private Object2IntMap<Relationship> sideEffectRelationships = new Object2IntOpenHashMap<>();
		private Relationship targetRelationship;
		private RepairSchemaGraphInstance instance;
		private Map<Relationship, NaturalNumberSet> oldCardinalities = new HashMap<>();

		public AbstractSchemaGraphOperation(Relationship targetRelationship, RepairSchemaGraphInstance instance) {
			super();
			this.targetRelationship = targetRelationship;
			this.instance = instance;
			this.sideEffectRelationships.defaultReturnValue(UNDEFINED);
		}

		abstract protected void searchSideEffectRelationships(SchemaGraphInstance sgi);

		@Override
		public void addPredecessor(SchemaGraphOperation sgo) {
			if (sgo == null) {
				throw new IllegalArgumentException(String.valueOf(sgo));
			}

			if (this.predecessors.contains(sgo)) {
				logger.warn("{} is already a predecessor of {}", sgo, this);
			} else {
				this.predecessors.add(sgo);
			}
		}

		@Override
		public List<SchemaGraphOperation> getDirectPredecessors() {
			return Collections.unmodifiableList(this.predecessors);
		}

		@Override
		public boolean hasPredecessor(SchemaGraphOperation sgo) {
			Set<SchemaGraphOperation> seenOperations = new HashSet<>();
			Queue<SchemaGraphOperation> unseenOperations = new LinkedList<>(this.predecessors);

			while (!unseenOperations.isEmpty()) {
				SchemaGraphOperation currentElement = unseenOperations.poll();
				if (currentElement.equals(sgo)) {
					return true;
				}
				seenOperations.add(currentElement);
				for (SchemaGraphOperation predecessor : currentElement.getDirectPredecessors()) {
					if (!seenOperations.contains(predecessor)) {
						unseenOperations.add(predecessor);
					}
				}
			}

			return false;
		}

		protected void addSideEffectRelationship(Relationship relationship, int type) {
			if (relationship == null)
				throw new IllegalArgumentException();

			if (this.sideEffectRelationships.containsKey(relationship)) {
				int curType = this.sideEffectRelationships.getInt(relationship);
				logger.warn("Tried to register {} as type {}, it is already registered as type {}.", relationship,
						type, curType);
				return;
			}

			this.sideEffectRelationships.put(relationship, type);
		}

		@Override
		public Collection<Relationship> getSideEffectRelationships() {
			return this.sideEffectRelationships.keySet();
		}

		protected Collection<Relationship> getSideEffectRelationships(int type) {
			List<Relationship> relationships = new ArrayList<>();
			ObjectIterator<Object2IntMap.Entry<Relationship>> iterator = this.sideEffectRelationships
					.object2IntEntrySet().iterator();
			while (iterator.hasNext()) {
				Object2IntMap.Entry<Relationship> next = iterator.next();
				if (next.getIntValue() == type) {
					relationships.add(next.getKey());
				}
			}
			return relationships;
		}

		@Override
		public Relationship getTargetRelationship() {
			return targetRelationship;
		}

		@Override
		public void execute() {
			executeOn(this.targetRelationship);
			for (Relationship sideEffectRelationship : this.sideEffectRelationships.keySet()) {
				executeOn(sideEffectRelationship);
			}
		}

		@Override
		public void executeOn(Relationship relationship) {
			if (this.oldCardinalities.containsKey(relationship)) {
				throw new IllegalStateException("This operation was already applied and not undone.");
			}

			NaturalNumberSet oldCardinality = this.instance.getActualCardinality(relationship);
			this.oldCardinalities.put(relationship, oldCardinality);

			NaturalNumberSet newCardinality;
			if (relationship.equals(getTargetRelationship())) {
				newCardinality = calculateNewTargetCardinality(oldCardinality);
			} else {
				int type = this.sideEffectRelationships.getInt(relationship);
				if (type == UNDEFINED) {
					throw new IllegalArgumentException("The given relationship is not subject to this operation: "
							+ String.valueOf(relationship));
				}
				newCardinality = calculateNewSideEffectCardinality(oldCardinality, type);
			}
			this.instance.setActualCardinality(relationship, newCardinality, this);
		}

		protected abstract NaturalNumberSet calculateNewTargetCardinality(NaturalNumberSet oldCardinality);

		protected abstract NaturalNumberSet calculateNewSideEffectCardinality(NaturalNumberSet oldCardinality, int type);

		@Override
		public void undoOn(Relationship relationship) {
			NaturalNumberSet oldCardinality = this.oldCardinalities.remove(relationship);
			if (oldCardinality == null) {
				throw new IllegalStateException("Cannot undo " + relationship + ", no previous cardinality found.");
			}
			this.instance.setActualCardinality(relationship, oldCardinality);
		}

		@Override
		public boolean isTargetConstraintViolated() {
			NaturalNumberSet actualCardinality = this.instance.getActualCardinality(this.targetRelationship);
			return !actualCardinality.isSubsetOf(getTargetCardinality());
		}

		protected RepairSchemaGraphInstance getInstance() {
			return instance;
		}

		@Override
		public Task executeAndReturnTask() {
			// Dummy implementation
			int violations = countViolations();
			executeOnBags();
			return new Task(getTaskName(), RepairModel.CATEGORY, violations,
					getTargetRelationship().toString());
		}
		
		protected String getTaskName() {
			String descriptorClassName = AbstractSchemaGraphOperationDescriptor.this.getClass().getSimpleName();
			StringBuilder sb = new StringBuilder();
			int pos = 0;
			while (pos < descriptorClassName.length()) {
				char c = descriptorClassName.charAt(pos);
				if (pos > 0 && Character.isUpperCase(c)) {
					sb.append(' ');
				}
				sb.append(c);
				pos++;
			}
			return sb.toString();
		}

		protected void executeOnBags() {
			logger.warn("#executeOnBags() not implemented for {}.", AbstractSchemaGraphOperationDescriptor.this
					.getClass().getSimpleName());
		}

		private int countViolations() {
			int violations = 0;
			NaturalNumberBag cardinalityBag = this.instance.getRealCardinality(targetRelationship);
			for (IntIterator i = cardinalityBag.getCountedCardinalities().iterator(); i.hasNext();) {
				int cardinality = i.nextInt();
				if (!getTargetCardinality().hasElement(cardinality)) {
					violations += cardinalityBag.getCountFor(cardinality);
				}
			}
			return violations;
		}

		@Override
		public String toString() {
			return AbstractSchemaGraphOperationDescriptor.this.getClass().getSimpleName() + "("
					+ getTargetRelationship() + ")";
		}

		/**
		 * Utility method to handle the side effects of the deletion of elements
		 * of a node: The deletion might make it necessary to remove links from
		 * other relationships. This method simulates this link removal on a
		 * relationship and its inverse.
		 * 
		 * @param numDeletedElements
		 *            is the number elements that have been deleted from the
		 *            node
		 * @param elementRelationships
		 *            is a collection of affected relationships
		 * @param areRelationshipsInverse
		 *            must be <tt>true</tt> if the given relationships's end nodes are the node
		 *            from which the elements have been removed
		 */
		protected void handleElementDeletionSideEffects(int numDeletedElements,
				Collection<Relationship> elementRelationships, boolean areRelationshipsInverse) {
			for (Relationship elementRelationship : elementRelationships) {
				Relationship inverse = elementRelationship.getInverse();
				if (areRelationshipsInverse) {
					Relationship temp = inverse;
					inverse = elementRelationship;
					elementRelationship = temp;
				}

				NaturalNumberBag oldCardinality = getInstance().getRealCardinality(elementRelationship);
				StatisticBagOperator statisticBagOperator = new StatisticBagOperator(oldCardinality);
				NaturalNumberBag newCardinality = statisticBagOperator.simulateElementRemoval(numDeletedElements);
				getInstance().setRealCardinality(elementRelationship, newCardinality);
				int removedLinks = StatisticBagOperator.determineLinkDiff(oldCardinality, newCardinality);

				statisticBagOperator = new StatisticBagOperator(getInstance().getRealCardinality(inverse));
				newCardinality = statisticBagOperator.simulateLinkRemoval(removedLinks);
				getInstance().setRealCardinality(inverse, newCardinality);
			}
		}

		/**
		 * Utility method to remove elements from a relationship that have no links in the
		 * given relationship.<br>
		 * Please note that this may make the graph inconsistent if deletion side effects are not handled.
		 * 
		 * @param relationship is the relationship from which start nodes shall be removed
		 * @return the number elements that have been removed
		 */
		protected int deleteUnlinkedStartElements(Relationship relationship) {
			NaturalNumberBag oldCardinality = getInstance().getRealCardinality(relationship);
			int numDeletedTuples = oldCardinality.getCountFor(0);
			NaturalNumberBag newCardinality = new NaturalNumberBag(oldCardinality);
			newCardinality.count(0, -numDeletedTuples);
			getInstance().setRealCardinality(relationship, newCardinality);
			return numDeletedTuples;
		}

		protected void addSingleLinkedElements(Relationship relationship, int numElements) {
			NaturalNumberBag oldCardinality;
			NaturalNumberBag newCardinality;
			oldCardinality = getInstance().getRealCardinality(relationship);
			newCardinality = new NaturalNumberBag(oldCardinality);
			newCardinality.count(1, numElements);
			getInstance().setRealCardinality(relationship, newCardinality);
		}

		protected int addEndElementsForUnlinkedStartElements(Relationship relationship) {
			int numDetachedElements = 0;
			NaturalNumberBag oldCardinality = getInstance().getRealCardinality(relationship);
			NaturalNumberBag newCardinality = new NaturalNumberBag();
			for (IntIterator i = oldCardinality.getCountedCardinalities().iterator(); i.hasNext();) {
				int cardinalityValue = i.nextInt();
				int count = oldCardinality.getCountFor(cardinalityValue);
				if (cardinalityValue == 0) {
					numDetachedElements += count;
				} else {
					newCardinality.count(cardinalityValue, count);
				}
			}
			newCardinality.count(1, numDetachedElements);
			getInstance().setRealCardinality(relationship, newCardinality);
			return numDetachedElements;
		}

		protected void handleElementAdditionSideEffects(Relationship relationship, int numAddedElements) {
			NaturalNumberBag oldCardinality;
			NaturalNumberBag newCardinality;
			oldCardinality = getInstance().getRealCardinality(relationship);
			newCardinality = new NaturalNumberBag(oldCardinality);
			newCardinality.count(0, numAddedElements);
			getInstance().setRealCardinality(relationship, newCardinality);

			// The inverse relationship is not affected then.
		}
		
		protected void aggregateEndElementsByStartElement(Relationship relationship) {
			int sumAggregatedElements = 0;
			int numAggregations = 0;
			NaturalNumberBag oldCardinality = getInstance().getRealCardinality(relationship);
			NaturalNumberBag newCardinality = new NaturalNumberBag();
			for (IntIterator i = oldCardinality.getCountedCardinalities().iterator(); i.hasNext();) {
				int cardinalityValue = i.nextInt();
				int count = oldCardinality.getCountFor(cardinalityValue);
				if (cardinalityValue > 1) {
					sumAggregatedElements += count * cardinalityValue;
					numAggregations += count;
				} else {
					newCardinality.count(cardinalityValue, count);
				}
			}
			newCardinality.count(1, numAggregations);
			getInstance().setRealCardinality(relationship, newCardinality);
			
			// Handle inverse.
			Relationship inverse = relationship.getInverse();
			oldCardinality = getInstance().getRealCardinality(inverse);
			StatisticBagOperator statisticBagOperator = new StatisticBagOperator(oldCardinality);
			// An extra adding operation would be nice. However, we just remove
			// less elements.
			// * sumAggregatedElements links deleted.
			// * numAggregation links have been established.
			newCardinality = statisticBagOperator.simulateLinkRemoval(sumAggregatedElements - numAggregations);
			int numNewDetachedValues = newCardinality.getCountFor(0) - oldCardinality.getCountFor(0);
			newCardinality.count(0, -numNewDetachedValues);
			getInstance().setRealCardinality(inverse, newCardinality);
		}
	}

}
