package de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.repair;

import it.unimi.dsi.fastutil.ints.IntIterator;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskListModel.Profile;
import de.uni_potsdam.hpi.isg.efes.csg.instance.SchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.EqualityRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;

public class AddMissingINDValues extends AbstractSchemaGraphOperationDescriptor<AddMissingINDValues.Operation> {

	private static final Logger logger = LogManager.getLogger(AddMissingINDValues.class.getSimpleName());

	private static final NaturalNumberSet TARGET_CARDINALITY = NaturalNumberSet.ONE;
	private static final List<Profile> PROFILES = Arrays.asList(Profile.HIGH_QUALITY);
	private static final Class<?>[] GRAPH_PATTERN = new Class<?>[] { Node.class, EqualityRelationship.class, Node.class };

	@Override
	protected Class<?>[] getGraphPattern() {
		return GRAPH_PATTERN;
	}

	@Override
	List<Profile> getProfiles() {
		return PROFILES;
	}

	@Override
	protected Operation createOperation(Constraint constraint, RepairSchemaGraphInstance sgInstance) {
		return new Operation(constraint.getRelationship(), sgInstance);
	}

	@Override
	protected NaturalNumberSet getTargetCardinality() {
		return TARGET_CARDINALITY;
	}

	public class Operation extends AbstractSchemaGraphOperationDescriptor<Operation>.AbstractSchemaGraphOperation {

		public Operation(Relationship targetRelationship, RepairSchemaGraphInstance instance) {
			super(targetRelationship, instance);
		}

		@Override
		public void searchSideEffectRelationships(SchemaGraphInstance sgi) {
			Collection<Relationship> relationships = sgi.getSchema().getAtomicRelationships(
					getTargetRelationship().getStartNode());
			
			for (Relationship relationship : relationships) {
				if (relationship.getContentType() == Relationship.ContentType.EQUALITY
						&& relationship.getEndNode().getContentType() == Node.ContentType.VALUE) {
					if (!relationship.equals(getTargetRelationship())) {
						addSideEffectRelationship(relationship, START_ATTRIBUTE_TO_ATTRIBUTE);
						addSideEffectRelationship(relationship.getInverse(), ATTRIBUTE_TO_START_ATTRIBUTE);
					}
			
				} else if (relationship.getContentType() == Relationship.ContentType.EXPLICIT
						&& relationship.getEndNode().getContentType() == Node.ContentType.IDENTITY) {
					addSideEffectRelationship(relationship, START_ATTRIBUTE_TO_TABLE);
				}
			}
		}

		@Override
		protected NaturalNumberSet calculateNewTargetCardinality(NaturalNumberSet oldCardinality) {
			return oldCardinality.intersect(NaturalNumberSet.ONE);
		}

		@Override
		protected NaturalNumberSet calculateNewSideEffectCardinality(NaturalNumberSet oldCardinality, int type) {
			switch (type) {
			case START_ATTRIBUTE_TO_ATTRIBUTE:
				return NaturalNumberSet.ZERO_OR_ONE;
			case ATTRIBUTE_TO_START_ATTRIBUTE:
				if (oldCardinality.isEmpty()) {
					return NaturalNumberSet.EMPTY_SET;
				}
				return NaturalNumberSet.ZERO_OR_ONE;
			case START_ATTRIBUTE_TO_TABLE:
				return oldCardinality.union(NaturalNumberSet.ZERO);
			default:
				throw new IllegalArgumentException("Unsupported side effect type: " + type);
			}
		}

		@Override
		protected void executeOnBags() {
			// Process fix.
			int numAddedValues = 0;
			NaturalNumberBag oldCardinality = getInstance().getRealCardinality(getTargetRelationship());
			oldCardinality.getCountFor(0);
			NaturalNumberBag newCardinality = new NaturalNumberBag();
			for (IntIterator i = oldCardinality.getCountedCardinalities().iterator(); i.hasNext();) {
				int cardinalityValue = i.nextInt();
				int count = oldCardinality.getCountFor(cardinalityValue);
				if (cardinalityValue == 0) {
					numAddedValues += count;
				} else {
					newCardinality.count(cardinalityValue, count);
				}
			}
			newCardinality.count(1, numAddedValues);
			getInstance().setRealCardinality(getTargetRelationship(), newCardinality);
			
			// Handle inverse.
			Relationship inverse = getTargetRelationship().getInverse();
			oldCardinality = getInstance().getRealCardinality(inverse);
			newCardinality = new NaturalNumberBag(oldCardinality);
			newCardinality.count(1, numAddedValues);
			getInstance().setRealCardinality(inverse, newCardinality);

			// Handle other foreign keys.
			executeOnOtherForeignKeys(numAddedValues);
			executeOnTableRelationship(numAddedValues);
		}

		private void executeOnOtherForeignKeys(int numAddedValues) {
			Collection<Relationship> relationships = getSideEffectRelationships(START_ATTRIBUTE_TO_ATTRIBUTE);
			for (Relationship relationship : relationships) {
				// Update the cardinality of the equality relationship.
				// Pessimistic: Assume, new values do not correspond to values in the related node.
				NaturalNumberBag oldCardinality = getInstance().getRealCardinality(relationship);
				NaturalNumberBag newCardinality = new NaturalNumberBag(oldCardinality);
				newCardinality.count(0, numAddedValues);
				getInstance().setRealCardinality(relationship, newCardinality);

				// No changes to the inverse due to the pessimistic assumption.
			}
		}

		private void executeOnTableRelationship(int numRemovedValues) {
			Collection<Relationship> relationships = getSideEffectRelationships(START_ATTRIBUTE_TO_TABLE);
			if (relationships.size() != 1) {
				logger.warn("Expected 1 table relationship for {}. found {}.", getTargetRelationship().getStartNode(),
						relationships.size());
			}

			for (Relationship relationship : relationships) {
				// Update attribute->table.
				NaturalNumberBag oldCardinality = getInstance().getRealCardinality(relationship);
				NaturalNumberBag newCardinality = new NaturalNumberBag(oldCardinality);
				newCardinality.count(0,numRemovedValues);
				getInstance().setRealCardinality(relationship, newCardinality);
			}
		}

	}

}
