package de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.repair;

import it.unimi.dsi.fastutil.ints.IntIterator;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskListModel.Profile;
import de.uni_potsdam.hpi.isg.efes.csg.instance.SchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.IdentityNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.ValueNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.EqualityRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.ExplicitAtomicRelationship;

/**
 * When a tuple has an attribute that is not nullable but not always a value is
 * provided, we have a (potential) constraint violation. This violation can be
 * fixed by impute missing values. It is assumed here, that the new values are
 * unique.<br>
 * Operates on the relationship <tt>table node -&gt; attribute node</tt>
 * 
 * @author Sebastian Kruse
 */
public class AddMissingUniqueValues extends AbstractSchemaGraphOperationDescriptor<AddMissingUniqueValues.Operation> {

	private static final Class<?>[] GRAPH_PATTERN = new Class<?>[] { IdentityNode.class,
			ExplicitAtomicRelationship.class, ValueNode.class };
	private static final List<Profile> PROFILES = Arrays.asList(Profile.HIGH_QUALITY);

	@SuppressWarnings("unused")
	private static final Logger logger = LogManager.getLogger(AddMissingUniqueValues.class.getSimpleName());

	@Override
	public String toString() {
		return "Add missing values";
	}

	@Override
	protected Class<?>[] getGraphPattern() {
		return GRAPH_PATTERN;
	}

	@Override
	List<Profile> getProfiles() {
		return PROFILES;
	}

	@Override
	protected Operation createOperation(Constraint constraint, RepairSchemaGraphInstance sgInstance) {
		return new Operation(constraint.getRelationship(), sgInstance);
	}

	@Override
	protected NaturalNumberSet getTargetCardinality() {
		return NaturalNumberSet.GREATER_OR_EQUAL_ONE;
	}

	public class Operation extends AbstractSchemaGraphOperationDescriptor<Operation>.AbstractSchemaGraphOperation {

		public Operation(Relationship targetRelationship, RepairSchemaGraphInstance instance) {
			super(targetRelationship, instance);
		}

		@Override
		protected NaturalNumberSet calculateNewTargetCardinality(NaturalNumberSet oldCardinality) {
			return oldCardinality.intersect(NaturalNumberSet.GREATER_OR_EQUAL_ONE);
		}

		@Override
		protected NaturalNumberSet calculateNewSideEffectCardinality(NaturalNumberSet oldCardinality, int type) {
			switch (type) {
			case INVERSE:
				return oldCardinality.union(NaturalNumberSet.ONE);
			case END_ATTRIBUTE_TO_ATTRIBUTE:
				return NaturalNumberSet.ZERO_OR_ONE;
			case ATTRIBUTE_TO_END_ATTRIBUTE:
				if (oldCardinality.isEmpty()) {
					return NaturalNumberSet.EMPTY_SET;
				}
				return oldCardinality.union(NaturalNumberSet.ONE);
			default:
				throw new IllegalArgumentException("Unsupported side effect type: " + type);
			}
		}

		@Override
		public void executeOnBags() {
			// Process fix.
			int addedElements = 0;
			NaturalNumberBag oldCardinality = getInstance().getRealCardinality(getTargetRelationship());
			NaturalNumberBag newCardinality = new NaturalNumberBag();
			for (IntIterator i = oldCardinality.getCountedCardinalities().iterator(); i.hasNext();) {
				int cardinalityValue = i.nextInt();
				int count = oldCardinality.getCountFor(cardinalityValue);
				if (cardinalityValue == 0) {
					addedElements += count;
				} else {
					newCardinality.count(cardinalityValue, count);
				}
			}
			newCardinality.count(1, addedElements);
			getInstance().setRealCardinality(getTargetRelationship(), newCardinality);

			// Handle inverse.
			Collection<Relationship> relationships = this.getSideEffectRelationships(INVERSE);
			if (relationships.size() != 1) {
				throw new IllegalArgumentException();
			}
			Relationship relationship = relationships.iterator().next();
			oldCardinality = getInstance().getRealCardinality(relationship);
			newCardinality = new NaturalNumberBag(oldCardinality);
			newCardinality.count(1, addedElements);
			getInstance().setRealCardinality(relationship, newCardinality);
			
			// Handle foreign keys.
			relationships = this.getSideEffectRelationships(END_ATTRIBUTE_TO_ATTRIBUTE);
			for (Relationship fkRelationship : relationships) {
				// Pessimistic assumption: All invented values are not in the other attribute. 
				oldCardinality = getInstance().getRealCardinality(fkRelationship);
				newCardinality = new NaturalNumberBag(oldCardinality);
				newCardinality.count(0, addedElements);
				getInstance().setRealCardinality(fkRelationship, newCardinality);
				// The inverse relationship is not affected then.
			}
		}

		@Override
		public void searchSideEffectRelationships(SchemaGraphInstance sgi) {
			addSideEffectRelationship(getTargetRelationship().getInverse(), INVERSE);

			Node attributeNode = getTargetRelationship().getEndNode();
			for (Relationship relationship : sgi.getSchema().getAtomicRelationships(attributeNode)) {
				if (relationship instanceof EqualityRelationship && relationship.getEndNode() instanceof ValueNode) {
					addSideEffectRelationship(relationship, END_ATTRIBUTE_TO_ATTRIBUTE);
					addSideEffectRelationship(relationship.getInverse(), ATTRIBUTE_TO_END_ATTRIBUTE);
				}
			}
		}

	}

}
