package de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.repair;

import it.unimi.dsi.fastutil.doubles.DoubleArrayList;
import it.unimi.dsi.fastutil.doubles.DoubleIterator;
import it.unimi.dsi.fastutil.doubles.DoubleList;
import it.unimi.dsi.fastutil.ints.Int2DoubleAVLTreeMap;
import it.unimi.dsi.fastutil.ints.Int2DoubleMap;
import it.unimi.dsi.fastutil.ints.Int2IntMap;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.fastutil.ints.IntList;
import it.unimi.dsi.fastutil.ints.IntSets;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberInterval;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.StatisticBagOperator;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskListModel.Profile;
import de.uni_potsdam.hpi.isg.efes.csg.instance.SchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.IdentityNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.ExplicitAtomicRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;

public class AggregateTuples extends AbstractSchemaGraphOperationDescriptor<AggregateTuples.Operation> {

	private static final Logger logger = LogManager.getLogger(AggregateTuples.class.getSimpleName());

	private static final NaturalNumberSet TARGET_CARDINALITY = NaturalNumberSet.ZERO_OR_ONE;
	private static final List<Profile> PROFILES = Arrays.asList(Profile.HIGH_QUALITY);
	private static final Class<?>[] GRAPH_PATTERN = new Class<?>[] { Node.class, ExplicitAtomicRelationship.class,
			IdentityNode.class };

	@Override
	protected Class<?>[] getGraphPattern() {
		return GRAPH_PATTERN;
	}

	@Override
	List<Profile> getProfiles() {
		return PROFILES;
	}

	@Override
	protected Operation createOperation(Constraint constraint, RepairSchemaGraphInstance sgInstance) {
		return new Operation(constraint.getRelationship(), sgInstance);
	}

	@Override
	protected NaturalNumberSet getTargetCardinality() {
		return TARGET_CARDINALITY;
	}

	public class Operation extends AbstractSchemaGraphOperationDescriptor<Operation>.AbstractSchemaGraphOperation {

		public Operation(Relationship targetRelationship, RepairSchemaGraphInstance instance) {
			super(targetRelationship, instance);
		}

		@Override
		public void searchSideEffectRelationships(SchemaGraphInstance sgi) {
			Collection<Relationship> relationships = sgi.getSchema().getAtomicRelationships(
					getTargetRelationship().getEndNode());
			for (Relationship relationship : relationships) {
				if (!relationship.isAtomic()) {
					continue;
				}

				if (relationship.isExplicit() && relationship.getEndNode().isSimpleIdentityNode()
						&& !relationship.equals(getTargetRelationship().getInverse())) {
					addSideEffectRelationship(relationship, END_TABLE_TO_TABLE);
					addSideEffectRelationship(relationship.getInverse(), TABLE_TO_END_TABLE);
				} else if (relationship.isExplicit() && relationship.getEndNode().isSimpleValueNode()
						&& !relationship.equals(getTargetRelationship().getInverse())) {
					addSideEffectRelationship(relationship, END_TABLE_TO_ATTRIBUTE);
					addSideEffectRelationship(relationship.getInverse(), ATTRIBUTE_TO_END_TABLE);
				}
			}
		}

		@Override
		protected NaturalNumberSet calculateNewTargetCardinality(NaturalNumberSet oldCardinality) {
			return NaturalNumberInterval.sign(oldCardinality);
		}

		@Override
		protected NaturalNumberSet calculateNewSideEffectCardinality(NaturalNumberSet oldCardinality, int type) {
			int minValue;
			switch (type) {
			case END_TABLE_TO_ATTRIBUTE:
			case END_TABLE_TO_TABLE:
				// The links to tables and attributes are aggregated to a
				// single end tuples, thus cardinalities can add up arbitrarily.
				if (oldCardinality.isSubsetOf(NaturalNumberSet.ZERO)) {
					return oldCardinality;
				}
				minValue = oldCardinality.getMin();
				return new NaturalNumberInterval(minValue);

			case ATTRIBUTE_TO_END_TABLE:
			case TABLE_TO_END_TABLE:
				// Attributes and tables that belonged to n tuples, belong to m
				// <= n tuples after the operation.
				if (oldCardinality.isSubsetOf(NaturalNumberSet.ZERO)) {
					return oldCardinality;
				}
				int maxValue = oldCardinality.getMax();
				// We need to check if 0 was part of the old cardinality.
				minValue = Math.min(oldCardinality.getMin(), 1);
				return new NaturalNumberInterval(minValue, maxValue);

			default:
				throw new IllegalArgumentException("Unsupported side effect type: " + type);
			}
		}

		@Override
		public void executeOnBags() {
			Relationship targetRelationship = getTargetRelationship();

			// Process fix.
			logger.trace("Fixing target relationship {}.", targetRelationship);
			int sumAggregatedElements = 0;
			int numAggregations = 0;
			NaturalNumberBag oldCardinality = getInstance().getRealCardinality(targetRelationship);
			NaturalNumberBag newCardinality = new NaturalNumberBag();
			for (IntIterator i = oldCardinality.getCountedCardinalities().iterator(); i.hasNext();) {
				int cardinalityValue = i.nextInt();
				int count = oldCardinality.getCountFor(cardinalityValue);
				if (cardinalityValue > 1) {
					sumAggregatedElements += count * cardinalityValue;
					numAggregations += count;
				} else {
					newCardinality.count(cardinalityValue, count);
				}
			}
			newCardinality.count(1, numAggregations);
			getInstance().setRealCardinality(targetRelationship, newCardinality);
			NaturalNumberBag oldTargetCardinality = oldCardinality;

			// Handle inverse: We required unique aggregation keys, so still for
			// each aggregated element, there can be only on key.
			Relationship inverse = targetRelationship.getInverse();
			logger.trace("Adjusting target inverse {}.", inverse);
			oldCardinality = getInstance().getRealCardinality(inverse);
			StatisticBagOperator statisticBagOperator = new StatisticBagOperator(oldCardinality);
			newCardinality = statisticBagOperator.simulateLinkRemoval(sumAggregatedElements);
			int numObsoleteTuples = newCardinality.getCountFor(0) - oldCardinality.getCountFor(0);
			newCardinality.count(0, -numObsoleteTuples);
			newCardinality.count(1, numAggregations);
			getInstance().setRealCardinality(inverse, newCardinality);

			int numRemovedTuples = numObsoleteTuples - numAggregations;
			Collection<Relationship> endTableToTableRelationships = getSideEffectRelationships(END_TABLE_TO_TABLE);
			for (Relationship endTableToTableRelationship : endTableToTableRelationships) {
				handleSideEffectOn(endTableToTableRelationship, numObsoleteTuples, oldTargetCardinality);
			}

			Collection<Relationship> endTableToAttributeRelationships = getSideEffectRelationships(END_TABLE_TO_ATTRIBUTE);
			for (Relationship endTableToAttributeRelationship : endTableToAttributeRelationships) {
				handleSideEffectOn(endTableToAttributeRelationship, numRemovedTuples, oldTargetCardinality);
			}

		}

		private void handleSideEffectOn(Relationship relationship, int numObsoleteTuples,
				NaturalNumberBag oldTargetCardinality) {
			// FIXME: This is a very simple implementation.
			logger.trace("Handling side effect on {}.", relationship);

			Relationship inverse = relationship.getInverse();
			NaturalNumberBag inverseCardinality = getInstance().getRealCardinality(inverse);
			NaturalNumberBag oldCardinality = getInstance().getRealCardinality(relationship);

			NaturalNumberBag aggregatedCardinalities = estimateAggregatedElementCardinalities(oldTargetCardinality,
					inverseCardinality, oldCardinality);
			logger.debug("Aggregated cardinalities for {}: {} (aggregator relationship: {}).", oldCardinality, aggregatedCardinalities, oldTargetCardinality);
			StatisticBagOperator sbo = new StatisticBagOperator(oldCardinality);
			NaturalNumberBag newCardinality = sbo.simulateElementRemoval(numObsoleteTuples);
			newCardinality.unionAll(aggregatedCardinalities);

			int linkDiff = StatisticBagOperator.determineLinkDiff(oldCardinality, newCardinality);
			logger.debug("Link diff of {} for {}.", linkDiff, relationship);
			if (linkDiff > 0) {
				int maxRemovableLinks = StatisticBagOperator.countLinks(inverseCardinality)
						- inverseCardinality.getCountFor(1);
				int maxRemovableLinks2 = inverseCardinality.getSum() - inverseCardinality.getCountFor(1);
				maxRemovableLinks = Math.min(maxRemovableLinks, maxRemovableLinks2);
				if (linkDiff > maxRemovableLinks) {
					int numReaddLinks = linkDiff - maxRemovableLinks;
					sbo = new StatisticBagOperator(newCardinality);
					newCardinality = sbo.simulateLinkInsertion(numReaddLinks, oldCardinality);
					linkDiff = maxRemovableLinks;
				}

				sbo = new StatisticBagOperator(inverseCardinality);
				NaturalNumberBag newInverseCardinality = sbo.simulateLinkRemoval(linkDiff, IntSets.singleton(1));
				getInstance().setRealCardinality(inverse, newInverseCardinality);

			} else if (linkDiff < 0) {
				sbo = new StatisticBagOperator(inverseCardinality);
				NaturalNumberBag newInverseCardinality = sbo.simulateLinkInsertion(-linkDiff);
				getInstance().setRealCardinality(inverse, newInverseCardinality);
			}

			getInstance().setRealCardinality(relationship, newCardinality);

		}

		private NaturalNumberBag estimateAggregatedElementCardinalities(NaturalNumberBag oldTargetCardinality,
				NaturalNumberBag inverseCardinality, NaturalNumberBag oldCardinality) {
			IntArrayList aggregationSizes = new IntArrayList(oldTargetCardinality.getCountedCardinalities());
			Collections.sort(aggregationSizes);
			aggregationSizes.rem(0);
			aggregationSizes.rem(1);

			int maxEndElements = inverseCardinality.getCardinality() - inverseCardinality.getCountFor(0);
			int numLinks = inverseCardinality.getSum();
			double duplicateProbability = 0;
			for (Int2IntMap.Entry entry : inverseCardinality.getCountsPerNumber().int2IntEntrySet()) {
				duplicateProbability += entry.getIntValue() * entry.getIntKey() * (entry.getIntKey() - 1) / 2;
			}
			duplicateProbability /= numLinks * (numLinks - 1) / 2;

			Int2DoubleMap baseVector = new Int2DoubleAVLTreeMap();
			for (Int2IntMap.Entry entry : oldCardinality.getCountsPerNumber().int2IntEntrySet()) {
				baseVector.put(entry.getIntKey(), entry.getIntValue());
			}
			Int2DoubleMap currentCombinationVector = new Int2DoubleAVLTreeMap(baseVector);
			int currentAggregationSize = 1;

			NaturalNumberBag aggregatedCardinalities = new NaturalNumberBag();
			for (IntIterator i = aggregationSizes.iterator(); i.hasNext();) {
				int aggregationSize = i.nextInt();
				while (currentAggregationSize < aggregationSize) {
					currentCombinationVector = compose(currentCombinationVector, baseVector);
					currentAggregationSize++;
				}
				int numAggregations = oldTargetCardinality.getCountFor(aggregationSize);

				logger.trace("Need to emit {}x {} (and reduce with {}).", numAggregations, currentCombinationVector,
						inverseCardinality);
				Int2DoubleMap scores = new Int2DoubleAVLTreeMap();
				for (Int2DoubleMap.Entry resultEntry : currentCombinationVector.int2DoubleEntrySet()) {
					int resultCardinality = resultEntry.getIntKey();
					double combinations = resultEntry.getDoubleValue();
					int minElements = (resultCardinality + currentAggregationSize - 1) / currentAggregationSize;
					int maxElements = Math.min(resultCardinality, maxEndElements);
					if (minElements == resultCardinality || duplicateProbability == 0) {
						logger.trace("Releasing {} with score {}.", resultCardinality, combinations);
						double score = scores.get(resultCardinality);
						scores.put(resultCardinality, score + combinations);
					} else {
						Int2DoubleMap partialCardinalityScores = new Int2DoubleAVLTreeMap();
						double sumPartialScores = 0d;
						for (int numElements = minElements; numElements <= maxElements; numElements++) {
							int duplicatePairs = resultCardinality - numElements;
							double duplicatePairRate = 2 * duplicatePairs / (double) (resultCardinality * (resultCardinality + 1));
							double partialScore = Math.pow(1 - Math.abs(duplicatePairRate - duplicateProbability), 2);
							partialCardinalityScores.put(numElements, partialScore);
							sumPartialScores += partialScore;
							logger.trace("Partial score for {} from {}: {}.", numElements, resultCardinality,
									partialScore);
						}
						for (int numElements = minElements; numElements <= maxElements; numElements++) {
							double partialScore = partialCardinalityScores.get(numElements);
							double probability = partialScore / sumPartialScores;
							logger.trace("Releasing {} with score {}.", numElements, combinations * probability);
							double score = scores.get(numElements);
							scores.put(numElements, score + probability * combinations);
						}
					}
				}
				double sumScores = 0;
				for (DoubleIterator scoreIterator = scores.values().iterator(); scoreIterator.hasNext();) {
					sumScores += scoreIterator.nextDouble();
				}

				double remainderSum = 0;
				DoubleList remainders = new DoubleArrayList();
				IntList cardinalityValues = new IntArrayList();
				int numAssignedElements = 0;
				for (Int2DoubleMap.Entry entry : scores.int2DoubleEntrySet()) {
					int cardinalityValue = entry.getIntKey();
					double score = entry.getDoubleValue();

					double newElementsValue = score / sumScores * numAggregations;
					int numLocallyNewElements = (int) Math.floor(newElementsValue);
					aggregatedCardinalities.count(cardinalityValue, numLocallyNewElements);
					numAssignedElements += numLocallyNewElements;

					double remainder = newElementsValue - numLocallyNewElements;
					remainderSum += remainder;
					remainders.add(remainderSum);
					cardinalityValues.add(cardinalityValue);
				}
				// Assign left-over elements by randomly.
				while (numAssignedElements < numAggregations) {
					logger.trace("Assigning from {}", remainders);
					if (remainders.isEmpty()) {
						logger.debug("");
					}
					double randomValue = Math.random() * remainderSum;
					int index = 0;
					double curRemainderSum = 0;
					for (DoubleIterator iterator = remainders.iterator(); iterator.hasNext();) {
						curRemainderSum = iterator.nextDouble();
						if (randomValue <= curRemainderSum) {
							int cardinalityValue = cardinalityValues.getInt(index);
							aggregatedCardinalities.count(cardinalityValue);
							numAssignedElements++;
							break;
						}
						index++;
					}

					remainderSum -= curRemainderSum;
					remainders.remove(index);
					cardinalityValues.remove(index);
				}
				logger.trace("{}x {}-aggregation of {} leads to {} (inverse {}).", numAggregations, aggregationSize,
						oldCardinality, aggregatedCardinalities, inverseCardinality);

			}
			return aggregatedCardinalities;
		}

		private Int2DoubleMap compose(Int2DoubleMap a, Int2DoubleMap b) {
			Int2DoubleMap result = new Int2DoubleAVLTreeMap();
			for (Int2DoubleMap.Entry aEntry : a.int2DoubleEntrySet()) {
				for (Int2DoubleMap.Entry bEntry : b.int2DoubleEntrySet()) {
					int newKey = aEntry.getIntKey() + bEntry.getIntKey();
					double baseValue = result.get(newKey);
					double offset = aEntry.getDoubleValue() * bEntry.getDoubleValue();
					result.put(newKey, baseValue + offset);
				}
			}
			return result;
		}

	}

}
