package de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.repair;

import it.unimi.dsi.fastutil.ints.IntIterator;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberInterval;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.StatisticBagOperator;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskListModel.Profile;
import de.uni_potsdam.hpi.isg.efes.csg.instance.SchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.IdentityNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.ValueNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.EqualityRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.ExplicitAtomicRelationship;

/**
 * deletes pre-aggregation values.
 * 
 * @author Sebastian Kruse
 */
public class AggregateValues extends AbstractSchemaGraphOperationDescriptor<AggregateValues.Operation> {

	@SuppressWarnings("unused")
	private static final Logger logger = LogManager.getLogger(AggregateValues.class.getSimpleName());

	private static final NaturalNumberSet TARGET_CARDINALITY = NaturalNumberSet.ZERO_OR_ONE;
	private static final List<Profile> PROFILES = Arrays.asList(Profile.HIGH_QUALITY);
	private static final Class<?>[] GRAPH_PATTERN = new Class<?>[] { IdentityNode.class,
			ExplicitAtomicRelationship.class, ValueNode.class };

	@Override
	protected Class<?>[] getGraphPattern() {
		return GRAPH_PATTERN;
	}

	@Override
	List<Profile> getProfiles() {
		return PROFILES;
	}

	@Override
	protected Operation createOperation(Constraint constraint, RepairSchemaGraphInstance sgInstance) {
		return new Operation(constraint.getRelationship(), sgInstance);
	}

	@Override
	protected NaturalNumberSet getTargetCardinality() {
		return TARGET_CARDINALITY;
	}

	public class Operation extends AbstractSchemaGraphOperationDescriptor<Operation>.AbstractSchemaGraphOperation {

		public Operation(Relationship targetRelationship, RepairSchemaGraphInstance instance) {
			super(targetRelationship, instance);
		}

		@Override
		public void searchSideEffectRelationships(SchemaGraphInstance sgi) {
			addSideEffectRelationship(getTargetRelationship().getInverse(), INVERSE);

			Collection<Relationship> relationships = sgi.getSchema().getAtomicRelationships(
					getTargetRelationship().getEndNode());
			for (Relationship relationship : relationships) {
				if (relationship instanceof EqualityRelationship && relationship.getEndNode() instanceof ValueNode) {
					addSideEffectRelationship(relationship, END_ATTRIBUTE_TO_ATTRIBUTE);
					addSideEffectRelationship(relationship.getInverse(), ATTRIBUTE_TO_END_ATTRIBUTE);
				}
			}
		}

		@Override
		protected NaturalNumberSet calculateNewTargetCardinality(NaturalNumberSet oldCardinality) {
			return NaturalNumberInterval.sign(oldCardinality);
		}

		@Override
		protected NaturalNumberSet calculateNewSideEffectCardinality(NaturalNumberSet oldCardinality, int type) {
			switch (type) {
			case INVERSE:
				if (oldCardinality.isEmpty()) {
					return NaturalNumberSet.EMPTY_SET;
				}
				return oldCardinality.union(NaturalNumberSet.GREATER_OR_EQUAL_ONE);
			case ATTRIBUTE_TO_END_ATTRIBUTE:
			case END_ATTRIBUTE_TO_ATTRIBUTE:
				if (oldCardinality.isEmpty()) {
					return NaturalNumberSet.EMPTY_SET;
				}
				return NaturalNumberSet.ZERO_OR_ONE;
			default:
				throw new IllegalArgumentException("Unsupported side effect type: " + type);
			}
		}

		@Override
		public void executeOnBags() {
			// Process fix.
			int sumAggregatedElements = 0;
			int numAggregations = 0;
			NaturalNumberBag oldCardinality = getInstance().getRealCardinality(getTargetRelationship());
			NaturalNumberBag newCardinality = new NaturalNumberBag();
			for (IntIterator i = oldCardinality.getCountedCardinalities().iterator(); i.hasNext();) {
				int cardinalityValue = i.nextInt();
				int count = oldCardinality.getCountFor(cardinalityValue);
				if (cardinalityValue > 1) {
					sumAggregatedElements += count * cardinalityValue;
					numAggregations += count;
				} else {
					newCardinality.count(cardinalityValue, count);
				}
			}
			newCardinality.count(1, numAggregations);
			getInstance().setRealCardinality(getTargetRelationship(), newCardinality);

			// Handle inverse.
			Collection<Relationship> relationships = this.getSideEffectRelationships(INVERSE);
			if (relationships.size() != 1) {
				throw new IllegalArgumentException();
			}
			Relationship relationship = relationships.iterator().next();
			oldCardinality = getInstance().getRealCardinality(relationship);
			StatisticBagOperator statisticBagOperator = new StatisticBagOperator(oldCardinality);
			// An extra adding operation would be nice. However, we just remove less elements.
			// * sumAggregatedElements links deleted.
			// * numAggregation links have been established.
			newCardinality = statisticBagOperator.simulateLinkRemoval(sumAggregatedElements - numAggregations);
			int numNewDetachedValues = newCardinality.getCountFor(0) - oldCardinality.getCountFor(0);
			newCardinality.count(0, -numNewDetachedValues);
			getInstance().setRealCardinality(relationship, newCardinality);
			
			// Handle foreign keys.
			relationships = this.getSideEffectRelationships(END_ATTRIBUTE_TO_ATTRIBUTE);
			for (Relationship fkRelationship : relationships) {
				oldCardinality = getInstance().getRealCardinality(fkRelationship);
				statisticBagOperator = new StatisticBagOperator(oldCardinality);
				newCardinality = statisticBagOperator.simulateSourceElementRemovalInEqualityLink(numNewDetachedValues);
				getInstance().setRealCardinality(fkRelationship, newCardinality);
				
				// The inverse relationship is not affected then.
				oldCardinality = getInstance().getRealCardinality(fkRelationship.getInverse());
				newCardinality = statisticBagOperator.simulateTargetElementRemovalInEqualityLink(oldCardinality, numNewDetachedValues);
				getInstance().setRealCardinality(fkRelationship.getInverse(), newCardinality);
			}
		}

	}

}
