package de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.repair;

import it.unimi.dsi.fastutil.ints.IntIterator;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberInterval;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.StatisticBagOperator;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskListModel.Profile;
import de.uni_potsdam.hpi.isg.efes.csg.instance.SchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.ValueNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.EqualityRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;

public class DeleteDanglingINDValues extends AbstractSchemaGraphOperationDescriptor<DeleteDanglingINDValues.Operation> {

	private static final Logger logger = LogManager.getLogger(DeleteDanglingINDValues.class.getSimpleName());

	private static final NaturalNumberSet TARGET_CARDINALITY = NaturalNumberSet.GREATER_OR_EQUAL_ONE;
	private static final List<Profile> PROFILES = Arrays.asList(Profile.LOW_EFFORT);
	private static final Class<?>[] GRAPH_PATTERN = new Class<?>[] { Node.class, EqualityRelationship.class, Node.class };

	@Override
	protected Class<?>[] getGraphPattern() {
		return GRAPH_PATTERN;
	}

	@Override
	List<Profile> getProfiles() {
		return PROFILES;
	}

	@Override
	protected Operation createOperation(Constraint constraint, RepairSchemaGraphInstance sgInstance) {
		return new Operation(constraint.getRelationship(), sgInstance);
	}

	@Override
	protected NaturalNumberSet getTargetCardinality() {
		return TARGET_CARDINALITY;
	}

	public class Operation extends AbstractSchemaGraphOperationDescriptor<Operation>.AbstractSchemaGraphOperation {

		public Operation(Relationship targetRelationship, RepairSchemaGraphInstance instance) {
			super(targetRelationship, instance);
		}

		@Override
		public void searchSideEffectRelationships(SchemaGraphInstance sgi) {
			Collection<Relationship> relationships = sgi.getSchema().getAtomicRelationships(
					getTargetRelationship().getStartNode());
			for (Relationship relationship : relationships) {
				if (relationship instanceof EqualityRelationship && relationship.getEndNode() instanceof ValueNode) {
					if (!relationship.equals(getTargetRelationship())) {
						addSideEffectRelationship(relationship, START_ATTRIBUTE_TO_ATTRIBUTE);
						addSideEffectRelationship(relationship.getInverse(), ATTRIBUTE_TO_START_ATTRIBUTE);
					}
				} else if (relationship.getContentType() == Relationship.ContentType.EXPLICIT
						&& relationship.getEndNode().getContentType() == Node.ContentType.IDENTITY) {
					addSideEffectRelationship(relationship, START_ATTRIBUTE_TO_TABLE);
					addSideEffectRelationship(relationship.getInverse(), TABLE_TO_START_ATTRIBUTE);
				}
			}
		}

		@Override
		protected NaturalNumberSet calculateNewTargetCardinality(NaturalNumberSet oldCardinality) {
			return oldCardinality.intersect(NaturalNumberSet.ONE);
		}

		@Override
		protected NaturalNumberSet calculateNewSideEffectCardinality(NaturalNumberSet oldCardinality, int type) {
			switch (type) {
			case START_ATTRIBUTE_TO_ATTRIBUTE:
				// This relationship is especially affected by the bag
				// semantics.
				return oldCardinality;
			case ATTRIBUTE_TO_START_ATTRIBUTE:
				if (oldCardinality.isEmpty()) {
					return NaturalNumberSet.EMPTY_SET;
				}
				return oldCardinality.union(NaturalNumberSet.ZERO);
			case START_ATTRIBUTE_TO_TABLE:
				// This relationship is especially affected by the bag
				// semantics.
				return oldCardinality;
			case TABLE_TO_START_ATTRIBUTE:
				if (oldCardinality.isEmpty()) {
					return NaturalNumberSet.EMPTY_SET;
				}
				return new NaturalNumberInterval(0, oldCardinality.getMax());
			default:
				throw new IllegalArgumentException("Unsupported side effect type: " + type);
			}
		}

		@Override
		protected void executeOnBags() {
			// Process fix.
			int numRemovedValues = 0;
			NaturalNumberBag oldCardinality = getInstance().getRealCardinality(getTargetRelationship());
			NaturalNumberBag newCardinality = new NaturalNumberBag();
			for (IntIterator i = oldCardinality.getCountedCardinalities().iterator(); i.hasNext();) {
				int cardinalityValue = i.nextInt();
				int count = oldCardinality.getCountFor(cardinalityValue);
				if (cardinalityValue == 0) {
					numRemovedValues += count;
				} else {
					newCardinality.count(cardinalityValue, count);
				}
			}
			getInstance().setRealCardinality(getTargetRelationship(), newCardinality);

			// Handle other foreign keys.
			executeOnOtherForeignKeys(numRemovedValues);
			executeOnTableRelationship(numRemovedValues);
		}

		private void executeOnOtherForeignKeys(int numRemovedValues) {
			Collection<Relationship> relationships = getSideEffectRelationships(START_ATTRIBUTE_TO_ATTRIBUTE);
			for (Relationship relationship : relationships) {
				// Update the cardinality of the equality relationship...
				NaturalNumberBag oldCardinality = getInstance().getRealCardinality(relationship);
				StatisticBagOperator statisticBagOperator = new StatisticBagOperator(oldCardinality);
				NaturalNumberBag newCardinality = statisticBagOperator
						.simulateSourceElementRemovalInEqualityLink(numRemovedValues);
				getInstance().setRealCardinality(relationship, newCardinality);

				// ...and its inverse.
				Relationship inverse = relationship.getInverse();
				oldCardinality = getInstance().getRealCardinality(relationship);
				newCardinality = statisticBagOperator.simulateTargetElementRemovalInEqualityLink(oldCardinality,
						numRemovedValues);
				getInstance().setRealCardinality(inverse, newCardinality);
			}
		}

		private void executeOnTableRelationship(int numRemovedValues) {
			Collection<Relationship> relationships = getSideEffectRelationships(START_ATTRIBUTE_TO_TABLE);
			if (relationships.size() != 1) {
				logger.warn("Expected 1 table relationship for {}. found {}.", getTargetRelationship().getStartNode(),
						relationships.size());
			}

			for (Relationship relationship : relationships) {
				// Update attribute->table.
				NaturalNumberBag oldCardinality = getInstance().getRealCardinality(relationship);
				StatisticBagOperator statisticBagOperator = new StatisticBagOperator(oldCardinality);
				NaturalNumberBag newCardinality = statisticBagOperator.simulateElementRemoval(numRemovedValues);
				getInstance().setRealCardinality(relationship, newCardinality);
				
				// Update table->attribute.
				Relationship inverse = relationship.getInverse();
				int numRemovedLinks = StatisticBagOperator.determineLinkDiff(oldCardinality, newCardinality);
				oldCardinality = getInstance().getRealCardinality(inverse);
				statisticBagOperator = new StatisticBagOperator(oldCardinality);
				newCardinality = statisticBagOperator.simulateLinkRemoval(numRemovedLinks);
				getInstance().setRealCardinality(inverse, newCardinality);
			}
		}

	}

}
