package de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.repair;

import it.unimi.dsi.fastutil.ints.IntIterator;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberInterval;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskListModel.Profile;
import de.uni_potsdam.hpi.isg.efes.csg.instance.SchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.IdentityNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.ExplicitAtomicRelationship;

public class DeleteDanglingTuples extends AbstractSchemaGraphOperationDescriptor<DeleteDanglingTuples.Operation> {

	@SuppressWarnings("unused")
	private static final Logger logger = LogManager.getLogger(DeleteDanglingTuples.class.getSimpleName());

	private static final NaturalNumberSet TARGET_CARDINALITY = NaturalNumberSet.GREATER_OR_EQUAL_ONE;
	private static final List<Profile> PROFILES = Arrays.asList(Profile.LOW_EFFORT);
	private static final Class<?>[] GRAPH_PATTERN = new Class<?>[] { IdentityNode.class,
			ExplicitAtomicRelationship.class, IdentityNode.class };

	@Override
	protected Class<?>[] getGraphPattern() {
		return GRAPH_PATTERN;
	}

	@Override
	List<Profile> getProfiles() {
		return PROFILES;
	}

	@Override
	protected Operation createOperation(Constraint constraint, RepairSchemaGraphInstance sgInstance) {
		return new Operation(constraint.getRelationship(), sgInstance);
	}

	@Override
	protected NaturalNumberSet getTargetCardinality() {
		return TARGET_CARDINALITY;
	}

	public class Operation extends AbstractSchemaGraphOperationDescriptor<Operation>.AbstractSchemaGraphOperation {

		public Operation(Relationship targetRelationship, RepairSchemaGraphInstance instance) {
			super(targetRelationship, instance);
		}

		@Override
		public void searchSideEffectRelationships(SchemaGraphInstance sgi) {
			Collection<Relationship> relationships = sgi.getSchema().getAtomicRelationships(
					getTargetRelationship().getStartNode());
			for (Relationship relationship : relationships) {
				if (!relationship.isAtomic())
					continue;

				if (relationship.isExplicit() && relationship.getEndNode().isSimpleValueNode()) {
					addSideEffectRelationship(relationship, START_TABLE_TO_ATTRIBUTE);
					addSideEffectRelationship(relationship.getInverse(), ATTRIBUTE_TO_START_TABLE);

				} else if (relationship.isExplicit() && relationship.getEndNode().isSimpleIdentityNode()
						&& !relationship.equals(getTargetRelationship())) {
					addSideEffectRelationship(relationship, START_TABLE_TO_TABLE);
					addSideEffectRelationship(relationship.getInverse(), TABLE_TO_START_TABLE);

				}
			}
		}

		@Override
		protected NaturalNumberSet calculateNewTargetCardinality(NaturalNumberSet oldCardinality) {
			return oldCardinality.intersect(NaturalNumberSet.GREATER_OR_EQUAL_ONE);
		}

		@Override
		protected NaturalNumberSet calculateNewSideEffectCardinality(NaturalNumberSet oldCardinality, int type) {
			switch (type) {
			case START_TABLE_TO_ATTRIBUTE:
				// This relationship is especially affected by the bag
				// semantics.
				return oldCardinality;
			case ATTRIBUTE_TO_START_TABLE:
				if (oldCardinality.isEmpty()) {
					return NaturalNumberSet.EMPTY_SET;
				}
				return new NaturalNumberInterval(0, oldCardinality.getMax());
			case START_TABLE_TO_TABLE:
				// This relationship is especially affected by the bag
				// semantics.
				return oldCardinality;
			case TABLE_TO_START_TABLE:
				if (oldCardinality.isEmpty()) {
					return NaturalNumberSet.EMPTY_SET;
				}
				return new NaturalNumberInterval(0, oldCardinality.getMax());
			default:
				throw new IllegalArgumentException("Unsupported side effect type: " + type);
			}
		}

		@Override
		protected void executeOnBags() {
			// Process fix: All tuples with cardinality zero are removed.
			int numRemovedValues = 0;
			NaturalNumberBag oldCardinality = getInstance().getRealCardinality(getTargetRelationship());
			NaturalNumberBag newCardinality = new NaturalNumberBag();
			for (IntIterator i = oldCardinality.getCountedCardinalities().iterator(); i.hasNext();) {
				int cardinalityValue = i.nextInt();
				int count = oldCardinality.getCountFor(cardinalityValue);
				if (cardinalityValue == 0) {
					numRemovedValues += count;
				} else {
					newCardinality.count(cardinalityValue, count);
				}
			}
			getInstance().setRealCardinality(getTargetRelationship(), newCardinality);
			
			// Inverse is not influences, since no links have been removed.

			// Handle other foreign keys.
			executeOnOtherRelatedTables(numRemovedValues);
			executeOnAttributeRelationships(numRemovedValues);
		}

		private void executeOnOtherRelatedTables(int numRemovedTuples) {
			Collection<Relationship> relationships = getSideEffectRelationships(START_TABLE_TO_TABLE);
			handleElementDeletionSideEffects(numRemovedTuples, relationships, false);
		}

		private void executeOnAttributeRelationships(int numRemovedTuples) {
			Collection<Relationship> relationships = getSideEffectRelationships(START_TABLE_TO_ATTRIBUTE);
			handleElementDeletionSideEffects(numRemovedTuples, relationships, false);
		}

	}

}
