package de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.repair;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.StatisticBagOperator;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskListModel.Profile;
import de.uni_potsdam.hpi.isg.efes.csg.instance.SchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.IdentityNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.ValueNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.EqualityRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.ExplicitAtomicRelationship;

public class DeleteDetachedValues extends AbstractSchemaGraphOperationDescriptor<DeleteDetachedValues.Operation> {
	
	@SuppressWarnings("unused")
	private static final Logger logger = LogManager.getLogger(DeleteDetachedValues.class.getSimpleName());

	private static final Class<?>[] GRAPH_PATTERN = new Class<?>[] { ValueNode.class, ExplicitAtomicRelationship.class,
			IdentityNode.class };
	private static final List<Profile> PROFILES = Arrays.asList(Profile.LOW_EFFORT);
	private static final NaturalNumberSet TARGET_CARDINALITY = NaturalNumberSet.GREATER_OR_EQUAL_ONE;

	@Override
	protected Class<?>[] getGraphPattern() {
		return GRAPH_PATTERN;
	}

	@Override
	List<Profile> getProfiles() {
		return PROFILES;
	}

	@Override
	protected Operation createOperation(Constraint constraint, RepairSchemaGraphInstance sgInstance) {
		return new Operation(constraint.getRelationship(), sgInstance);
	}

	@Override
	protected NaturalNumberSet getTargetCardinality() {
		return TARGET_CARDINALITY;
	}

	public class Operation extends AbstractSchemaGraphOperationDescriptor<Operation>.AbstractSchemaGraphOperation {

		public Operation(Relationship targetRelationship, RepairSchemaGraphInstance instance) {
			super(targetRelationship, instance);
		}

		@Override
		public void searchSideEffectRelationships(SchemaGraphInstance sgi) {
			Node attributeNode = getTargetRelationship().getStartNode();

			// Type 0: Foreign key relationships of deleted values might be impaired.
			for (Relationship relationship : sgi.getSchema().getAtomicRelationships(attributeNode)) {
				if (!relationship.equals(getTargetRelationship()) && relationship instanceof EqualityRelationship
						&& relationship.getEndNode() instanceof ValueNode) {
					addSideEffectRelationship(relationship.getInverse(), ATTRIBUTE_TO_START_ATTRIBUTE);
				}
			}
		}

		@Override
		protected NaturalNumberSet calculateNewTargetCardinality(NaturalNumberSet oldCardinality) {
			return oldCardinality.intersect(NaturalNumberSet.GREATER_OR_EQUAL_ONE);
		}

		
		@Override
		protected NaturalNumberSet calculateNewSideEffectCardinality(NaturalNumberSet oldCardinality, int type) {
			switch (type) {
			case ATTRIBUTE_TO_START_ATTRIBUTE:
				if (oldCardinality.isEmpty()) {
					return NaturalNumberSet.EMPTY_SET;
				}
				return oldCardinality.union(NaturalNumberSet.ZERO);
			default:
				throw new IllegalArgumentException("Unsupported side effect type: " + type);
			}
		}
		
		@Override
		protected void executeOnBags() {
			// Handle target relationship
			NaturalNumberBag oldCardinality = getInstance().getRealCardinality(getTargetRelationship());
			int numDetachedValues = oldCardinality.getCountFor(0);
			NaturalNumberBag newCardinality = new NaturalNumberBag(oldCardinality);
			newCardinality.count(0, -numDetachedValues);
			getInstance().setRealCardinality(getTargetRelationship(), newCardinality);
			
			// Handle deletion of values towards an equality-connected value node.
			// Handle foreign keys.
			Collection<Relationship> relationships = this.getSideEffectRelationships(START_ATTRIBUTE_TO_ATTRIBUTE);
			for (Relationship fkRelationship : relationships) {
				oldCardinality = getInstance().getRealCardinality(fkRelationship);
				StatisticBagOperator statisticBagOperator = new StatisticBagOperator(oldCardinality);
				newCardinality = statisticBagOperator.simulateSourceElementRemovalInEqualityLink(numDetachedValues);
				getInstance().setRealCardinality(fkRelationship, newCardinality);
				
				// The inverse relationship is not affected then.
				Relationship fkInverse = fkRelationship.getInverse();
				oldCardinality = getInstance().getRealCardinality(fkInverse);
				newCardinality = statisticBagOperator.simulateTargetElementRemovalInEqualityLink(oldCardinality, numDetachedValues);
				getInstance().setRealCardinality(fkInverse, newCardinality);
			}
		}

	}

}
