package de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.repair;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberInterval;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskListModel.Profile;
import de.uni_potsdam.hpi.isg.efes.csg.instance.SchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.IdentityNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.ValueNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.ExplicitAtomicRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;

/**
 * When a tuple has an attribute that is not nullable but not always, a value is
 * provided, we have a (potential) constraint violation. This violation can be
 * fixed by deleting the tuples, that violate the constraints.<br>
 * Operates on the relationship <tt>table node -&gt; attribute node</tt>
 * 
 * @author Sebastian Kruse
 */
public class DeleteTuplesWithMissingValue extends
		AbstractSchemaGraphOperationDescriptor<DeleteTuplesWithMissingValue.Operation> {

	private static final Class<?>[] GRAPH_PATTERN = new Class<?>[] { IdentityNode.class,
			ExplicitAtomicRelationship.class, ValueNode.class };
	private static final List<Profile> PROFILES = Arrays.asList(Profile.LOW_EFFORT);

	@SuppressWarnings("unused")
	private static final Logger logger = LogManager.getLogger(DeleteTuplesWithMissingValue.class.getSimpleName());

	@Override
	public String toString() {
		return "DeleteTupleWithViolatingValuesOperation";
	}

	@Override
	protected Class<?>[] getGraphPattern() {
		return GRAPH_PATTERN;
	}

	@Override
	List<Profile> getProfiles() {
		return PROFILES;
	}

	@Override
	protected Operation createOperation(Constraint constraint, RepairSchemaGraphInstance sgInstance) {
		return new Operation(constraint.getRelationship(), sgInstance);
	}

	@Override
	protected NaturalNumberSet getTargetCardinality() {
		return NaturalNumberSet.GREATER_OR_EQUAL_ONE;
	}

	public class Operation extends AbstractSchemaGraphOperationDescriptor<Operation>.AbstractSchemaGraphOperation {

		public Operation(Relationship targetRelationship, RepairSchemaGraphInstance instance) {
			super(targetRelationship, instance);
		}

		@Override
		protected NaturalNumberSet calculateNewTargetCardinality(NaturalNumberSet oldCardinality) {
			return oldCardinality.intersect(getTargetCardinality());
		}

		@Override
		protected NaturalNumberSet calculateNewSideEffectCardinality(NaturalNumberSet oldCardinality, int type) {
			switch (type) {
			/*
			 * case INVERSE: if (oldCardinality.isEmpty()) { return
			 * NaturalNumberSet.EMPTY_SET; } return new NaturalNumberInterval(0,
			 * oldCardinality.getMax());
			 */
			case ATTRIBUTE_TO_START_TABLE:
			case TABLE_TO_START_TABLE:
				if (oldCardinality.isEmpty()) {
					return NaturalNumberSet.EMPTY_SET;
				}
				return new NaturalNumberInterval(0, oldCardinality.getMax());
			default:
				throw new IllegalArgumentException("Unsupported side effect type: " + type);
			}
		}

		@Override
		public void searchSideEffectRelationships(SchemaGraphInstance sgi) {
			// addSideEffectRelationship(getTargetRelationship().getInverse(),
			// INVERSE);

			Node tableNode = getTargetRelationship().getStartNode();
			for (Relationship relationship : sgi.getSchema().getAtomicRelationships(tableNode)) {
				if (!relationship.isAtomic() || relationship.equals(getTargetRelationship()))
					continue;

				if (relationship.isExplicit() && relationship.getEndNode().isSimpleValueNode()) {
					addSideEffectRelationship(relationship.getInverse(), ATTRIBUTE_TO_START_TABLE);
				} else if (relationship.isExplicit() && relationship.getEndNode().isSimpleIdentityNode()) {
					addSideEffectRelationship(relationship.getInverse(), TABLE_TO_START_TABLE);
				}
			}
		}

		@Override
		protected void executeOnBags() {
			// Process fix.
			int numDeletedTuples = processFix();
			executeOnOtherAttributes(numDeletedTuples);
			executeOnOtherTables(numDeletedTuples);
		}

		private int processFix() {
			Relationship relationship = getTargetRelationship();
			int numDeletedTuples = deleteUnlinkedStartElements(relationship);
			return numDeletedTuples;
		}

		protected void executeOnOtherAttributes(int deletedTuples) {
			Collection<Relationship> allAttributeRelationships = getSideEffectRelationships(ATTRIBUTE_TO_START_TABLE);
			handleElementDeletionSideEffects(deletedTuples, allAttributeRelationships, true);			
		}
		
		protected void executeOnOtherTables(int deletedTuples) {
			Collection<Relationship> tableRelationships = getSideEffectRelationships(TABLE_TO_START_TABLE);
			handleElementDeletionSideEffects(deletedTuples, tableRelationships, true);			
		}
	}

}
