package de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.repair;

import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.DataFitReport.Section;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.structure.StructureFitInspector;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.structure.StructureFitInspector.ReportSection;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskList;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskList.Task;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskListModel;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskListModel.Estimate;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskListModel.Profile;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.repair.FallbackFix.Operation;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.IntegrationScenario;
import de.uni_potsdam.hpi.isg.utils.util.CollectionUtil;
import de.uni_potsdam.hpi.isg.utils.util.MoreMath;
import de.uni_potsdam.hpi.isg.utils.util.Permutations;

public class RepairModel implements TaskListModel.Submodel {

	private static Logger logger = LogManager.getLogger(RepairModel.class.getSimpleName());
	private static final boolean RANDOM_VIOLATION_SELECTION = false;

	private static final Collection<SchemaGraphOperation.Descriptor<?>> repairOperations = new ArrayList<>();

	private boolean checkConsistency = false;

	public static final String CATEGORY = "Structure Repairing";
	static {
		// Register operations here.
		repairOperations.add(new DeleteTuplesWithMissingValue());
		repairOperations.add(new DeleteDetachedValues());
		repairOperations.add(new KeepAnyValue());
		repairOperations.add(new DeleteDanglingINDValues());
		repairOperations.add(new DeleteDanglingTuples());
		repairOperations.add(new UnlinkAllButOneTuple());

		repairOperations.add(new AggregateValues());
		repairOperations.add(new AddMissingUniqueValues());
		repairOperations.add(new CreateTuplesForDetachedValues());
		repairOperations.add(new AddMissingINDValues());
		repairOperations.add(new AddMissingTuples());
		repairOperations.add(new AggregateTuples());
	}

	@SuppressWarnings("unused")
	private IntegrationScenario integrationScenario;
	private Profile profile;
	private RepairSchemaGraphInstance instance;

	public RepairModel() {
		super();
	}

	@Override
	public void setIntegrationScenario(IntegrationScenario integrationScenario) {
		this.integrationScenario = integrationScenario;
	}

	public void setProfile(Profile profile) {
		this.profile = profile;
	}

	/**
	 * From the registered repair tasks, this function chooses an appropriate
	 * one for each violation and applies it.
	 * 
	 * @param section
	 */
	public TaskList createTaskList(Section section, Estimate estimate) {
		StructureFitInspector.ReportSection structureFitSection = (ReportSection) section;

		this.instance = new RepairSchemaGraphInstance(structureFitSection.getSchemaGraphInstance());

		Object2IntMap<Node> nodeSizes = new Object2IntOpenHashMap<>();
		for (Node node : this.instance.getSchema().getAtomicNodes().values()) {
			if (node.getOperationType() == Node.Operation.NONE) {
				int size = this.instance.getRealSize(node);
				nodeSizes.put(node, size);
			}
		}

		if (checkConsistency) {
			this.instance.checkBagConsistency();
		}

		logger.info("Searching fixing operations.");
		Collection<Constraint> violatedConstraints;
		while (!(violatedConstraints = this.instance.getViolatedConstraints()).isEmpty()) {
			Constraint violatedConstraint = fetchViolatedConstraint(violatedConstraints);
			SchemaGraphOperation fixingOperation = findFixingOperation(violatedConstraint);
			integrateFix(violatedConstraint, fixingOperation);
		}

		TaskList taskList = buildTaskList();

		int sumOldSizes = 0;
		int sumLosses = 0;
		for (Node node : this.instance.getSchema().getAtomicNodes().values()) {
			if (node.getOperationType() == Node.Operation.NONE && node.getContentType() == Node.ContentType.VALUE) {
				int oldSize = nodeSizes.getInt(node);
				int newSize = this.instance.getRealSize(node);
				int loss = Math.max(0, oldSize - newSize);
				if (oldSize != newSize || newSize != 0) {
					logger.debug("Sources provide {} values for {}, final integration result will have {}.", oldSize,
							node, newSize);
				}
				sumOldSizes += oldSize;
				sumLosses += loss;
				if (loss > 0) {
					estimate.addComment(String.format(
							"Loss: Sources provide %d values for %s, final integration result will have %d.", oldSize,
							node, newSize));
				}
			}
		}
		estimate.addComment(String.format("Loss rate: %.2f%%", (100d * sumLosses / (double) sumOldSizes)));

		return taskList;
	}

	private Constraint fetchViolatedConstraint(
			Collection<Constraint> violatedConstraints) {
		Constraint violatedConstraint;
		if (RANDOM_VIOLATION_SELECTION) {
			List<Constraint> violationList = new ArrayList<>(violatedConstraints);
			Random random = new Random();
			int index = random.nextInt(violationList.size());
			violatedConstraint = violationList.get(index);
		} else {
			violatedConstraint = CollectionUtil.getAny(violatedConstraints);
		}
		return violatedConstraint;
	}

	private TaskList buildTaskList() {

		// XXX: This method cannot deal with repair cycles, although detecting
		// them.
		// This could be done by searching cycles and break them at an arbitrary
		// point.

		Collection<SchemaGraphOperation> operations = this.instance.getAllOperations();
		List<SchemaGraphOperation> repairOrder = new ArrayList<>();
		if (RANDOM_VIOLATION_SELECTION) {
			Random random = new Random();
			List<SchemaGraphOperation> operationList = new ArrayList<>(operations);
			while (!operationList.isEmpty()) {
				int index = random.nextInt(operationList.size());
				SchemaGraphOperation currentOperation = operationList.get(index);
				if (repairOrder.containsAll(currentOperation.getDirectPredecessors())) {
					repairOrder.add(currentOperation);
					operationList.remove(index);
				}
			}

		} else {
			while (!operations.isEmpty()) {
				for (Iterator<SchemaGraphOperation> i = operations.iterator(); i.hasNext();) {
					SchemaGraphOperation currentOperation = i.next();
					if (repairOrder.containsAll(currentOperation.getDirectPredecessors())) {
						repairOrder.add(currentOperation);
						i.remove();
					}
				}
			}
		}

		logger.info("Simulating repair tasks.");
		TaskList taskList = new TaskList();
		for (SchemaGraphOperation operation : repairOrder) {
			logger.debug("Executing {}.", operation);
			Task task = operation.executeAndReturnTask();
			if (task.getRepetitions() != 0) {
				taskList.add(task);
			}

			if (checkConsistency) {
				this.instance.checkBagConsistency();
			}
		}

		return taskList;
	}

	/**
	 * A fixing operation can conflict or interact with other fixing operations.
	 * This method has to take responsibility of integrating the operation with
	 * other operations and execute it properly.
	 * 
	 * @param violatedConstraint
	 * @param fixingOperation
	 */
	private void integrateFix(Constraint violatedConstraint, SchemaGraphOperation fixingOperation) {
		Relationship relationship = violatedConstraint.getRelationship();
		NaturalNumberSet actualCardinality = this.instance.getActualCardinality(relationship);
		logger.info("Using {} to fix {} (is {}).", fixingOperation, violatedConstraint, actualCardinality);
		fixingOperation.execute();

		// Check if any operations have side effects on the observed fields.
		Collection<SchemaGraphOperation> fixingOperations = this.instance.getFixingOperations(relationship);
		Collection<SchemaGraphOperation> sideEffectOperations = this.instance.getSideEffectOperations(relationship);
		evaluateOrder(fixingOperations, sideEffectOperations, relationship);
		// NB: We assume that multiple fixes for one constraint are deflating,
		// i.e., they do not cause new violations on the fixed constraint.

		// Check if operation has side effects on already fixed fields.
		for (Relationship sideEffectRelationship : fixingOperation.getSideEffectRelationships()) {
			Collection<SchemaGraphOperation> sideEffectFixingOperations = this.instance
					.getFixingOperations(sideEffectRelationship);
			evaluateOrder(sideEffectFixingOperations, Collections.singleton(fixingOperation), sideEffectRelationship);
		}
		logger.debug("Now it is {}.", this.instance.getActualCardinality(relationship));

	}

	/**
	 * It is necessary to determine if the side effect operations cause the
	 * constraint violations fixed by the fixing operation. If so, they must be
	 * executed before the fixing operation. Otherwise, their execution order is
	 * irrelevant concerning the fixed constraint.
	 * 
	 * @param fixingOperations
	 * @param sideEffectOperations
	 * @param conflictRelationship
	 */
	private void evaluateOrder(Collection<SchemaGraphOperation> fixingOperations,
			Collection<SchemaGraphOperation> sideEffectOperations, Relationship conflictRelationship) {

		if (fixingOperations.isEmpty() || sideEffectOperations.isEmpty()) {
			return;
		}
		logger.trace("Need to evaluate order between {} and {} on {}", fixingOperations, sideEffectOperations,
				conflictRelationship);

		// If we have more than one fixing operation, we can run this method for
		// each fixing operation separately and regard the other fixing
		// operations as side effect operations.
		if (fixingOperations.size() > 1) {
			for (SchemaGraphOperation fixingOperation : fixingOperations) {
				Collection<SchemaGraphOperation> newSideEffectOperations = new ArrayList<>(sideEffectOperations);
				newSideEffectOperations.addAll(fixingOperations);
				newSideEffectOperations.remove(fixingOperation);
				Collection<SchemaGraphOperation> newFixingOperations = Collections.singletonList(fixingOperation);
				evaluateOrder(newFixingOperations, newSideEffectOperations, conflictRelationship);
			}
			return;
		}
		
		
		SchemaGraphOperation fixingOperation = CollectionUtil.getAny(fixingOperations);
		if (!fixingOperation.getTargetRelationship().equals(conflictRelationship)) {
			throw new IllegalArgumentException(fixingOperation.getTargetRelationship() + " does not fix " + conflictRelationship + ".");
		}
		List<List<SchemaGraphOperation>> nonViolatingTrailingOperations = new ArrayList<>();

		// Build every possible execution order of operations on the conflict
		// relationship.
		List<SchemaGraphOperation> allOperations = new ArrayList<>();
		allOperations.addAll(fixingOperations);
		allOperations.addAll(sideEffectOperations);
		List<List<SchemaGraphOperation>> executionOrders = Permutations.generateAllPermutations(allOperations);

		// Execute in each order.
		for (List<SchemaGraphOperation> executionOrder : executionOrders) {
			this.instance.reset(conflictRelationship);
			for (SchemaGraphOperation operation : executionOrder) {
				operation.executeOn(conflictRelationship);
			}

			// Check if we have a non-violating execution order.
			if (!fixingOperation.isTargetConstraintViolated()) {
				// If so, store, which operations followed the fixing operation.
				List<SchemaGraphOperation> trailingOperations = extractTrailingOperations(executionOrder,
						fixingOperation);
				nonViolatingTrailingOperations.add(trailingOperations);
			}
		}

		// Now, for all non-violating execution orders, check that the order of
		// the trailing operations is irrelevant and that operations may be left
		// out. According execution orders must be present as well.
		List<List<SchemaGraphOperation>> independentOperationCandidates = new ArrayList<>(
				nonViolatingTrailingOperations);
		Collections.sort(independentOperationCandidates, new Comparator<List<?>>() {

			@Override
			public int compare(List<?> l1, List<?> l2) {
				return Integer.compare(l1.size(), l2.size());
			}
		});

		// In principle, we might yield multiple sets of independent operations
		// that are overlapping, yet, not included in each other. However, we
		// can
		// only use one such a set. So, check the candidates in descending size
		// and take the first matching.
		Collection<List<SchemaGraphOperation>> checkedCandidates = new ArrayList<>();
		List<SchemaGraphOperation> independentOperations = new ArrayList<>();
		CheckCandidates: while (!independentOperationCandidates.isEmpty()) {
			List<SchemaGraphOperation> candidate = independentOperationCandidates.remove(independentOperationCandidates
					.size() - 1);

			// Avoid redundant checks.
			for (List<SchemaGraphOperation> checkedCandidate : checkedCandidates) {
				if (checkedCandidate.size() == candidate.size() && checkedCandidate.containsAll(candidate)) {
					continue CheckCandidates;
				}
			}
			checkedCandidates.add(candidate);

			// Calculate the number of subset candidates there must be if we
			// have a valid candidate.
			int expectedSubsets = 0;
			int trailingOperations = candidate.size();
			int leadingElements = allOperations.size() - trailingOperations - 1;
			for (int i = 0; i <= trailingOperations; i++) {
				expectedSubsets += MoreMath.binomialCoefficient(trailingOperations, i)
						* MoreMath.faculty(leadingElements + i) * MoreMath.faculty(trailingOperations - i);
			}

			// Count the number of candidates that are subsets (incl. the
			// current candidate).
			int numSubsets = 1;
			for (List<SchemaGraphOperation> remainingCandidate : independentOperationCandidates) {
				if (candidate.containsAll(remainingCandidate)) {
					numSubsets++;
				}
			}

			logger.trace("Test result for {}: expected {}, found {}", candidate, expectedSubsets, numSubsets);
			if (numSubsets == expectedSubsets) {
				independentOperations.addAll(candidate);
				break;
			}
		}

		if (!independentOperations.isEmpty()) {
			logger.debug("Independent of {}: {}.", fixingOperation, independentOperations);
		}
		List<SchemaGraphOperation> precedingOperations = new ArrayList<>(sideEffectOperations);
		precedingOperations.removeAll(independentOperations);
		if (!precedingOperations.isEmpty()) {
			logger.debug("Found predecessors of {}: {} (on {}).", fixingOperation, precedingOperations, conflictRelationship);
		}

		for (SchemaGraphOperation predecessor : precedingOperations) {
			if (predecessor.hasPredecessor(fixingOperation)) {
				throw new RuntimeException("There seems to be a repair cycle in between " + fixingOperation + " and "
						+ predecessor + " (detected on " + conflictRelationship + ")");
			}
			fixingOperation.addPredecessor(predecessor);
		}

		// Finally, execute all operations in a correct order.
		this.instance.reset(conflictRelationship);
		for (SchemaGraphOperation sideEffectOperation : sideEffectOperations) {
			sideEffectOperation.executeOn(conflictRelationship);
		}
		for (SchemaGraphOperation fixingOperation2 : fixingOperations) {
			fixingOperation2.executeOn(conflictRelationship);
		}
	}

	private List<SchemaGraphOperation> extractTrailingOperations(List<SchemaGraphOperation> operations,
			SchemaGraphOperation separatorOperation) {
		int separatorIndex = operations.indexOf(separatorOperation);
		return operations.subList(separatorIndex + 1, operations.size());
	}

	/**
	 * Searches a suitable fixing operation for the violated constraint and if
	 * it finds one, this operation will be associated to the violation.
	 * 
	 * @param violatedConstraint
	 * @return
	 */
	private SchemaGraphOperation findFixingOperation(Constraint violatedConstraint) {
		for (SchemaGraphOperation.Descriptor<?> opDescriptor : repairOperations) {
			if (opDescriptor.matchesProfile(this.profile)
					&& opDescriptor.isOperationApplicableTo(violatedConstraint, this.instance)) {
				SchemaGraphOperation operation = opDescriptor.applyTo(violatedConstraint, this.instance);
				return operation;
			}
		}

		// Use fallback, if no fix is appropriate.
		logger.error(
				"Could not find a fixing operation for {} in the profile {}. The constraint violation will be removed.",
				violatedConstraint, profile);
		FallbackFix fallbackFix = new FallbackFix(violatedConstraint.getCardinality());
		Operation fallbackFixOperation = fallbackFix.applyTo(violatedConstraint, instance);
		return fallbackFixOperation;
	}

	@Override
	public Collection<Class<? extends Section>> getSupportedReportSectionClasses() {
		List<Class<? extends Section>> supportedReportSectionClasses = new ArrayList<>();
		supportedReportSectionClasses.add(StructureFitInspector.ReportSection.class);
		return supportedReportSectionClasses;
	}

	public void setCheckConsistency(boolean checkConsistency) {
		this.checkConsistency = checkConsistency;
	}

}
