package de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.repair;

import it.unimi.dsi.fastutil.Stack;
import it.unimi.dsi.fastutil.objects.ObjectArrayList;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.instance.DataStoreAdapter;
import de.uni_potsdam.hpi.isg.efes.csg.instance.SchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.instance.VirtualSchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;

public class RepairSchemaGraphInstance implements SchemaGraphInstance {

	private static Logger logger = LogManager.getLogger(RepairSchemaGraphInstance.class.getSimpleName());

	private VirtualSchemaGraphInstance vsgInstance;

	private Map<Relationship, Stack<SchemaGraphOperation>> commandStacks;
	private Map<Relationship, Collection<SchemaGraphOperation>> fixingOperations, sideEffectOperations;

	public RepairSchemaGraphInstance(VirtualSchemaGraphInstance vsgInstance) {
		super();
		this.vsgInstance = vsgInstance;
		this.fixingOperations = new HashMap<>();
		this.sideEffectOperations = new HashMap<>();
		this.commandStacks = new HashMap<>();
	}

	public void register(SchemaGraphOperation operation) {
		add(this.fixingOperations, operation.getTargetRelationship(), operation);
		for (Relationship sideEffectRelationship : operation.getSideEffectRelationships()) {
			add(this.sideEffectOperations, sideEffectRelationship, operation);
		}
	}

	private void add(Map<Relationship, Collection<SchemaGraphOperation>> operationStore,
			Relationship relationship, SchemaGraphOperation operation) {
		Collection<SchemaGraphOperation> collection = operationStore.get(relationship);
		if (collection == null) {
			collection = new ArrayList<>();
			operationStore.put(relationship, collection);
		} else if (collection.contains(operation)) {
			logger.warn("Avoided to insert {} a second time.");
			return;
		}
		collection.add(operation);
	}

	public Collection<Relationship> getViolatedRelationships() {
		return this.vsgInstance.getViolatedRelationships();
	}

	@Override
	public Collection<Constraint> getViolatedConstraints() {
		return this.vsgInstance.getViolatedConstraints();
	}

	@Override
	public void setActualCardinality(Relationship relationship, NaturalNumberSet cardinality) {
		logger.debug("Set cardinality of {} to {} (without operation).", relationship, cardinality);
		this.vsgInstance.setActualCardinality(relationship, cardinality);
	}

	/**
	 * In contrast to
	 * {@link #setActualCardinality(Relationship, NaturalNumberSet)}, this
	 * method associates the cardinality change to the given operation, so that
	 * it can be undone.
	 * 
	 * @param relationship
	 *            the relationship whose cardinality is to be changed
	 * @param cardinality
	 *            the new cardinality of the relationship
	 * @param operation
	 *            the operation that makes this change request
	 */
	public void setActualCardinality(Relationship relationship, NaturalNumberSet cardinality,
			SchemaGraphOperation operation) {
		logger.debug("Set cardinality of {} to {} (by {}).", relationship, cardinality, operation);
		Stack<SchemaGraphOperation> stack = getCommandStackFor(relationship);
		stack.push(operation);
		this.vsgInstance.setActualCardinality(relationship, cardinality);
	}

	private Stack<SchemaGraphOperation> getCommandStackFor(Relationship relationship) {
		Stack<SchemaGraphOperation> stack = this.commandStacks.get(relationship);
		if (stack == null) {
			stack = new ObjectArrayList<>();
			this.commandStacks.put(relationship, stack);
		}
		return stack;
	}

	@Override
	public NaturalNumberSet getActualCardinality(Relationship relationship) {
		return this.vsgInstance.getActualCardinality(relationship);
	}

	public NaturalNumberSet getPrescribedCardinality(Relationship violatedRelationship) {
		return this.vsgInstance.getPrescribedCardinality(violatedRelationship);
	}

	@Override
	public int getRealSize(Node node) {
		return this.vsgInstance.getRealSize(node);
	}

	public Collection<SchemaGraphOperation> getFixingOperations(Relationship relationship) {
		Collection<SchemaGraphOperation> collection = this.fixingOperations.get(relationship);
		if (collection == null) {
			collection = Collections.emptyList();
		}
		return collection;
	}

	public Collection<SchemaGraphOperation> getSideEffectOperations(Relationship relationship) {
		Collection<SchemaGraphOperation> collection = this.sideEffectOperations.get(relationship);
		if (collection == null) {
			collection = Collections.emptyList();
		}
		return collection;
	}

	public NaturalNumberBag getRealCardinality(Relationship relationship) {
		return this.vsgInstance.getRealCardinality(relationship);
	}

	public void setRealCardinality(Relationship relationship, NaturalNumberBag cardinality) {
		NaturalNumberBag oldCardinality = this.vsgInstance.getRealCardinality(relationship);
		logger.debug("Set cardinality of {} from {} to {}.", relationship, oldCardinality, cardinality);
		this.vsgInstance.setRealCardinality(relationship, cardinality);
	}

	public void checkBagConsistency() {
		BasicSchemaGraph schema = getSchema();
		boolean errorDetected = false;
		for (Node node : schema.getAtomicNodes().values()) {
			if (node.getOperationType() != Node.Operation.NONE) {
				logger.warn("Non-atomic node {}.", node);
			}
			logger.trace("Checking node {}.", node);
			Collection<Relationship> relationships = schema.getAtomicRelationships(node);
			int nodeSize = -1;
			Relationship firstRelationship = null;
			for (Relationship relationship : relationships) {
				logger.trace(" Checking relationship {}.", relationship);
				NaturalNumberBag realCardinality = getRealCardinality(relationship);
				int cardinality = realCardinality.getCardinality();
				if (firstRelationship == null) {
					firstRelationship = relationship;
					nodeSize = cardinality;
				} else if (nodeSize != cardinality) {
					logger.error("Inconsistent node sizes of {}: {} for {} and {} for {}", node, nodeSize,
							firstRelationship, cardinality, relationship);
					errorDetected = true;
				}
			}
		}

		Set<Relationship> handledRelationships = new HashSet<>();
		for (Relationship relationship : schema.getAllRelationships()) {
			Relationship inverse = relationship.getInverse();
			if (!(handledRelationships.add(relationship) | handledRelationships.add(inverse)))
				continue;
			NaturalNumberBag cardinality = getRealCardinality(relationship);
			int sum = cardinality.getSum();
			NaturalNumberBag inverseCardinality = getRealCardinality(inverse);
			int inverseSum = inverseCardinality.getSum();
			
			if (sum != inverseSum) {
				logger.error("{} and {} disagree in their sums: {} != {}.", relationship, inverse, sum, inverseSum);
				errorDetected = true;
			}
		}
		
		if (errorDetected) {
			throw new IllegalStateException("Schema graph inconsistency detected.");
		} else {
			logger.debug("No inconsistencies detected.");
		}
	}

	@Override
	public BasicSchemaGraph getSchema() {
		return this.vsgInstance.getSchema();
	}

	public void reset(Relationship relationship) {
		logger.trace("Resetting {} (currently {}).", relationship, getActualCardinality(relationship));
		Stack<SchemaGraphOperation> stack = getCommandStackFor(relationship);
		while (!stack.isEmpty()) {
			SchemaGraphOperation operation = stack.pop();
			operation.undoOn(relationship);
			logger.trace("Undoing {} changed cardinality to {}.", operation, getActualCardinality(relationship));
		}
	}

	public Collection<SchemaGraphOperation> getAllOperations() {
		List<SchemaGraphOperation> allOperations = new ArrayList<>();
		for (Collection<SchemaGraphOperation> operations : this.fixingOperations.values()) {
			allOperations.addAll(operations);
		}
		return allOperations;
	}

	@Override
	public boolean hasDataStoreAdapter() {
		return this.vsgInstance.hasDataStoreAdapter();
	}
	
	@Override
	public DataStoreAdapter getDataStoreAdapter() {
		return this.vsgInstance.getDataStoreAdapter();
	}

	@Override
	public NaturalNumberSet getCardinality(Relationship relationship) {
		return getRealCardinality(relationship).toInterval();
	}

	@Override
	public boolean areOverlapping(Node node1, Node node2) {
		// TODO: Implement if needed.
		throw new UnsupportedOperationException();
	}

}
