package de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.repair;

import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.fastutil.ints.IntSets;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberInterval;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.StatisticBagOperator;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskListModel.Profile;
import de.uni_potsdam.hpi.isg.efes.csg.instance.SchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.IdentityNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.ExplicitAtomicRelationship;

/**
 * deletes pre-aggregation values.
 * 
 * @author Sebastian Kruse
 */
public class StrictlyAggregateTuples extends AbstractSchemaGraphOperationDescriptor<StrictlyAggregateTuples.Operation> {

	private static final Logger logger = LogManager.getLogger(StrictlyAggregateTuples.class.getSimpleName());

	private static final NaturalNumberSet TARGET_CARDINALITY = NaturalNumberSet.ZERO_OR_ONE;
	private static final List<Profile> PROFILES = Arrays.asList(Profile.HIGH_QUALITY);
	private static final Class<?>[] GRAPH_PATTERN = new Class<?>[] { IdentityNode.class,
			ExplicitAtomicRelationship.class, IdentityNode.class };

	@Override
	protected Class<?>[] getGraphPattern() {
		return GRAPH_PATTERN;
	}

	@Override
	List<Profile> getProfiles() {
		return PROFILES;
	}

	@Override
	public boolean isOperationApplicableTo(Constraint violatedConstraint, SchemaGraphInstance sgInstance) {
		if (!super.isOperationApplicableTo(violatedConstraint, sgInstance)) {
			return false;
		}

		// Additionally, for an aggregation we must ensure that the tuples to be
		// aggregated only have a single corresponding source tuple otherwise,
		// the aggregation is not clearly defined.
		Relationship inverse = violatedConstraint.getRelationship().getInverse();
		NaturalNumberSet inverseCardinality = sgInstance.getActualCardinality(inverse);
		return inverseCardinality.isSubsetOf(NaturalNumberSet.ZERO_OR_ONE);
	}

	@Override
	protected Operation createOperation(Constraint constraint, RepairSchemaGraphInstance sgInstance) {
		return new Operation(constraint.getRelationship(), sgInstance);
	}

	@Override
	protected NaturalNumberSet getTargetCardinality() {
		return TARGET_CARDINALITY;
	}

	public class Operation extends AbstractSchemaGraphOperationDescriptor<Operation>.AbstractSchemaGraphOperation {

		public Operation(Relationship targetRelationship, RepairSchemaGraphInstance instance) {
			super(targetRelationship, instance);
		}

		@Override
		public void searchSideEffectRelationships(SchemaGraphInstance sgi) {
			addSideEffectRelationship(getTargetRelationship().getInverse(), INVERSE);

			Collection<Relationship> relationships = sgi.getSchema().getAtomicRelationships(
					getTargetRelationship().getEndNode());
			for (Relationship relationship : relationships) {
				if (!relationship.isAtomic()) {
					continue;
				}

				if (relationship.isExplicit() && relationship.getEndNode().isSimpleIdentityNode()
						&& !relationship.equals(getTargetRelationship().getInverse())) {
					addSideEffectRelationship(relationship, END_TABLE_TO_TABLE);
					addSideEffectRelationship(relationship.getInverse(), TABLE_TO_END_TABLE);
				} else if (relationship.isExplicit() && relationship.getEndNode().isSimpleValueNode()) {
					addSideEffectRelationship(relationship, END_TABLE_TO_ATTRIBUTE);
					addSideEffectRelationship(relationship.getInverse(), ATTRIBUTE_TO_END_TABLE);
				}
			}
		}

		@Override
		protected NaturalNumberSet calculateNewTargetCardinality(NaturalNumberSet oldCardinality) {
			return NaturalNumberInterval.sign(oldCardinality);
		}

		@Override
		protected NaturalNumberSet calculateNewSideEffectCardinality(NaturalNumberSet oldCardinality, int type) {
			int minValue;
			switch (type) {
			case INVERSE:
				// Just check this condition.
				if (!oldCardinality.isSubsetOf(NaturalNumberSet.ZERO_OR_ONE)) {
					throw new RuntimeException("Inverse cardinality is " + oldCardinality
							+ " but must be subset of 0..1.");
				}

			case END_TABLE_TO_ATTRIBUTE:
			case END_TABLE_TO_TABLE:
				// The links to tables and attributes are aggregated to a
				// single end tuples, thus cardinalities can add up arbitrarily.
				if (oldCardinality.isSubsetOf(NaturalNumberSet.ZERO)) {
					return oldCardinality;
				}
				minValue = oldCardinality.getMin();
				return new NaturalNumberInterval(minValue);

			case ATTRIBUTE_TO_END_TABLE:
			case TABLE_TO_END_TABLE:
				// Attributes and tables that belonged to n tuples, belong to m
				// <= n tuples after the operation.
				if (oldCardinality.isSubsetOf(NaturalNumberSet.ZERO)) {
					return oldCardinality;
				}
				int maxValue = oldCardinality.getMax();
				// We need to check if 0 was part of the old cardinality.
				minValue = Math.min(oldCardinality.getMin(), 1);
				return new NaturalNumberInterval(minValue, maxValue);

			default:
				throw new IllegalArgumentException("Unsupported side effect type: " + type);
			}
		}

		@Override
		public void executeOnBags() {
			Relationship targetRelationship = getTargetRelationship();
//			checkForUniqueAggregationKeys(targetRelationship);

			// Process fix.
			int sumAggregatedElements = 0;
			int numAggregations = 0;
			NaturalNumberBag oldCardinality = getInstance().getRealCardinality(targetRelationship);
			NaturalNumberBag newCardinality = new NaturalNumberBag();
			for (IntIterator i = oldCardinality.getCountedCardinalities().iterator(); i.hasNext();) {
				int cardinalityValue = i.nextInt();
				int count = oldCardinality.getCountFor(cardinalityValue);
				if (cardinalityValue > 1) {
					sumAggregatedElements += count * cardinalityValue;
					numAggregations += count;
				} else {
					newCardinality.count(cardinalityValue, count);
				}
			}
			newCardinality.count(1, numAggregations);
			getInstance().setRealCardinality(targetRelationship, newCardinality);

			// Handle inverse: We required unique aggregation keys, so still for
			// each aggregated element, there can be only on key.
			Relationship inverse = targetRelationship.getInverse();
			oldCardinality = getInstance().getRealCardinality(inverse);
			StatisticBagOperator statisticBagOperator = new StatisticBagOperator(oldCardinality);
			newCardinality = statisticBagOperator.simulateLinkRemoval(sumAggregatedElements);
			int numObsoleteTuples = newCardinality.getCountFor(0) - oldCardinality.getCountFor(0);
			newCardinality.count(0, -numObsoleteTuples);
			newCardinality.count(1, numAggregations);
			getInstance().setRealCardinality(inverse, newCardinality);

			int numRemovedTuples = numObsoleteTuples - numAggregations;
			logger.debug("Need to remove {} tuples.", numRemovedTuples);
			Collection<Relationship> endTableToTableRelationships = getSideEffectRelationships(END_TABLE_TO_ATTRIBUTE);
			for (Relationship endTableToTableRelationship : endTableToTableRelationships) {
				handleSideEffectOn(endTableToTableRelationship, numRemovedTuples, oldCardinality);
			}
			
		}

		private void handleSideEffectOn(Relationship relationship, int numRemovedTuples, NaturalNumberBag oldTargetCardinality) {
			// XXX: This is a very simple implementation.
			if (numRemovedTuples == 0) {
				logger.debug("Handling no side effect because no tuples have been removed.");
				return;
			}
			
			Relationship inverse = relationship.getInverse();
			NaturalNumberBag inverseCardinality = getInstance().getRealCardinality(inverse);
			NaturalNumberBag oldCardinality = getInstance().getRealCardinality(relationship);
			
			if (numRemovedTuples > 0) {
				// Tuples have been removed.
				StatisticBagOperator sbo = new StatisticBagOperator(oldCardinality);
				NaturalNumberBag newCardinality = sbo.simulateElementRemoval(numRemovedTuples);
				int numRemovedLinks = StatisticBagOperator.determineLinkDiff(oldCardinality, newCardinality);
				int maxRemovableLinks = StatisticBagOperator.countLinks(inverseCardinality) - inverseCardinality.getCountFor(1);
				if (numRemovedLinks > maxRemovableLinks) {
					int numReaddLinks = numRemovedLinks - maxRemovableLinks;
					sbo = new StatisticBagOperator(newCardinality);
					newCardinality = sbo.simulateLinkInsertion(numReaddLinks, oldCardinality);
					numRemovedLinks = maxRemovableLinks;
					
				}
				getInstance().setRealCardinality(relationship, newCardinality);
				
				sbo = new StatisticBagOperator(inverseCardinality);
				newCardinality = sbo.simulateLinkRemoval(numRemovedLinks, IntSets.singleton(1));
				getInstance().setRealCardinality(inverse, newCardinality);

			} else {
				// Tuples have been added.
				int numAddedTuples = -numRemovedTuples;
				StatisticBagOperator sbo = new StatisticBagOperator(oldCardinality);
				NaturalNumberBag newCardinality = sbo.simulateElementInsertion(numAddedTuples);
				getInstance().setRealCardinality(relationship, newCardinality);
				
				int numAddedLinks = StatisticBagOperator.determineLinkDiff(newCardinality, oldCardinality);
				sbo = new StatisticBagOperator(inverseCardinality);
				newCardinality = sbo.simulateLinkInsertion(numAddedLinks);
				getInstance().setRealCardinality(inverse, newCardinality);;
			}
			
			
		}

		@SuppressWarnings("unused")
		private void checkForUniqueAggregationKeys(Relationship targetRelationship) {
			// Sanity check: is the aggregation currently possible?
			Relationship inverse = targetRelationship.getInverse();
			NaturalNumberBag cardinality = getInstance().getRealCardinality(inverse);
			IntIterator iterator = cardinality.getCountedCardinalities().iterator();
			while (iterator.hasNext()) {
				int cardinalityValue = iterator.nextInt();
				if (cardinalityValue > 1) {
					throw new IllegalStateException("Cannot aggregate " + targetRelationship
							+ ", inverse cardinality is " + cardinality + ".");
				}
			}
		}

	}

}
