package de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.repair;

import it.unimi.dsi.fastutil.ints.IntIterator;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberInterval;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.StatisticBagOperator;
import de.uni_potsdam.hpi.isg.efes.csg.effort.tasklist.TaskListModel.Profile;
import de.uni_potsdam.hpi.isg.efes.csg.instance.SchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.IdentityNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.ExplicitAtomicRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;

public class UnlinkAllButOneTuple extends AbstractSchemaGraphOperationDescriptor<UnlinkAllButOneTuple.Operation> {
	
	@SuppressWarnings("unused")
	private static final Logger logger = LogManager.getLogger(UnlinkAllButOneTuple.class.getSimpleName());

	private static final NaturalNumberSet TARGET_CARDINALITY = NaturalNumberSet.ZERO_OR_ONE;
	private static final List<Profile> PROFILES = Arrays.asList(Profile.LOW_EFFORT);
	private static final Class<?>[] GRAPH_PATTERN = new Class<?>[] { Node.class,
			ExplicitAtomicRelationship.class, IdentityNode.class };

	@Override
	protected Class<?>[] getGraphPattern() {
		return GRAPH_PATTERN;
	}

	@Override
	List<Profile> getProfiles() {
		return PROFILES;
	}

	@Override
	protected Operation createOperation(Constraint constraint, RepairSchemaGraphInstance sgInstance) {
		return new Operation(constraint.getRelationship(), sgInstance);
	}

	@Override
	protected NaturalNumberSet getTargetCardinality() {
		return TARGET_CARDINALITY;
	}

	public class Operation extends AbstractSchemaGraphOperationDescriptor<Operation>.AbstractSchemaGraphOperation {

		public Operation(Relationship targetRelationship, RepairSchemaGraphInstance instance) {
			super(targetRelationship, instance);
		}

		@Override
		public void searchSideEffectRelationships(SchemaGraphInstance sgi) {
			// We delete links, so there are side effects on the inverse.
			Relationship inverse = getTargetRelationship().getInverse();
			addSideEffectRelationship(inverse, INVERSE);

			
			Collection<Relationship> relationships = sgi.getSchema().getAtomicRelationships(
					getTargetRelationship().getEndNode());
			for (Relationship relationship : relationships) {
				if (!relationship.isAtomic()) {
					continue;
				}
			}
		}

		@Override
		protected NaturalNumberSet calculateNewTargetCardinality(NaturalNumberSet oldCardinality) {
			return NaturalNumberInterval.sign(oldCardinality);
		}

		@Override
		protected NaturalNumberSet calculateNewSideEffectCardinality(NaturalNumberSet oldCardinality, int type) {
			switch (type) {
			case INVERSE:
				if (oldCardinality.isEmpty()) {
					return NaturalNumberSet.EMPTY_SET;
				}
				return new NaturalNumberInterval(0, oldCardinality.getMax());
			default:
				throw new IllegalArgumentException("Unsupported side effect type: " + type);
			}
		}
		
		@Override
		protected void executeOnBags() {
			// Process fix: Delete all but one link for each tuple.
			int numRemovedLinks = 0;
			int numAffectedTuples = 0;
			NaturalNumberBag oldCardinality = getInstance().getRealCardinality(getTargetRelationship());
			NaturalNumberBag newCardinality = new NaturalNumberBag();
			for (IntIterator i = oldCardinality.getCountedCardinalities().iterator(); i.hasNext();) {
				int cardinalityValue = i.nextInt();
				int count = oldCardinality.getCountFor(cardinalityValue);
				if (cardinalityValue > 1) {
					numRemovedLinks += count * (cardinalityValue - 1);
					numAffectedTuples += count;
				} else {
					newCardinality.count(cardinalityValue, count);
				}
			}
			newCardinality.count(1, numAffectedTuples);
			getInstance().setRealCardinality(getTargetRelationship(), newCardinality);

			// Handle inverse.
			Collection<Relationship> relationships = this.getSideEffectRelationships(INVERSE);
			if (relationships.size() != 1) {
				throw new IllegalStateException();
			}
			Relationship relationship = relationships.iterator().next();
			oldCardinality = getInstance().getRealCardinality(relationship);
			// Simulate the removal of links.
			StatisticBagOperator statisticBagOperator = new StatisticBagOperator(oldCardinality);
			newCardinality = statisticBagOperator.simulateLinkRemoval(numRemovedLinks);
			// We instantly delete values, that are now detached by not keeping them.
			getInstance().setRealCardinality(relationship, newCardinality);
		}

	}

}
