package de.uni_potsdam.hpi.isg.efes.pipeline;

import java.util.ArrayList;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.datafit.DataFitModel;
import de.uni_potsdam.hpi.isg.efes.csg.datafit.DataFitReport;
import de.uni_potsdam.hpi.isg.efes.csg.effort.EffortEstimate;
import de.uni_potsdam.hpi.isg.efes.csg.effort.EffortModel;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.BasicNodeMapping;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.ForeignKeyMatcher;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.NodeCorrespondence;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.TableNodeMatcher;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.DataSet;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.IntegrationScenario;
import de.uni_potsdam.hpi.isg.efes.csg.util.MetaDataDetector;
import de.uni_potsdam.hpi.isg.utils.util.MultiLevelIterator;

public class Pipeline {

	private static Logger logger = LogManager.getLogger(Pipeline.class.getSimpleName());

	private List<DataFitModel> dataFitModels;

	private EffortModel effortModel;

	private DataFitReport dataFitReport;

	private EffortEstimate effortEstimate;

	public Pipeline() {
		this.dataFitModels = new ArrayList<>();
	}

	public void addDataFitModel(DataFitModel dataFitModel) {
		if (this.dataFitModels.add(dataFitModel)) {
			logger.warn("Preventing from inserting {} a second time.", dataFitModel);
			return;
		}
		this.dataFitModels.add(dataFitModel);
	}

	public void setEffortModel(EffortModel effortModel) {
		this.effortModel = effortModel;
	}

	public void preprocess(IntegrationScenario integrationScenario) {
		DataSet targetDataSet = integrationScenario.getTargetDataSet();
		for (DataSet sourceDataSet : integrationScenario.getSourceDataSets()) {
			BasicNodeMapping mapping = integrationScenario.getMappings().get(sourceDataSet);

			// Match foreign keys.
//			ForeignKeyMatcher fkMatcher = new ForeignKeyMatcher();
//			fkMatcher.enrich(sourceDataSet.getSchema(), mapping, targetDataSet.getSchema());
			
			// Run table node matcher if necessary.
			boolean hasTableCorrespondences = false;
			Iterable<NodeCorrespondence> allCorrespondences = new MultiLevelIterator<>(mapping.getNodeCorrespondences()
					.values()).wrapWithIterable();
			for (NodeCorrespondence correspondence : allCorrespondences) {
				hasTableCorrespondences |= (correspondence.getSourceNode().isSimpleIdentityNode());
				hasTableCorrespondences |= (correspondence.getTargetNode().isSimpleIdentityNode());
			}
			if (!hasTableCorrespondences) {
				logger.info("Running table node matcher on {} and {}.", sourceDataSet, targetDataSet);
				TableNodeMatcher tableNodeMatcher = new TableNodeMatcher();
				tableNodeMatcher.enrich(sourceDataSet.getSchema(), mapping, targetDataSet.getSchema());
			} else {
				logger.info("Already found table node mapping for {}, so the table node matcher will not be run.",
						sourceDataSet);
			}

			ForeignKeyMatcher fkMatcher = new ForeignKeyMatcher();
			fkMatcher.enrichComplex(sourceDataSet.getSchema(), targetDataSet.getSchema(), mapping);
		}
		

		// Detect source's meta data if necessary.
		MetaDataDetector metaDataDetector = new MetaDataDetector(targetDataSet.getSchema());
		metaDataDetector.markMetaData();
	}

	public void process(IntegrationScenario integrationScenario, String estimationUnit) {
		if (this.dataFitModels.isEmpty() || this.effortModel == null) {
			throw new IllegalStateException("No data fit and/or effort model registered in the pipeline.");
		}

		logger.info("Processing scenario {}.", integrationScenario);

		preprocess(integrationScenario);
		configureModels(integrationScenario);
		runDataFitModels();
		runEffortModel(estimationUnit);
	}

	public void rerunEffortModel(String estimationUnit) {
		if (this.dataFitReport == null) {
			throw new IllegalStateException("No data fit report to rerun effort model on.");
		}
		runEffortModel(estimationUnit);
	}

	private void runEffortModel(String estimationUnit) {
		this.effortEstimate = this.effortModel.evaluate(this.dataFitReport, estimationUnit);
	}

	private void runDataFitModels() {
		this.dataFitReport = new DataFitReport();
		for (DataFitModel dataFitModel : this.dataFitModels) {
			logger.info("Running {}.", dataFitModel);
			dataFitModel.evaluateScenarioTo(this.dataFitReport);
		}
	}

	private void configureModels(IntegrationScenario integrationScenario) {
		for (DataFitModel dataFitModel : this.dataFitModels) {
			dataFitModel.setIntegrationScenario(integrationScenario);
		}
		this.effortModel.setIntegrationScenario(integrationScenario);
	}

	public DataFitReport getDataFitReport() {
		return dataFitReport;
	}

	public EffortEstimate getEffortEstimate() {
		return effortEstimate;
	}

}
