package de.uni_potsdam.hpi.isg.efes.csg.integration;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.SchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.instance.SchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.NodeCorrespondence;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.NodeMapping;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.RelationshipCorrespondence;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.RelationshipMapping;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.UnionRelationship;

public class IntegratedMapping implements RelationshipMapping {

	private static final Logger logger = LogManager.getLogger(IntegratedMapping.class.getSimpleName());

	private Map<String, RelationshipMapping> mappings = new HashMap<>();
	private IntegratedSchemaGraph source;
	private IntegratedSchemaGraphInstance instance;

	public IntegratedMapping(Collection<RelationshipMapping> integratedMappings) {
		super();

		List<SchemaGraphInstance> sourceInstances = new ArrayList<>();
		boolean createInstance = true;
		this.source = new IntegratedSchemaGraph();
		for (RelationshipMapping mapping : integratedMappings) {
			SchemaGraph sourceSchema = mapping.getSourceSchema();
			if (!(sourceSchema instanceof BasicSchemaGraph)) {
				throw new IllegalArgumentException("Given source must contain a BasicSchemaGraph.");
			}
			BasicSchemaGraph basicSchemaGraph = (BasicSchemaGraph) sourceSchema;
			this.source.add(basicSchemaGraph);

			String sourceName = basicSchemaGraph.getName();
			this.mappings.put(sourceName, mapping);

			SchemaGraphInstance sourceInstance = mapping.getSourceInstance();
			if (sourceInstance != null) {
				sourceInstances.add(sourceInstance);
			} else {
				createInstance = false;
			}
		}

		if (createInstance) {
			this.instance = new IntegratedSchemaGraphInstance(sourceInstances);
		}

	}

	@Override
	public Collection<NodeCorrespondence> getCorrespondencesForTargetNode(Node targetNode) {
		List<NodeCorrespondence> result = new ArrayList<>(mappings.size());
		for (NodeMapping mapping : this.mappings.values()) {
			result.addAll(mapping.getCorrespondencesForTargetNode(targetNode));
		}
		return result;
	}

	@Override
	public Collection<RelationshipCorrespondence> getCorrespondencesForTargetRelationship(
			Relationship targetRelationship) {
		List<RelationshipCorrespondence> result = new ArrayList<>(mappings.size());
		for (RelationshipMapping mapping : this.mappings.values()) {
			result.addAll(mapping.getCorrespondencesForTargetRelationship(targetRelationship));
		}
		return result;
	}

	@Override
	public Relationship translateTargetRelationship(Relationship targetRelationship) {
		List<Relationship> translatedRelationships = new ArrayList<>();
		for (RelationshipMapping mapping : this.mappings.values()) {
			Relationship partialRelationship = mapping.translateTargetRelationship(targetRelationship);
			logger.trace(" Partial translation {} to {}.", targetRelationship, partialRelationship);
			if (partialRelationship != null) {
				translatedRelationships.add(partialRelationship);
			}
		}
		if (translatedRelationships.isEmpty()) {
			logger.trace("Could not translate {}.", targetRelationship);
			return null;
		}
		Relationship translatedRelationship = UnionRelationship.union(translatedRelationships);
		logger.trace("Translated {} to {}.", targetRelationship, translatedRelationship);
		return translatedRelationship;
	}
	
	@Override
	public void validateDomainsAndCodomains() throws IllegalStateException {
		for (RelationshipMapping mapping : this.mappings.values()) {
			mapping.validateDomainsAndCodomains();
		}
	}

	public NaturalNumberSet estimateSourceCardinality(Relationship targetRelationship) {
		Relationship sourceRelationship = translateTargetRelationship(targetRelationship);
		if (sourceRelationship == null) {
			return NaturalNumberSet.EMPTY_SET;
		}
		NaturalNumberSet cardinality = this.source.getCardinality(sourceRelationship);
		return cardinality;
	}

	@Override
	public boolean isBackedByInstance() {
		return getSourceInstance() != null;
	}

	@Override
	public SchemaGraphInstance getSourceInstance() {
		return this.instance;
	}

	@Override
	public NaturalNumberBag getRealSourceCardinality(Relationship targetRelationship) {
		Relationship sourceRelationship = translateTargetRelationship(targetRelationship);
		if (sourceRelationship == null) {
			logger.debug("Could not translate {}.", targetRelationship);
			return NaturalNumberBag.EMPTY_BAG;
		}
		return this.instance.getRealCardinality(sourceRelationship);
	}

	@Override
	public Map<Relationship, List<RelationshipCorrespondence>> getRelationshipCorrespondences() {
		throw new UnsupportedOperationException();
	}

	@Override
	public SchemaGraph getSourceSchema() {
		return this.source;
	}

	@Override
	public Map<Node, List<NodeCorrespondence>> getNodeCorrespondences() {
		throw new UnsupportedOperationException();
	}

}
