package de.uni_potsdam.hpi.isg.efes.csg.integration;

import java.text.Collator;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.instance.DataStoreAdapter;
import de.uni_potsdam.hpi.isg.efes.csg.instance.DataStoreAdapter.KeyValueCount;
import de.uni_potsdam.hpi.isg.efes.csg.instance.SchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node.ContentType;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.UnionNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.UnionRelationship;

public class IntegratedSchemaGraphInstance implements SchemaGraphInstance {

	private static final Logger logger = LogManager
			.getLogger(IntegratedSchemaGraphInstance.class.getSimpleName());

	private Map<String, SchemaGraphInstance> instances;
	private boolean hasDataStoreAdapter;

	public IntegratedSchemaGraphInstance(
			Collection<SchemaGraphInstance> schemaGraphInstances) {
		this.instances = new HashMap<>();
		this.hasDataStoreAdapter = true;
		for (SchemaGraphInstance schemaGraphInstance : schemaGraphInstances) {
			String schemaName = schemaGraphInstance.getSchema().getName();
			this.instances.put(schemaName, schemaGraphInstance);
			this.hasDataStoreAdapter &= schemaGraphInstance
					.hasDataStoreAdapter();
		}
	}

	@Override
	public boolean hasDataStoreAdapter() {
		return this.hasDataStoreAdapter;
	}

	@Override
	public DataStoreAdapter getDataStoreAdapter() {
		throw new UnsupportedOperationException();
	}

	@Override
	public NaturalNumberBag getRealCardinality(Relationship relationship) {
		logger.trace("#getRealCardinality({}).", relationship);
		switch (relationship.getOperationType()) {
		case UNION:
			UnionRelationship unionRelationship = (UnionRelationship) relationship;
			Map<SchemaGraphInstance, List<Relationship>> relationshipsByInstance = new HashMap<>();
			for (Relationship baseRelationship : unionRelationship
					.getBaseRelationships()) {
				SchemaGraphInstance instance = getInstance(baseRelationship);
				List<Relationship> relationships = relationshipsByInstance
						.get(instance);
				if (relationships == null) {
					relationships = new ArrayList<>(1);
					relationshipsByInstance.put(instance, relationships);
				}
				relationships.add(baseRelationship);
			}

			if (unionRelationship.getStartNode() != null
					&& unionRelationship.getStartNode().getContentType() == Node.ContentType.VALUE) {
				// Need to consider possibly same values from different data
				// sets.

				if (unionRelationship.getEndNode() != null
						&& unionRelationship.getEndNode().getContentType() == Node.ContentType.VALUE) {
					NaturalNumberBag cardinality = mergeKeyValuesAndSumCardinalities(relationshipsByInstance);
					return cardinality;

				} else {
					NaturalNumberBag cardinality = mergeKeysAndSumCardinalities(relationshipsByInstance);
					return cardinality;
				}

			} else {
				// No same start values possible across sources. Can just query
				// each source in isolation.
				NaturalNumberBag aggregatedCardinality = separatelyQueryAndAggregate(relationshipsByInstance);
				return aggregatedCardinality;
			}

		default:
			SchemaGraphInstance instance = getInstance(relationship);
			return instance.getRealCardinality(relationship);
		}
	}

	private NaturalNumberBag separatelyQueryAndAggregate(
			Map<SchemaGraphInstance, List<Relationship>> relationshipsByInstance) {
		logger.trace("#separatelyQueryAndAggregate({}).",
				relationshipsByInstance);
		NaturalNumberBag aggregatedCardinality = new NaturalNumberBag();
		for (Map.Entry<SchemaGraphInstance, List<Relationship>> entry : relationshipsByInstance
				.entrySet()) {
			SchemaGraphInstance instance = entry.getKey();
			List<Relationship> relationships = entry.getValue();
			Relationship sourceRelationship = UnionRelationship
					.union(relationships);
			NaturalNumberBag partialCardinality = instance
					.getRealCardinality(sourceRelationship);
			logger.trace(" Partial cardinality {} of {}.", partialCardinality,
					sourceRelationship);
			aggregatedCardinality.unionAll(partialCardinality);
		}
		return aggregatedCardinality;
	}

	private NaturalNumberBag mergeKeysAndSumCardinalities(
			Map<SchemaGraphInstance, List<Relationship>> relationshipsByInstance) {
		logger.trace("#mergeKeysAndSumCardinalities({}).",
				relationshipsByInstance);
		List<Iterator<KeyValueCount>> keyCardinalityIterators = new ArrayList<>();
		for (Map.Entry<SchemaGraphInstance, List<Relationship>> entry : relationshipsByInstance
				.entrySet()) {
			SchemaGraphInstance instance = entry.getKey();
			List<Relationship> relationships = entry.getValue();
			Relationship sourceRelationship = UnionRelationship
					.union(relationships);
			Iterator<KeyValueCount> keyCardinalityIterator = instance
					.getDataStoreAdapter().queryKeysWithCardinality(
							sourceRelationship);
			keyCardinalityIterators.add(keyCardinalityIterator);
		}
		Iterator<KeyValueCount> compoundIterator = new WrappedKeyValueCardinalityIterator(
				keyCardinalityIterators);

		NaturalNumberBag cardinality = new NaturalNumberBag();
		while (compoundIterator.hasNext()) {
			KeyValueCount valueCount = compoundIterator.next();
			cardinality.count(valueCount.getCount());
		}
		logger.trace("Merged cardinality is {}.", cardinality);
		return cardinality;
	}

	private NaturalNumberBag mergeKeyValuesAndSumCardinalities(
			Map<SchemaGraphInstance, List<Relationship>> relationshipsByInstance) {
		logger.trace("#mergeKeyValuesAndSumCardinalities({}).",
				relationshipsByInstance);
		List<Iterator<KeyValueCount>> keyCardinalityIterators = new ArrayList<>();
		for (Map.Entry<SchemaGraphInstance, List<Relationship>> entry : relationshipsByInstance
				.entrySet()) {
			SchemaGraphInstance instance = entry.getKey();
			List<Relationship> relationships = entry.getValue();
			Relationship sourceRelationship = UnionRelationship
					.union(relationships);
			Iterator<KeyValueCount> keyCardinalityIterator = instance
					.getDataStoreAdapter().queryKeyValuesWithCardinality(
							sourceRelationship);
			keyCardinalityIterators.add(keyCardinalityIterator);
		}
		Iterator<KeyValueCount> compoundIterator = new WrappedKeyValueCardinalityIterator(
				keyCardinalityIterators);

		NaturalNumberBag cardinality = new NaturalNumberBag();
		while (compoundIterator.hasNext()) {
			KeyValueCount valueCount = compoundIterator.next();
			cardinality.count(valueCount.getCount());
		}
		logger.trace("Merged cardinality is {}.", cardinality);
		return cardinality;
	}

	private SchemaGraphInstance getInstance(Relationship relationship) {
		String schemaName = BasicSchemaGraph.getSchemaNameFor(relationship);
		SchemaGraphInstance instance = this.instances.get(schemaName);
		if (instance == null) {
			throw new IllegalArgumentException(
					"No (umambiguous) instance found for " + relationship + ".");
		}
		return instance;
	}

	@Override
	public int getRealSize(Node node) {
		Map<String, Collection<Node>> nodesBySchema = new HashMap<>();
		separateUnionedNodesBySchema(node, nodesBySchema);

		int sumIdentities = 0;
		List<Iterator<String>> iteratorsPerSchema = new ArrayList<>();
		for (Map.Entry<String, Collection<Node>> entry : nodesBySchema
				.entrySet()) {

			// Get the correct schema instance.
			String schemaName = entry.getKey();
			SchemaGraphInstance instance = this.instances.get(schemaName);

			// Separate value and identity nodes.
			Collection<Node> identityNodes = new ArrayList<>();
			Collection<Node> valueNodes = new ArrayList<>();
			for (Node schemaNode : entry.getValue()) {
				if (schemaNode.getContentType() == ContentType.IDENTITY) {
					identityNodes.add(schemaNode);
				} else {
					valueNodes.add(schemaNode);
				}
			}

			// Evaluate identity nodes.
			for (Node identityNode : identityNodes) {
				int realSize = instance.getRealSize(identityNode);
				sumIdentities += realSize;
			}

			// Evaluate value nodes I: Load sorted values.
			if (!valueNodes.isEmpty()) {
				Node nodeForSchema = UnionNode.union(valueNodes);
				Iterator<String> iteratorForSource = instance
						.getDataStoreAdapter().getValues(nodeForSchema);
				iteratorsPerSchema.add(iteratorForSource);
			}
		}

		// Evaluate value nodes II: Merge values.
		Iterator<String> compoundIterator = new WrappedStringIterator(
				iteratorsPerSchema);
		int sumValues = 0;
		while (compoundIterator.hasNext()) {
			compoundIterator.next();
			sumValues++;
		}

		// Finally, we account for the fact, that there are situations where we
		// have a mix of values and identities in union nodes. These two sets
		// are distinct.
		return sumIdentities + sumValues;

	}

	private void separateUnionedNodesBySchema(Node node,
			Map<String, Collection<Node>> nodesBySchema) {
		if (node.getOperationType() == Node.Operation.NONE) {
			String schemaName = BasicSchemaGraph.getSchemaNameFor(node);
			Collection<Node> nodesForSchema = nodesBySchema.get(schemaName);
			if (nodesForSchema == null) {
				nodesForSchema = new ArrayList<>();
				nodesBySchema.put(schemaName, nodesForSchema);
			}
			nodesForSchema.add(node);

		} else if (node.getOperationType() == Node.Operation.UNION) {
			UnionNode unionNode = (UnionNode) node;
			List<Node> parentNodes = unionNode.getParentNodes();
			for (Node parentNode : parentNodes) {
				separateUnionedNodesBySchema(parentNode, nodesBySchema);
			}

		} else {
			throw new IllegalArgumentException(
					"Cannot distribute size query for " + node + ".");
		}
	}

	@Override
	public NaturalNumberSet getPrescribedCardinality(Relationship relationship) {
		throw new UnsupportedOperationException();
	}

	@Override
	public BasicSchemaGraph getSchema() {
		throw new UnsupportedOperationException();
	}

	@Override
	public NaturalNumberSet getActualCardinality(Relationship relationship) {
		return getRealCardinality(relationship).toInterval();
	}

	@Override
	public Collection<Constraint> getViolatedConstraints() {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setActualCardinality(Relationship relationship,
			NaturalNumberSet cardinality) {
		throw new UnsupportedOperationException();
	}

	private static class WrappedKeyValueCardinalityIterator extends
			WrappedIterator<KeyValueCount> {

		// Use the defaut locale to emulate postgres sorting.
		private Collator collator;

		public WrappedKeyValueCardinalityIterator(
				List<Iterator<KeyValueCount>> iterators) {
			super(iterators);
			Locale locale = Locale.getDefault();
			this.collator = Collator.getInstance(locale);
		}

		@Override
		public void remove() {
			throw new UnsupportedOperationException();
		}

		@Override
		protected boolean shallGroup(KeyValueCount e1, KeyValueCount e2) {
			return e1.keyValueEqual(e2);
		}

		@Override
		protected KeyValueCount aggregate(List<KeyValueCount> iteratorStates) {
			KeyValueCount aggregate = null;
			for (KeyValueCount keyValueCount : iteratorStates) {
				if (aggregate == null) {
					aggregate = keyValueCount;
				} else {
					aggregate.setCount(aggregate.getCount()
							+ keyValueCount.getCount());
				}
			}
			return aggregate;
		}

		@Override
		protected int compare(KeyValueCount e1, KeyValueCount e2) {
			int result = this.collator.compare(e1.getKey(), e2.getKey());
			if (result == 0
					&& !(e1.getValue() == null || e2.getValue() == null)) {
				result = this.collator.compare(e1.getValue(), e2.getValue());
			}
			return result;
		}

	}

	private static class WrappedStringIterator extends WrappedIterator<String> {

		public WrappedStringIterator(List<Iterator<String>> iterators) {
			super(iterators);
		}

		@Override
		protected boolean shallGroup(String e1, String e2) {
			return e1.equals(e2);
		}

		@Override
		protected String aggregate(List<String> states) {
			return states.get(0);
		}

		@Override
		protected int compare(String e1, String e2) {
			return e1.compareTo(e2);
		}

	}

	@Override
	public NaturalNumberSet getCardinality(Relationship relationship) {
		return getRealCardinality(relationship).toInterval();
	}

	@Override
	public boolean areOverlapping(Node node1, Node node2) {
		// TODO: Implement if needed.
		throw new UnsupportedOperationException();
	}

}
