package de.uni_potsdam.hpi.isg.efes.csg.mapping;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.instance.SchemaGraphInstance;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.EmptyRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.UnionRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.DataSet;

public class BasicRelationshipMapping extends BasicNodeMapping implements RelationshipMapping {
	
	private static final Logger logger = LogManager.getLogger(BasicRelationshipMapping.class.getSimpleName());

	private final Map<Relationship, List<RelationshipCorrespondence>> correspondencesByTargetRelationship = new HashMap<>();

	@Deprecated
	public BasicRelationshipMapping(DataSet source, DataSet target) {
		super(source, target);
	}

	public BasicRelationshipMapping(NodeMapping mapping) {
		super(mapping);
	}

	public void addCorrespondence(final RelationshipCorrespondence correspondence) {
		if (correspondence == null)
			throw new IllegalArgumentException();
	
		
		List<RelationshipCorrespondence> correspondences = this.correspondencesByTargetRelationship.get(correspondence
				.getTargetRelationship());
		if (correspondences == null) {
			correspondences = new ArrayList<>();
			this.correspondencesByTargetRelationship.put(correspondence.getTargetRelationship(), correspondences);
			logger.info("Register correspondence \n\t{} to \n\t{}.", correspondence.getSourceRelationship(), correspondence.getTargetRelationship());
		} else if (correspondences.contains(correspondence)) {
			logger.debug("Tried to reinsert existing correspondence {}.", correspondence);
			return;
		}
	
		correspondences.add(correspondence);
	}

	public Collection<RelationshipCorrespondence> getCorrespondencesForTargetRelationship(Relationship targetRelationship) {
		Collection<RelationshipCorrespondence> correspondences = this.correspondencesByTargetRelationship
				.get(targetRelationship);
		return correspondences == null ? Collections.<RelationshipCorrespondence> emptyList() : correspondences;
	}

	/**
	 * Experimental feature: Translate a atomic source relationship to a complex
	 * target relationship. This is useful when a single target relationship
	 * corresponds to multiple source relationships.<br>
	 * Currently only support atomic relationships.
	 * 
	 * @return a single relationship or <code>null</code> if no source
	 *         relationship exists
	 */
	public Relationship translateTargetRelationship(Relationship targetRelationship) {
		logger.trace("Translating {}.", targetRelationship);

		if (!targetRelationship.isAtomic()) {
			throw new IllegalArgumentException("Can only translate atomic relationships.");
		}
	
		Collection<RelationshipCorrespondence> correspondences = getCorrespondencesForTargetRelationship(targetRelationship);
		List<Relationship> sourceRelationships = new ArrayList<>();
		for (RelationshipCorrespondence correspondence : correspondences) {
			sourceRelationships.add(correspondence.getSourceRelationship());
		}
		logger.trace("Corresponding relationships: {}", sourceRelationships);
		
		Relationship translatedRelationship = UnionRelationship.union(sourceRelationships);
		
		logger.debug("Translated {} to {}.", targetRelationship, translatedRelationship);
		
		return translatedRelationship;
	}

	public NaturalNumberSet estimateSourceCardinality(Relationship targetRelationship) {
		Relationship sourceRelationship = translateTargetRelationship(targetRelationship);
		if (sourceRelationship == null) {
			return NaturalNumberSet.EMPTY_SET;
		}
		NaturalNumberSet cardinality = this.getSourceSchema().getCardinality(sourceRelationship);
		return cardinality;
	}

	public NaturalNumberBag getRealSourceCardinality(Relationship targetRelationship) {
		Relationship sourceRelationship = translateTargetRelationship(targetRelationship);
		if (sourceRelationship == null) {
			return NaturalNumberBag.EMPTY_BAG;
		}
		SchemaGraphInstance instance = this.getSourceInstance();
		if (instance == null) {
			throw new IllegalStateException("No source instance available.");
		}
		NaturalNumberBag cardinalityBag = instance.getRealCardinality(sourceRelationship);
		return cardinalityBag;
	}

	public Map<Relationship, List<RelationshipCorrespondence>> getRelationshipCorrespondences() {
		return Collections.unmodifiableMap(this.correspondencesByTargetRelationship);
	}
	
	@Override
	public void validateDomainsAndCodomains() throws IllegalStateException {
		
		boolean isValidationFailed = false;
		
		for (Map.Entry<Relationship, List<RelationshipCorrespondence>> entry : this.correspondencesByTargetRelationship.entrySet()) {
			Relationship targetRelationship = entry.getKey();
			List<RelationshipCorrespondence> relationshipCorrespondences = entry.getValue();

			isValidationFailed |= validateForTargetRelationship(targetRelationship,
					relationshipCorrespondences);
			
		}
		
		if (isValidationFailed) {
			throw new IllegalStateException("Relationship mapping validation failed.");
		}
	}

	private boolean validateForTargetRelationship(Relationship targetRelationship,
			List<RelationshipCorrespondence> relationshipCorrespondences) {
		
		boolean isValidationFailed = false;
		
		Collection<Node> startNodes = new HashSet<>();
		for (NodeCorrespondence nodeCorrespondences : getCorrespondencesForTargetNode(targetRelationship.getStartNode())) {
			startNodes.add(nodeCorrespondences.getSourceNode());
		}
		Collection<Node> endNodes = new HashSet<>();
		for (NodeCorrespondence nodeCorrespondences : getCorrespondencesForTargetNode(targetRelationship.getEndNode())) {
			endNodes.add(nodeCorrespondences.getSourceNode());
		}
		
		for (RelationshipCorrespondence relationshipCorrespondence : relationshipCorrespondences) {
			Relationship sourceRelationship = relationshipCorrespondence.getSourceRelationship();
			startNodes.remove(sourceRelationship.getStartNode());
			endNodes.remove(sourceRelationship.getEndNode());
		}
		
		if (!startNodes.isEmpty()) {
			logger.error("No corresponding relationship with start nodes {} for {}: {}", startNodes, targetRelationship, relationshipCorrespondences);
//			isValidationFailed = true;
			for (Node startNode : startNodes) {
				relationshipCorrespondences.add(new RelationshipCorrespondence(EmptyRelationship.createWithStartNode(startNode), targetRelationship));
			}
		}
		if (!endNodes.isEmpty()) {
			logger.error("No corresponding relationship with end nodes {} for {}: {}", endNodes, targetRelationship, relationshipCorrespondences);
//			isValidationFailed = true;
			for (Node endNode : endNodes) {
				relationshipCorrespondences.add(new RelationshipCorrespondence(EmptyRelationship.createWithEndNode(endNode), targetRelationship));
			}
		}
		return isValidationFailed;
	}

}
