package de.uni_potsdam.hpi.isg.efes.csg.mapping;

import java.util.Collection;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.ValueNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.ExplicitAtomicRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.util.RelationalSchemaGraphUtil;
import de.uni_potsdam.hpi.isg.utils.util.CollectionUtil;

/**
 * Adds correspondences for foreign keys according to correspondences of the primary keys.
 */
public class ForeignKeyMatcher {

	private static final Logger logger = LogManager.getLogger(ForeignKeyMatcher.class.getSimpleName());

	/**
	 * Adds {@link NodeCorrespondence} based on a foreign key analysis.
	 * <p>
	 * If the target attribute <tt>S(B)</tt> references <tt>R(A)</tt> and <tt>T(C)</tt> corresponds to <tt>R(A)</tt>, this method puts a correspondence towards
	 * <tt>S(B)</tt> as well.
	 * 
	 * @param source
	 * @param mapping
	 * @param target
	 */
	public void enrichSimple(BasicSchemaGraph source, BasicNodeMapping mapping, BasicSchemaGraph target) {

		for (Node targetValueNode : target.getAtomicNodes().values()) {
			// Look for value nodes.
			if (!(targetValueNode instanceof ValueNode)) {
				continue;
			}

			// Ensure that the node does not have a correspondence yet.
			Collection<NodeCorrespondence> correpsondences = mapping.getCorrespondencesForTargetNode(targetValueNode);
			if (!correpsondences.isEmpty()) {
				continue;
			}

			// See if the node is referencing another value node.
			Relationship identityRelationship = null;
			Collection<Relationship> relationships = target.getAtomicRelationships(targetValueNode);
			for (Relationship relationship : relationships) {
				if (!relationship.isExplicit()
						&& target.getConstraints(relationship).toConstraint().getCardinality()
								.isSubsetOf(NaturalNumberSet.ONE)) {
					identityRelationship = relationship;
					continue;
				}
			}
			if (identityRelationship == null) {
				continue;
			}

			// Copy all correspondences from the referenced node.
			Node referencedNode = identityRelationship.getEndNode();
			for (NodeCorrespondence blueprintCorrespondence : mapping.getCorrespondencesForTargetNode(referencedNode)) {
				NodeCorrespondence copiedCorrespondence = new NodeCorrespondence(
						blueprintCorrespondence.getSourceNode(), targetValueNode);
				mapping.addCorrespondence(copiedCorrespondence);
				logger.info("Adding {} to the node mapping.", copiedCorrespondence);
			}
		}
	}

	/**
	 * Adds {@link NodeCorrespondence} based on a foreign key analysis.
	 * <p>
	 * If the target attribute <tt>S(B)</tt> references <tt>R(A)</tt> and <tt>T(C)</tt> corresponds to <tt>R(A)</tt>, this method puts a correspondence towards
	 * <tt>S(B)</tt> as well but also checks if in place of <tt>R(A)</tt> there is a more suitable source foreign key.
	 * 
	 * @param source
	 * @param mapping
	 * @param target
	 */
	public void enrichComplex(BasicSchemaGraph sourceSchema, BasicSchemaGraph targetSchema, BasicNodeMapping nodeMapping) {

		// Go over all target tables.
		for (Node targetTable : targetSchema.getAtomicNodes().values()) {
			Collection<NodeCorrespondence> tableCorrespondences = nodeMapping
					.getCorrespondencesForTargetNode(targetTable);
			if (!targetTable.isSimpleIdentityNode() || tableCorrespondences.isEmpty()) {
				continue;
			}

			// Find non-mapped, non-metadata FKs.
			for (Node targetFkNode : RelationalSchemaGraphUtil.findAttributeNodes(targetSchema, targetTable)) {
				if (targetFkNode.isMetaData()
						|| !RelationalSchemaGraphUtil.isForeignKeyNode(targetFkNode, targetSchema)
						|| !nodeMapping.getCorrespondencesForTargetNode(targetFkNode).isEmpty()) {
					continue;
				}

				// Collect PK.
				Node targetPkNode = CollectionUtil.getAny(RelationalSchemaGraphUtil.findReferencedAttributeNodes(
						targetSchema, targetFkNode));

				// Check if PK is mapped and -if so- collect it.
				Collection<NodeCorrespondence> pkCorrespondences = nodeMapping
						.getCorrespondencesForTargetNode(targetPkNode);
				if (pkCorrespondences.isEmpty()) {
					continue;
				}

				// Try to find a link between source tables and source PKs.
				for (NodeCorrespondence tableCorrespondence : tableCorrespondences) {
					Node sourceTable = tableCorrespondence.getSourceNode();
					for (NodeCorrespondence pkCorrespondence : pkCorrespondences) {
						Node sourcePk = pkCorrespondence.getSourceNode();

						// Option 1: source table = referencing table
						if (sourceSchema.contains(new ExplicitAtomicRelationship(sourceTable, sourcePk))) {
							NodeCorrespondence newCorrespondence = new NodeCorrespondence(sourcePk, targetFkNode);
							nodeMapping.addCorrespondence(newCorrespondence);
							logger.info("Adding temporary correspondence {} due to graph analysis.", newCorrespondence);
							continue;
						}

						// Option 2: source table is also referencing, find appropriate FK
						for (Node sourceFkNode : RelationalSchemaGraphUtil
								.findAttributeNodes(sourceSchema, sourceTable)) {
							if (RelationalSchemaGraphUtil.findReferencedAttributeNodes(sourceSchema, sourceFkNode)
									.contains(sourcePk)) {
								NodeCorrespondence newCorrespondence = new NodeCorrespondence(sourceFkNode,
										targetFkNode);
								nodeMapping.addCorrespondence(newCorrespondence);
								logger.info("Adding temporary correspondence {} due to graph analysis.",
										newCorrespondence);
							}
						}

					}
				}

			}
		}

	}
}
