package de.uni_potsdam.hpi.isg.efes.csg.mapping;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.EmptyRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship.ContentType;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.IntegrationScenario;
import de.uni_potsdam.hpi.isg.efes.csg.search.ConstraintOrder;
import de.uni_potsdam.hpi.isg.efes.csg.util.PartialOrder;
import de.uni_potsdam.hpi.isg.efes.csg.util.PartialOrderQueue;
import de.uni_potsdam.hpi.isg.utils.util.CollectionUtil;

/**
 * Matches relationships from a source and target schema from a set of {@link NodeCorrespondence} objects.
 * 
 * @author Sebastian Kruse
 */
public class RelationshipMatcher {

	private static Logger logger = LogManager.getLogger(RelationshipMatcher.class.getSimpleName());

	private BasicSchemaGraph sourceSchema, targetSchema;
	private NodeMapping inputMapping;
	private BasicRelationshipMapping resultMapping;
	private Set<Relationship> handledRelationships;

	public RelationshipMatcher(IntegrationScenario integrationScenario, int sourceIndex) {
		this(integrationScenario.getSourceDataSets().get(sourceIndex).getSchema(), integrationScenario
				.getTargetDataSet().getSchema(), integrationScenario.getMappings().get(
				integrationScenario.getSourceDataSets().get(sourceIndex)));
	}

	public RelationshipMatcher(BasicSchemaGraph sourceSchema, BasicSchemaGraph targetSchema, NodeMapping mapping) {
		this.sourceSchema = sourceSchema;
		this.targetSchema = targetSchema;
		this.inputMapping = mapping;
		this.handledRelationships = new HashSet<>();
	}

	public RelationshipMapping run() {
		if (this.inputMapping instanceof RelationshipMapping) {
			logger.warn("The given mapping already is a RelationshipMapping.");
		}

		this.handledRelationships.clear();
		this.resultMapping = new BasicRelationshipMapping(this.inputMapping);

		for (Relationship targetRelationship : this.targetSchema.getAllRelationships()) {
			matchWithInverse(targetRelationship);
		}

		reconcileAlternativeRelationship();

		this.resultMapping.validateDomainsAndCodomains();
		
		return this.resultMapping;
	}

	/**
	 * This method implements the following idea: If we have a target node corresponds to multiple source nodes, this can likely cause union relationships. To
	 * avoid unwanted relationships, we check for each relationship if either of the source nodes is a better fit (globally seen).
	 */
	private void reconcileAlternativeRelationship() {

		Map<Node, Map<Relationship, List<Relationship>>> reconciliationClustersByTargetEndNode = new HashMap<>();
		Set<Node> targetReconciliationNodes = new HashSet<>();

		// Phase 1: Find ambiguous mappings that are candidates for reconciliation.
		for (Map.Entry<Relationship, List<RelationshipCorrespondence>> entry : this.resultMapping
				.getRelationshipCorrespondences().entrySet()) {
			Relationship targetRelationship = entry.getKey();
			List<RelationshipCorrespondence> correspondences = entry.getValue();

			// A correspondence set is a candidate if we the correspondences have the same start node but different, non-null end nodes.
			Map<Node, Collection<Node>> connectedNodes = new HashMap<>();
			for (RelationshipCorrespondence correspondence : correspondences) {
				Relationship sourceRelationship = correspondence.getSourceRelationship();
				if (!sourceRelationship.isEmpty()) {
					CollectionUtil.addChecked(sourceRelationship.getStartNode(), sourceRelationship.getEndNode(),
							connectedNodes);
				}
			}

			for (Map.Entry<Node, Collection<Node>> connectedNodesEntry : connectedNodes.entrySet()) {
				Node startNode = connectedNodesEntry.getKey();
				Collection<Node> endNodes = connectedNodesEntry.getValue();
				// store these relationships for reconciliation, registered with the target end node
				List<Relationship> reconciliationCluster = new ArrayList<>();
				for (RelationshipCorrespondence correspondence : correspondences) {
					Relationship sourceRelationship = correspondence.getSourceRelationship();
					if (startNode.equals(sourceRelationship.getStartNode())
							&& endNodes.contains(sourceRelationship.getEndNode())) {
						reconciliationCluster.add(sourceRelationship);
					}
				}
				Node targetEndNode = targetRelationship.getEndNode();
				Map<Relationship, List<Relationship>> clusters = reconciliationClustersByTargetEndNode
						.get(targetEndNode);
				if (clusters == null) {
					clusters = new HashMap<>();
					reconciliationClustersByTargetEndNode.put(targetEndNode, clusters);
				}
				clusters.put(targetRelationship, reconciliationCluster);

				if (endNodes.size() > 1) {
					targetReconciliationNodes.add(targetEndNode);
				}
			}

		}

		// Phase 2: Check if every ambiguous node fits better than all others in a certain context.
		for (Node targetReconciliationNode : targetReconciliationNodes) {
			Map<Relationship, List<Relationship>> reconciliationClusters = reconciliationClustersByTargetEndNode
					.get(targetReconciliationNode);
			logger.debug("Reconcile {} with {}.", targetReconciliationNode, reconciliationClusters);

			Set<Node> nonPreferredNodes = new HashSet<>();
			for (List<Relationship> reconciliationCluster : reconciliationClusters.values()) {
				for (Relationship sourceRelationship : reconciliationCluster) {
					nonPreferredNodes.add(sourceRelationship.getEndNode());
				}
			}

			// Find a preferred source node for each cluster.
			Map<Relationship, Map<Node, Node>> preferredEndNodes = new HashMap<>();
			for (Map.Entry<Relationship, List<Relationship>> reconciliationClustersEntry : reconciliationClusters
					.entrySet()) {

				Relationship targetRelationship = reconciliationClustersEntry.getKey();
				List<Relationship> reconciliationCluster = reconciliationClustersEntry.getValue();

				PartialOrderQueue<ConstraintPair> partialOrderQueue = new PartialOrderQueue<>(ConstraintPair.ORDER);
				for (Relationship sourceRelationship : reconciliationCluster) {
					Constraint constraint = this.sourceSchema.getConstraints(sourceRelationship).toConstraint();
					Constraint inverseConstraint = this.sourceSchema.getConstraints(sourceRelationship.getInverse())
							.toConstraint();
					partialOrderQueue.add(new ConstraintPair(constraint, inverseConstraint));
				}
				partialOrderQueue.retainSmallest();

				List<ConstraintPair> retainedConstraintPairs = partialOrderQueue.toList();
				if (retainedConstraintPairs.size() < reconciliationCluster.size()) {
					logger.debug("Reduced {} to {}.", reconciliationCluster, retainedConstraintPairs);
				}

				Node preferredEndNode = null;
				for (ConstraintPair retainedConstraintPair : retainedConstraintPairs) {
					Node endNode = retainedConstraintPair.base.getRelationship().getEndNode();
					if (preferredEndNode == null) {
						preferredEndNode = endNode;
					} else if (!endNode.equals(preferredEndNode)) {
						preferredEndNode = null;
					}
				}
				if (preferredEndNode != null) {
					nonPreferredNodes.remove(preferredEndNode);
					Map<Node, Node> endNodePreferences = preferredEndNodes.get(targetRelationship);
					if (endNodePreferences == null) {
						endNodePreferences = new HashMap<>();
						preferredEndNodes.put(targetRelationship, endNodePreferences);
					}
					Node startNode = CollectionUtil.getAny(reconciliationCluster).getStartNode();
					endNodePreferences.put(startNode, preferredEndNode);
				}
				logger.debug("Preferred end node of current cluster: {}", preferredEndNode);

			}
			// If every source node is preferable in some context, we can trim our relationship mapping.
			if (nonPreferredNodes.isEmpty()) {
				for (Map.Entry<Relationship, Map<Node, Node>> endNodePreferenceEntry : preferredEndNodes.entrySet()) {
					Relationship targetRelationship = endNodePreferenceEntry.getKey();
					Map<Node, Node> endNodePreferences = endNodePreferenceEntry.getValue();
					boolean isRelationshipAffected = false;
					
					Set<Node> removedNodes = new HashSet<>();
					for (Iterator<RelationshipCorrespondence> i = this.resultMapping
							.getCorrespondencesForTargetRelationship(targetRelationship).iterator(); i.hasNext();) {
						RelationshipCorrespondence correspondence = i.next();
						Relationship sourceRelationship = correspondence.getSourceRelationship();
						Node sourceStartNode = sourceRelationship.getStartNode();
						Node sourceEndNode = sourceRelationship.getEndNode();
						Node preferredEndNode = endNodePreferences.get(sourceStartNode);
						if (preferredEndNode != null && !preferredEndNode.equals(sourceEndNode)) {
							i.remove();
							removedNodes.add(sourceEndNode);
							logger.info("Reconcile {}: remove {} due to global reconciliation.",
									correspondence.getTargetRelationship(), sourceRelationship);
							isRelationshipAffected = true;
						}
					}

					for (Iterator<RelationshipCorrespondence> i = this.resultMapping
							.getCorrespondencesForTargetRelationship(targetRelationship.getInverse()).iterator(); i
							.hasNext();) {
						RelationshipCorrespondence inverseCorrespondence = i.next();
						Relationship sourceRelationship = inverseCorrespondence.getSourceRelationship();
						Node sourceStartNode = sourceRelationship.getStartNode();
						Node sourceEndNode = sourceRelationship.getEndNode();
						Node preferredStartNodeNode = endNodePreferences.get(sourceEndNode);
						if (preferredStartNodeNode != null && !preferredStartNodeNode.equals(sourceStartNode)) {
							i.remove();
							removedNodes.add(sourceStartNode);
							logger.info("Reconcile {}: remove {} due to global reconciliation.",
									inverseCorrespondence.getTargetRelationship(),
									inverseCorrespondence.getSourceRelationship());
							isRelationshipAffected = true;
						}
					}
					
					for (Node removedNode : removedNodes) {
						this.resultMapping.addCorrespondence(new RelationshipCorrespondence(EmptyRelationship.createWithEndNode(removedNode), targetRelationship));
						this.resultMapping.addCorrespondence(new RelationshipCorrespondence(EmptyRelationship.createWithStartNode(removedNode), targetRelationship.getInverse()));
					}
					
					if (isRelationshipAffected) {
						logger.info("New correspondences: {}", this.resultMapping.getCorrespondencesForTargetRelationship(targetRelationship));
					}
				}
			}
		}
	}
	
	private void matchWithInverse(Relationship targetRelationship) {
		Relationship inverseTargetRelationship = targetRelationship.getInverse();
		boolean relationshipProcessed = this.handledRelationships.contains(targetRelationship);
		boolean inverseProcessed = this.handledRelationships.contains(inverseTargetRelationship);

		if (relationshipProcessed ^ inverseProcessed) throw new IllegalStateException(
				"Relationships and their inverse not processed together.");

		if (!relationshipProcessed) {
			this.handledRelationships.add(targetRelationship);
			this.handledRelationships.add(inverseTargetRelationship);
		} else {
			logger.trace("Avoiding duplicate handling of {}", targetRelationship);
			return;
		}

		logger.debug("Trying to match {}.", targetRelationship);

		List<Node> startSourceNodes = findSourceNodes(targetRelationship.getStartNode());
		List<Node> endSourceNodes = findSourceNodes(targetRelationship.getEndNode());
		logger.trace("Start nodes: {}, end nodes: {}", startSourceNodes, endSourceNodes);
		boolean isIdentityRelationship = targetRelationship.getContentType() == ContentType.EQUALITY;
		
		
		if (startSourceNodes.isEmpty() && endSourceNodes.isEmpty()) {
			// No relationship.
		} else if (!startSourceNodes.isEmpty() && endSourceNodes.isEmpty()) {
			for (Node startSourceNode : startSourceNodes) {
				RelationshipCorrespondence correspondence = new RelationshipCorrespondence(
						EmptyRelationship.createWithStartNode(startSourceNode), targetRelationship);
				this.resultMapping.addCorrespondence(correspondence);
				correspondence = new RelationshipCorrespondence(EmptyRelationship.createWithEndNode(startSourceNode),
						targetRelationship.getInverse());
				this.resultMapping.addCorrespondence(correspondence);
			}
		} else if (startSourceNodes.isEmpty() && !endSourceNodes.isEmpty()) {
			for (Node endSourceNode : endSourceNodes) {
				RelationshipCorrespondence correspondence = new RelationshipCorrespondence(
						EmptyRelationship.createWithEndNode(endSourceNode), targetRelationship);
				this.resultMapping.addCorrespondence(correspondence);
				correspondence = new RelationshipCorrespondence(EmptyRelationship.createWithStartNode(endSourceNode),
						targetRelationship.getInverse());
				this.resultMapping.addCorrespondence(correspondence);
			}

		} else {
			// Both start and end source nodes exist.
			logger.trace("Need further evaluation of {}: {} source nodes and {} target nodes.", targetRelationship,
					startSourceNodes.size(), endSourceNodes.size());

			// Search constraints pairs (i.e. bidirectional constraint pairs)
			// between nodes.
			Map<NodePair, Collection<Constraint>> sourceRelationships = findSourceRelationships(startSourceNodes,
					endSourceNodes, isIdentityRelationship);

			// If we have no pairs, add an empty (two half) relationship.
			if (sourceRelationships.isEmpty()) {
				for (Node startSourceNode : startSourceNodes) {
					EmptyRelationship sourceRelationship = EmptyRelationship.createWithStartNode(startSourceNode);
					RelationshipCorrespondence correspondence = new RelationshipCorrespondence(sourceRelationship,
							targetRelationship);
					this.resultMapping.addCorrespondence(correspondence);
					correspondence = new RelationshipCorrespondence(sourceRelationship.getInverse(),
							inverseTargetRelationship);
					this.resultMapping.addCorrespondence(correspondence);
				}
				for (Node endSourceNode : endSourceNodes) {
					EmptyRelationship sourceRelationship = EmptyRelationship.createWithEndNode(endSourceNode);
					RelationshipCorrespondence correspondence = new RelationshipCorrespondence(sourceRelationship,
							targetRelationship);
					this.resultMapping.addCorrespondence(correspondence);
					correspondence = new RelationshipCorrespondence(sourceRelationship.getInverse(),
							inverseTargetRelationship);
					this.resultMapping.addCorrespondence(correspondence);
				}
				return;
			}

			// Otherwise, remove unstable paths.
			Map<NodePair, Collection<ConstraintPair>> constraintPairs = new HashMap<>();
			Map<Node, Collection<Constraint>> constraintsPerNode = new HashMap<>();
			for (Entry<NodePair, Collection<Constraint>> entry : sourceRelationships.entrySet()) {
				NodePair nodePair = entry.getKey();
				Collection<Constraint> sourceConstraints = entry.getValue();
				List<ConstraintPair> sourceRelationshipPairs = addInvertSourceRelationships(nodePair, sourceConstraints, isIdentityRelationship);
				constraintPairs.put(nodePair, sourceRelationshipPairs);

				for (ConstraintPair constraintPair : sourceRelationshipPairs) {
					addConstraintToStartNode(constraintsPerNode, constraintPair.base);
					addConstraintToStartNode(constraintsPerNode, constraintPair.invert);
				}

				if (!sourceRelationshipPairs.isEmpty()) {
					logger.debug("Found relationships for {}:", targetRelationship);
					for (ConstraintPair x : sourceRelationshipPairs)
						logger.debug(" -> {}", x);
				}

			}

			if (constraintPairs.size() > 1) logger.debug("Multiple constraints found: {}", constraintPairs);

			// Reduce set of constraint pairs and add them retained constraints
			// to the mapping.
			for (Entry<NodePair, Collection<ConstraintPair>> entry : constraintPairs.entrySet()) {
				for (ConstraintPair constraintPair : entry.getValue()) {
					if (!hasBetterConstraint(constraintPair.base, constraintsPerNode)
							|| !hasBetterConstraint(constraintPair.invert, constraintsPerNode)) {
						RelationshipCorrespondence correspondence = new RelationshipCorrespondence(
								constraintPair.base.getRelationship(), targetRelationship);
						this.resultMapping.addCorrespondence(correspondence);

						correspondence = new RelationshipCorrespondence(constraintPair.invert.getRelationship(),
								targetRelationship.getInverse());
						this.resultMapping.addCorrespondence(correspondence);
					} else {
						logger.trace("Dropping {} in favor of another constraint pair.", constraintPair);
					}
				}
			}

		}
	}

	private void addConstraintToStartNode(Map<Node, Collection<Constraint>> constraintsPerNode, Constraint base) {
		Node startNode = base.getRelationship().getStartNode();
		Collection<Constraint> nodeConstraints = constraintsPerNode.get(startNode);
		if (nodeConstraints == null) {
			nodeConstraints = new ArrayList<>();
			constraintsPerNode.put(startNode, nodeConstraints);
		}
		nodeConstraints.add(base);
	}

	private boolean hasBetterConstraint(Constraint constraintUnderTest,
			Map<Node, Collection<Constraint>> constraintsPerNode) {
		Node startNode = constraintUnderTest.getRelationship().getStartNode();
		Collection<Constraint> nodeConstraints = constraintsPerNode.get(startNode);
		for (Constraint constraint : nodeConstraints) {
			if (ConstraintOrder.INSTANCE.compare(constraint, constraintUnderTest) == PartialOrder.Type.SMALLER) return true;
		}
		return false;
	}

	/** TODO: Find a better implementation for this method. 
	 * @param isIdentityRelationship */
	private List<ConstraintPair> addInvertSourceRelationships(NodePair sourceNodePair,
			Collection<Constraint> sourceRelationships, boolean isIdentityRelationship) {
		List<ConstraintPair> constraintPairs = new ArrayList<>();

		// Is it necessary that a relationship in two directions is found? I
		// think not. It might be, that the path search algorithm drops the
		// inverse of a good path in favor of another one.
		// This issue is neglected for now.
		Collection<Constraint> invertSourceRelationships = this.sourceSchema.findCardinality(
				sourceNodePair.getEndNode(), sourceNodePair.getStartNode(), isIdentityRelationship);
		invertSourceRelationships = new ArrayList<>(invertSourceRelationships);

		// Put the invert source relationships together with the original ones.
		InvertSourceRelationshipLoop: for (Constraint constraint : sourceRelationships) {
			Relationship relationship = constraint.getRelationship();
			Relationship invertedRelationship = relationship.getInverse();
			for (Iterator<Constraint> i = invertSourceRelationships.iterator(); i.hasNext();) {
				Constraint inverseConstraint = i.next();
				Relationship inverse = inverseConstraint.getRelationship();
				if (invertedRelationship.equals(inverse)) {
					constraintPairs.add(new ConstraintPair(constraint, inverseConstraint));
					i.remove();
					continue InvertSourceRelationshipLoop;
				}

			}

			// If not found, we add manually.
			Constraint inverseConstraint = this.sourceSchema.getConstraints(invertedRelationship).toConstraint();
			constraintPairs.add(new ConstraintPair(constraint, inverseConstraint));
			logger.debug("Could not find inverse of {}, so adding it with {}.", constraint, inverseConstraint);
		}

		// Assign left-over inverse relationships.
		for (Constraint inverseConstraint : invertSourceRelationships) {
			Relationship inverse = inverseConstraint.getRelationship();
			Constraint constraint = this.sourceSchema.getConstraints(inverse.getInverse()).toConstraint();
			constraintPairs.add(new ConstraintPair(constraint, inverseConstraint));
			logger.debug("Adding left-over node pair {} and {}", constraint, inverseConstraint);
		}

		// Remove superfluous pairs? These should actually not exist as the
		// search algorithm is already optimizing. But let's do it to be sure.
		PartialOrderQueue<ConstraintPair> partialOrderQueue = new PartialOrderQueue<>(ConstraintPair.ORDER);
		partialOrderQueue.addAll(constraintPairs);
		partialOrderQueue.retainSmallest();

		return partialOrderQueue.toList();
	}

	private List<Node> findSourceNodes(Node targetNode) {
		Collection<NodeCorrespondence> correspondences = this.resultMapping.getCorrespondencesForTargetNode(targetNode);
		List<Node> sourceNodes = new ArrayList<>();
		for (NodeCorrespondence correspondence : correspondences) {
			sourceNodes.add(correspondence.getSourceNode());
		}
		return sourceNodes;
	}

	private Map<NodePair, Collection<Constraint>> findSourceRelationships(List<Node> startSourceNodes,
			List<Node> endSourceNodes, boolean searchIdentityRelationship) {
		Map<NodePair, Collection<Constraint>> result = new HashMap<>();

		for (Node startSourceNode : startSourceNodes) {
			for (Node endSourceNode : endSourceNodes) {
				Collection<Constraint> constraints = this.sourceSchema.findCardinality(startSourceNode, endSourceNode, searchIdentityRelationship);
				if (!constraints.isEmpty()) {
					NodePair nodePair = new NodePair(startSourceNode, endSourceNode);
					result.put(nodePair, constraints);
				}
			}
		}

		return result;
	}

	/**
	 * Helper class to distinguish gathered relationships in source schemas by their end and start node.
	 */
	public static class NodePair {

		private Node startNode;
		private Node endNode;

		public NodePair(Node startNode, Node endNode) {
			super();
			this.startNode = startNode;
			this.endNode = endNode;
		}

		public NodePair invert() {
			return new NodePair(this.endNode, this.startNode);
		}

		public Node getStartNode() {
			return startNode;
		}

		public Node getEndNode() {
			return endNode;
		}

		@Override
		public String toString() {
			return "NodePair [startNode=" + startNode + ", endNode=" + endNode + "]";
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((endNode == null) ? 0 : endNode.hashCode());
			result = prime * result + ((startNode == null) ? 0 : startNode.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj) return true;
			if (obj == null) return false;
			if (getClass() != obj.getClass()) return false;
			NodePair other = (NodePair) obj;
			if (endNode == null) {
				if (other.endNode != null) return false;
			} else if (!endNode.equals(other.endNode)) return false;
			if (startNode == null) {
				if (other.startNode != null) return false;
			} else if (!startNode.equals(other.startNode)) return false;
			return true;
		}

	}

	public static class ConstraintPair {

		private Constraint base, invert;

		public ConstraintPair(Constraint base, Constraint invert) {
			super();
			this.base = base;
			this.invert = invert;
		}

		public static final PartialOrder<ConstraintPair> ORDER = new PartialOrder<RelationshipMatcher.ConstraintPair>() {

			@Override
			public PartialOrder.Type compare(ConstraintPair a, ConstraintPair b) {
				ConstraintOrder constraintOrder = ConstraintOrder.INSTANCE;
				PartialOrder.Type baseOrder = constraintOrder.compare(a.base, b.base);
				PartialOrder.Type invertOrder = constraintOrder.compare(a.invert, b.invert);

				if (baseOrder == invertOrder) return baseOrder;

				if ((baseOrder == PartialOrder.Type.EQUAL || baseOrder == PartialOrder.Type.SMALLER)
						&& (invertOrder == PartialOrder.Type.EQUAL || invertOrder == PartialOrder.Type.SMALLER)) {
					return PartialOrder.Type.SMALLER;
				}

				if ((baseOrder == PartialOrder.Type.EQUAL || baseOrder == PartialOrder.Type.GREATER)
						&& (invertOrder == PartialOrder.Type.EQUAL || invertOrder == PartialOrder.Type.GREATER)) {
					return PartialOrder.Type.GREATER;
				}

				return PartialOrder.Type.UNORDERED;
			}

		};

		@Override
		public String toString() {
			return "ConstraintPair [base=" + base + ", invert=" + invert + "]";
		}

	}

}
