package de.uni_potsdam.hpi.isg.efes.csg.mapping;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import uk.ac.shef.wit.simmetrics.similaritymetrics.AbstractStringMetric;
import uk.ac.shef.wit.simmetrics.similaritymetrics.Levenshtein;
import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.IdentityNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.util.RelationalSchemaToCSGConverter;
import de.uni_potsdam.hpi.isg.utils.util.CollectionUtil;

public class TableNodeMatcher {

	private static final Logger logger = LogManager
			.getLogger(TableNodeMatcher.class.getSimpleName());

	public void enrich(BasicSchemaGraph source, BasicNodeMapping mapping,
			BasicSchemaGraph target) {

		// First pass: Match tables with attributes involved in correspondences.
		for (Node targetTableNode : target.getAtomicNodes().values()) {
			if (!(targetTableNode instanceof IdentityNode)) {
				continue;
			}

			Collection<TableLink> candidates = findLinkedTables(target,
					targetTableNode, mapping, source);
			Node bestNode = findMostSimilarNamedTable(targetTableNode,
					candidates);

			if (bestNode != null) {
				NodeCorrespondence newCorrespondence = new NodeCorrespondence(
						bestNode, targetTableNode);
				mapping.addCorrespondence(newCorrespondence);
				logger.info("Matched {} to {}.", targetTableNode, bestNode);
			} else {
				logger.debug("No corresponding node for {}.", targetTableNode);
			}
		}

//		// Second pass: Match join tables.
//		for (Node targetTableNode : target.getAtomicNodes().values()) {
//			if (!(targetTableNode instanceof IdentityNode)) {
//				continue;
//			}
//
//			// Tables that are already matched are no problem.
//			if (!mapping.getCorrespondencesForTargetNode(targetTableNode)
//					.isEmpty()) {
//				continue;
//			}
//
//			// Heuristic for join tables:
//			// It must reference two matched tables.
//			List<Node> referencedTableNodes = findReferencedTableNodes(target,
//					targetTableNode);
//			for (Iterator<Node> i = referencedTableNodes.iterator(); i
//					.hasNext();) {
//				Node referencedTableNode = i.next();
//				if (mapping
//						.getCorrespondencesForTargetNode(referencedTableNode)
//						.isEmpty()) {
//					i.remove();
//				}
//			}
//
//			if (referencedTableNodes.size() == 2) {
//				logger.info("Seems to be a valid join table: {}",
//						targetTableNode);
//				// Now we need to determine to which node this node shall be
//				// matched.
//				// Pick the appropriate source table nodes.
//				Node sourceTableNode1 = CollectionUtil
//						.getAny(mapping
//								.getCorrespondencesForTargetNode(referencedTableNodes
//										.get(0))).getSourceNode();
//				Node sourceTableNode2 = CollectionUtil
//						.getAny(mapping
//								.getCorrespondencesForTargetNode(referencedTableNodes
//										.get(1))).getSourceNode();
//				logger.info("Now regarding {} and {}.", sourceTableNode1, sourceTableNode2);
//			}
//		}
	}

	/**
	 * From the candidates, select the one, that has the most similar name to
	 * the target table.
	 */
	private Node findMostSimilarNamedTable(Node targetTableNode,
			Collection<TableLink> candidates) {
		Node bestNode = null;
		if (candidates.size() == 1) {
			bestNode = CollectionUtil.getAny(candidates).getSourceTableNode();
		} else {
			// AbstractStringMetric metric = new Levenshtein();
			// float maxSimilarity = -1;
			// for (TableLink candidate : candidates) {
			// Node candidateNode = candidate.getSourceTableNode();
			// float similarity = metric.getSimilarity(candidateNode.getName(),
			// targetTableNode.getName());
			// if (maxSimilarity < similarity) {
			// maxSimilarity = similarity;
			// bestNode = candidateNode;
			// }
			// }
			AbstractStringMetric metric = new Levenshtein();
			float maxScore = Float.NEGATIVE_INFINITY;
			for (TableLink candidate : candidates) {

				Node sourceTableNode = candidate.getSourceTableNode();
				String sourceTableName = RelationalSchemaToCSGConverter
						.extractTableName(sourceTableNode);
				String targetTableName = RelationalSchemaToCSGConverter
						.extractTableName(targetTableNode);
				float tableSimilarity = metric.getSimilarity(sourceTableName,
						targetTableName);

				float sumAttributeSimilarity = 0;
				for (NodeCorrespondence correspondence : candidate
						.getAttributeCorrespondences()) {
					String sourceAttributeName = CollectionUtil
							.getAny(RelationalSchemaToCSGConverter
									.extractAttributeNames(correspondence
											.getSourceNode()));
					String targetAttributeName = CollectionUtil
							.getAny(RelationalSchemaToCSGConverter
									.extractAttributeNames(correspondence
											.getTargetNode()));
					float attributeSimilarity = metric.getSimilarity(
							sourceAttributeName, targetAttributeName);
					sumAttributeSimilarity += attributeSimilarity;
				}

				float attributeTablePunishment = 0;
				if (candidate.getAttributeCorrespondences().size() == 1) {
					NodeCorrespondence correspondence = CollectionUtil
							.getAny(candidate.getAttributeCorrespondences());
					String targetAttributeName = CollectionUtil
							.getAny(RelationalSchemaToCSGConverter
									.extractAttributeNames(correspondence
											.getTargetNode()));
					attributeTablePunishment = metric.getSimilarity(
							sourceTableName, targetAttributeName);
				}

				float score = tableSimilarity + sumAttributeSimilarity
						- attributeTablePunishment;
				if (score > maxScore) {
					maxScore = score;
					bestNode = sourceTableNode;
				}
			}
		}
		return bestNode;
	}

	/**
	 * For the table node from the target, find those table nodes from the
	 * sources, that have the most corresponding attributes with the target
	 * table's attributes.
	 */
	private Collection<TableLink> findLinkedTables(BasicSchemaGraph target,
			Node targetTableNode, BasicNodeMapping mapping,
			BasicSchemaGraph source) {

		Map<Node, TableLink> tableLinksPerSourceTable = new HashMap<>();
		List<Node> targetAttributeNodes = findAttributeNodes(target,
				targetTableNode);
		for (Node targetAttributeNode : targetAttributeNodes) {
			// Find corresponding nodes.
			Collection<NodeCorrespondence> correspondences = mapping
					.getCorrespondencesForTargetNode(targetAttributeNode);
			for (NodeCorrespondence correspondence : correspondences) {
				Node sourceNode = correspondence.getSourceNode();
				Node sourceTableNode = findTableNode(source, sourceNode);
				TableLink tableLink = tableLinksPerSourceTable
						.get(sourceTableNode);
				if (tableLink == null) {
					tableLink = new TableLink(sourceTableNode);
					tableLinksPerSourceTable.put(sourceNode, tableLink);
				}
				tableLink.addAttributeCorrespondence(correspondence);
			}
		}

		return tableLinksPerSourceTable.values();
	}

	/**
	 * Returns the table node for a given attribute node.
	 * 
	 * @param graph
	 * @param node
	 * @return
	 */
	private Node findTableNode(BasicSchemaGraph graph, Node node) {
		if (node instanceof IdentityNode) {
			return node;
		}
		Collection<Relationship> relationships = graph
				.getAtomicRelationships(node);
		for (Relationship relationship : relationships) {
			if (relationship.isAtomic()
					&& relationship.getEndNode() instanceof IdentityNode)
				return relationship.getEndNode();
		}
		throw new RuntimeException(
				"There should have been a node found actually.");
	}

	/**
	 * Returns the attribute nodes for a given table node.
	 * 
	 * @param schemaGraph
	 * @param tableNode
	 * @return
	 */
	private List<Node> findAttributeNodes(BasicSchemaGraph schemaGraph,
			Node tableNode) {
		List<Node> result = new ArrayList<>();
		Collection<Relationship> relationships = schemaGraph
				.getAtomicRelationships(tableNode);
		for (Relationship relationship : relationships) {
			if (relationship.isAtomic()) {
				Node endNode = relationship.getEndNode();
				if (endNode.isSimpleValueNode()) {
					result.add(endNode);
				}
			}
		}
		return result;
	}

	@SuppressWarnings("unused")
	private List<Node> findReferencedTableNodes(BasicSchemaGraph graph,
			Node referencingTableNode) {

		List<Node> referencedTableNodes = new ArrayList<>();
		List<Node> attribteNodes = findAttributeNodes(graph,
				referencingTableNode);
		for (Node attributeNode : attribteNodes) {
			for (Relationship relationship : graph
					.getAtomicRelationships(attributeNode)) {
				if (relationship.isExplicit()) {
					continue;
				}

				NaturalNumberSet cardinality = graph.getConstraints(
						relationship).calculateEffectiveCardinality();
				if (cardinality.isSubsetOf(NaturalNumberSet.ONE)) {
					Node referencedAttributeNode = relationship.getEndNode();
					referencedTableNodes.add(findTableNode(graph,
							referencedAttributeNode));
				}
			}
		}
		return referencedTableNodes;
	}

	private static class TableLink {

		private Node sourceTableNode;
		private List<NodeCorrespondence> attributeCorrespondences;

		public TableLink(Node sourceTableNode) {
			super();
			this.sourceTableNode = sourceTableNode;
			this.attributeCorrespondences = new ArrayList<>();
		}

		public Node getSourceTableNode() {
			return sourceTableNode;
		}

		void addAttributeCorrespondence(NodeCorrespondence correspondence) {
			this.attributeCorrespondences.add(correspondence);
		}

		public List<NodeCorrespondence> getAttributeCorrespondences() {
			return attributeCorrespondences;
		}

	}
}
