package de.uni_potsdam.hpi.isg.efes.csg.scenario.adapter;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.BasicNodeMapping;
import de.uni_potsdam.hpi.isg.efes.csg.mapping.NodeCorrespondence;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.DataSet;
import de.uni_potsdam.hpi.isg.efes.csg.scenario.IntegrationScenario;
import de.uni_potsdam.hpi.isg.efes.csg.util.RelationalSchemaToCSGConverter;
import de.uni_potsdam.hpi.isg.efes.spicy.SpicyConfiguration;
import de.uni_potsdam.hpi.isg.efes.spicy.SpicyConfiguration.RelationalDataset;
import de.uni_potsdam.hpi.isg.efes.spicy.SpicyConfiguration.RelationalPath;
import de.uni_potsdam.hpi.isg.efes.spicy.extensions.TableMappingConfiguration;
import de.uni_potsdam.hpi.isg.efes.spicy.extensions.TableMappingConfiguration.TableMapping;
import de.uni_potsdam.hpi.isg.utils.db.Database;
import de.uni_potsdam.hpi.isg.utils.settings.DatabaseSettings;

public class SpicyConfigurationScenarioBuilder {

	private static final Logger logger = LogManager
			.getLogger(SpicyConfigurationScenarioBuilder.class.getSimpleName());
	private List<SpicyConfiguration> spicyConfigurations = new ArrayList<SpicyConfiguration>();
	private List<TableMappingConfiguration> tableMappingConfigurations = new ArrayList<>();

	public void add(SpicyConfiguration spicyConfiguration) {
		if (spicyConfiguration == null)
			throw new IllegalArgumentException();

		this.spicyConfigurations.add(spicyConfiguration);
	}

	public void add(TableMappingConfiguration tableMappingConfiguration) {
		this.tableMappingConfigurations.add(tableMappingConfiguration);
	}

	public IntegrationScenario build() throws SQLException {
		if (this.spicyConfigurations.isEmpty()) {
			throw new IllegalStateException();
		}

		DataSet targetDataSet = null;
		SpicyConfiguration.RelationalDataset target = null;
		IntegrationScenario result = null;
		for (SpicyConfiguration spicyConfiguration : this.spicyConfigurations) {
			RelationalDataset newTarget = spicyConfiguration.getTarget();
			if (result == null) {
				targetDataSet = createDataSet(newTarget);
				result = new IntegrationScenario(targetDataSet);
				target = newTarget;
			} else if (!target.equals(newTarget)) {
				throw new IllegalStateException(
						"Spicy configurations have different targets.");
			}
			DataSet sourceDataSet = createDataSet(spicyConfiguration
					.getSource());
			BasicNodeMapping mapping = createMapping(sourceDataSet,
					targetDataSet, spicyConfiguration.getCorrespondences());
			result.addSource(sourceDataSet, mapping);
		}

		// Process additional table mappings.
		for (TableMappingConfiguration tableMappingConfiguration : this.tableMappingConfigurations) {
			for (TableMapping tableMapping : tableMappingConfiguration
					.getTableMappings()) {
				String sourceNodeName = tableMapping.getSource();

				// Find the source for the source node name.
				DataSet source = null;
				Node sourceNode = null;
				for (DataSet dataSet : result.getSourceDataSets()) {
					BasicSchemaGraph schema = dataSet.getSchema();
					if (schema.getAtomicNodes().containsKey(sourceNodeName)) {
						source = dataSet;
						sourceNode = schema.getAtomicNodes()
								.get(sourceNodeName);
						break;
					}
				}

				if (source == null) {
					throw new IllegalArgumentException(
							String.format(
									"Could not find the node '%s' in the source data set.",
									tableMapping.getSource()));
				}

				BasicSchemaGraph targetSchema = targetDataSet.getSchema();
				Node targetNode = targetSchema.getAtomicNodes().get(
						tableMapping.getTarget());
				if (targetNode == null) {
					throw new IllegalArgumentException(
							String.format(
									"Could not find the node '%s' in the target data set.",
									tableMapping.getTarget()));
				}

				BasicNodeMapping mapping = result.getMappings().get(source);
				mapping.addCorrespondence(new NodeCorrespondence(sourceNode,
						targetNode));
			}
		}

		return result;
	}

	private DataSet createDataSet(RelationalDataset spicyDataSet)
			throws SQLException {
		String dbUri = spicyDataSet.getDbUri();
		DatabaseSettings settings = DatabaseSettings.from(dbUri,
				spicyDataSet.getDbLogin(), spicyDataSet.getDbPassword());
		Database<?> db = Database.newInstance(settings.toProperties());
		DataSet dataSet = new DataSet(null, DataSet.Type.RELATIONAL, db);
		return dataSet;
	}

	private BasicNodeMapping createMapping(DataSet source, DataSet target,
			List<SpicyConfiguration.Correspondence> correspondences) {
		BasicNodeMapping result = new BasicNodeMapping(source);

		for (SpicyConfiguration.Correspondence spicyCorrespondence : correspondences) {
			if (spicyCorrespondence.getSourcePaths().size() > 1) {
				logger.warn("Cannot handle all correspondences (yet) in {}",
						spicyCorrespondence.getSourcePaths());
			}

			String rawSourcePath = spicyCorrespondence.getSourcePaths().get(0);
			Node sourceNode = findNodeForPath(source, rawSourcePath);

			String rawTargetPath = spicyCorrespondence.getTargetPath();
			Node targetNode = findNodeForPath(target, rawTargetPath);

			NodeCorrespondence correspondence = new NodeCorrespondence(
					sourceNode, targetNode);
			result.addCorrespondence(correspondence);
		}
		return result;
	}

	private Node findNodeForPath(DataSet dataSet, String rawPath) {
		RelationalPath sourcePath = SpicyConfiguration
				.resolveRelationalPath(rawPath);
		String nodeName = RelationalSchemaToCSGConverter.createColumnNodeName(
				dataSet.getSchema(), sourcePath.getTable(),
				sourcePath.getAttribute());
		Node node = dataSet.getSchema().getNode(nodeName);
		return node;
	}

}
