package de.uni_potsdam.hpi.isg.efes.csg;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.ConstraintCollection;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.CrossProductNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Datatype;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.IdentityNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node.Operation;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.ValueNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.CoalesceRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.EqualityRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.ExplicitAtomicRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.search.ConstraintOrder;
import de.uni_potsdam.hpi.isg.efes.csg.search.PartialOrderGraphSearcher;
import de.uni_potsdam.hpi.isg.utils.util.CollectionUtil;
import de.uni_potsdam.hpi.isg.utils.util.MultiLevelIterator;

public class BasicSchemaGraph implements SchemaGraph {

	private static Logger logger = LogManager.getLogger(BasicSchemaGraph.class.getSimpleName());

	private String name;
	private Map<String, Node> atomicNodes = new HashMap<>();
	private Map<Node, Collection<Relationship>> atomicRelationships = new HashMap<>();
	private Map<Node, Collection<Relationship>> virtualRelationships = new HashMap<>();
	private Map<Relationship, ConstraintCollection> constraintCollections = new HashMap<>();
	private PartialOrderGraphSearcher graphSearcher;

	public BasicSchemaGraph(String name) {
		this.name = name;
	}

	public boolean add(Node node) {
		if (node == null) {
			throw new IllegalArgumentException();

		} else if (node.getOperationType() != Operation.NONE) {
			boolean anyNodeAdded = false;
			for (Node parentNode : node.getParentNodes()) {
				anyNodeAdded |= add(parentNode);
			}
			return anyNodeAdded;
		}

		assertNamePrefixCorrect(node);
		return atomicNodes.put(node.getName(), node) != null;
	}

	public IdentityNode addIdentityNode(String name) {
		IdentityNode node = new IdentityNode(name);
		add(node);
		return node;
	}

	public ValueNode addValueNode(String name, Datatype datatype) {
		ValueNode node = new ValueNode(name, datatype);
		add(node);
		return node;
	}

	public boolean add(Relationship relationship) {
		if (!relationship.isAtomic()) {
			logger.debug("Adding non-atomic relationship {}.", relationship);
			for (Relationship atomicRelationship : relationship.getAtomicRelationships()) {
				add(atomicRelationship);
			}
			return addVirtualRelationship(relationship);
		}

		if (contains(relationship)) {
			logger.debug("Not adding {}, because it is already present in the schema graph.", relationship);
			return false;
		}

		Node startNode = relationship.getStartNode();
		assertNodeContained(startNode);

		Node endNode = relationship.getEndNode();
		assertNodeContained(endNode);

		CollectionUtil.add(startNode, relationship, this.atomicRelationships);

		return true;
	}

	private boolean addVirtualRelationship(Relationship relationship) {
		return CollectionUtil.addChecked(relationship.getStartNode(), relationship, this.virtualRelationships);
	}

	public void addExplicitRelationships(Node node1, Node node2) {
		Relationship relationship1to2 = new ExplicitAtomicRelationship(node1, node2);
		add(relationship1to2);
		add(relationship1to2.getInverse());
	}

	public void addEqualityRelationships(Node node1, Node node2) {
		Relationship relationship1to2 = new EqualityRelationship(node1, node2);
		add(relationship1to2);
		add(relationship1to2.getInverse());
	}

	public void addConstraint(Node startNode, Node endNode, NaturalNumberSet cardinality) {
		Relationship relationship = getAtomicRelationship(startNode, endNode);
		addConstraint(relationship, cardinality);
	}

	public void addConstraint(Relationship relationship, NaturalNumberSet cardinality) {
		assertRelationshipContained(relationship);

		ConstraintCollection constraintCollection = this.constraintCollections.get(relationship);
		if (constraintCollection == null) {
			constraintCollection = new ConstraintCollection(this, relationship);
			this.constraintCollections.put(relationship, constraintCollection);
		}
		constraintCollection.add(cardinality);
	}

	public Relationship getAtomicRelationship(String startNodeName, String endNodeName) {
		Node startNode = getNode(startNodeName);
		Node endNode = getNode(endNodeName);
		return getAtomicRelationship(startNode, endNode);
	}

	public Relationship getAtomicRelationship(Node startNode, Node endNode) {
		assertNodeContained(startNode);
		assertNodeContained(endNode);

		Collection<Relationship> candidateRelationships = atomicRelationships.get(startNode);
		for (Relationship relationship : candidateRelationships) {
			if (relationship.isAtomic() && relationship.getEndNode().equals(endNode)) {
				return relationship;
			}
		}

		return null;
	}
	
	@Deprecated
	public Collection<Constraint> findCardinality(Node startNode, Node endNode) {
		return findCardinality(startNode, endNode, false);
	}

	public Collection<Constraint> findCardinality(Node startNode, Node endNode, boolean searchIdentityRelationship) {
		if (searchIdentityRelationship)
			System.out.println();
		if (this.graphSearcher == null) {
			this.graphSearcher = new PartialOrderGraphSearcher(this, ConstraintOrder.INSTANCE);
		}
		return this.graphSearcher.searchPaths(startNode, endNode, searchIdentityRelationship);
	}

	public String getName() {
		return name;
	}

	public Map<String, Node> getAtomicNodes() {
		return atomicNodes;
	}

	public Node getNode(String name) {
		Node node = atomicNodes.get(name);
		if (node == null) {
			throw new IllegalArgumentException("No node for name " + name + " in " + this.name + ".");
		}
		return node;
	}

	/**
	 * @deprecated Use {@link #getConstraints(Relationship)} or
	 *             {@link #getCardinality(Relationship)} instead.
	 * @return the internal collection of constraints (not necessarily
	 *         complete).
	 */
	public Map<Relationship, ConstraintCollection> getConstraintCollections() {
		return constraintCollections;
	}

	public ConstraintCollection getConstraints(Relationship relationship) {
		ConstraintCollection constraintCollection = constraintCollections.get(relationship);
		if (constraintCollection == null) {
			logger.debug("No constraint collection for {}", relationship);
			constraintCollection = new ConstraintCollection(this, relationship);
			constraintCollections.put(relationship, constraintCollection);
		}
		return constraintCollection;
	}

	@Override
	public boolean areOverlapping(Node node1, Node node2) {
		if (node1 == null || node2 == null) {
			logger.debug("One of the nodes is null.");
			return false;
		}

		if (node1.equals(node2)) {
			return true;
		}

		return node1.getContentType() == Node.ContentType.VALUE && node2.getContentType() == Node.ContentType.VALUE;
	}

	public NaturalNumberSet getCardinality(Relationship relationship) {
		return getConstraints(relationship).toConstraint().getCardinality();
	}

	/**
	 * @deprecated I should offer a method that directly uses this collecion.
	 *             Otherwise,
	 * @return
	 */
	public Map<Node, Collection<Relationship>> getRelationships() {
		return atomicRelationships;
	}

	public Collection<Relationship> getAtomicRelationships(Node node) {
		if (node == null)
			throw new IllegalArgumentException();

		Collection<Relationship> nodeRelationships = atomicRelationships.get(node);
		return nodeRelationships == null ? Collections.<Relationship> emptyList() : nodeRelationships;
	}

	@Override
	public Collection<Relationship> getVirtualRelationships(Node node) {
		if (node == null)
			throw new IllegalArgumentException();

		Collection<Relationship> nodeRelationships = virtualRelationships.get(node);
		return nodeRelationships == null ? Collections.<Relationship> emptyList() : nodeRelationships;
	}

	public Iterable<Relationship> getAllRelationships() {
		return new MultiLevelIterator<>(this.atomicRelationships.values()).wrapWithIterable();
	}

	public boolean contains(Relationship relationship) {
		switch (relationship.getOperationType()) {
		case IDENTITY:
			return containsNode(relationship.getStartNode());

		case NONE:
			if (relationship.getContentType() == Relationship.ContentType.EMPTY) {
				Node testNode = relationship.getStartNode();
				testNode = testNode == null ? relationship.getEndNode() : testNode;
				return containsNode(testNode);
			}
			Collection<Relationship> relationships = getAtomicRelationships(relationship.getStartNode());
			return relationships.contains(relationship);

		default:
			for (Relationship atomicRelationship : relationship.getAtomicRelationships()) {
				if (!contains(atomicRelationship)) {
					return false;
				}
			}
			return true;
		}
	}

	public boolean containsNode(Node node) {
		if (node == null)
			return false;
		switch (node.getOperationType()) {
		case NONE:
			return node.equals(getNode(node.getName()));
		case UNION:
			for (Node parentNode : node.getParentNodes()) {
				if (!containsNode(parentNode)) {
					return false;
				}
			}
			return true;
		case CROSS_PRODUCT:
			CrossProductNode crossProductNode = (CrossProductNode) node;
			CoalesceRelationship projectingRelationship = crossProductNode.getProjectingRelationship();
			for (Relationship relationship : projectingRelationship.getAtomicRelationships()) {
				if (!contains(relationship)) {
					return false;
				}
			}
			return true;
		default:
			throw new IllegalArgumentException("Unknown operation type of node " + node);
		}
	}

	/**
	 * Adds information to the node name so that it becomes unique across
	 * schemas.
	 * 
	 * @param nodeName is the name of the node without schema information
	 * @return the node name enriched with schema information
	 */
	public String createSchemaSpecificName(String nodeName) {
		return this.name + "." + nodeName;
	}

	@Override
	public String toString() {
		return "SchemaGraph[name=" + name + ", " + atomicNodes.size() + " nodes, "
				+ countValues(this.atomicRelationships) + " relationships, " + this.constraintCollections.size()
				+ " constraints]";
	}

	private int countValues(Map<?, ? extends Collection<?>> multimap) {
		int result = 0;
		for (Collection<?> values : multimap.values()) {
			result += values.size();
		}
		return result;
	}

	private void assertRelationshipContained(Relationship relationship) {
		if (!contains(relationship))
			throw new IllegalArgumentException(String.valueOf(relationship));

	}

	private void assertNodeContained(Node node) {
		if (!node.equals(atomicNodes.get(node.getName()))) {
			throw new IllegalArgumentException("Edge references a node that is not contained in this graph.");
		}
	}

	private void assertNamePrefixCorrect(Node node) {
		int dotIndex = node.getName().indexOf('.');
		if (dotIndex >= 0) {
			String prefix = node.getName().substring(0, dotIndex);
			if (prefix.equals(this.name)) {
				return;
			}
		}
		throw new IllegalArgumentException("Malformed node name: " + node);
	}

	/**
	 * Returns the schema name of the node or <tt>null</tt> if no or multiple
	 * schemas are present in the node.
	 * 
	 * @param node
	 *            is the node to test
	 * @return the schema name or <tt>null</tt>
	 */
	public static String getSchemaNameFor(Node node) {
		if (node == null) {
			return null;
		}
		if (node.getOperationType() != Node.Operation.NONE) {
			String schemaName = null;
			for (Node parentNode : node.getParentNodes()) {
				String newSchemaName = getSchemaNameFor(parentNode);
				if (schemaName == null) {
					schemaName = newSchemaName;
				} else if (!schemaName.equals(newSchemaName)) {
					return null;
				}
			}
			return schemaName;
		}
		String nodeName = node.getName();
		int dotIndex = nodeName.indexOf('.');
		return nodeName.substring(0, dotIndex);
	}

	/**
	 * Returns the node name without the schema information.
	 * 
	 * @param node
	 *            is the node whose name is needed
	 * @return the node name
	 */
	public static String getSchemalessNameFor(Node node) {
		if (node == null || node.getOperationType() != Node.Operation.NONE) {
			throw new IllegalArgumentException(String.valueOf(node));
		}
		String nodeName = node.getName();
		int dotIndex = nodeName.indexOf('.');
		if (dotIndex < 0) {
			throw new IllegalStateException("Node has illegal name: " + nodeName);
		}
		return nodeName.substring(dotIndex + 1);
	}

	public static String getSchemaNameFor(Relationship relationship) {
		String startSchemaName = null;
		Node testNode = relationship.getStartNode();
		if (testNode != null) {
			startSchemaName = BasicSchemaGraph.getSchemaNameFor(testNode);
			if (startSchemaName == null) {
				return null;
			}
		}

		String endSchemaName = null;
		testNode = relationship.getEndNode();
		if (testNode != null) {
			endSchemaName = BasicSchemaGraph.getSchemaNameFor(testNode);
			if (endSchemaName == null) {
				return null;
			}
		}

		if (startSchemaName != null && endSchemaName != null && !startSchemaName.equals(endSchemaName)) {
			return null;
		}

		return startSchemaName == null ? endSchemaName : startSchemaName;
	}
}
