package de.uni_potsdam.hpi.isg.efes.csg.constraints;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

public class NaturalNumberInterval implements NaturalNumberSet {

	private static final Logger logger = LogManager.getLogger(NaturalNumberSet.class.getSimpleName());
	
	private int min;
	private int max;

	public NaturalNumberInterval(int min) {
		this(min, INFINITE);
	}

	public NaturalNumberInterval(int min, int max) {
		this.min = min;
		this.max = max;
	}

	@Override
	public boolean isEmpty() {
		return false;
	}

	public NaturalNumberSet add(NaturalNumberSet other) {
		if (this.min == 0 && this.max == 0)
			return other;

		return other.addInterval(this);
	}

	public NaturalNumberSet addInterval(NaturalNumberInterval interval) {
		int min = this.min + interval.getMin();
		int max = (this.max == INFINITE || interval.getMax() == INFINITE) ? INFINITE : this.max
				+ interval.getMax();

		return new NaturalNumberInterval(min, max);
	}
	
	@Override
	public NaturalNumberSet inclExclAdd(NaturalNumberSet other) {
		return other.inclExclAddInterval(this);
	}
	
	@Override
	public NaturalNumberSet inclExclAddInterval(NaturalNumberInterval other) {
		int newMin = Math.max(this.getMin(), other.getMin());
		int newMax = add(this.getMax(), other.getMax());
		return new NaturalNumberInterval(newMin, newMax);
	}

	@Override
	public boolean hasElement(int element) {
		return max(element, this.max) == this.max && min(element, this.min) == this.min;
	}
	
	public NaturalNumberSet compose(NaturalNumberSet other) {
		if (this.min == 1 && this.max == 1)
			return other;

		if (this.min == 0 && this.max == INFINITE)
			return this;

		if (this.min == 0 && this.max == 0)
			return this;

		return other.composeInterval(this);
	}

	public NaturalNumberSet composeInterval(NaturalNumberInterval interval) {
		if (interval.getMin() > 1)
			throw new UnsupportedOperationException(
					"This method assumes the second interval to start with 0 or 1.");

		int min = this.min * interval.getMin();
		int max = (this.max == INFINITE || interval.getMax() == INFINITE) ? INFINITE : this.max
				* interval.getMax();

		return new NaturalNumberInterval(min, max);
	}

	public int getMin() {
		return min;
	}

	public int getMax() {
		return max;
	}

	@Override
	public boolean isSubsetOf(NaturalNumberSet other) {
		return other.isSupersetOfInterval(this);
	}

	public boolean isSubsetOfInterval(NaturalNumberInterval other) {
		if (this.min < other.getMin())
			return false;
		if (other.getMax() == INFINITE)
			return true;
		if (this.max == INFINITE)
			return false;
		return this.max <= other.getMax();
	}

	@Override
	public boolean isSupersetOf(NaturalNumberSet other) {
		return other.isSubsetOfInterval(this);
	}

	@Override
	public boolean isSupersetOfInterval(NaturalNumberInterval other) {
		return this.min <= other.getMin()
				&& ((other.getMax() != INFINITE && this.max >= other.getMax()) || this.max == INFINITE);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + max;
		result = prime * result + min;
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		NaturalNumberInterval other = (NaturalNumberInterval) obj;
		if (max != other.max)
			return false;
		if (min != other.min)
			return false;
		return true;
	}

	@Override
	public String toString() {
		return min + ".." + (max == INFINITE ? "*" : max);

	}

	@Override
	public NaturalNumberSet intersect(NaturalNumberSet other) {
		return other.intersectInterval(this);
	}

	@Override
	public NaturalNumberSet intersectInterval(NaturalNumberInterval other) {
		int newMin, newMax;
		if (this.max == INFINITE) {
			if (this.min == 0) {
				return other;
			} else if (other.getMax() == INFINITE) {
				return (this.min >= other.getMin()) ? this : other;
			} else {
				newMax = other.getMax();
			}
		} else if (other.getMax() == INFINITE) {
			if (other.getMin() == 0) {
				return this;
			} else {
				newMax = this.max;
			}
		} else {
			newMax = Math.min(this.max, other.getMax());
		}

		newMin = Math.max(this.min, other.getMin());
		return newMin <= newMax ? new NaturalNumberInterval(newMin, newMax)
				: NaturalNumberSet.EMPTY_SET;
	}

	@Override
	public NaturalNumberSet union(NaturalNumberSet other) {
		return other.unionInterval(this);
	}

	@Override
	public NaturalNumberSet unionInterval(NaturalNumberInterval other) {
		if (this.max == INFINITE && (other.getMax() == INFINITE || other.getMax() + 1 >= this.min)) {
			// Case 1: This interval is unbounded.
			return new NaturalNumberInterval(min(this.min, other.getMin()));

		} else if (other.getMax() == INFINITE && this.getMax() + 1 >= other.getMin()) {
			// Case 2: Other interval is unbounded.
			return new NaturalNumberInterval(min(this.min, other.getMin()));

		} else if (this.getMax() + 1 >= other.getMin() && this.getMin() <= other.getMax() + 1) {
			// Case 3: Overlap of unbounded intervals.
			return new NaturalNumberInterval(min(this.min, other.getMin()), max(this.getMax(),
					other.getMax()));
		
		} else {
			logger.warn("Can only approximate union({}, {}) using intervals.", this, other);
			int min = min(this.getMin(), other.getMin());
			int max = max(this.getMax(), other.getMax());
			return new NaturalNumberInterval(min, max);
		}
	}

	public static NaturalNumberSet sign(NaturalNumberSet set) {
		if (NaturalNumberSet.ZERO.equals(set)) {
			return NaturalNumberSet.ZERO;

		} else if (NaturalNumberSet.ZERO.isSubsetOf(set)) {
			return NaturalNumberSet.ZERO_OR_ONE;
		
		} else if (set.isEmpty()) {
			return NaturalNumberSet.EMPTY_SET;
		
		} else {
			return NaturalNumberSet.ONE;
		}
	}

	public static int min(int a, int b) {
		if (a == INFINITE)
			return b;
		if (b == INFINITE)
			return a;
		return Math.min(a, b);
	}

	public static int max(int a, int b) {
		if (a == INFINITE || b == INFINITE)
			return INFINITE;
		return Math.max(a, b);
	}
	
	public static int multiply(int a, int b) {
		if (a == INFINITE || b == INFINITE)
			return INFINITE;
		return a * b;
	}
	
	public static int add(int a, int b) {
		if (a == INFINITE || b == INFINITE)
			return INFINITE;
		return a + b;
	}

}
