package de.uni_potsdam.hpi.isg.efes.csg.constraints;

import it.unimi.dsi.fastutil.doubles.DoubleArrayList;
import it.unimi.dsi.fastutil.doubles.DoubleIterator;
import it.unimi.dsi.fastutil.doubles.DoubleList;
import it.unimi.dsi.fastutil.ints.Int2IntMap;
import it.unimi.dsi.fastutil.ints.Int2IntRBTreeMap;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.fastutil.ints.IntList;
import it.unimi.dsi.fastutil.ints.IntSet;
import it.unimi.dsi.fastutil.ints.IntSets;

import java.util.Arrays;
import java.util.Collections;
import java.util.Random;

import org.apache.commons.math3.util.ArithmeticUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.utils.util.MoreMath;

public class StatisticBagOperator {

	private static final Logger logger = LogManager.getLogger(StatisticBagOperator.class.getSimpleName());

	private NaturalNumberBag baseDistribution;

	public StatisticBagOperator(NaturalNumberBag baseDistribution) {
		super();
		this.baseDistribution = baseDistribution;
	}

	public NaturalNumberBag simulateLinkRemoval(int numTakenLinks) {
		return simulateLinkRemoval(numTakenLinks, IntSets.EMPTY_SET);
	}

	public NaturalNumberBag simulateLinkRemoval(int numTakenLinks, IntSet excludedCardinalities) {
		logger.debug("Remove {}/{} links from {}.", numTakenLinks, countLinks(baseDistribution), this.baseDistribution);

		if (numTakenLinks == 0) {
			return new NaturalNumberBag(this.baseDistribution);

		} else if (numTakenLinks < 0 || numTakenLinks > countLinks(this.baseDistribution)) {
			throw new IllegalArgumentException("Cannot remove " + String.valueOf(numTakenLinks) + " links from "
					+ this.baseDistribution);
		}

		int numRemainingLinks = countLinks(this.baseDistribution) - numTakenLinks;
		Int2IntMap retainedLinksPerCardinalityValue = retainLinks(numRemainingLinks, excludedCardinalities);
		logger.debug("Retained {}.", retainedLinksPerCardinalityValue);

		NaturalNumberBag newDistribution = new NaturalNumberBag();
		for (Int2IntMap.Entry entry : retainedLinksPerCardinalityValue.int2IntEntrySet()) {
			int cardinalityValue = entry.getIntKey();
			int numRetainedLinks = entry.getIntValue();
			int numPriorElements = this.baseDistribution.getCountFor(cardinalityValue);

			addNewCardinalityValues(cardinalityValue, numPriorElements, numRetainedLinks, newDistribution);
		}
		return newDistribution;

	}

	public static int countLinks(NaturalNumberBag distribution) {
		return distribution.getSum();
	}

	public static int determineLinkDiff(NaturalNumberBag oldDistribution, NaturalNumberBag newDistribution) {
		int numOldLinks = countLinks(oldDistribution);
		int numNewLinks = countLinks(newDistribution);
		logger.trace("determineLinkDiff({}, {})", numOldLinks, numNewLinks);
		return numOldLinks - numNewLinks;
	}

	/**
	 * Distributes all retained elements among the blocks of elements that have
	 * the same cardinality value.
	 * 
	 * @param remainingLinks
	 *            number of elements that will be retained from the base
	 *            distribution
	 * @param excludedCardinalities
	 * @return a map with entries like <i>(cardinality value, number of retained
	 *         links of elements of this cardinality)</i>
	 */
	private Int2IntMap retainLinks(int remainingLinks, IntSet excludedCardinalities) {
		Int2IntMap retainedElements = new Int2IntRBTreeMap();

		// Excluded cardinalities will be retained anyway, so exclude
		// them from the following algorithm.
		int numExcludedLinks = 0;
		for (IntIterator i = excludedCardinalities.iterator(); i.hasNext();) {
			int excludedCardinalityValue = i.nextInt();
			int numLinks = this.baseDistribution.getCountFor(excludedCardinalityValue) * excludedCardinalityValue;
			if (numLinks > 0) {
				retainedElements.put(excludedCardinalityValue, numLinks);
			}
			numExcludedLinks += numLinks;
		}

		// Calculate how many elements to retain from every cardinality value
		// block.
		remainingLinks -= numExcludedLinks;
		int numIncludedLinks = countLinks(this.baseDistribution) - numExcludedLinks;
		double retainFactor = remainingLinks / (double) numIncludedLinks;
		double remainderSum = 0;
		DoubleList remainders = new DoubleArrayList();
		IntList cardinalityValues = new IntArrayList();
		int numAssignedElements = 0;
		for (Int2IntMap.Entry entry : this.baseDistribution.getCountsPerNumber().int2IntEntrySet()) {
			int cardinalityValue = entry.getIntKey();
			int numOccurrences = entry.getIntValue();

			// Skip over retained elements.
			if (excludedCardinalities.contains(cardinalityValue)) {
				continue;
			}

			int numRelationshipElements = cardinalityValue * numOccurrences;

			double retainValue = numRelationshipElements * retainFactor;
			int numLocallyRetainedElements = (int) Math.floor(retainValue);
			retainedElements.put(cardinalityValue, numLocallyRetainedElements);
			numAssignedElements += numLocallyRetainedElements;

			double remainder = retainValue - numLocallyRetainedElements;
			if (remainder > 0) {
				remainderSum += remainder;
				remainders.add(remainder);
				cardinalityValues.add(cardinalityValue);
			}
		}
		// Assign left-over elements by randomly.
		while (numAssignedElements < remainingLinks) {
			double randomValue = Math.random() * remainderSum;
			int index = 0;
			double curRemainderSum = 0;
			boolean assignSuccess = false;
			for (DoubleIterator iterator = remainders.iterator(); iterator.hasNext();) {
				curRemainderSum += iterator.nextDouble();
				if (randomValue <= curRemainderSum) {
					int cardinalityValue = cardinalityValues.getInt(index);
					logger.debug("Assigning left-over element to {}.", cardinalityValue);
					int alreadyRetainedElements = retainedElements.get(cardinalityValue);
					retainedElements.put(cardinalityValue, alreadyRetainedElements + 1);
					numAssignedElements++;
					assignSuccess = true;
					break;
				}
				index++;
			}

			// Remove the remainder so that an element is not picked twice.
			if (assignSuccess) {
				remainderSum -= curRemainderSum;
				remainders.remove(index);
				cardinalityValues.remove(index);
			}
		}
		return retainedElements;
	}

	private void addNewCardinalityValues(int cardinalityValue, int numPriorElements, int numRetainedLinks,
			NaturalNumberBag newDistribution) {

		if (numPriorElements <= 0) {
			throw new IllegalArgumentException("Cannot simulate retaining without prior elements.");
		}

		if (numRetainedLinks > numPriorElements * cardinalityValue) {
			throw new IllegalArgumentException("Cannot fit " + numRetainedLinks + " into " + cardinalityValue + "=>"
					+ numPriorElements + ".");
		}

		// Initialize the distribution structure.
		int[] linksPerElement = new int[numPriorElements];
		int index = 0;
		while (numRetainedLinks > 0) {
			int numAssignedLinks = Math.min(cardinalityValue, numRetainedLinks);
			linksPerElement[index++] = numAssignedLinks;
			numRetainedLinks -= numAssignedLinks;
		}

		int[] bestDistribution = null;
		double bestDistributionFrequency = -1;

		boolean hasNextDistribution = true;
		if (linksPerElement[0] <= linksPerElement[linksPerElement.length - 1]) {
			// Under this condition there is only a single distribution thus no
			// competition.
			bestDistribution = linksPerElement;
			hasNextDistribution = false;
		}
		while (hasNextDistribution) {

			// Calculate the frequency of the current distribution.
			double logFrequency = 0;
			for (int links : linksPerElement) {
				logFrequency += ArithmeticUtils.binomialCoefficientLog(cardinalityValue, links);
			}
			int lastElement = linksPerElement[0];
			int count = 1;
			IntList counts = new IntArrayList();
			for (index = 1; index < linksPerElement.length; index++) {
				int newElement = linksPerElement[index];
				if (newElement == lastElement) {
					count++;
				} else {
					counts.add(count);
					count = 1;
				}
				lastElement = newElement;
			}
			counts.add(count);
			Collections.sort(counts);
			int sumCount = 0;
			for (IntIterator iterator = counts.iterator(); iterator.hasNext();) {
				count = iterator.nextInt();
				logger.trace("Calculating log binomial coefficient of {} and {}.", sumCount + count, sumCount);
				logFrequency += MoreMath.logBinomialCoefficient(sumCount + count, sumCount);
				sumCount += count;
			}

			logger.trace("Log frequency is {}.", logFrequency);
			logger.trace("Score {} for {}.", logFrequency, Arrays.toString(linksPerElement));

			// Evaluate if thats the best frequency seen so far.
			if (logFrequency > bestDistributionFrequency) {
				bestDistributionFrequency = logFrequency;
				bestDistribution = linksPerElement.clone();
			}

			// Try to 'smear' to obtain a new distribution.
			int lowestValue = linksPerElement[linksPerElement.length - 1];
			for (index = linksPerElement.length - 2; index >= 0; index--) {
				if (linksPerElement[index] >= lowestValue + 2)
					break;
			}

			if (index < 0) {
				hasNextDistribution = false;
			} else {
				int targetIndex = index + 1;
				while (linksPerElement[targetIndex] > linksPerElement[index] - 2) {
					targetIndex++;
				}
				linksPerElement[index]--;
				linksPerElement[targetIndex]++;
				hasNextDistribution = true;
			}
		}
		while (hasNextDistribution)
			;

		for (int links : bestDistribution) {
			newDistribution.count(links);
		}
	}

	public NaturalNumberBag simulateElementRemoval(int numElements) {
		logger.debug("Remove {}/{} elements from {}.", numElements, this.baseDistribution.getCardinality(),
				this.baseDistribution);

		// Calculate how many elements to retain from every cardinality value
		// block.
		int numRetainedElements = this.baseDistribution.getCardinality() - numElements;
		NaturalNumberBag retainedElements = new NaturalNumberBag();
		double retainFactor = numRetainedElements / (double) this.baseDistribution.getCardinality();
		double remainderSum = 0;
		DoubleList remainders = new DoubleArrayList();
		IntList cardinalityValues = new IntArrayList();
		double numAssignedElements = 0;
		for (Int2IntMap.Entry entry : this.baseDistribution.getCountsPerNumber().int2IntEntrySet()) {
			int cardinalityValue = entry.getIntKey();
			int numOccurrences = entry.getIntValue();

			double retainValue = numOccurrences * retainFactor;
			int numLocallyRetainedElements = (int) Math.floor(retainValue);
			retainedElements.count(cardinalityValue, numLocallyRetainedElements);
			numAssignedElements += numLocallyRetainedElements;

			double remainder = retainValue - numLocallyRetainedElements;
			remainderSum += remainder;
			remainders.add(remainderSum);
			cardinalityValues.add(cardinalityValue);
		}
		// Assign left-over elements by randomly.
		while (numAssignedElements < numRetainedElements) {
			double randomValue = Math.random() * remainderSum;
			int index = 0;
			double curRemainderSum = 0;
			for (DoubleIterator iterator = remainders.iterator(); iterator.hasNext();) {
				curRemainderSum = iterator.nextDouble();
				if (randomValue <= curRemainderSum) {
					int cardinalityValue = cardinalityValues.getInt(index);
					retainedElements.count(cardinalityValue);
					numAssignedElements++;
					break;
				}
				index++;
			}

			remainderSum -= curRemainderSum;
			remainders.remove(index);
			cardinalityValues.remove(index);
		}

		logger.debug("Retained {}.", retainedElements);

		return retainedElements;
	}

	/**
	 * Simulates the removal of elements in a node with an equality
	 * relationship. This method should be used together with
	 * {@link #simulateTargetElementRemovalInEqualityLink(NaturalNumberBag, int)}
	 * <br>
	 * Assume, e.g., a scenario with two node <tt>A</tt> and <tt>B</tt>, where
	 * there is an equality relationship between those. Further assume, from
	 * <tt>A</tt> are elements removed. This method returns an update of the new
	 * cardinality from <tt>A</tt> to <tt>B</tt>.
	 * 
	 * @param numRemovedElements
	 * @return
	 */
	public NaturalNumberBag simulateSourceElementRemovalInEqualityLink(int numRemovedElements) {

		int removedEqualElements = calculateRemovedEqualElements(numRemovedElements);
		int removedUnequalElements = numRemovedElements;

		NaturalNumberBag newCardinality = new NaturalNumberBag(this.baseDistribution);
		newCardinality.count(0, -removedUnequalElements);
		newCardinality.count(1, -removedEqualElements);

		return newCardinality;
	}

	/**
	 * This method updates cardinality of an equality relationship after the
	 * removal of elements in the target node. This method should be used
	 * together with {@link #simulateSourceElementRemovalInEqualityLink(int)}.
	 * 
	 * @param inverseCardinality
	 * @param numRemovedElements
	 * @return
	 */
	public NaturalNumberBag simulateTargetElementRemovalInEqualityLink(NaturalNumberBag inverseCardinality,
			int numRemovedElements) {

		NaturalNumberBag newCardinality = new NaturalNumberBag(inverseCardinality);
		int numRemovedEqualElements = calculateRemovedEqualElements(numRemovedElements);
		newCardinality.count(0, numRemovedEqualElements);
		newCardinality.count(1, -numRemovedEqualElements);

		return newCardinality;
	}

	private int calculateRemovedEqualElements(int numRemovedElements) {
		int equalElements = this.baseDistribution.getCountFor(1);
		double equalElementsRatio = equalElements / (double) this.baseDistribution.getCardinality();
		int removedEqualElements = (int) Math.round(equalElementsRatio * numRemovedElements);
		return removedEqualElements;
	}

	public NaturalNumberBag simulateLinkInsertion(int numAddedLinks) {
		return simulateLinkInsertion(numAddedLinks, null);
	}

	public NaturalNumberBag simulateLinkInsertion(int numLinksToAdd, NaturalNumberBag boundary) {
		// XXX: This is a very simple implementation, it could be improved.
		Random random = new Random();
		NaturalNumberBag newCardinality = new NaturalNumberBag(this.baseDistribution);
		IntList cardinalityKeys = new IntArrayList(newCardinality.getCountedCardinalities());
		while (numLinksToAdd > 0) {
			// Randomly select an element to add the link to.
			int randInt = random.nextInt(cardinalityKeys.size());
			int oldCardinalityKey = cardinalityKeys.getInt(randInt);

			// Should also check boundary here for a more accurate simulation.
			// ...

			int newCardinalityKey = oldCardinalityKey + 1;
			newCardinality.count(newCardinalityKey);
			newCardinality.count(oldCardinalityKey, -1);
			if (newCardinality.getCountFor(oldCardinalityKey) == 0) {
				cardinalityKeys.remove(randInt);
			}
			if (newCardinality.getCountFor(newCardinalityKey) == 1) {
				cardinalityKeys.add(newCardinalityKey);
			}

			numLinksToAdd--;
		}

		logger.debug("Added {} links to {}: {}.", numLinksToAdd, this.baseDistribution, newCardinality);

		return newCardinality;
	}

	public NaturalNumberBag simulateElementInsertion(int numTuplesToAdd) {
		// XXX: This is a very simple implementation, it could be improved.
		Random random = new Random();
		NaturalNumberBag newCardinality = new NaturalNumberBag(this.baseDistribution);
		int numElements = this.baseDistribution.getCardinality();
		IntList cardinalityKeys = new IntArrayList(newCardinality.getCountedCardinalities());

		while (numTuplesToAdd > 0) {
			// Randomly select a cardinality to that a new element will be
			// added.
			int randInt = random.nextInt(numElements);
			int sum = 0;

			for (IntIterator iterator = cardinalityKeys.iterator(); iterator.hasNext();) {
				int oldSum = sum;
				int cardinalityKey = iterator.nextInt();
				int numCardinalityElements = this.baseDistribution.getCountFor(cardinalityKey);
				sum += numCardinalityElements;
				if (randInt >= oldSum && randInt < sum) {
					newCardinality.count(cardinalityKey);
					numTuplesToAdd--;
					break;
				}
			}

		}

		logger.debug("Added {} elements to {}: {}.", numTuplesToAdd, this.baseDistribution, newCardinality);

		return newCardinality;
	}

	@Override
	public String toString() {
		return "StatisticBagOperator [baseDistribution=" + baseDistribution + "]";
	}

}
