package de.uni_potsdam.hpi.isg.efes.csg.instance;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node.ContentType;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.UnionNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.CompositeRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.EmptyRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.UnionRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.util.RelationalSchemaToCSGConverter;
import de.uni_potsdam.hpi.isg.utils.db.PostgresqlDatabase;
import de.uni_potsdam.hpi.isg.utils.util.CollectionUtil;

public class PostgresqlAdapter implements DataStoreAdapter {

	private static final Logger logger = LogManager.getLogger(PostgresqlAdapter.class.getSimpleName());

	private static final String TUPLE_ID_ATTR = "_efes_tuple_id";
	private static final String TUPLE_ID_SELECT_EXPR = "nextval('tuple_id_seq') AS _efes_tuple_id";
	private BasicSchemaGraph schemaGraph;
	private PostgresqlDatabase db;

	@Override
	public void setSchemaGraph(BasicSchemaGraph schemaGraph) {
		this.schemaGraph = schemaGraph;
	}

	@Override
	public void setDataStore(Object dataStore) throws IllegalArgumentException {
		if (dataStore == null || !(dataStore instanceof PostgresqlDatabase)) {
			throw new IllegalArgumentException(String.valueOf(dataStore));
		}

		try {
			this.db = (PostgresqlDatabase) dataStore;
			if (!this.db.isOpen()) {
				this.db.open();
			}

			this.db.execute("CREATE TEMPORARY SEQUENCE tuple_id_seq;");
		} catch (SQLException e) {
			logger.catching(e);
		}
	}

	/**
	 * Queries the cardinality of the given relationship in the underlying database.<br>
	 * <i>Please note, that currently it is assumed that only atomic, composite, and union relationships are queried.</i>
	 * 
	 */
	@Override
	public NaturalNumberBag queryCardinality(Relationship relationship) {
		logger.debug("Querying relationship {}.", relationship);

		if (this.schemaGraph == null || this.db == null) {
			throw new IllegalStateException();
		}

		if (!this.schemaGraph.contains(relationship)) {
			throw new IllegalArgumentException(String.valueOf(relationship));
		}

		// If we have to query an identity relationship, we should rather be
		// doing a size query.
		if (relationship.getOperationType() == Relationship.OperationType.IDENTITY) {
			int nodeSize = querySize(relationship.getEndNode());
			NaturalNumberBag cardinality = new NaturalNumberBag();
			cardinality.count(1, nodeSize);
			logger.debug("Result: {}", cardinality);
			return cardinality;
		}

		List<Path> paths = findPaths(relationship);
		if (paths.isEmpty()) {
			return NaturalNumberBag.EMPTY_BAG;
		}

		String sqlQuery = createSqlQuery(paths);
		logger.trace("Executing query: \n{}", sqlQuery);
		NaturalNumberBag cardinalities = new NaturalNumberBag();
		try (Statement stmt = db.getConnection().createStatement()) {
			ResultSet resultSet = stmt.executeQuery(sqlQuery);
			while (resultSet.next()) {
				int cardinality = resultSet.getInt(1);
				int occurrences = resultSet.getInt(2);
				cardinalities.count(cardinality, occurrences);
			}
		} catch (Exception e) {
			logger.error("Error on following query:\n{}", sqlQuery);
			logger.catching(e);
		}

		logger.debug("Result: {}", cardinalities);
		return cardinalities;
	}

	private List<Path> findPaths(Relationship relationship) {
		Path path;
		List<Path> paths = new ArrayList<>(1);
		switch (relationship.getOperationType()) {
		case IDENTITY:
		case NONE:
		case COMPOSITE:
			path = findPath(relationship);
			if (path != null) {
				paths.add(path);
			}
			break;

		case UNION:
			UnionRelationship unionRelationship = (UnionRelationship) relationship;
			for (Relationship baseRelationship : unionRelationship.getBaseRelationships()) {
				path = findPath(baseRelationship);
				if (path != null) {
					paths.add(path);
				}
			}
			break;

		default:
			throw new IllegalArgumentException("Cannot query " + relationship + ".");

		}
		return paths;
	}

	private Path findPath(Relationship relationship) {
		switch (relationship.getOperationType()) {
		case NONE:
		case IDENTITY:
			return findPath(Collections.singletonList(relationship));

		case COMPOSITE:
			CompositeRelationship compositeRelationship = (CompositeRelationship) relationship;
			return findPath(compositeRelationship.getComposedRelationships());

		default:
			throw new IllegalArgumentException(String.valueOf(relationship));
		}
	}

	private Path findPath(List<Relationship> composedRelationships) {
		// Find the composed nodes.
		List<Node> composedNodes = new ArrayList<>();
		for (Relationship composedRelationship : composedRelationships) {
			if (composedNodes.isEmpty()) composedNodes.add(composedRelationship.getStartNode());
			composedNodes.add(composedRelationship.getEndNode());
		}

		Node startNode = composedNodes.get(0);
		if (startNode == null) {
			// If so, we are dealing with an left-open relationship.
			return null;
		}

		// Parse the nodes of the relationships into linked tables.
		Path path = new Path();
		Table curTable = new Table(RelationalSchemaToCSGConverter.extractTableName(startNode));
		List<String> startAttributes = RelationalSchemaToCSGConverter.extractAttributeNames(startNode);
		curTable.addStartAttributes(startAttributes);
		path.add(curTable);

		for (int index = 0; index < composedRelationships.size(); index++) {
			Relationship curRelationship = composedRelationships.get(index);
			Node curNode = composedNodes.get(index + 1);

			if (curNode == null) {
				break;
			}

			// If we encounter a new table, we need to create a new table object
			// and set start and end attributes of the old and new table
			// appropriately.
			String tableName = RelationalSchemaToCSGConverter.extractTableName(curNode);
			if (!curTable.getName().equals(tableName)) {
				curTable.addEndAttributes(RelationalSchemaToCSGConverter.extractAttributeNames(curRelationship
						.getStartNode()));
				curTable = new Table(tableName);
				path.add(curTable);
				startAttributes = RelationalSchemaToCSGConverter.extractAttributeNames(curNode);
				curTable.addStartAttributes(startAttributes);
			}
		}
		// Add information of last node.
		Relationship lastRelationship = composedRelationships.get(composedRelationships.size() - 1);
		Node lastNode = lastRelationship.getEndNode();
		if (lastNode == null) {
			// If so, we are dealing with an right-open relationship.
			curTable.setIsRightOpen(true);
			return path;
		} else {
			curTable.addEndAttributes(RelationalSchemaToCSGConverter.extractAttributeNames(lastNode));
		}

		return path;
	}

	private Collection<Table> findTablesWithTupleQuery(Collection<Path> paths) {
		Map<String, Table> result = new HashMap<>();
		for (Path path : paths) {
			Table table = path.getFirstTable();
			List<String> startAttributes = table.getStartAttributes();
			List<String> endAttributes = table.getEndAttributes();
			collectAttributes(result, table, startAttributes, endAttributes, true);
			table = path.getLastTable();
			startAttributes = table.getStartAttributes();
			endAttributes = table.getEndAttributes();
			collectAttributes(result, table, endAttributes, startAttributes, true);
		}
		for (Path path : paths) {
			for (Table table : path.getTables()) {
				if (!table.getEndAttributes().isEmpty() && !table.getStartAttributes().isEmpty()) {
					collectAttributes(result, table, Collections.<String>emptyList(), table.getStartAttributes(), false);
					collectAttributes(result, table, Collections.<String>emptyList(), table.getEndAttributes(), false);
				}
			}
		}
		return result.values();
	}

	private void collectAttributes(Map<String, Table> result, Table table, List<String> attributes1,
			List<String> attributes2, boolean addTableIfNotExists) {
		if (attributes1.isEmpty()) {
			String tableName = table.getName();
			Table resultTable = result.get(tableName);
			if (resultTable == null) {
				if (!addTableIfNotExists) return;
				resultTable = new Table(tableName);
				result.put(tableName, resultTable);
			}
			for (String endColumnName : attributes2) {
				if (!resultTable.getOtherAttributes().contains(endColumnName)) {
					resultTable.addOtherAttribute(endColumnName);
				}
			}
		}
	}

	private String createSqlQuery(List<Path> paths) {
		StringBuilder sb = new StringBuilder(1024);

		createRelationshipWithStatements(paths, sb);

		String with = sb.length() == 0 ? "WITH " : ",\n";

		// Turn relationship into a cardinality bag.
		sb.append(with).append("cardlist AS (SELECT count(val) AS CARD FROM (");

		String union = "";
		for (int pathId = 0; pathId < paths.size(); pathId++) {
			sb.append(union).append("SELECT key, val FROM relationship").append(pathId);
			union = " UNION ";
		}

		sb.append(") AS combined_relationship GROUP BY key)\n");

		// Group the cardinality bag.
		sb.append("SELECT card, count(*) FROM cardlist GROUP BY card;");

		return sb.toString();
	}

	private void createRelationshipWithStatements(List<Path> paths, StringBuilder sb) {
		Collection<Table> tupleTables = findTablesWithTupleQuery(paths);

		// WITH <table> AS (...) to ensure stable tuple IDs
		String separator;
		String with = "WITH ";
		for (Table tupleTable : tupleTables) {
			if (!tupleTable.getStartAttributes().isEmpty()
					&& (!tupleTable.getEndAttributes().isEmpty() || tupleTable.isRightOpen())) continue;

			sb.append(with).append(tupleTable.getName()).append(" AS (SELECT ").append(TUPLE_ID_SELECT_EXPR);
			for (String attribute : tupleTable.getOtherAttributes()) {
				sb.append(", ").append(attribute);
			}
			sb.append(" FROM ").append(tupleTable.getName()).append(")");
			with = ",\n";
		}

		// WITH relationship<n> AS (...)
		for (int pathId = 0; pathId < paths.size(); pathId++) {
			Path path = paths.get(pathId);

			// Find start and end of the relationship to query.
			Table firstTable = path.getFirstTable();
			boolean startWithTuples = firstTable.getStartAttributes().isEmpty();
			List<String> startAttributes = startWithTuples ? Collections.singletonList(TUPLE_ID_ATTR) : firstTable
					.getStartAttributes();
			String startTableName = firstTable.getName();

			Table lastTable = path.getLastTable();
			boolean isRightOpen = lastTable.isRightOpen();
			boolean endWithTuples = !isRightOpen && lastTable.getEndAttributes().isEmpty();
			List<String> endAttributes;
			if (isRightOpen) {
				endAttributes = Collections.emptyList();
			} else if (endWithTuples) {
				endAttributes = Collections.singletonList(TUPLE_ID_ATTR);
			} else {
				endAttributes = lastTable.getEndAttributes();
			}
			String endTableName = lastTable.getName();

			// Build relationship query
			sb.append(with).append("relationship").append(pathId).append(" AS (\n");
			// SELECT clause
			sb.append("SELECT encode(CAST(");
			separator = "";
			for (int i = 0; i < startAttributes.size(); i++) {
				sb.append(separator).append("replace(").append(startTableName).append(".")
						.append(startAttributes.get(i)).append("||'', '\\', '\\\\')");
				separator = "||'***'||";
			}
			sb.append("||'' AS bytea), 'base64') AS key, ");
			separator = "";
			if (endAttributes.isEmpty()) {
				sb.append("NULL::text");
			} else {
				sb.append("encode(CAST(");
				for (int i = 0; i < endAttributes.size(); i++) {
					sb.append(separator).append("replace(").append(endTableName).append(".")
							.append(endAttributes.get(i)).append("||'', '\\', '\\\\')");
					separator = "||'***'||";
				}
				sb.append("||'' AS bytea), 'base64')");
			}
			sb.append(" AS val\n");
			// FROM clause
			sb.append("FROM\n");
			Table prevTable = null;
			for (Table table : path.getTables()) {
				if (prevTable != null) {
					sb.append("LEFT OUTER JOIN ");
				}
				sb.append("(SELECT DISTINCT ");
				separator = "";
				List<String> selectElements = new ArrayList<>();
				if (table.getStartAttributes().isEmpty()) {
					// This condition must only hold for the first table.
					if (prevTable != null) {
						throw new IllegalStateException();
					}
					selectElements.add(TUPLE_ID_ATTR);
				} else {
					selectElements.addAll(table.getStartAttributes());
				}
				if (table.getEndAttributes().isEmpty() && endWithTuples) {
					// This condition must only hold for the last table.
					if (!selectElements.contains(TUPLE_ID_ATTR)) {
						selectElements.add(TUPLE_ID_ATTR);
					}
				}
				for (String endAttribute : table.getEndAttributes()) {
					if (!selectElements.contains(endAttribute)) selectElements.add(endAttribute);
				}
				separator = "";
				for (String selectElement : selectElements) {
					sb.append(separator).append(selectElement);
					separator = ", ";
				}
				sb.append("\nFROM ").append(table.getName());

				separator = "\nWHERE ";
				for (String attribute : table.getStartAttributes()) {
					sb.append(separator).append(attribute).append(" IS NOT NULL");
					separator = " AND ";
				}
				sb.append(") AS ").append(table.getName());

				if (prevTable != null) {
					sb.append("\nON ");
					separator = "";
					for (int i = 0; i < prevTable.getEndAttributes().size(); i++) {
						String prevTableAttribute = prevTable.getEndAttributes().get(i);
						String curTableAttriubte = table.getStartAttributes().get(i);
						sb.append(separator).append(prevTable.getName()).append(".").append(prevTableAttribute)
								.append(" = ").append(table.getName()).append(".").append(curTableAttriubte);
						separator = "AND ";
					}
				}
				sb.append("\n");

				prevTable = table;
			}
			sb.append(")");

			with = ",\n";
		}
	}

	@Override
	public int querySize(Node node) {
		if (this.schemaGraph == null || this.db == null) {
			throw new IllegalStateException();
		}

		if (node == null || !this.schemaGraph.containsNode(node)) {
			throw new IllegalArgumentException(String.valueOf(node));
		}

		switch (node.getOperationType()) {
		case NONE:
			String sqlQuery;
			String tableName = RelationalSchemaToCSGConverter.extractTableName(node);
			switch (node.getContentType()) {
			case IDENTITY:
				sqlQuery = String.format("SELECT count(*) FROM %s;", tableName);
				break;

			case VALUE:
				String attributeName = RelationalSchemaToCSGConverter.extractAttributeNames(node).get(0);
				sqlQuery = String.format("SELECT count(distinct %1$s) FROM %2$s WHERE %1$s IS NOT NULL;",
						attributeName, tableName);
				break;

			default:
				throw new IllegalArgumentException("Unprocessable content type of node " + node);

			}
			int size = -1;
			try (Statement stmt = db.getConnection().createStatement()) {
				ResultSet resultSet = stmt.executeQuery(sqlQuery);
				if (!resultSet.next()) {
					throw new IllegalStateException("Result set for size query is empty.");
				}
				size = resultSet.getInt(1);
			} catch (Exception e) {
				throw new RuntimeException(tableName);
			}

			return size;

		case UNION:

			// If we have identity nodes, we can just add up their sizes.
			if (node.getContentType() == ContentType.IDENTITY) {
				int sumPartialSizes = 0;
				for (Node parentNode : ((UnionNode) node).getParentNodes()) {
					sumPartialSizes += querySize(parentNode);
				}
				return sumPartialSizes;
			}

			// Otherwise, we use a little trick: We query the cardinality of an
			// empty
			// relationship, because we can already do this. The node size
			// is the number of unlinked elements.
			
			List<Relationship> emptyRelationships = new ArrayList<>();
			for (Node parentNode : ((UnionNode) node).getParentNodes()) {
				emptyRelationships.add(EmptyRelationship.createWithStartNode(parentNode));
			}
			Relationship emptyRelationship = UnionRelationship.union(emptyRelationships);
			NaturalNumberBag cardinality = queryCardinality(emptyRelationship);
			return cardinality.getCountFor(0);

		default:
			throw new IllegalArgumentException("Unprocessable operation type of node " + node);

		}
	}

	@Override
	public Iterator<String> getValues(Node node) {
		if (this.schemaGraph == null || this.db == null) {
			throw new IllegalStateException();
		}
		if (node == null || !this.schemaGraph.containsNode(node)) {
			throw new IllegalArgumentException(String.valueOf(node));
		}
		if (node.getContentType() != Node.ContentType.VALUE) {
			throw new IllegalArgumentException("Node must have value content.");
		}

		String sqlQuery;
		switch (node.getOperationType()) {
		case NONE: {
			String tableName = RelationalSchemaToCSGConverter.extractTableName(node);
			String attributeName = RelationalSchemaToCSGConverter.extractAttributeNames(node).get(0);
			sqlQuery = String.format("SELECT DISTINCT %1$s FROM %2$s WHERE %1$s IS NOT NULL ORDER BY %1$s;",
					attributeName, tableName);
			break;
		}

		case UNION: {
			// SELECT DISTINCT val
			// FROM (SELECT val1 AS val FROM table1 UNION ...) AS values
			// WHERE val IS NOT NULL
			// ORDER BY val
			StringBuilder sb = new StringBuilder(512);
			sb.append("SELECT DISTINCT val \nFROM (");
			String union = "";
			for (Node parentNode : node.getParentNodes()) {
				if (!parentNode.isSimpleValueNode()) {
					throw new IllegalArgumentException("Expected a straight union of simple value nodes, found " + node);
				}
				String tableName = RelationalSchemaToCSGConverter.extractTableName(parentNode);
				String attributeName = CollectionUtil.getAny(RelationalSchemaToCSGConverter
						.extractAttributeNames(parentNode));
				sb.append(union).append("SELECT cast(").append(attributeName).append(" AS text) AS val\n");
				sb.append("FROM ").append(tableName).append("\n");
				sb.append("WHERE ").append(attributeName).append(" IS NOT NULL");
				union = "\nUNION\n";
			}
			sb.append(") AS values\nORDER BY val");
			sqlQuery = sb.toString();
			break;
		}

		default:
			throw new IllegalArgumentException("Unsupported operation type of " + node);

		}

		try {
			Statement stmt = db.getConnection().createStatement();
			Iterator<String> keyIterator = new KeyIterator(stmt, sqlQuery);
			return keyIterator;
		} catch (Exception e) {
			logger.error("Error with following query for {}:\n{}", this.schemaGraph.getName(), sqlQuery);
			throw new RuntimeException(e);
		}
	}

	@Override
	public Iterator<KeyValueCount> queryKeysWithCardinality(Relationship relationship) {
		logger.debug("Querying keys with cardinalities of {}.", relationship);

		if (this.schemaGraph == null || this.db == null) {
			throw new IllegalStateException();
		}

		if (!this.schemaGraph.contains(relationship)) {
			throw new IllegalArgumentException(String.valueOf(relationship));
		}

		List<Path> paths = findPaths(relationship);
		if (paths.isEmpty()) {
			return Collections.emptyIterator();
		}
		Collection<Table> tupleTables = findTablesWithTupleQuery(paths);

		String sqlQuery = createKeysWithCardinalitySqlQuery(paths, tupleTables);
		logger.trace("Executing query: \n{}", sqlQuery);
		try {
			Statement stmt = db.getConnection().createStatement();
			return new KeyValueCountIterator(stmt, sqlQuery, false);
		} catch (Exception e) {
			logger.error("Error with following query for {}:\n{}", this.schemaGraph.getName(), sqlQuery);
			throw new RuntimeException(e);
		}
	}

	private String createKeysWithCardinalitySqlQuery(List<Path> paths, Collection<Table> tupleTables) {

		StringBuilder sb = new StringBuilder(1024);
		createRelationshipWithStatements(paths, sb);

		// Group by key and count the values.
		sb.append("\n");
		sb.append("SELECT count(val) AS card, key FROM (");
		String union = "";
		for (int pathId = 0; pathId < paths.size(); pathId++) {
			sb.append(union).append("SELECT key, val FROM relationship").append(pathId);
			union = " UNION ";
		}

		sb.append(") AS combined_relationship GROUP BY key ORDER BY key;");

		return sb.toString();
	}

	@Override
	public Iterator<KeyValueCount> queryKeyValuesWithCardinality(Relationship relationship) {
		logger.debug("Querying keys with cardinalities of {}.", relationship);

		if (this.schemaGraph == null || this.db == null) {
			throw new IllegalStateException();
		}

		if (!this.schemaGraph.contains(relationship)) {
			throw new IllegalArgumentException(String.valueOf(relationship));
		}

		List<Path> paths = findPaths(relationship);
		if (paths.isEmpty()) {
			return Collections.emptyIterator();
		}
		Collection<Table> tupleTables = findTablesWithTupleQuery(paths);

		String sqlQuery = createKeyValuesWithCardinalitySqlQuery(paths, tupleTables);
		logger.trace("Executing query: \n{}", sqlQuery);
		try {
			Statement stmt = db.getConnection().createStatement();
			return new KeyValueCountIterator(stmt, sqlQuery, true);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	private String createKeyValuesWithCardinalitySqlQuery(List<Path> paths, Collection<Table> tupleTables) {
		StringBuilder sb = new StringBuilder(1024);

		createRelationshipWithStatements(paths, sb);

		// Group by key and count the values.
		sb.append("\n").append("SELECT 1 AS card, key, val FROM (");
		String union = "";
		for (int pathId = 0; pathId < paths.size(); pathId++) {
			sb.append(union).append("SELECT key, val FROM relationship").append(pathId);
			union = " UNION ";
		}
		sb.append(") AS combined_relationship GROUP BY key, val ORDER BY key, val;");

		return sb.toString();
	}

	private static class Path {

		private List<Table> tables;

		public Path() {
			super();
			this.tables = new ArrayList<>(1);
		}

		public void add(Table table) {
			this.tables.add(table);
		}

		public List<Table> getTables() {
			return Collections.unmodifiableList(tables);
		}

		public Table getFirstTable() {
			return this.tables.get(0);
		}

		public Table getLastTable() {
			return this.tables.get(this.tables.size() - 1);
		}

		@Override
		public String toString() {
			return "Path [tables=" + tables + "]";
		}

	}

	private static class Table {

		private String name;

		private List<String> startAttributes, endAttributes, otherAttributes;

		private boolean isRightOpen = false;

		public Table(String name) {
			this.name = name;
			this.startAttributes = new ArrayList<>();
			this.endAttributes = new ArrayList<>();
			this.otherAttributes = new ArrayList<>();
		}

		@SuppressWarnings("unused")
		public void addStartAttribute(String attributeName) {
			this.startAttributes.add(attributeName);
		}

		public void addStartAttributes(List<String> attributeNames) {
			this.startAttributes.addAll(attributeNames);
		}

		@SuppressWarnings("unused")
		public void addEndAttribute(String attributeName) {
			this.endAttributes.add(attributeName);
		}

		public void setIsRightOpen(boolean isRightOpen) {
			this.isRightOpen = isRightOpen;
		}

		public boolean isRightOpen() {
			return isRightOpen;
		}

		public void addEndAttributes(List<String> attributeNames) {
			this.endAttributes.addAll(attributeNames);
		}

		public void addOtherAttribute(String attributeName) {
			this.otherAttributes.add(attributeName);
		}

		public List<String> getOtherAttributes() {
			return this.otherAttributes;
		}

		public String getName() {
			return name;
		}

		public List<String> getStartAttributes() {
			return startAttributes;
		}

		public List<String> getEndAttributes() {
			return endAttributes;
		}

		@Override
		public String toString() {
			return "Table [name=" + name + ", startAttributes=" + startAttributes + ", endAttributes=" + endAttributes
					+ "]";
		}

	}

	private static class KeyValueCountImpl implements KeyValueCount {

		private String value;
		private int count;
		private String key;

		public KeyValueCountImpl(String key, int count) {
			this(key, null, count);
		}

		public KeyValueCountImpl(String key, String value, int count) {
			this.key = key;
			this.value = value;
			this.count = count;
		}

		public int getCount() {
			return count;
		}

		@Override
		public void setCount(int count) {
			this.count = count;
		}

		@Override
		public String getValue() {
			return value;
		}

		@Override
		public String getKey() {
			return key;
		}

		@Override
		public boolean keyValueEqual(KeyValueCount kc) {
			if (this == kc) return true;
			if (kc == null) return false;
			if (key == null) {
				if (kc.getKey() != null) return false;
			} else if (!key.equals(kc.getKey())) return false;
			if (value == null) {
				if (kc.getValue() != null) return false;
			} else if (!value.equals(kc.getValue())) return false;
			return true;
		}

		@Override
		public String toString() {
			if (this.value == null) {
				return count + "x\"" + key + "\"";
			} else {
				return count + "x\"" + key + "\"->\"" + value + "\"";
			}
		}

	}

	private static class KeyValueCountIterator extends JDBCStatementIterator<KeyValueCount> {

		private boolean hasValue;

		public KeyValueCountIterator(Statement stmt, String sqlQuery, boolean hasValue) throws SQLException {
			super(stmt, sqlQuery);
			this.hasValue = hasValue;
		}

		@Override
		protected KeyValueCount readNext(ResultSet resultSet) throws SQLException {
			int count = resultSet.getInt(1);
			String key = resultSet.getString(2);
			if (this.hasValue) {
				String value = resultSet.getString(3);
				return new KeyValueCountImpl(key, value, count);
			} else {
				return new KeyValueCountImpl(key, count);
			}
		}

	}

	private static class KeyIterator extends JDBCStatementIterator<String> {

		public KeyIterator(Statement stmt, String sqlQuery) throws SQLException {
			super(stmt, sqlQuery);
		}

		@Override
		protected String readNext(ResultSet resultSet) throws SQLException {
			String value = resultSet.getString(1);
			return value;
		}

	}

	public static class Factory implements DataStoreAdapter.Factory {

		@Override
		public DataStoreAdapter createAdapter() {
			return new PostgresqlAdapter();
		}

	}

}
