package de.uni_potsdam.hpi.isg.efes.csg.instance;

import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberBag;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;

/**
 * This class represents instances of a {@link BasicSchemaGraph}. However,
 * instead of containing actual data, it is thought to only capture a few
 * aspects of actual instances, e.g. actual cardinalities of relationships and
 * nodes.
 * 
 * @author Sebastian Kruse
 */
public class VirtualSchemaGraphInstance implements SchemaGraphInstance {

	private Logger logger = LogManager.getLogger(VirtualSchemaGraphInstance.class.getSimpleName());

	private static boolean CACHE_ENABLED = false;

	private BasicSchemaGraph schemaGraph;
	private Map<Relationship, NaturalNumberSet> actualCardinalities;
	private Set<Relationship> violatedRelationships;
	private DataStoreAdapter dataStoreAdapter;

	// Real data statistics.
	private Object2IntMap<Node> realNoadSizeCache;
	private Map<Relationship, NaturalNumberBag> realCardinalityCache;

	public VirtualSchemaGraphInstance(BasicSchemaGraph schemaGraph) {
		super();
		this.schemaGraph = schemaGraph;
		this.actualCardinalities = new HashMap<>();
		this.violatedRelationships = new HashSet<>();
		this.realCardinalityCache = new HashMap<>();
		this.realNoadSizeCache = new Object2IntOpenHashMap<>();
		this.realNoadSizeCache.defaultReturnValue(-1);
	}

	/**
	 * Tries to acquire a data store adapter for the given data store and
	 * associate it to this schema graph instance.
	 * 
	 * @param dataStore
	 * @return whether the data store adapter could be successfully acquired
	 */
	public boolean acquireDataStoreAdapter(Object dataStore) {
		if (dataStore == null) {
			throw new IllegalArgumentException();
		}

		if (this.dataStoreAdapter != null)
			return true;

		try {
			this.dataStoreAdapter = DataStoreAdapters.createFor(this.schemaGraph, dataStore);
		} catch (Exception e) {
			logger.warn("An exception occured when trying to acquire a data store adapter.");
			logger.catching(e);
		}
		return this.dataStoreAdapter != null;
	}

	public boolean hasDataStoreAdapter() {
		return this.dataStoreAdapter != null;
	}

	@Override
	public DataStoreAdapter getDataStoreAdapter() {
		return this.dataStoreAdapter;
	}

	public void setActualCardinality(Relationship relationship, NaturalNumberSet cardinality) {
		if (!this.schemaGraph.contains(relationship)) {
			throw new IllegalArgumentException();
		}

		NaturalNumberSet oldCardinality = this.actualCardinalities.get(relationship);
		if (cardinality.equals(oldCardinality)) {
			logger.trace("Not changing cardinality of {}, because it already was {}.", relationship, cardinality);
		}

		NaturalNumberSet prescribedCardinality = getPrescribedCardinality(relationship);
		if (cardinality.isSubsetOf(prescribedCardinality)) {
			this.violatedRelationships.remove(relationship);
			this.actualCardinalities.put(relationship, cardinality);
		} else {
			this.violatedRelationships.add(relationship);
			this.actualCardinalities.put(relationship, cardinality);
		}
	}

	/**
	 * Use this method with caution. It does not interact with the data store
	 * and must be explicitly set beforehand.
	 * 
	 * @param relationship
	 * @return
	 */
	public NaturalNumberSet getActualCardinality(Relationship relationship) {
		NaturalNumberSet actualCardinality = this.actualCardinalities.get(relationship);
		if (actualCardinality == null) {
			logger.debug("Missing actual cardinality for {}, querying it now.", relationship);
			NaturalNumberBag realCardinality = getRealCardinality(relationship);
			actualCardinality = realCardinality.toInterval();
			setActualCardinality(relationship, actualCardinality);
		}

		return actualCardinality;
	}

	public NaturalNumberSet getPrescribedCardinality(Relationship relationship) {
		return this.schemaGraph.getCardinality(relationship);
	}

	public Collection<Relationship> getViolatedRelationships() {
		return Collections.unmodifiableCollection(this.violatedRelationships);
	}

	@Override
	public Collection<Constraint> getViolatedConstraints() {
		List<Constraint> violatedConstraints = new ArrayList<>();
		for (Relationship violatedRelationship : this.violatedRelationships) {
			Constraint constraint = this.schemaGraph.getConstraints(violatedRelationship).toConstraint();
			violatedConstraints.add(constraint);
		}
		return violatedConstraints;
	}

	/**
	 * @deprecated Use {@link #getSchema()} instead.
	 * @return
	 */
	public BasicSchemaGraph getBaseSchemaGraph() {
		return getSchema();
	}

	public NaturalNumberBag getRealCardinality(Relationship relationship) {
		if (relationship == null || !this.schemaGraph.contains(relationship)) {
			throw new IllegalArgumentException(String.valueOf(relationship));
		}

		NaturalNumberBag cardinality = this.realCardinalityCache.get(relationship);
		logger.trace("Cached cardinality of {}: {}", relationship, cardinality);
		if (cardinality == null) {
			if (relationship.getStartNode() == null) {
				cardinality = NaturalNumberBag.EMPTY_BAG;
			} else if (relationship.getEndNode() == null) {
				int startNodeSize = getRealSize(relationship.getStartNode());
				cardinality = new NaturalNumberBag();
				cardinality.count(0, startNodeSize);
			} else {
				if (hasDataStoreAdapter()) {
					cardinality = this.dataStoreAdapter.queryCardinality(relationship);
				} else {
					logger.warn("Neither a cached cardinality bag nor a data store adapter is available for {}. "
							+ "Assuming empty bag.", relationship);
					cardinality = NaturalNumberBag.EMPTY_BAG;
				}
			}
			if (CACHE_ENABLED) {
				setRealCardinality(relationship, cardinality);
			}
		}

		return cardinality;
	}

	public int getRealSize(Node node) {
		if (node == null) {
			throw new NullPointerException("node must not be null.");
//		} else if (!node.isSimpleIdentityNode() && !node.isSimpleValueNode()) {
//			throw new IllegalArgumentException(String.format("Cannot query size of non-atomic node %s.", node));
//		} else if (!this.schemaGraph.getNode(node.getName()).equals(node)) {
//			throw new IllegalArgumentException("Node not contained in schema.");
//		}
		} else if (!this.schemaGraph.containsNode(node)) {
			throw new IllegalArgumentException(String.format("Node not contained in schema graph: %s.", node));
			}
		int size = this.realNoadSizeCache.getInt(node);
		if (size < 0) {
			if (hasDataStoreAdapter()) {
				size = this.dataStoreAdapter.querySize(node);
			} else {
				logger.warn("Neither a cached node size nor a data store adapter is available for {}. "
						+ "Assuming empty node.", node);
				size = 0;
			}
			if (CACHE_ENABLED) {
				this.realNoadSizeCache.put(node, size);
			}
		}

		return size;
	}

	public void setRealCardinality(Relationship relationship, NaturalNumberBag cardinality) {
		logger.trace("Setting cardinality of {} to {}.", relationship, cardinality);
		this.realCardinalityCache.put(relationship, cardinality);

		// Use this opportunity to set the node size.
		// We assume that this instance is put into a valid state before the
		// node size is queried again.
		int numNodes = cardinality.getCardinality();
		Node startNode = relationship.getStartNode();
		this.realNoadSizeCache.put(startNode, numNodes);
	}

	@Override
	public BasicSchemaGraph getSchema() {
		return schemaGraph;
	}

	@Override
	public NaturalNumberSet getCardinality(Relationship relationship) {
		return getRealCardinality(relationship).toInterval();
	}

	@Override
	public boolean areOverlapping(Node node1, Node node2) {
		// TODO: Implement if needed.
		throw new UnsupportedOperationException();
	}

}
