package de.uni_potsdam.hpi.isg.efes.csg.nodes;

import java.util.HashSet;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.uni_potsdam.hpi.isg.utils.db.metamodel.Column;

abstract public class Datatype {

	private static final Pattern RELATIONAL_INT_PATTERN = Pattern.compile("int\\((\\d+)\\)");
	private static final Pattern RELATIONAL_TEXT_PATTERN = Pattern.compile("text");
	private static final Pattern RELATIONAL_VARCHAR_PATTERN = Pattern.compile("varchar\\((\\d+)\\)");
	private static final Pattern RELATIONAL_CHAR_PATTERN = Pattern.compile("char\\((\\d+)\\)");
	private static final Pattern RELATIONAL_TIMESTAMP_PATTERN = Pattern.compile("timestamp");
	private static final Pattern RELATIONAL_BOOLEAN_PATTERN = Pattern.compile("boolean");

	public static enum GeneralType {
		STRING, INTEGER, FLOAT, BOOLEAN, UNKNOWN, TIMESTAMP
	}

	private GeneralType generalType;

	private Set<GeneralType> moreSpecificTypes;

	private Datatype(GeneralType generalType, GeneralType... moreSpecificTypes) {
		this.generalType = generalType;
		this.moreSpecificTypes = new HashSet<>();
		for (GeneralType moreSpecificType : moreSpecificTypes) {
			this.moreSpecificTypes.add(moreSpecificType);
		}
	}

	public boolean canBeConvertedFrom(GeneralType generalType) {
		return this.getGeneralType() == generalType || this.moreSpecificTypes.contains(generalType);
	}

	public GeneralType getGeneralType() {
		return generalType;
	}

	public static class Integer extends Datatype {

		private int precision;

		public Integer(int precision) {
			super(GeneralType.INTEGER, GeneralType.BOOLEAN);
			this.precision = precision;
		}

		public int getPrecision() {
			return precision;
		}

		@Override
		public java.lang.String toString() {
			return "Integer(" + precision + ")";
		}
		
		
	}

	public static class Boolean extends Datatype {

		public Boolean() {
			super(GeneralType.BOOLEAN);
		}

		@Override
		public java.lang.String toString() {
			return "Boolean";
		}


	}
	
	public static class Timestamp extends Datatype {
		
		public Timestamp() {
			super(GeneralType.TIMESTAMP);
		}
		
		@Override
		public java.lang.String toString() {
			return "Timestamp";
		}
		
	}

	public static class String extends Datatype {

		private int length;

		public String(int length) {
			super(GeneralType.STRING, GeneralType.BOOLEAN, GeneralType.TIMESTAMP, GeneralType.FLOAT, GeneralType.INTEGER);
			this.length = length;
		}

		public int getLength() {
			return length;
		}

		public void setLength(int length) {
			this.length = length;
		}

		@Override
		public java.lang.String toString() {
			if (length < 0) {
				return "String";
			} else {
				return "String(" + length + ")";
			}
		}

	}

	public static class Unknown extends Datatype {

		public Unknown() {
			super(GeneralType.UNKNOWN);
		}

		@Override
		public boolean canBeConvertedFrom(GeneralType generalType) {
			return false;
		}

		@Override
		public java.lang.String toString() {
			return "Unknown";
		}
		
		

	}

	/**
	 * Creates a datatype from the relational schema according to the
	 * specification in {@link Column}.
	 * 
	 * @param dataType
	 *            the textual description of the data type
	 * @return a matching datatype or {@link Datatype.GeneralType.UNKNOWN} if no
	 *         matching type is available
	 */
	public static Datatype createForRelationalDataType(java.lang.String dataType) {
		Matcher matcher;
		if ((matcher = RELATIONAL_BOOLEAN_PATTERN.matcher(dataType)).matches()) {
			return new Datatype.Boolean();

		} else if ((matcher = RELATIONAL_INT_PATTERN.matcher(dataType)).matches()) {
			int precision = java.lang.Integer.valueOf(matcher.group(1));
			return new Datatype.Integer(precision);
		
		} else if ((matcher = RELATIONAL_TEXT_PATTERN.matcher(dataType)).matches()) {
			return new Datatype.String(-1);
		
		} else if ((matcher = RELATIONAL_VARCHAR_PATTERN.matcher(dataType)).matches()) {
			int length = java.lang.Integer.valueOf(matcher.group(1));
			return new Datatype.String(length);
		
		} else if ((matcher = RELATIONAL_CHAR_PATTERN.matcher(dataType)).matches()) {
			int length = java.lang.Integer.valueOf(matcher.group(1));
			return new Datatype.String(length);
			
		} else if ((matcher = RELATIONAL_TIMESTAMP_PATTERN.matcher(dataType)).matches()) {
			return new Datatype.Timestamp();
		}
		
		return new Datatype.Unknown();
	}

}
