package de.uni_potsdam.hpi.isg.efes.csg.nodes;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.utils.util.CollectionUtil;
import de.uni_potsdam.hpi.isg.utils.util.Strings;

/**
 * This virtual node represents the union of other nodes.
 * 
 * @author Sebastian Kruse
 */
public class UnionNode extends AbstractNode {
	
	private static Logger logger = LogManager.getLogger(UnionNode.class.getSimpleName());

	private List<Node> parentNodes;
	private ContentType contentType;

	private UnionNode(String name, List<Node> parentNodes) {
		super(name);
		if (parentNodes.size() < 2) {
			throw new IllegalArgumentException("Union nodes should only be used for two or more base nodes.");
		}
		this.parentNodes = new ArrayList<>(parentNodes);
		Collections.sort(parentNodes, new Comparator<Node>() {

			@Override
			public int compare(Node n1, Node n2) {
				return n1.getName().compareTo(n2.getName());
			}});

		this.contentType = Node.ContentType.EMPTY;
		for (Node parentNode : parentNodes) {
			this.contentType = this.contentType.mix(parentNode.getContentType());
		}
	}

	@Override
	public ContentType getContentType() {
		return this.contentType;
	}

	@Override
	public Operation getOperationType() {
		return Node.Operation.UNION;
	}

	public boolean contains(Node node) {
		return this.equals(node) || this.parentNodes.contains(node);
	}
	
	@Override
	public List<Node> getParentNodes() {
		return parentNodes;
	}
	
	public static Node union(Collection<Node> nodes) {
		if (nodes.isEmpty()) {
			throw new IllegalArgumentException(String.valueOf(nodes));
		}
		
		if (nodes.size() == 1) {
			return nodes.iterator().next();
		}
		
		List<Node> deduplicatedNodes = new ArrayList<>(nodes.size());
		collectNodes(nodes, deduplicatedNodes);
		
		if (deduplicatedNodes.isEmpty()) {
			return null;
		} else if (deduplicatedNodes.size() == 1) {
			return CollectionUtil.getAny(deduplicatedNodes);
		}
		
		StringBuilder name = new StringBuilder();
		Collections.sort(deduplicatedNodes);
		name.append("{");
		Strings.join(deduplicatedNodes, ",", name);
		name.append("}");
		return new UnionNode(name.toString(), deduplicatedNodes);
	}

	private static void collectNodes(Collection<Node> nodes, List<Node> deduplicatedNodes) {
		for (Node node : nodes) {
			if (node == null) {
				logger.debug("Skipping a null node.");
				continue;
			} else if (node.getOperationType() == Node.Operation.UNION) {
				UnionNode unionNode = (UnionNode) node;
				collectNodes(unionNode.getParentNodes(), deduplicatedNodes);
			}
			if (!deduplicatedNodes.contains(node)) {
				deduplicatedNodes.add(node);
			}
		}
	}
	
	@Override
	public String toString() {
		return "" + getName() + "";
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		result = prime * result + ((parentNodes == null) ? 0 : parentNodes.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) return true;
		if (!super.equals(obj)) return false;
		if (getClass() != obj.getClass()) return false;
		UnionNode other = (UnionNode) obj;
		if (parentNodes == null) {
			if (other.parentNodes != null) return false;
		} else if (!parentNodes.equals(other.parentNodes)) return false;
		return true;
	}
	
	
	
	

}
