package de.uni_potsdam.hpi.isg.efes.csg.relationships;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.ConstraintStorage;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberInterval;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.CrossProductNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.utils.util.Strings;

public class CoalesceRelationship extends AbstractComplexRelationship {

	private List<Relationship> relationships;
	private Node startNode, endNode;

	public CoalesceRelationship(Relationship... relationships) {
		this(Arrays.asList(relationships));
	}

	public CoalesceRelationship(List<Relationship> relationships) {
		if (relationships == null || relationships.isEmpty())
			throw new IllegalArgumentException(String.valueOf(relationships));
		this.endNode = null;
		List<Node> startNodes = new ArrayList<>();
		for (Relationship relationship : relationships) {
			if (this.endNode == null)
				this.endNode = relationship.getEndNode();
			else if (!this.endNode.equals(relationship.getEndNode()))
				// TODO: Allow different nodes in coalesce relationships.
				throw new IllegalArgumentException("Different coalesce nodes not supported yet.");
			startNodes.add(relationship.getStartNode());
		}
		this.relationships = relationships;
		this.startNode = new CrossProductNode(constructStartNodeName(), startNodes, this);
	}

	private String constructStartNodeName() {
		StringBuilder result = new StringBuilder();
		String separator = "";
		for (Relationship relationship : this.relationships) {
			result.append(separator).append(relationship.getStartNode().getName());
			separator = " x ";
		}
		return result.toString();
	}

	public Node getStartNode() {
		return this.startNode;
	}

	public Node getEndNode() {
		return this.endNode;
	}

	public Relationship compose(Relationship relationship) {
		return new CoalesceRelationship(this, relationship);
	}

	@Override
	public NaturalNumberSet inferCardinality(ConstraintStorage constraintStorage) {
		int minMax = -1;
		for (Relationship relationship : relationships) {
			NaturalNumberSet parentCardinality = constraintStorage.getCardinality(relationship);
			minMax = NaturalNumberInterval.min(minMax, parentCardinality.getMax());
		}
		return new NaturalNumberInterval(1, minMax);
	}

	@Override
	public boolean contains(Node node) {
		for (Relationship relationship : this.relationships) {
			if (relationship.contains(node)) {
				return true;
			}
		}
		return false;
	}

	@Override
	public ContentType getContentType() {
		return Relationship.ContentType.EXPLICIT;
	}

	@Override
	public OperationType getOperationType() {
		return Relationship.OperationType.COALESCE;
	}

	@Override
	public Inverse getInverse() {
		return (Inverse) super.getInverse();
	}

	@Override
	protected Relationship createInverse() {
		return new Inverse();
	}

	@Override
	public Collection<Relationship> collectAtomicRelationships() {
		List<Relationship> atomicRelationships = new ArrayList<>();
		for (Relationship relationship : this.relationships) {
			atomicRelationships.addAll(relationship.getAtomicRelationships());
		}
		return atomicRelationships;
	}

	@Override
	public String toString() {
		return "[" + Strings.join(this.relationships, "|><|") + "]";
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((relationships == null) ? 0 : relationships.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		CoalesceRelationship other = (CoalesceRelationship) obj;
		if (relationships == null) {
			if (other.relationships != null)
				return false;
		} else if (!relationships.equals(other.relationships))
			return false;
		return true;
	}

	public class Inverse extends AbstractComplexRelationship {

		@Override
		public boolean contains(Node node) {
			return getInverse().contains(node);
		}

		@Override
		public Node getStartNode() {
			return getInverse().getEndNode();
		}

		@Override
		public Node getEndNode() {
			return getInverse().getStartNode();
		}

		@Override
		public Relationship compose(Relationship relationship) {
			return new CompositeRelationship(this, relationship);
		}

		@Override
		public NaturalNumberSet inferCardinality(ConstraintStorage constraintStorage) {
			int minProd = 1, maxProd = 1;
			for (Relationship inverseBaseRelationship : getInverse().relationships) {
				Relationship baseRelationship = inverseBaseRelationship.getInverse();
				NaturalNumberSet cardinality = constraintStorage.getCardinality(baseRelationship);
				minProd *= cardinality.getMin();
				maxProd *= cardinality.getMax();
			}
			return new NaturalNumberInterval(minProd, maxProd);
		}

		@Override
		public CoalesceRelationship getInverse() {
			return CoalesceRelationship.this;
		}

		@Override
		protected Relationship createInverse() {
			throw new UnsupportedOperationException();
		}

		@Override
		public int getComplexity() {
			return getInverse().getComplexity();
		}

		@Override
		public int hashCode() {
			return getInverse().hashCode() + 1;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			Inverse other = (Inverse) obj;
			return this.getInverse().equals(other.getInverse());
		}

		@Override
		public String toString() {
			List<Relationship> inverseRelationships = new ArrayList<>();
			for (Relationship relationship : CoalesceRelationship.this.relationships) {
				inverseRelationships.add(relationship.getInverse());
			}
			Collections.reverse(inverseRelationships);
			return "[" + Strings.join(inverseRelationships, "<|>") + "]";
		}

		@Override
		public Collection<Relationship> collectAtomicRelationships() {
			List<Relationship> atomicRelationships = new ArrayList<>();
			for (Relationship relationship : getInverse().getAtomicRelationships()) {
				atomicRelationships.add(relationship.getInverse());
			}
			return atomicRelationships;
		}

		@Override
		public ContentType getContentType() {
			return getInverse().getContentType();
		}

		@Override
		public OperationType getOperationType() {
			return Relationship.OperationType.FORK;
		}

	}

}
