package de.uni_potsdam.hpi.isg.efes.csg.relationships;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.ConstraintStorage;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.utils.util.Strings;

public class CompositeRelationship extends AbstractComplexRelationship {

	private List<Relationship> relationships;

	public CompositeRelationship(Relationship... relationships) {
		initialize(Arrays.asList(relationships));
	}

	protected CompositeRelationship(CompositeRelationship inverse, List<Relationship> relationships) {
		super(inverse);
		initialize(relationships);
	}

	public CompositeRelationship(List<Relationship> relationships) {
		initialize(relationships);
	}

	private void initialize(List<Relationship> relationships) {
		this.relationships = new ArrayList<>(relationships);
	}

	public Node getStartNode() {
		return relationships.get(0).getStartNode();
	}

	public Node getEndNode() {
		return relationships.get(relationships.size() - 1).getEndNode();
	}

	public Relationship compose(Relationship relationship) {
		List<Relationship> newRelationships = new ArrayList<Relationship>();
		newRelationships.addAll(this.relationships);
		newRelationships.add(relationship);
		return new CompositeRelationship(newRelationships);
	}

	public List<Relationship> getComposedRelationships() {
		return Collections.unmodifiableList(relationships);
	}

	public boolean contains(Node node) {
		for (Relationship relationship : this.relationships) {
			if (relationship.contains(node))
				return true;
		}
		return false;
	}

	@Override
	protected Relationship createInverse() {
		List<Relationship> inverseRelationships = new ArrayList<>();
		for (int i = this.relationships.size() - 1; i >= 0; i--) {
			inverseRelationships.add(this.relationships.get(i).getInverse());
		}
		return new CompositeRelationship(this, inverseRelationships);
	}

	@Override
	public int getComplexity() {
		int complexity = 0;
		for (Relationship relationship : this.relationships) {
			complexity += relationship.getComplexity();
		}
		return complexity;
	}

	@Override
	public NaturalNumberSet inferCardinality(ConstraintStorage constraintStorage) {
		NaturalNumberSet result = NaturalNumberSet.ONE;
		for (Relationship relationship : relationships) {
			NaturalNumberSet effectiveCardinality = constraintStorage.getCardinality(relationship);
			result = result.compose(effectiveCardinality);
		}
		return result;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((relationships == null) ? 0 : relationships.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		CompositeRelationship other = (CompositeRelationship) obj;
		if (relationships == null) {
			if (other.relationships != null)
				return false;
		} else if (!relationships.equals(other.relationships))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return "[" + Strings.join(this.relationships, "o") + "]";
	}

	@Override
	public Collection<Relationship> collectAtomicRelationships() {
		List<Relationship> atomicRelationships = new ArrayList<>();
		for (Relationship relationship : this.relationships) {
			atomicRelationships.addAll(relationship.getAtomicRelationships());
		}
		return atomicRelationships;
	}

	@Override
	public ContentType getContentType() {
		return Relationship.ContentType.EXPLICIT;
	}

	@Override
	public OperationType getOperationType() {
		return Relationship.OperationType.COMPOSITE;
	}

}
