package de.uni_potsdam.hpi.isg.efes.csg.relationships;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;


import de.uni_potsdam.hpi.isg.efes.csg.constraints.ConstraintStorage;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberInterval;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.CrossProductNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;

public class CompoundRelationship extends AbstractComplexRelationship {

	private CrossProductNode startNode;
	private CrossProductNode endNode;
	private List<Relationship> baseRelationships;
	private Relationship.ContentType contentType;

	public CompoundRelationship(CrossProductNode startNode, CrossProductNode endNode,
			List<Relationship> baseRelationships) {
		initialize(startNode, endNode, baseRelationships);
	}

	protected CompoundRelationship(CrossProductNode startNode, CrossProductNode endNode,
			List<Relationship> baseRelationships, CompoundRelationship inverse) {
		super(inverse);
		initialize(startNode, endNode, baseRelationships);
	}

	private void initialize(CrossProductNode startNode, CrossProductNode endNode, List<Relationship> baseRelationships) {
		// Sanity checks
		if (startNode == null || endNode == null || baseRelationships == null || baseRelationships.isEmpty())
			throw new NullPointerException();

		List<Node> startNodes = startNode.getParentNodes();
		List<Node> endNodes = endNode.getParentNodes();

		if (startNodes.size() != endNodes.size() || endNodes.size() != baseRelationships.size())
			throw new IllegalArgumentException();

		this.contentType = Relationship.ContentType.EMPTY;
		for (int i = 0; i < baseRelationships.size(); i++) {
			Node curStartNode = startNodes.get(i);
			Node curEndNode = endNodes.get(i);
			Relationship relationship = baseRelationships.get(i);
			this.contentType = this.contentType.mix(relationship.getContentType());
			
			if (!curStartNode.equals(relationship.getStartNode()) || !curEndNode.equals(relationship.getEndNode())) {
				throw new IllegalArgumentException();
			}
		}

		this.startNode = startNode;
		this.endNode = endNode;
		this.baseRelationships = baseRelationships;
	}

	@Override
	public boolean contains(Node node) {
		return this.startNode.contains(node) || this.endNode.contains(node);
	}

	@Override
	public Node getStartNode() {
		return this.startNode;
	}

	@Override
	public Node getEndNode() {
		return this.endNode;
	}

	@Override
	public Relationship compose(Relationship relationship) {
		return new CompositeRelationship(this, relationship);
	}

	@Override
	public NaturalNumberSet inferCardinality(ConstraintStorage constraintStorage) {
		int maxProd = 1;
		for (Relationship baseRelationship : this.baseRelationships) {
			NaturalNumberSet effectiveCardinality = constraintStorage.getCardinality(baseRelationship);
			int newMax = effectiveCardinality.getMax();
			if (maxProd != 0 && newMax != 0
					&& (maxProd == NaturalNumberSet.INFINITE || newMax == NaturalNumberSet.INFINITE)) {
				maxProd = NaturalNumberSet.INFINITE;
			} else {
				maxProd *= newMax;
			}
		}
		return new NaturalNumberInterval(0, maxProd);
	}

	@Override
	protected Relationship createInverse() {
		return new CompoundRelationship(endNode, startNode, invertAll(baseRelationships), this);
	}

	@Override
	public Collection<Relationship> collectAtomicRelationships() {
		List<Relationship> atomicRelationships = new ArrayList<>();
		for (Relationship relationship : this.baseRelationships) {
			atomicRelationships.addAll(relationship.getAtomicRelationships());
		}
		return atomicRelationships;
	}

	@Override
	public ContentType getContentType() {
		return this.contentType;
	}

	@Override
	public OperationType getOperationType() {
		return Relationship.OperationType.COMPOUND;
	}

	@Override
	public String toString() {
		return startNode + "=>" + endNode;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((baseRelationships == null) ? 0 : baseRelationships.hashCode());
		result = prime * result + ((endNode == null) ? 0 : endNode.hashCode());
		result = prime * result + ((startNode == null) ? 0 : startNode.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		CompoundRelationship other = (CompoundRelationship) obj;
		if (baseRelationships == null) {
			if (other.baseRelationships != null)
				return false;
		} else if (!baseRelationships.equals(other.baseRelationships))
			return false;
		if (endNode == null) {
			if (other.endNode != null)
				return false;
		} else if (!endNode.equals(other.endNode))
			return false;
		if (startNode == null) {
			if (other.startNode != null)
				return false;
		} else if (!startNode.equals(other.startNode))
			return false;
		return true;
	}

}
