package de.uni_potsdam.hpi.isg.efes.csg.relationships;

import it.unimi.dsi.fastutil.ints.Int2ObjectMap;
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap;
import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.ConstraintStorage;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.UnionNode;
import de.uni_potsdam.hpi.isg.utils.util.CollectionUtil;
import de.uni_potsdam.hpi.isg.utils.util.Permutations;

public class UnionRelationship extends AbstractComplexRelationship {

	private Node startNode;
	private Node endNode;
	private List<Relationship> baseRelationships;
	private Relationship.ContentType contentType;

	public UnionRelationship(List<Relationship> baseRelationships) {
		initialize(baseRelationships);
	}

	protected UnionRelationship(List<Relationship> baseRelationships, UnionRelationship inverse) {
		super(inverse);
		initialize(baseRelationships);
	}

	private void initialize(List<Relationship> baseRelationships) {
		// Sanity checks
		if (baseRelationships == null || baseRelationships.isEmpty())
			throw new NullPointerException();

		this.contentType = Relationship.ContentType.EMPTY;
		List<Node> startNodes = new ArrayList<>();
		List<Node> endNodes = new ArrayList<>();
		for (Relationship baseRelationship : baseRelationships) {
			startNodes.add(baseRelationship.getStartNode());
			endNodes.add(baseRelationship.getEndNode());
			this.contentType = this.contentType.mix(baseRelationship.getContentType());
		}
		this.startNode = UnionNode.union(startNodes);
		this.endNode = UnionNode.union(endNodes);
		this.baseRelationships = baseRelationships;
	}

	@Override
	public boolean contains(Node node) {
		return this.startNode.contains(node) || this.endNode.contains(node);
	}

	@Override
	public Node getStartNode() {
		return this.startNode;
	}

	@Override
	public Node getEndNode() {
		return this.endNode;
	}

	@Override
	public Relationship compose(Relationship relationship) {
		return new CompositeRelationship(this, relationship);
	}

	public List<Relationship> getBaseRelationships() {
		return baseRelationships;
	}

	@Override
	public NaturalNumberSet inferCardinality(ConstraintStorage constraintStorage) {
		// XXX: This method treats the overlapping relationship like a equality
		// relationship, which it is de facto when evaluating it on schemas but
		// not on data. Adapt this function if necessary.

		Map<Node, Collection<Relationship>> relationshipsPerStartNode = new HashMap<>();
		for (Relationship relationship : this.baseRelationships) {
			CollectionUtil.add(relationship.getStartNode(), relationship, relationshipsPerStartNode);
		}

		Object2IntMap<Node> endNodeToClusterNumber = new Object2IntOpenHashMap<>();
		int unusedClusterNumber = 0;
		for (Relationship relationship : this.baseRelationships) {
			Node endNode = relationship.getEndNode();
			int endNodeClusterNumber = -1;
			for (Object2IntMap.Entry<Node> entry : endNodeToClusterNumber.object2IntEntrySet()) {
				Node node = entry.getKey();
				int clusterNumber = entry.getIntValue();
				if (constraintStorage.areOverlapping(endNode, node)) {
					endNodeClusterNumber = clusterNumber;
					break;
				}
			}
			if (endNodeClusterNumber < 0) {
				endNodeClusterNumber = unusedClusterNumber++;
			}
			endNodeToClusterNumber.put(endNode, endNodeClusterNumber);
		}

		Collection<Collection<Node>> domainDisjointGroups = new ArrayList<>();
		for (Node node : relationshipsPerStartNode.keySet()) {
			Collection<Node> group = null;

			ExistingGroupLoop: for (Collection<Node> existingGroup : domainDisjointGroups) {
				for (Node groupedNode : existingGroup) {
					// Compare the correspondences.
					if (constraintStorage.areOverlapping(node, groupedNode)) {
						group = existingGroup;
						break ExistingGroupLoop;
					}
				}
			}
			if (group == null) {
				group = new ArrayList<>();
				domainDisjointGroups.add(group);
			}
			group.add(node);
		}

		Int2ObjectMap<NaturalNumberSet> cardinalities = new Int2ObjectOpenHashMap<>();
		NaturalNumberSet overallCardinality = NaturalNumberSet.EMPTY_SET;
		for (Collection<Node> domainDisjointGroup : domainDisjointGroups) {
			Collection<List<Node>> allSourceCombinations = Permutations.powerSet((List<Node>) domainDisjointGroup);

			for (List<Node> sourceCombination : allSourceCombinations) {
				if (sourceCombination.isEmpty())
					continue;

				cardinalities.clear();

				for (Node node : sourceCombination) {
					Collection<Relationship> relationships = relationshipsPerStartNode.get(node);

					for (Relationship relationship : relationships) {
						NaturalNumberSet relationshipCardinality = constraintStorage.getCardinality(relationship);

						Node endNode = relationship.getEndNode();
						int clusterNumber = endNodeToClusterNumber.getInt(endNode);
						NaturalNumberSet clusterCardinality = cardinalities.get(clusterNumber);
						if (clusterCardinality == null) {
							clusterCardinality = relationshipCardinality;
						} else {
							clusterCardinality = clusterCardinality.inclExclAdd(relationshipCardinality);
						}
						cardinalities.put(clusterNumber, clusterCardinality);
					}
				}
				NaturalNumberSet sourceCombinationCardinality = NaturalNumberSet.ZERO;
				for (NaturalNumberSet clusterCardinality : cardinalities.values()) {
					sourceCombinationCardinality = sourceCombinationCardinality.add(clusterCardinality);
				}

				overallCardinality = overallCardinality.union(sourceCombinationCardinality);

			}
		}

		// We might yield unrealistic outcomes if we are dealing with an
		// equality relationship. Thus, limit if applicable.
		if (this.contentType == Relationship.ContentType.EQUALITY) {
			overallCardinality = overallCardinality.intersect(NaturalNumberSet.ZERO_OR_ONE);
		}

		return overallCardinality;
	}

	@Override
	protected Relationship createInverse() {
		return new UnionRelationship(invertAll(baseRelationships), this);
	}

	@Override
	public Collection<Relationship> collectAtomicRelationships() {
		List<Relationship> atomicRelationships = new ArrayList<>();
		for (Relationship relationship : this.baseRelationships) {
			atomicRelationships.addAll(relationship.getAtomicRelationships());
		}
		return atomicRelationships;
	}

	@Override
	public ContentType getContentType() {
		return this.contentType;
	}

	@Override
	public OperationType getOperationType() {
		return Relationship.OperationType.UNION;
	}

	@Override
	public String toString() {
		return startNode + "=>" + endNode;
	}

	public static Relationship union(Collection<Relationship> relationships) {
		if (relationships.isEmpty()) {
			return null;
		} else if (relationships.size() == 1) {
			return relationships.iterator().next();
		} else {
			List<Relationship> nonUnionRelationships = new ArrayList<>();
			collectNonUnionRelationships(relationships, nonUnionRelationships);
			return new UnionRelationship(nonUnionRelationships);
		}
	}

	private static void collectNonUnionRelationships(Collection<Relationship> relationships,
			List<Relationship> nonUnionRelationships) {
		for (Relationship relationship : relationships) {
			if (relationship.getOperationType() == Relationship.OperationType.UNION) {
				UnionRelationship unionRelationship = (UnionRelationship) relationship;
				collectNonUnionRelationships(unionRelationship.getBaseRelationships(), nonUnionRelationships);
			} else {
				nonUnionRelationships.add(relationship);
			}
		}
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((baseRelationships == null) ? 0 : baseRelationships.hashCode());
		result = prime * result + ((endNode == null) ? 0 : endNode.hashCode());
		result = prime * result + ((startNode == null) ? 0 : startNode.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		UnionRelationship other = (UnionRelationship) obj;
		if (baseRelationships == null) {
			if (other.baseRelationships != null)
				return false;
		} else if (!baseRelationships.equals(other.baseRelationships))
			return false;
		if (endNode == null) {
			if (other.endNode != null)
				return false;
		} else if (!endNode.equals(other.endNode))
			return false;
		if (startNode == null) {
			if (other.startNode != null)
				return false;
		} else if (!startNode.equals(other.startNode))
			return false;
		return true;
	}

}
