package de.uni_potsdam.hpi.isg.efes.csg.search;

import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.ConstraintCollection;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.ConstraintImpl;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.IdentityRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.util.PartialOrder;
import de.uni_potsdam.hpi.isg.efes.csg.util.PartialOrderQueue;
import de.uni_potsdam.hpi.isg.efes.csg.util.PartialOrder.Type;

public class PartialOrderGraphSearcher {

	private static Logger logger = LogManager.getLogger(PartialOrderGraphSearcher.class.getSimpleName());

	private BasicSchemaGraph graph;
	private PartialOrderQueue<Constraint> queue;
	private Map<Node, List<Constraint>> visitors;
	private PartialOrder<Constraint> order;

	public PartialOrderGraphSearcher(BasicSchemaGraph graph, PartialOrder<Constraint> order) {
		this.graph = graph;
		this.queue = new PartialOrderQueue<Constraint>(order);
		this.order = order;
		this.visitors = new Object2ObjectOpenHashMap<Node, List<Constraint>>(graph.getAtomicNodes().size());
	}

	public Collection<Constraint> searchPaths(Node startNode, Node endNode) {
		return searchPaths(startNode, endNode, false);
	}
	
	/**
	 * Searches paths in the schema graph between the start and end node.
	 * Minimal paths should be returned, i.e., there should be no path in the
	 * result set that is greater than another one in the result set (according
	 * to the given partial order).
	 * 
	 * @param startNode
	 * @param endNode
	 * @return
	 */
	public Collection<Constraint> searchPaths(Node startNode, Node endNode, boolean onlyIdentityRelationships) {
		if (startNode == null || endNode == null)
			throw new NullPointerException();

		if (!this.graph.containsNode(startNode)) {
			throw new IllegalArgumentException("Node not found in graph: " + startNode);
		}
		if (!this.graph.containsNode(endNode)) {
			throw new IllegalArgumentException("Node not found in graph: " + endNode);
		}

		this.queue.clear();
		this.visitors.clear();

		Constraint startAnnotation = new ConstraintImpl(new IdentityRelationship(startNode), NaturalNumberSet.ONE);
		this.visitors.put(startNode, Collections.<Constraint> singletonList(startAnnotation));
		this.queue.add(startAnnotation);

		while (!this.queue.isEmpty()) {
			takeSearchStep(endNode, onlyIdentityRelationships);
		}

		return getVisitors(endNode);
	}

	/**
	 * Choose the most-promising node and expand paths pointing to this node in
	 * order to get closer to the endNode.
	 * 
	 * @param endNode
	 * @param onlyIdentityRelationships 
	 */
	private void takeSearchStep(Node endNode, boolean onlyIdentityRelationships) {
		// Take the most-promising node/path/constraint from the queue.
		Constraint bestConstraint = this.queue.poll();
		Relationship baseRelationship = bestConstraint.getRelationship();
		Node baseNode = baseRelationship.getEndNode();

		// Find expanding relationships for the chose node.
		List<Constraint> candidates = findFollowUpCardinalityAnnotations(baseNode, onlyIdentityRelationships);

		// Try to expand with each of the relationships.
		for (Constraint candidate : candidates) {
			tryToExpand(bestConstraint, candidate, endNode);
		}
	}

	/**
	 * Retrieve outgoing relationships w/ cardinality from the given node in the
	 * schema graph.
	 * @param onlyIdentityRelationships 
	 */
	private List<Constraint> findFollowUpCardinalityAnnotations(Node baseNode, boolean onlyIdentityRelationships) {
		// Get relationships from the graph.
		Collection<Relationship> relationships = new ArrayList<>();
		relationships.addAll(this.graph.getAtomicRelationships(baseNode));
		relationships.addAll(this.graph.getVirtualRelationships(baseNode));
		

		// Collect the constraints for the relationships.
		List<Constraint> result = new ArrayList<Constraint>();
		for (Relationship relationship : relationships) {
			// Sanity check
			if (!relationship.getStartNode().equals(baseNode)) {
				throw new IllegalStateException("Expected node " + baseNode + ", found " + relationship.getStartNode());
			}
			
			if (onlyIdentityRelationships && relationship.getContentType() != Relationship.ContentType.EQUALITY) {
				continue;
			}
			ConstraintCollection constraints = this.graph.getConstraints(relationship);
			result.add(constraints.toConstraint());
		}

		return result;
	}

	/**
	 * Tries to expand the base constraint with the candidate constraint.
	 * However, the expansion is not necessarily succeeding (cycles, better
	 * paths through the target node, ...).
	 * 
	 * @param base
	 * @param candidate
	 * @param endNode
	 */
	private void tryToExpand(Constraint base, Constraint candidate, Node endNode) {
		// Check that the new end node is not already visited by the base path.
		if (base.getRelationship().contains(candidate.getRelationship().getEndNode())) {
			logger.trace("Cycle detected: " + base.getRelationship() + " o " + candidate.getRelationship());
			return;
		}

		// Construct the new path.
		Constraint expandedConstraint = base.compose(candidate);

		logger.trace("Expanding {} ({}) with {}", base.getRelationship().getEndNode(), base.getCardinality(), candidate);

		// Optimization: Check if base the candidate will yield a result that
		// could be used in the end.
		List<Constraint> currentSolutions = getVisitors(endNode);
		for (Constraint currentSolution : currentSolutions) {
			if (order.compare(currentSolution, expandedConstraint) == Type.SMALLER) {
				logger.trace("Dropping {} in favor of {}.", expandedConstraint, currentSolution);
				return;
			}
		}

		Node targetNode = candidate.getRelationship().getEndNode();
		List<Constraint> visitors = getVisitors(targetNode);

		// Consolidate new path with existing visitors.
		for (Iterator<Constraint> i = visitors.iterator(); i.hasNext();) {
			Constraint visitor = i.next();
			// NaturalNumberSet visitorCardinality = visitor.getCardinality();
			PartialOrder.Type visitorOrder = this.order.compare(visitor, expandedConstraint);

			// TODO: It can be that we have cardinalities worse than the current
			// one.
			// We need to delete those with their according relationships.
			// But if they are worse, they will not have been expanded yet.
			// if (newCardinality.isSubsetOf(visitorCardinality)) {
			// if (!newCardinality.equals(visitorCardinality)) {
			if (visitorOrder == PartialOrder.Type.GREATER) {
				// This means, we have a better path already.
				// Delete this one. Probably rarely the case.
				i.remove();
				this.queue.remove(visitor);
			} else if (visitorOrder == PartialOrder.Type.SMALLER) {
				// In this case, there can also not be a worse cardinality. As
				// it would be violating this existing cardinality.
				return;
			}
		}

		this.queue.add(expandedConstraint);
		visitors.add(expandedConstraint);

	}

	private List<Constraint> getVisitors(Node targetNode) {
		List<Constraint> visitors = this.visitors.get(targetNode);
		if (visitors == null) {
			visitors = new ArrayList<Constraint>();
			this.visitors.put(targetNode, visitors);
		}
		return visitors;
	}

}
