package de.uni_potsdam.hpi.isg.efes.csg.util;

import java.util.ArrayList;
import java.util.Collection;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.SchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Datatype;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.ValueNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.utils.util.CollectionUtil;

public class MetaDataDetector {

	private static final Logger logger = LogManager.getLogger(MetaDataDetector.class.getSimpleName());

	private SchemaGraph schemaGraph;

	public MetaDataDetector(SchemaGraph schemaGraph) {
		this.schemaGraph = schemaGraph;
	}

	public void markMetaData() {
		Collection<Node> pkNodes = collectAndMarkMetaDataPrimaryKeys();
		collectAndMarkMetaDataForeignKeys(pkNodes);
		markMetaDataTables();
	}

	private Collection<Node> collectAndMarkMetaDataPrimaryKeys() {
		Collection<Node> primaryKeyNodes = new ArrayList<>();

		for (Node node : this.schemaGraph.getAtomicNodes().values()) {

			// Only regard table nodes.
			if (!node.isSimpleIdentityNode()) {
				continue;
			}

			Collection<Relationship> attributeRelationships = findAttributeRelationships(node);
			Collection<Relationship> pkCandidates = new ArrayList<>(2);
			for (Relationship attributeRelationship : attributeRelationships) {

				ValueNode valueNode = (ValueNode) attributeRelationship.getEndNode();

				// If we already have meta data set, skip this table.
				if (valueNode.isMetaData()) {
					pkCandidates.clear();
					break;
				}

				// The attribute node must be of type integer.
				Datatype datatype = valueNode.getDatatype();
				if (datatype.getGeneralType() != Datatype.GeneralType.INTEGER) {
					continue;
				}

				// The relationship must be 1:1.
				NaturalNumberSet cardinality = this.schemaGraph.getCardinality(attributeRelationship);
				if (!cardinality.equals(NaturalNumberSet.ONE)) {
					continue;
				}
				cardinality = this.schemaGraph.getConstraints(attributeRelationship.getInverse())
						.calculateEffectiveCardinality();
				if (!cardinality.equals(NaturalNumberSet.ONE)) {
					continue;
				}

				// When the attribute node has passed this tests, it is a
				// candidate.
				pkCandidates.add(attributeRelationship);
			}

			// If we have multiple candidates, try to find the most plausible
			// one.
			if (pkCandidates.size() > 1) {
				Collection<Relationship> filteredCandidates = new ArrayList<>(pkCandidates.size());
				for (Relationship candidateRelationship : pkCandidates) {
					String nodeName = candidateRelationship.getEndNode().getName();
					if (nodeName.toLowerCase().endsWith("id")) {
						filteredCandidates.add(candidateRelationship);
					}
				}
				if (!filteredCandidates.isEmpty()) {
					pkCandidates = filteredCandidates;
				}
			}
			if (pkCandidates.size() > 1) {
				Collection<Relationship> filteredCandidates = new ArrayList<>(pkCandidates.size());
				for (Relationship candidateRelationship : pkCandidates) {
					String nodeName = candidateRelationship.getEndNode().getName();
					if (nodeName.equalsIgnoreCase("id")) {
						filteredCandidates.add(candidateRelationship);
					}
				}
				if (!filteredCandidates.isEmpty()) {
					pkCandidates = filteredCandidates;
				}
			}

			// If we did not find a candidate, try again with softened
			// constraint.
			if (pkCandidates.isEmpty()) {
				for (Relationship attributeRelationship : attributeRelationships) {

					ValueNode valueNode = (ValueNode) attributeRelationship.getEndNode();

					if (valueNode.isMetaData()) {
						pkCandidates.clear();
						break;
					}

					// The attribute node must be of type integer.
					Datatype datatype = valueNode.getDatatype();
					if (datatype.getGeneralType() != Datatype.GeneralType.INTEGER) {
						continue;
					}

					String attributeName = CollectionUtil.getAny(RelationalSchemaToCSGConverter
							.extractAttributeNames(valueNode));
					if (attributeName.equalsIgnoreCase("id")) {
						pkCandidates.add(attributeRelationship);
					}
					// When the attribute node has passed this tests, it is a
					// candidate.
				}
			}

			// Evaluate the candidates.
			if (pkCandidates.isEmpty()) {
				logger.debug("No synthetic PK found for {}.", node);
			} else if (pkCandidates.size() > 1) {
				logger.debug("Multiple potential synthetic PKs found for {}: {}.", node, pkCandidates);
			} else {
				Relationship pkRelationship = pkCandidates.iterator().next();
				Node pkNode = pkRelationship.getEndNode();

				primaryKeyNodes.add(pkNode);
				pkNode.setIsMetaData(true);
				logger.debug("{} is assumed to be a synthetic primary key node.", pkNode);
			}

		}

		return primaryKeyNodes;
	}

	/**
	 * Alternative implementation of
	 * {@link #collectAndMarkMetaDataPrimaryKeys()}. Does not take into account
	 * that each table might have multiple PK candidates.
	 * 
	 * @return the alleged primary key nodes
	 */
	@SuppressWarnings("unused")
	private Collection<Node> collectAndMarkMetaDataPrimaryKeys2() {
		Collection<Node> primaryKeyNodes = new ArrayList<>();

		TestNodeLoop: for (Node node : this.schemaGraph.getAtomicNodes().values()) {

			// Deal with previously set meta data.
			if (node.isMetaData()) {
				primaryKeyNodes.add(node);
				continue;
			}

			// The node must be a atomic node with values.
			if (!node.isSimpleValueNode()) {
				continue;
			}

			// The node must be of type integer.
			ValueNode valueNode = (ValueNode) node;
			Datatype datatype = valueNode.getDatatype();
			if (datatype.getGeneralType() != Datatype.GeneralType.INTEGER) {
				continue;
			}

			// The node must have a 1:1 relationship with its table node.
			Collection<Relationship> tableRelationships = findTableRelationships(node);
			for (Relationship tableRelationship : tableRelationships) {
				NaturalNumberSet cardinality = this.schemaGraph.getCardinality(tableRelationship);
				if (!cardinality.equals(NaturalNumberSet.ONE)) {
					continue TestNodeLoop;
				}
				cardinality = this.schemaGraph.getConstraints(tableRelationship.getInverse())
						.calculateEffectiveCardinality();
				if (!cardinality.equals(NaturalNumberSet.ONE)) {
					continue TestNodeLoop;
				}
			}

			// When the node has passed this tests, it is assumed to be
			// metadata.
			primaryKeyNodes.add(node);
			node.setIsMetaData(true);

			logger.debug("{} is assumed to be a synthetic primary key node.", node);
		}

		return primaryKeyNodes;
	}

	private void collectAndMarkMetaDataForeignKeys(Collection<Node> pkNodes) {
		for (Node pkNode : pkNodes) {
			Collection<Relationship> relationships = this.schemaGraph.getAtomicRelationships(pkNode);
			for (Relationship relationship : relationships) {
				if (relationship.isAtomic() && relationship.getContentType() == Relationship.ContentType.EQUALITY) {
					Node fkNode = relationship.getEndNode();
					fkNode.setIsMetaData(true);
					logger.debug("Assuming {} to be a synthetic FK (depending on {}).", fkNode, pkNode);
				}
			}

		}
	}

	private void markMetaDataTables() {
		for (Node tableNode : this.schemaGraph.getAtomicNodes().values()) {

			// Only regard table nodes.
			if (!tableNode.isSimpleIdentityNode()) {
				continue;
			}

			// Check if all attributes are meta data.
			boolean allAttributesAreMetaData = true;
			Collection<Relationship> attributeRelationships = findAttributeRelationships(tableNode);
			for (Relationship attributeRelationship : attributeRelationships) {

				Node attributeNode = attributeRelationship.getEndNode();
				if (!attributeNode.isMetaData()) {
					allAttributesAreMetaData = false;
					break;
				}
			}

			tableNode.setIsMetaData(allAttributesAreMetaData);
			if (allAttributesAreMetaData) {
				logger.debug("{} is assumed to be a meta data table.", tableNode);
			}
		}
	}

	private Collection<Relationship> findTableRelationships(Node node) {
		Collection<Relationship> result = new ArrayList<>(1);
		Collection<Relationship> relationships = this.schemaGraph.getAtomicRelationships(node);
		for (Relationship relationship : relationships) {
			Node endNode = relationship.getEndNode();
			if (relationship.isAtomic() && relationship.isExplicit() && endNode.isSimpleIdentityNode()) {
				result.add(relationship);
			}
		}

		if (result.size() != 1) {
			logger.warn("Expected one table node connected to {}, found {}.", node, result.size());
		}

		return result;
	}

	private Collection<Relationship> findAttributeRelationships(Node tableNode) {
		Collection<Relationship> result = new ArrayList<>(1);
		Collection<Relationship> relationships = this.schemaGraph.getAtomicRelationships(tableNode);
		for (Relationship relationship : relationships) {
			Node endNode = relationship.getEndNode();
			if (relationship.isAtomic() && relationship.isExplicit() && endNode.isSimpleValueNode()) {
				result.add(relationship);
			}
		}

		return result;
	}

}
