package de.uni_potsdam.hpi.isg.efes.csg.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.IdentityNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;

public class RelationalSchemaGraphUtil {

	private static final Logger logger = LogManager
			.getLogger(RelationalSchemaGraphUtil.class);

	/**
	 * Returns nodes of tables that are directly referenced via a foreign key
	 * relationship by the given node.
	 * 
	 * @param graph
	 * @param referencingTableNode
	 * @return
	 */
	public static List<Node> findReferencedTableNodes(BasicSchemaGraph graph,
			Node referencingTableNode) {

		List<Node> referencedTableNodes = new ArrayList<>();
		List<Node> attribteNodes = findAttributeNodes(graph,
				referencingTableNode);
		for (Node attributeNode : attribteNodes) {
			for (Relationship relationship : graph
					.getAtomicRelationships(attributeNode)) {
				if (relationship.isExplicit()) {
					continue;
				}

				NaturalNumberSet cardinality = graph.getConstraints(
						relationship).calculateEffectiveCardinality();
				if (cardinality.isSubsetOf(NaturalNumberSet.ONE)) {
					Node referencedAttributeNode = relationship.getEndNode();
					referencedTableNodes.add(findTableNode(graph,
							referencedAttributeNode));
				}
			}
		}
		return referencedTableNodes;
	}

	/**
	 * Returns nodes of attributes that are directly referenced via a foreign
	 * key relationship by the given attribute node.
	 * 
	 * @param graph
	 * @param referencingTableNode
	 * @return
	 */
	public static List<Node> findReferencedAttributeNodes(
			BasicSchemaGraph graph, Node referencingAttributeNode) {

		List<Node> referencedAttributeNodes = new ArrayList<>();
		for (Relationship relationship : graph
				.getAtomicRelationships(referencingAttributeNode)) {
			if (relationship.isExplicit()) {
				continue;
			}

			NaturalNumberSet cardinality = graph.getConstraints(relationship)
					.calculateEffectiveCardinality();
			if (cardinality.isSubsetOf(NaturalNumberSet.ONE)) {
				Node referencedAttributeNode = relationship.getEndNode();
				referencedAttributeNodes.add(referencedAttributeNode);
			}
		}
		return referencedAttributeNodes;
	}

	public static List<Relationship> findAttributeRelationships(
			IdentityNode tableNode, BasicSchemaGraph targetSchema) {
		List<Relationship> attributeRelationships = new ArrayList<>();
		Collection<Relationship> relationships = targetSchema
				.getAtomicRelationships(tableNode);
		for (Relationship relationship : relationships) {
			if (relationship.isAtomic() && relationship.isExplicit()
					&& relationship.getEndNode().isSimpleValueNode()) {
				attributeRelationships.add(relationship);
			}
		}
		return attributeRelationships;
	}

	/**
	 * Tells if the given attribute node represents a primary key attribute.
	 * 
	 * @param attributeRelationship
	 * @param schema
	 * @return
	 */
	public static boolean isPrimaryKey(Relationship attributeRelationship,
			BasicSchemaGraph schema) {
		NaturalNumberSet cardinality = schema
				.getCardinality(attributeRelationship);
		NaturalNumberSet inverseCardinality = schema
				.getCardinality(attributeRelationship.getInverse());

		return cardinality.equals(NaturalNumberSet.ONE)
				&& inverseCardinality.equals(NaturalNumberSet.ONE);
	}

	/**
	 * Checks if the given node has an equality relationship with another node
	 * with cardinality 1.
	 */
	public static boolean isForeignKeyNode(Node attributeNode,
			BasicSchemaGraph targetSchema) {
		Collection<Relationship> metaDataRelationships = targetSchema
				.getAtomicRelationships(attributeNode);
		for (Relationship metaDataRelationship : metaDataRelationships) {
			if (metaDataRelationship.isAtomic()
					&& metaDataRelationship.getContentType() == Relationship.ContentType.EQUALITY) {
				NaturalNumberSet cardinality = targetSchema
						.getCardinality(metaDataRelationship);
				if (cardinality.equals(NaturalNumberSet.ONE)) {
					return true;
				}
			}
		}
		return false;
	}

	public static void extractTableNames(Relationship relationship,
			Set<String> tableNames) {
		Collection<Relationship> atomicRelationships = relationship
				.getAtomicRelationships();
		for (Relationship atomicRelationship : atomicRelationships) {
			extractTableName(atomicRelationship.getStartNode(), tableNames);
			extractTableName(atomicRelationship.getEndNode(), tableNames);
		}
	}

	public static void extractTableName(Node node, Set<String> tableNames) {
		if (node == null) {
			return;
		}

		if (node.getOperationType() != Node.Operation.NONE) {
			logger.warn("Expected only atomic nodes, found {}.", node);
			return;
		}

		String tableName = RelationalSchemaToCSGConverter
				.extractTableName(node);
		tableNames.add(tableName);
	}

	/**
	 * Returns the table node for a given attribute node.
	 * 
	 * @param graph
	 * @param node
	 * @return
	 */
	public static Node findTableNode(BasicSchemaGraph graph, Node node) {
		if (node instanceof IdentityNode) {
			return node;
		}
		Collection<Relationship> relationships = graph
				.getAtomicRelationships(node);
		for (Relationship relationship : relationships) {
			if (relationship.isAtomic()
					&& relationship.getEndNode() instanceof IdentityNode)
				return relationship.getEndNode();
		}
		throw new RuntimeException(
				"There should have been a node found actually.");
	}

	/**
	 * Returns the attribute nodes for a given table node.
	 * 
	 * @param schemaGraph
	 * @param tableNode
	 * @return
	 */
	public static List<Node> findAttributeNodes(BasicSchemaGraph schemaGraph,
			Node tableNode) {
		List<Node> result = new ArrayList<>();
		Collection<Relationship> relationships = schemaGraph
				.getAtomicRelationships(tableNode);
		for (Relationship relationship : relationships) {
			if (relationship.isAtomic()) {
				Node endNode = relationship.getEndNode();
				if (endNode.isSimpleValueNode()) {
					result.add(endNode);
				}
			}
		}
		return result;
	}

}
