package de.uni_potsdam.hpi.isg.efes.csg.util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberSet;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.CrossProductNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Datatype;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.ValueNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.CoalesceRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.CompoundRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.utils.db.metamodel.Column;
import de.uni_potsdam.hpi.isg.utils.db.metamodel.Constraint;
import de.uni_potsdam.hpi.isg.utils.db.metamodel.ForeignKeyConstraint;
import de.uni_potsdam.hpi.isg.utils.db.metamodel.NotNullConstraint;
import de.uni_potsdam.hpi.isg.utils.db.metamodel.RelationalSchema;
import de.uni_potsdam.hpi.isg.utils.db.metamodel.Table;
import de.uni_potsdam.hpi.isg.utils.db.metamodel.UniqueConstraint;

public class RelationalSchemaToCSGConverter {

	private static Logger logger = LogManager.getLogger(RelationalSchemaToCSGConverter.class.getSimpleName());

	public BasicSchemaGraph convert(RelationalSchema schema) {
		BasicSchemaGraph csg = new BasicSchemaGraph(schema.getName());
		convertStructure(schema, csg);
		convertConstraints(schema, csg);
		return csg;
	}

	private void convertStructure(RelationalSchema schema, BasicSchemaGraph csg) {
		for (Table table : schema.getTables().values()) {
			String tableNodeName = createTableNodeName(csg, table);
			Node tableNode = csg.addIdentityNode(tableNodeName);

			for (Column column : table.getColumns().values()) {
				String columnNodeName = createColumnNodeName(csg, table, column);
				Datatype datatType = Datatype.createForRelationalDataType(column.getDataType());
				ValueNode columnNode = csg.addValueNode(columnNodeName, datatType);

				csg.addExplicitRelationships(tableNode, columnNode);
				csg.addConstraint(tableNode, columnNode, NaturalNumberSet.ZERO_OR_ONE);
				csg.addConstraint(columnNode, tableNode, NaturalNumberSet.GREATER_OR_EQUAL_ONE);
			}
		}
	}

	private void convertConstraints(RelationalSchema schema, BasicSchemaGraph csg) {
		for (Constraint constraint : schema.getConstraints()) {
			switch (constraint.getType()) {
			case NOT_NULL: {
				NotNullConstraint nnConstraint = (NotNullConstraint) constraint;
				convertNotNullConstraint(csg, nnConstraint);
				break;
			}
			case UNIQUE: {
				UniqueConstraint uniqueConstraint = (UniqueConstraint) constraint;
				convertUniqueConstraint(csg, uniqueConstraint);
				break;
			}
			case FOREIGN_KEY:
				ForeignKeyConstraint fkConstraint = (ForeignKeyConstraint) constraint;
				convertForeignKeyConstraint(csg, fkConstraint);
				break;
			case FUNCTIONAL_DEPENDENCY:
				logger.warn("Functional dependencies not handled yet.");
				break;
			case INCLUSION_DEPENDENCY:
				logger.warn("Inclusion dependencies not handled yet.");
				break;
			default:
				logger.warn("Unhandled constraint type: " + constraint.getType());
				break;
			}
		}
	}

	private void convertNotNullConstraint(BasicSchemaGraph csg, NotNullConstraint nnConstraint) {
		String tableName = createTableNodeName(csg, nnConstraint.getTable());
		Node tableNode = csg.getAtomicNodes().get(tableName);
		Node columnNode = csg.getAtomicNodes().get(
				createColumnNodeName(csg, nnConstraint.getTable(), nnConstraint.getColumn()));
		csg.addConstraint(tableNode, columnNode, NaturalNumberSet.GREATER_OR_EQUAL_ONE);
	}

	private void convertUniqueConstraint(BasicSchemaGraph csg, UniqueConstraint uniqueConstraint) {
		Table table = uniqueConstraint.getTable();
		String tableName = createTableNodeName(csg, table);
		Node tableNode = csg.getAtomicNodes().get(tableName);

		List<Column> columns = uniqueConstraint.getColumns();
		List<Node> columnNodes = getColumnNodes(csg, table, columns);

		if (columnNodes.size() > 1) {
			List<Relationship> relationships = collectRelationships(csg, tableNode, columnNodes);
			CoalesceRelationship relationship = new CoalesceRelationship(relationships);
			csg.add(relationship.getStartNode());
			csg.add(relationship);
			csg.add(relationship.getInverse());
			csg.addConstraint(relationship, NaturalNumberSet.ZERO_OR_ONE);
		} else {
			Node columnNode = columnNodes.get(0);
			csg.addConstraint(columnNode, tableNode, NaturalNumberSet.ZERO_OR_ONE);
		}
	}

	private void convertForeignKeyConstraint(BasicSchemaGraph csg, ForeignKeyConstraint fkConstraint) {

		String sourceTableName = createTableNodeName(csg, fkConstraint.getSourceTable());
		Node sourceTableNode = csg.getAtomicNodes().get(sourceTableName);
		List<Node> sourceColumnNodes = getColumnNodes(csg, fkConstraint.getSourceTable(),
				fkConstraint.getSourceColumns());

		String targetTableName = createTableNodeName(csg, fkConstraint.getTargetTable());
		Node targetTableNode = csg.getAtomicNodes().get(targetTableName);
		List<Node> targetColumnNodes = getColumnNodes(csg, fkConstraint.getTargetTable(),
				fkConstraint.getTargetColumns());

		List<Relationship> inclusionRelationships = createEqualityRelationships(csg, sourceColumnNodes,
				targetColumnNodes);

		Relationship inclusionRelationship;
		if (fkConstraint.getArity() == 1) {
			inclusionRelationship = inclusionRelationships.get(0);
		} else {
			List<Relationship> sourceRelationships = collectRelationships(csg, sourceTableNode, sourceColumnNodes);
			CoalesceRelationship.Inverse sourceRelationship = new CoalesceRelationship(sourceRelationships)
					.getInverse();
			csg.add(sourceRelationship.getEndNode());
			csg.add(sourceRelationship);
			csg.add(sourceRelationship.getInverse());

			List<Relationship> targetRelationships = collectRelationships(csg, targetTableNode, targetColumnNodes);
			CoalesceRelationship targetRelationship = new CoalesceRelationship(targetRelationships);
			csg.add(targetRelationship.getStartNode());
			csg.add(targetRelationship);
			csg.add(targetRelationship.getInverse());

			inclusionRelationship = new CompoundRelationship((CrossProductNode) sourceRelationship.getEndNode(),
					(CrossProductNode) targetRelationship.getStartNode(), inclusionRelationships);
			csg.add(inclusionRelationship);
			csg.add(inclusionRelationship.getInverse());
		}

		csg.addConstraint(inclusionRelationship, NaturalNumberSet.ONE);
	}

	private List<Relationship> createEqualityRelationships(BasicSchemaGraph csg, List<Node> startNodes,
			List<Node> endNodes) {
		List<Relationship> result = new ArrayList<>();
		for (int i = 0; i < startNodes.size(); i++) {
			Node startNode = startNodes.get(i);
			Node endNode = endNodes.get(i);
			csg.addEqualityRelationships(startNode, endNode);
			result.add(csg.getAtomicRelationship(startNode, endNode));
		}
		return result;
	}

	private List<Relationship> collectRelationships(BasicSchemaGraph csg, Node tableNode, List<Node> columnNodes) {
		List<Relationship> result = new ArrayList<>();
		for (Node columnNode : columnNodes) {
			result.add(csg.getAtomicRelationship(columnNode, tableNode));
		}
		return result;
	}

	private List<Node> getColumnNodes(BasicSchemaGraph csg, Table table, List<Column> columns) {
		List<Node> columnNodes = new ArrayList<>();
		for (Column column : columns) {
			columnNodes.add(csg.getAtomicNodes().get(createColumnNodeName(csg, table, column)));
		}
		return columnNodes;
	}

	public static String createTableNodeName(BasicSchemaGraph schemaGraph, Table table) {
		return schemaGraph.createSchemaSpecificName(table.getName());
	}

	public static String createColumnNodeName(BasicSchemaGraph schemaGraph, Table table, Column column) {
		return createColumnNodeName(schemaGraph, table.getName(), column.getName());
	}

	public static String createColumnNodeName(BasicSchemaGraph schemaGraph, String tableName, String columnName) {
		return schemaGraph.createSchemaSpecificName(tableName + "." + columnName);
	}

	public static String extractTableName(Node node) {
		String nodeName = node.getName();
		switch (node.getContentType()) {
		case IDENTITY:
			switch (node.getOperationType()) {
			case NONE:
				// Expected pattern: <schema>.<table>
				int dotIndex = nodeName.indexOf('.');
				return nodeName.substring(dotIndex + 1);
			default:
				break;
			}
			break;

		case VALUE:
			switch (node.getOperationType()) {
			case NONE:
				// Expected pattern: <schema>.<table>.<column>
				int firstDotIndex = nodeName.indexOf('.');
				int secondDotIndex = nodeName.indexOf('.', firstDotIndex + 1);
				return nodeName.substring(firstDotIndex + 1, secondDotIndex);
			case CROSS_PRODUCT:
				CrossProductNode crossProductNode = (CrossProductNode) node;
				String tableName = null;
				for (Node parentNode : crossProductNode.getParentNodes()) {
					String newTableName = extractTableName(parentNode);
					if (tableName != null && !tableName.equals(newTableName)) {
						throw new IllegalArgumentException("Inconsistent tables found in " + node);
					}
					tableName = newTableName;
				}
				return tableName;
			default:
				break;
			}
			break;

		default:
			break;
		}

		throw new IllegalArgumentException("Cannot extract a table name from " + node);
	}

	public static List<String> extractAttributeNames(Node node) {
		switch (node.getContentType()) {
		case IDENTITY:
			return Collections.emptyList();

		case VALUE:
			switch (node.getOperationType()) {
			case NONE:
				// Expected pattern: <schema>.<table>.<column>
				String nodeName = node.getName();
				int dotPos = nodeName.lastIndexOf('.');
				String attributeName = node.getName().substring(dotPos + 1);
				return Collections.singletonList(attributeName);
			case CROSS_PRODUCT:
				CrossProductNode crossProductNode = (CrossProductNode) node;
				List<Node> parentNodes = crossProductNode.getParentNodes();
				List<String> attributeNames = new ArrayList<>(2);
				for (Node parentNode : parentNodes) {
					attributeNames.addAll(extractAttributeNames(parentNode));
				}
				return attributeNames;

			default:
				break;
			}
			break;

		default:
			break;
		}

		throw new IllegalArgumentException("Cannot extract a attribute names from " + node);
	}

}
