package de.uni_potsdam.hpi.isg.efes.csg;

import org.junit.Assert;
import org.junit.Test;

import de.uni_potsdam.hpi.isg.efes.csg.constraints.NaturalNumberInterval;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Datatype;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.IdentityNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.Node;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.ValueNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.EqualityRelationship;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.ExplicitAtomicRelationship;

public class SchemaGraphTest {

	@Test
	public void testAddNode() {
		BasicSchemaGraph schemaGraph = new BasicSchemaGraph("test-graph");
		Node node = new IdentityNode("A");

		schemaGraph.add(node);
		Assert.assertTrue("Node not contained in graph.", schemaGraph
				.getAtomicNodes().get("A").equals(node));
	}

	@Test
	public void testDoubleAddNodeFails() {
		BasicSchemaGraph schemaGraph = new BasicSchemaGraph("test-graph");
		Node nodeA = new IdentityNode("A");
		Node nodeB = new IdentityNode("B");

		schemaGraph.add(nodeA);
		schemaGraph.add(nodeB);
		Assert.assertTrue("Node A not contained in graph.", schemaGraph
				.getAtomicNodes().get("A").equals(nodeA));
		Assert.assertTrue("Node B not contained in graph.", schemaGraph
				.getAtomicNodes().get("B").equals(nodeB));

		boolean exceptionCaught = false;
		try {
			schemaGraph.add(nodeA);
		} catch (IllegalArgumentException e) {
			exceptionCaught = true;
		}
		Assert.assertTrue("No exception caught.", exceptionCaught);
	}

	@Test
	public void testAddEdge() {
		BasicSchemaGraph schemaGraph = new BasicSchemaGraph("test-graph");
		Node nodeA = new IdentityNode("A");
		Node nodeB = new IdentityNode("B");
		Relationship edge = new EqualityRelationship(nodeA, nodeB);

		schemaGraph.add(nodeA);
		schemaGraph.add(nodeB);
		schemaGraph.add(edge);
	}

	@Test
	public void testEdgeWithMissingNodeFails1() {
		BasicSchemaGraph schemaGraph = new BasicSchemaGraph("test-graph");
		Node nodeA = new IdentityNode("A");
		Node nodeB = new IdentityNode("B");
		Relationship edge = new EqualityRelationship(nodeA, nodeB);

		schemaGraph.add(nodeA);

		boolean exceptionCaught = false;
		try {
			schemaGraph.add(edge);
		} catch (IllegalArgumentException e) {
			exceptionCaught = true;
		}

		Assert.assertTrue("No exception caught.", exceptionCaught);
	}

	@Test
	public void testEdgeWithMissingNodeFails2() {
		BasicSchemaGraph schemaGraph = new BasicSchemaGraph("test-graph");
		Node nodeA = new IdentityNode("A");
		Node nodeB = new IdentityNode("B");
		Relationship edge = new EqualityRelationship(nodeA, nodeB);

		schemaGraph.add(nodeB);

		boolean exceptionCaught = false;
		try {
			schemaGraph.add(edge);
		} catch (IllegalArgumentException e) {
			exceptionCaught = true;
		}

		Assert.assertTrue("No exception caught.", exceptionCaught);
	}

	@Test
	public void testBuildingCompleteGraph() {
		// build relational schema R(A, B), S(C, D)
		// with FK R(B) references S(C)
		// and PK R(A) and S(C)
		// and NOT NULL S(D)

		BasicSchemaGraph schemaGraph = new BasicSchemaGraph("test-graph");

		// R(A, B)
		IdentityNode r = new IdentityNode("R");
		schemaGraph.add(r);
		ValueNode a = new ValueNode("A", new Datatype.Unknown());
		schemaGraph.add(a);
		ExplicitAtomicRelationship aToR = new ExplicitAtomicRelationship(a, r);
		schemaGraph.add(aToR);
		schemaGraph.addConstraint(aToR, new NaturalNumberInterval(1, 1));
		ExplicitAtomicRelationship rToA = new ExplicitAtomicRelationship(r, a);
		schemaGraph.add(rToA);
		schemaGraph.addConstraint(rToA, new NaturalNumberInterval(1, 1));
		ValueNode b = new ValueNode("B", new Datatype.Unknown());
		schemaGraph.add(b);
		Relationship bToR = new ExplicitAtomicRelationship(b, r);
		schemaGraph.add(bToR);
		schemaGraph.addConstraint(bToR, new NaturalNumberInterval(1));
		Relationship rToB = new ExplicitAtomicRelationship(r, b);
		schemaGraph.add(rToB);
		schemaGraph.addConstraint(rToB, new NaturalNumberInterval(0, 1));

		// S(C, D)
		IdentityNode s = new IdentityNode("S");
		schemaGraph.add(s);
		ValueNode c = new ValueNode("C", new Datatype.Unknown());
		schemaGraph.add(c);
		ExplicitAtomicRelationship cToS = new ExplicitAtomicRelationship(c, s);
		schemaGraph.add(cToS);
		schemaGraph.addConstraint(cToS, new NaturalNumberInterval(1, 1));
		ExplicitAtomicRelationship sToC = new ExplicitAtomicRelationship(s, c);
		schemaGraph.add(sToC);
		schemaGraph.addConstraint(sToC, new NaturalNumberInterval(1, 1));
		ValueNode d = new ValueNode("D", new Datatype.Unknown());
		schemaGraph.add(d);
		Relationship dToS = new ExplicitAtomicRelationship(d, s);
		schemaGraph.add(dToS);
		schemaGraph.addConstraint(dToS, new NaturalNumberInterval(1));
		Relationship sToD = new ExplicitAtomicRelationship(s, d);
		schemaGraph.add(sToD);
		schemaGraph.addConstraint(sToD, new NaturalNumberInterval(1, 1));
		
		// FK R(B) references S(C)
		Relationship bToC = new EqualityRelationship(b, c);
		schemaGraph.add(bToC);
		schemaGraph.addConstraint(bToC, new NaturalNumberInterval(1, 1));
		Relationship cToB = new EqualityRelationship(c, b);
		schemaGraph.add(cToB);
		schemaGraph.addConstraint(cToB, new NaturalNumberInterval(0, 1));
		
		// Expected: 1 path, C = 0..1
		// System.out.println(schemaGraph.findCardinality(a, d));
	}

}
