package de.uni_potsdam.hpi.isg.efes.csg.util;

import java.util.Collection;
import java.util.Map;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import de.uni_potsdam.hpi.isg.efes.csg.BasicSchemaGraph;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.Constraint;
import de.uni_potsdam.hpi.isg.efes.csg.constraints.ConstraintCollection;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.IdentityNode;
import de.uni_potsdam.hpi.isg.efes.csg.nodes.ValueNode;
import de.uni_potsdam.hpi.isg.efes.csg.relationships.Relationship;
import de.uni_potsdam.hpi.isg.utils.db.metamodel.Column;
import de.uni_potsdam.hpi.isg.utils.db.metamodel.ForeignKeyConstraint;
import de.uni_potsdam.hpi.isg.utils.db.metamodel.NotNullConstraint;
import de.uni_potsdam.hpi.isg.utils.db.metamodel.RelationalSchema;
import de.uni_potsdam.hpi.isg.utils.db.metamodel.Table;
import de.uni_potsdam.hpi.isg.utils.db.metamodel.UniqueConstraint;

public class RelationalSchemaToCSGConverterTest {

	private RelationalSchema relSchema;

	@Before
	public void buildRelationalSchema() {
		relSchema = new RelationalSchema("test_schema");

		// Table album
		Table album = relSchema.addTable("album");
		Column id = album.addColumn("id");
		relSchema.addConstraint(new NotNullConstraint(album, id));
		relSchema.addConstraint(new UniqueConstraint(album, id));

		Column albumName = album.addColumn("name");
		relSchema.addConstraint(new NotNullConstraint(album, albumName));

		Column barcode = album.addColumn("barcode");
		relSchema.addConstraint(new UniqueConstraint(album, barcode));

		// Table track
		Table track = relSchema.addTable("track");
		Column album_id = track.addColumn("album_id");
		relSchema.addConstraint(new NotNullConstraint(track, album_id));
		relSchema.addConstraint(new ForeignKeyConstraint(track,
				new Column[] { album_id }, album, new Column[] { id }));

		Column position = track.addColumn("position");
		relSchema.addConstraint(new NotNullConstraint(track, position));
		relSchema
				.addConstraint(new UniqueConstraint(track, album_id, position));

		Column trackName = track.addColumn("name");
		relSchema.addConstraint(new NotNullConstraint(track, trackName));

		track.addColumn("length");

		// Table tag
		Table tag = relSchema.addTable("tag");
		Column tagAlbum_id = tag.addColumn("album_id");
		relSchema.addConstraint(new NotNullConstraint(tag, tagAlbum_id));

		Column track_position = tag.addColumn("track_position");
		relSchema.addConstraint(new NotNullConstraint(tag, track_position));
		relSchema.addConstraint(new ForeignKeyConstraint(tag, new Column[] {
				tagAlbum_id, track_position }, track, new Column[] { album_id,
				position }));

		Column value = tag.addColumn("value");
		relSchema.addConstraint(new NotNullConstraint(tag, value));
	}

	@Test
	public void testName() {
		RelationalSchemaToCSGConverter converter = new RelationalSchemaToCSGConverter();
		BasicSchemaGraph schemaGraph = converter.convert(relSchema);

		Assert.assertEquals("test_schema", schemaGraph.getName());
	}

	@Test
	public void testTables() {
		RelationalSchemaToCSGConverter converter = new RelationalSchemaToCSGConverter();
		BasicSchemaGraph schemaGraph = converter.convert(relSchema);
		Assert.assertEquals(IdentityNode.class, schemaGraph.getNode("album")
				.getClass());
		Assert.assertEquals(IdentityNode.class, schemaGraph.getNode("track")
				.getClass());
		Assert.assertEquals(IdentityNode.class, schemaGraph.getNode("tag")
				.getClass());
	}

	@Test
	public void testAttributes() {
		RelationalSchemaToCSGConverter converter = new RelationalSchemaToCSGConverter();
		BasicSchemaGraph schemaGraph = converter.convert(relSchema);
		Assert.assertEquals(ValueNode.class, schemaGraph.getNode("album.id")
				.getClass());
		Assert.assertEquals(ValueNode.class, schemaGraph.getNode("album.name")
				.getClass());
		Assert.assertEquals(ValueNode.class,
				schemaGraph.getNode("album.barcode").getClass());
		Assert.assertEquals(ValueNode.class,
				schemaGraph.getNode("track.album_id").getClass());
		Assert.assertEquals(ValueNode.class,
				schemaGraph.getNode("track.position").getClass());
		Assert.assertEquals(ValueNode.class, schemaGraph.getNode("track.name")
				.getClass());
		Assert.assertEquals(ValueNode.class, schemaGraph
				.getNode("track.length").getClass());
		Assert.assertEquals(ValueNode.class, schemaGraph
				.getNode("tag.album_id").getClass());
		Assert.assertEquals(ValueNode.class,
				schemaGraph.getNode("tag.track_position").getClass());
		Assert.assertEquals(ValueNode.class, schemaGraph.getNode("tag.value")
				.getClass());
	}

	@Test
	public void testDefaultConstaints() {
		RelationalSchemaToCSGConverter converter = new RelationalSchemaToCSGConverter();
		BasicSchemaGraph schemaGraph = converter.convert(relSchema);

		Assert.assertEquals(ValueNode.class, schemaGraph.getNode("album.id")
				.getClass());
		Assert.assertEquals(ValueNode.class, schemaGraph.getNode("album.name")
				.getClass());
		Assert.assertEquals(ValueNode.class,
				schemaGraph.getNode("album.barcode").getClass());
		Assert.assertEquals(ValueNode.class,
				schemaGraph.getNode("track.album_id").getClass());
		Assert.assertEquals(ValueNode.class,
				schemaGraph.getNode("track.position").getClass());
		Assert.assertEquals(ValueNode.class, schemaGraph.getNode("track.name")
				.getClass());
		Assert.assertEquals(ValueNode.class, schemaGraph
				.getNode("track.length").getClass());
		Assert.assertEquals(ValueNode.class, schemaGraph
				.getNode("tag.album_id").getClass());
		Assert.assertEquals(ValueNode.class,
				schemaGraph.getNode("tag.track_position").getClass());
		Assert.assertEquals(ValueNode.class, schemaGraph.getNode("tag.value")
				.getClass());
	}

	@Test
	@Deprecated
	public void testPrintSchemaGraph() {
		RelationalSchemaToCSGConverter converter = new RelationalSchemaToCSGConverter();
		BasicSchemaGraph schemaGraph = converter.convert(relSchema);

		Map<Relationship, ConstraintCollection> constraintCollections = schemaGraph
				.getConstraintCollections();
		for (ConstraintCollection collection : constraintCollections.values()) {
			System.out.println(collection);
		}
		// System.out.println(schemaGraph);

		Collection<Constraint> paths;
		//
		// schemaGraph.getNode("album.barcode"),
		// schemaGraph.getNode("tag.value"));
		// System.out.println(paths);
		// paths = schemaGraph.findCardinality(
		// schemaGraph.getNode("tag.value"),
		// schemaGraph.getNode("album.barcode"));
		// System.out.println(paths);
		// paths = schemaGraph.findCardinality(
		paths = schemaGraph.findCardinality(
				schemaGraph.getNode("tag"), schemaGraph.getNode("track"));
		System.out.println(paths);

	}

}
