package de.uni_potsdam.hpi.isg.efes.spicy;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import de.uni_potsdam.hpi.isg.efes.spicy.SpicyConfiguration.Correspondence;
import de.uni_potsdam.hpi.isg.efes.spicy.SpicyConfiguration.RelationalDataset;

/**
 * Parses configuration files of the ++Spicy tool and returns their content as
 * {@link SpicyConfiguration} object.
 * 
 * @author Sebastian Kruse
 */
public class SpicyParser {

	private static Logger logger = LogManager.getLogger(SpicyParser.class
			.getSimpleName());

	private DocumentBuilderFactory documentBuilderFactory;

	private DocumentBuilder documentBuilder;

	public SpicyParser() {
		this.documentBuilderFactory = DocumentBuilderFactory.newInstance();
		try {
			documentBuilder = this.documentBuilderFactory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			logger.error(e);
			throw new RuntimeException(e);
		}
	}

	public SpicyConfiguration parse(File mappingTaskFile) throws SAXException,
			IOException {
		return parse(new FileInputStream(mappingTaskFile));
	}

	public SpicyConfiguration parse(InputStream mappingTaskFileStream)
			throws SAXException, IOException {
		SpicyConfiguration configuration = new SpicyConfiguration();
		Document document = this.documentBuilder.parse(mappingTaskFileStream);

		Node targetNode = getUniqueNode(document, "target");
		configuration.setTarget(parseRelationalDatasource(targetNode));

		Node sourceNode = getUniqueNode(document, "source");
		configuration.setSource(parseRelationalDatasource(sourceNode));

		Node correspondencesNode = getUniqueNode(document, "correspondences");
		configuration
				.setCorrespondences(parseCorrespondences(correspondencesNode));

		return configuration;
	}

	private List<Correspondence> parseCorrespondences(Node correspondencesNode) {
		List<Correspondence> correspondences = new ArrayList<Correspondence>();

		NodeList childNodes = correspondencesNode.getChildNodes();
		for (int i = 0; i < childNodes.getLength(); i++) {
			Node childNode = childNodes.item(i);
			if ("correspondence".equals(childNode.getNodeName())) {
				correspondences.add(parseCorrespondence(childNode));
			}
		}

		return correspondences;
	}

	private Correspondence parseCorrespondence(Node correspondenceNode) {
		Correspondence correspondence = new Correspondence();
		Map<String, Node> childNodes = childNodesToMap(correspondenceNode);
		Node sourcePathsNode = childNodes.get("source-paths");
		NodeList sourcePathsNodeChildren = sourcePathsNode.getChildNodes();
		for (int i = 0; i < sourcePathsNodeChildren.getLength(); i++) {
			Node sourcePathNode = sourcePathsNodeChildren.item(i);
			if ("source-path".equals(sourcePathNode.getNodeName())) {
				correspondence.addSourcePath(sourcePathNode.getTextContent());
			}
		}
		Node confidenceNode = childNodes.get("confidence");
		correspondence.setConfidence(Double.parseDouble(confidenceNode
				.getTextContent()));

		Node targetPathNode = childNodes.get("target-path");
		correspondence.setTargetPath(targetPathNode.getTextContent());

		Node transformationFunctionNode = childNodes
				.get("transformation-function");
		correspondence.setTransformationFunction(transformationFunctionNode
				.getTextContent());

		return correspondence;
	}

	private RelationalDataset parseRelationalDatasource(Node node) {
		Map<String, Node> children = childNodesToMap(node);
		if (!"relational".equals(children.get("type").getTextContent())) {
			throw new RuntimeException();
		}
		Node relationalNode = children.get("relational");
		children = childNodesToMap(relationalNode);

		RelationalDataset dataset = new RelationalDataset();
		dataset.setDbDriver(children.get("driver").getTextContent());
		dataset.setDbUri(children.get("uri").getTextContent());
		dataset.setDbLogin(children.get("login").getTextContent());
		dataset.setDbPassword(children.get("password").getTextContent());

		return dataset;
	}

	private Map<String, Node> childNodesToMap(Node node) {
		Map<String, Node> children = new HashMap<String, Node>();
		NodeList childList = node.getChildNodes();
		for (int i = 0; i < childList.getLength(); i++) {
			Node child = childList.item(i);
			children.put(child.getNodeName(), child);
		}
		return children;
	}

	private Node getUniqueNode(Document document, String nodeName) {
		NodeList nodeList = document.getElementsByTagName(nodeName);
		if (nodeList.getLength() == 0) {
			logger.warn("No node with the name " + nodeName + " found.");
			return null;
		}

		if (nodeList.getLength() > 1)
			logger.warn(
					"Multiple nodes with the name {} found, will return only first.",
					nodeName);
		return nodeList.item(0);
	}

}
