package de.uni_potsdam.hpi.isg.utils.db;

import java.io.BufferedReader;
import java.io.IOException;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import de.uni_potsdam.hpi.isg.utils.datatypes.Utf8String;
import de.uni_potsdam.hpi.isg.utils.db.metamodel.SchemaExtractor;
import de.uni_potsdam.hpi.isg.utils.settings.DatabaseSettings;

/**
 * Databases encapsulate the database operations needed for our tasks.
 * Database-specific implementations shall be done in subclasses.
 */
public abstract class Database<Self extends Database<Self>> {

	public static final boolean IS_VERBOSE = false;

	/** Properties key for the database type. */
	public static final String PROP_DB_TYPE = "db.type";

	/** Properties key for the database name. */
	public static final String PROP_DB_NAME = "db.name";

	/** How many rows to preload in memory on DB queries. */
	public static final int MAX_FETCH_SIZE = 10000;

	/** Connection to the database. */
	public Connection connection;

	protected Properties properties;

	protected Database(Properties properties) throws SQLException {
		this.properties = properties;
	}

	/**
	 * Executes commands independent of the database, e.g.,
	 * <tt>CREATE DATABASE</tt> commands. It can therefore be used without
	 * calling {@link #open()} beforehand.
	 */
	public abstract void executeWithoutDb(String command) throws SQLException;
	
	/**
	 * @return a schema extractor that can create a metamodel of this database.
	 */
	public abstract SchemaExtractor<Self> getSchemaExtractor();

	/**
	 * Creates the database represented by this {@link Database} instance.
	 */
	public void create() throws SQLException {
		create(false);
	}

	/**
	 * Creates the database represented by this {@link Database} instance.
	 * 
	 * @param dbMayExist
	 *            whether the database may already exist
	 */
	public void create(boolean dbMayExist) throws SQLException {
		executeWithoutDb("CREATE DATABASE "
				+ (dbMayExist ? "IF NOT EXISTS " : "") + getDbName());
	}

	/**
	 * Creates the database represented by this {@link Database} instance.
	 */
	public void drop() throws SQLException {
		executeWithoutDb("DROP DATABASE IF EXISTS " + getDbName());
	}

	/**
	 * Open a new connection to the database. Should only be used when the
	 * {@link Database} instance is fresh or the old connection has been closed.
	 * 
	 * @return <code>this</code> object
	 * 
	 * @see Database#newInstance(Properties)
	 * @see #close()
	 */
	public Database<Self> open() throws SQLException {
		log("Connecting...");
		connection = createConnection();
		connection.setAutoCommit(true);
		return this;
	}

	private void log(String string) {
		if (IS_VERBOSE) {
			if (string.length() > 500)
				string = string.substring(0, 500) + "...";
			System.out.println(getDbName() + "> " + string);
		}
	}

	private String getDbName() {
		return properties.getProperty(PROP_DB_NAME);
	}

	/**
	 * Close the database connection.
	 */
	public void close() throws SQLException {
		log("Closing...");
		if (isOpen()) {
			connection.close();
		}
	}

	/**
	 * Tells whether the database is opened and ready for operation.
	 */
	public boolean isOpen() {
		try {
			return connection != null && !connection.isClosed();
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	/** Create a new connection to the database. */
	abstract protected Connection createConnection() throws SQLException;

	/**
	 * Give access to the connection, so that arbitrary DB operations are
	 * possible.
	 */
	public Connection getConnection() {
		return connection;
	}

	/** Find all tables of the current scheme. */
	abstract public List<String> findTables() throws SQLException;

	/**
	 * Find all preprocessed tables of the current scheme (txxx_cxxx).
	 * 
	 * @throws SQLException
	 */
	public List<String> findPreprocessedTables() throws SQLException {
		List<String> tables = findTables();
		for (Iterator<String> i = tables.iterator(); i.hasNext();) {
			if (!i.next().matches("t\\d{3}_c\\d{3}")) {
				i.remove();
			}
		}
		return tables;
	}

	/** Find all columns of a given table. */
	abstract public List<String> findColumns(String table) throws SQLException;

	/**
	 * Read and delete first value from the column
	 * 
	 * @param rowNumber
	 * @throws SQLException
	 **/
	abstract public Utf8String readValue(String table, String column,
			Integer rowNumber) throws SQLException;

	/**
	 * Create a table "table_column" that contains all sorted distinct values of
	 * "column".
	 */
	abstract public void createSortedTable(String table, String column)
			throws SQLException;

	/**
	 * Renames a table.
	 */
	abstract public void renameTable(String oldName, String newName)
			throws SQLException;

	public void execute(String query) throws SQLException {
		Statement stmt = connection.createStatement();
		log(query);
		stmt.execute(query);
		// connection.commit();
		stmt.close();
	}

	public ResultSet executeQuery(String query) throws SQLException {
		// connection.setAutoCommit(false);
		Statement stmt = connection.createStatement();
		stmt.setFetchSize(MAX_FETCH_SIZE);
		return stmt.executeQuery(query);
	}

	public Integer getNumberOfDistinctValues(String table, String column)
			throws SQLException {
		Integer result;
		Statement statement = connection.createStatement();
		String request = "select count(" + column + ") from " + table + ";";
		System.out.println(request);
		ResultSet resultSet = statement.executeQuery(request);
		resultSet.next();
		result = resultSet.getInt(1);
		return result;
	}

	public Integer getMaxValueLength(String table, String column)
			throws SQLException {
		Integer result;
		Statement statement = connection.createStatement();
		String request = "select max(length(" + column + ")) from " + table
				+ ";";
		System.out.println(request);
		ResultSet resultSet = statement.executeQuery(request);
		resultSet.next();
		result = resultSet.getInt(1);
		return result;
	}

	/**
	 * Reads an SQL script from the given reader and execute it.<br>
	 * <i>NB: Only line comments (<tt>-- Comment</tt>) are allowed and they MUST
	 * NOT be mixed with statements (<tt>Statement -- Comment</tt>)</i>
	 */
	public void runScript(BufferedReader reader) throws IOException,
			SQLException {
		StringBuilder sb = new StringBuilder();
		String line;
		while ((line = reader.readLine()) != null) {
			line = line.trim();
			if (line.isEmpty() || line.startsWith("--"))
				continue;

			sb.append(line).append(" ");
			if (line.endsWith(";")) {
				String command = sb.toString();
				execute(command);
				sb.setLength(0);
			}
		}
	}
	
	public String getName() {
		return this.properties.getProperty(PROP_DB_NAME);
	}

	/**
	 * Creates a {@link Database} instance as specified by the given properties.
	 * This instance will not be opened automatically.
	 * 
	 * @see #create(boolean)
	 * @see #open()
	 * @see #drop()
	 * @see DatabaseSettings#toProperties()
	 */
	public static Database<?> newInstance(Properties properties)
			throws SQLException {
		Database<?> db;
		System.out.println(properties);
		switch (properties.getProperty(PROP_DB_TYPE)) {
		case "postgresql":
			db = new PostgresqlDatabase(properties);
			break;
		case "mysql":
			db = new MySqlDatabase(properties);
			break;
		default:
			throw new IllegalArgumentException("Unknown DB!");
		}
		// db.open();
		return db;
	}

}
