package de.uni_potsdam.hpi.isg.utils.db.metamodel;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.uni_potsdam.hpi.isg.utils.db.PostgresqlDatabase;

public class PostgresqlSchemaExtractor implements SchemaExtractor<PostgresqlDatabase> {

	private static Logger logger = LogManager.getLogger(PostgresqlSchemaExtractor.class.getSimpleName());

	private static final String DEFAULT_SCHEMA_NAME = "public";

	/**
	 * SQL query template to find out whether a schema is existent in the
	 * database.<br>
	 * Parameter 1: name of the schema.<br>
	 * Result: single row, single attribute, boolean, if schema exists.
	 */
	private static final String SCHEMA_EXISTENCE_SQL = "SELECT COUNT(*) > 0 AS schema_exists "
			+ "FROM information_schema.schemata " + "WHERE schema_name = '%s'";

	/**
	 * SQL query template to find the tables of a schema.<br>
	 * Parameter 1: name of the schema.<br>
	 * Result: single attribute 'table_name', all tables.
	 */
	private static final String TABLE_EXTRACTION_SQL = "SELECT table_name " + "FROM information_schema.tables "
			+ "WHERE table_schema = '%s'";

	/**
	 * SQL query template to find the columns of a schema.<br>
	 * Parameter 1: name of the schema.<br>
	 * Result: (table_name, column_name, data_type, is_nullable) for each
	 * column.
	 */
	private static final String COLUMN_EXTRACTION_SQL = "SELECT table_name, column_name, data_type, "
			+ "is_nullable, COALESCE(character_maximum_length, numeric_precision, -1) AS data_type_length "
			+ "FROM information_schema.columns " + "WHERE table_schema = '%s'";

	/**
	 * SQL query template to find the unique and PK constraints of a schema.<br>
	 * Parameter 1: name of the schema.<br>
	 * Result: (table_name, column_names (";" separated) for constraint.
	 */
	private static final String UNIQUE_CONSTRAINT_EXTRACTION_SQL = "SELECT tc.table_name, "
			+ "string_agg(kcu.column_name, ';' ORDER BY kcu.ordinal_position) AS column_names "
			+ "FROM information_schema.table_constraints tc, " + "information_schema.key_column_usage kcu "
			+ "WHERE tc.table_schema = '%s' " + "AND tc.constraint_schema = kcu.constraint_schema "
			+ "AND tc.constraint_name = kcu.constraint_name " + "AND tc.constraint_type IN ('UNIQUE', 'PRIMARY KEY') "
			+ "GROUP BY tc.constraint_name, tc.table_name;";

	/**
	 * SQL query template to find the FK constraints in a schema.<br>
	 * Paramter: name of the schema<br>
	 * Result: (target_table_name, target_column_names, source_table_name,
	 * source_column_names)
	 */
	private static final String FK_CONSTRAINT_EXTRACTION_SQL = "SELECT pkcu.table_name AS target_table_name, "
			+ "string_agg(pkcu.column_name, ';' ORDER BY pkcu.ordinal_position) AS target_column_names, "
			+ "fkcu.table_name AS source_table_name, "
			+ "string_agg(fkcu.column_name, ';' ORDER BY pkcu.ordinal_position) AS source_column_names "
			+ "FROM information_schema.referential_constraints rc, " + "information_schema.key_column_usage pkcu, "
			+ "information_schema.key_column_usage fkcu " + "WHERE rc.constraint_schema = fkcu.constraint_schema "
			+ "AND rc.constraint_name = fkcu.constraint_name "
			+ "AND rc.unique_constraint_schema = pkcu.constraint_schema "
			+ "AND rc.unique_constraint_name = pkcu.constraint_name "
			+ "AND fkcu.position_in_unique_constraint = pkcu.ordinal_position " + "AND fkcu.table_schema = '%1$s' "
			+ "AND pkcu.table_schema = '%1$s' " + "GROUP BY rc.constraint_name, fkcu.table_name, pkcu.table_name";

	private PostgresqlDatabase db;

	/*
	 * tables, columns, column_domain_usage, check_constraint,
	 * constraint_column_usage
	 */

	public PostgresqlSchemaExtractor(PostgresqlDatabase db) {
		setDatabase(db);
	}

	public void setDatabase(PostgresqlDatabase db) {
		this.db = db;
	}

	public RelationalSchema extractSchema(String schemaName) throws SQLException {
		if (schemaName == null) {
			logger.info("No schema name given, using public as default.");
			schemaName = DEFAULT_SCHEMA_NAME;
		}

		boolean dbWasAlreadyOpen;
		if (!(dbWasAlreadyOpen = this.db.isOpen())) {
			this.db.open();
		}

		checkSchemaExistence(schemaName);
		String schemaId = this.db.getName() + "-" + schemaName;
		RelationalSchema resultSchema = new RelationalSchema(schemaId);
		extractTables(resultSchema, schemaName);
		extractColumns(resultSchema, schemaName);
		extractUniqueConstraints(resultSchema, schemaName);
		extractForeignKeyConstraints(resultSchema, schemaName);

		if (!dbWasAlreadyOpen)
			db.close();

		return resultSchema;
	}

	private void extractUniqueConstraints(RelationalSchema resultSchema, String schemaName) throws SQLException {
		String uniqueConstraintExtractionSQL = String.format(UNIQUE_CONSTRAINT_EXTRACTION_SQL, schemaName);

		try (ResultSet resultSet = db.executeQuery(uniqueConstraintExtractionSQL)) {
			while (resultSet.next()) {
				String tableName = resultSet.getString("table_name");
				String columnNameList = resultSet.getString("column_names");

				Table table = resultSchema.getTables().get(tableName);
				Column[] columns = resolveColumnNames(columnNameList, table);

				Constraint uniqueConstraint = new UniqueConstraint(table, columns);
				resultSchema.addConstraint(uniqueConstraint);
			}
		}
	}

	private Column[] resolveColumnNames(String columnNameList, Table table) {
		String[] columnNames = columnNameList.split(";");
		Column[] columns = new Column[columnNames.length];
		for (int i = 0; i < columnNames.length; i++) {
			String columnName = columnNames[i];
			columns[i] = table.getColumns().get(columnName);
		}
		return columns;
	}

	private void extractForeignKeyConstraints(RelationalSchema resultSchema, String schemaName) throws SQLException {
		String fkConstraintExtractionSQL = String.format(FK_CONSTRAINT_EXTRACTION_SQL, schemaName);

		try (ResultSet resultSet = db.executeQuery(fkConstraintExtractionSQL)) {
			while (resultSet.next()) {
				String sourceTableName = resultSet.getString("source_table_name");
				String targetTableName = resultSet.getString("target_table_name");
				String sourceColumnNameList = resultSet.getString("source_column_names");
				String targetColumnNameList = resultSet.getString("target_column_names");

				Table sourceTable = resultSchema.getTables().get(sourceTableName);
				Column[] sourceColumns = resolveColumnNames(sourceColumnNameList, sourceTable);
				Table targetTable = resultSchema.getTables().get(targetTableName);
				Column[] targetColumns = resolveColumnNames(targetColumnNameList, targetTable);

				Constraint fkConstraint = new ForeignKeyConstraint(sourceTable, sourceColumns, targetTable,
						targetColumns);
				resultSchema.addConstraint(fkConstraint);
			}
		}
	}

	private void extractColumns(RelationalSchema resultSchema, String schemaName) throws SQLException {
		String columnExtractionSQL = String.format(COLUMN_EXTRACTION_SQL, schemaName);
		Map<String, Table> tables = resultSchema.getTables();

		try (ResultSet resultSet = db.executeQuery(columnExtractionSQL)) {
			while (resultSet.next()) {
				String tableName = resultSet.getString("table_name");
				String columnName = resultSet.getString("column_name");
				String dataType = resultSet.getString("data_type");
				String isNullable = resultSet.getString("is_nullable");
				int dataTypeLength = resultSet.getInt("data_type_length");

				String normedDataType;
				if ("boolean".equals(dataType)) {
					normedDataType = "boolean";
				} else if ("integer".equals(dataType) || "smallint".equals(dataType) || "bigint".equals(dataType)) {
					normedDataType = String.format("int(%d)", dataTypeLength);
				} else if ("character".equals(dataType)) {
					normedDataType = String.format("char(%d)", dataTypeLength);
				} else if ("character varying".equals(dataType) || "text".equals(dataType)) {
					if (dataTypeLength > 0) {
						normedDataType = String.format("varchar(%d)", dataTypeLength);
					} else {
						normedDataType = "text";
					}
				} else if (dataType.contains("time") || dataType.contains("date")) {
					normedDataType = "timestamp";
				} else {
					normedDataType = dataType;
				}
				// TODO: Handle floating point data types.

				Table table = tables.get(tableName);
				Column column = new Column(columnName, normedDataType);
				table.addColumn(column);

				if (isNullable.equalsIgnoreCase("NO")) {
					resultSchema.addConstraint(new NotNullConstraint(table, column));
				}
			}
		}
	}

	private void extractTables(RelationalSchema resultSchema, String schemaName) throws SQLException {
		String tableExtractionSQL = String.format(TABLE_EXTRACTION_SQL, schemaName);
		try (ResultSet resultSet = db.executeQuery(tableExtractionSQL)) {
			while (resultSet.next()) {
				String tableName = resultSet.getString("table_name");
				resultSchema.addTable(tableName);
			}
		}
	}

	private void checkSchemaExistence(String schemaName) throws SQLException {
		final String existenceSQL = String.format(SCHEMA_EXISTENCE_SQL, schemaName);
		try (ResultSet resultSet = db.executeQuery(existenceSQL)) {
			if (!resultSet.next())
				throw new RuntimeException("Expected a result row.");
			final boolean schemaExists = resultSet.getBoolean("schema_exists");
			if (!schemaExists) {
				throw new IllegalArgumentException();
			}
		}
		;
	}

}
