package de.uni_potsdam.hpi.isg.utils.files;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

public class AsyncFileQueueInputStream extends InputStream {

	private List<File> files;

	private Thread readerThread;

	private int readerPosition, userPosition;

	private BlockingQueue<Buffer> freeBuffers, readyBuffers;

	private Buffer userBuffer;

	volatile boolean isReaderThreadRunning;

	public AsyncFileQueueInputStream(List<File> files, int bufferSize,
			int numBuffers) {
		this.files = files;

		this.readerPosition = -1;
		this.userPosition = -1;

		// Add an extra position, no need for blocking due to an upper buffer limit.
		this.readyBuffers = new LinkedBlockingQueue<Buffer>(numBuffers + 1);
		this.freeBuffers = new LinkedBlockingQueue<Buffer>(numBuffers + 1);
		for (int i = 0; i < numBuffers; i++) {
			freeBuffers.add(new Buffer(bufferSize));
		}

		startReader();
	}

	private void startReader() {
		isReaderThreadRunning = true;
		readerThread = new Thread(new ReaderJob());
		readerThread.start();
	}

	public File moveToNextFile() {
		userPosition++;
		return userPosition < files.size() ? files.get(userPosition) : null;
	}

	@Override
	public int read() throws IOException {
		if (!findCurrentUserBuffer())
			return -1;

		int readByte = userBuffer.bytes[userBuffer.offset++];
		if (userBuffer.getReadableBytes() == 0) {
			freeUserBuffer();
		}

		return readByte;
	}

	@Override
	public int read(byte[] b, int off, int len) throws IOException {
		if (!findCurrentUserBuffer())
			return -1;

		int readBytes = Math.min(len, userBuffer.getReadableBytes());
		System.arraycopy(userBuffer.bytes, userBuffer.offset, b, off, readBytes);
		userBuffer.offset += readBytes;
		if (userBuffer.getReadableBytes() == 0)
			freeUserBuffer();

		return readBytes;
	}

	private boolean findCurrentUserBuffer() {
		try {
			while (userBuffer == null) {
				userBuffer = isReaderThreadRunning ? readyBuffers.take()
						: readyBuffers.poll();

				if (userBuffer.associatedFilePos < userPosition) {
					System.out.println("Warning: a buffer was ignored!");
					freeBuffers.put(userBuffer);
					userBuffer = null;
				}
			}
		} catch (InterruptedException e) {
			throw new RuntimeException(e);
		}

		return userBuffer.associatedFilePos == userPosition;
	}

	private void freeUserBuffer() {
		try {
			freeBuffers.put(userBuffer);
			userBuffer = null;
		} catch (InterruptedException e) {
			throw new RuntimeException(e);
		}
	}

	private boolean moveReaderToNextFile() {
		readerPosition++;
		return readerPosition < files.size();
	}

	private void loadFileToBuffers() throws InterruptedException {
		InputStream inputStream = null;
		try {
			File file = files.get(readerPosition);
			inputStream = new FileInputStream(file);
			int readBytes;
			do {
				readBytes = fillBuffer(inputStream);
			} while (readBytes > 0);
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			close(inputStream);
		}
	}

	private int fillBuffer(InputStream inputStream)
			throws InterruptedException, IOException {
		int readBytes;
		Buffer buffer = freeBuffers.take();
		readBytes = readAsMuchAsPossible(inputStream, buffer.bytes);
		if (readBytes > 0) {
			buffer.length = readBytes;
			buffer.offset = 0;
			buffer.associatedFilePos = readerPosition;
			readyBuffers.put(buffer);
		} else {
			freeBuffers.put(buffer);
		}
		return readBytes;
	}

	private int readAsMuchAsPossible(InputStream inputStream,
			byte[] buffer) throws IOException {
		int offset = 0;
		int readBytes = 0;
		while (offset < buffer.length && readBytes >= 0) {
			offset += readBytes;
			readBytes = inputStream
					.read(buffer, offset, buffer.length - offset);
		}
		if (offset == 0 && readBytes == -1)
			return -1;
		return offset;
	}

	private void close(InputStream inputStream) {
		if (inputStream != null) {
			try {
				inputStream.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}
	
	public void stopReading() {
		if (readerThread != null && readerThread.isAlive()) {
			readerThread.interrupt();
		}
	}
	
	private class ReaderJob implements Runnable {

		@Override
		public void run() {
			try {
				while (moveReaderToNextFile()) {
					loadFileToBuffers();
				}
			} catch (InterruptedException e) {
				e.printStackTrace();
			}

			isReaderThreadRunning = false;

			// Finally add an empty block to avoid endless waiting in the read
			// functions.
			try {
				Buffer buffer = new Buffer(0);
				buffer.associatedFilePos = files.size();
				buffer.length = 0;
				buffer.offset = 0;
				readyBuffers.put(buffer);
			} catch (InterruptedException e) {
				throw new RuntimeException(e);
			}
		}

	}

	private static class Buffer {
		int associatedFilePos;
		byte[] bytes;
		int offset;
		int length;

		Buffer(int size) {
			bytes = new byte[size];
		}

		int getReadableBytes() {
			return length - offset;
		}
	}

}
