package de.uni_potsdam.hpi.isg.utils.files;

import it.unimi.dsi.fastutil.ints.IntArrayList;
import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.fastutil.ints.IntList;
import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Arrays;

/**
 * Takes a column file, compresses it, and writes a binary output file.
 */
public class DictionaryCompressor {

	private static final int INPUT_BUFFER_SIZE = 0x1000000;
	private static final int OUTPUT_BUFFER_SIZE = 0x1000000;

	public static final int NO_VALUE = -1;

	public static final String INT_COLUMN_FILE_EXTENSION = ".intcol";

	public static final FileFilter INT_COLUMUN_FILE_FILTER = new FileFilter() {

		@Override
		public boolean accept(File file) {
			return file.isFile()
					&& file.getName().endsWith(INT_COLUMN_FILE_EXTENSION);
		}
	};

	private Object2IntMap<String> dictionary;

	private IntList encodedColumn;

	/**
	 * Creates a new {@link DictionaryCompressor} for the given input file.
	 */
	public DictionaryCompressor(int expectedColumnSize) {
		super();
		this.dictionary = new Object2IntOpenHashMap<String>();
		this.dictionary.defaultReturnValue(NO_VALUE);
		this.encodedColumn = new IntArrayList(expectedColumnSize);
	}

	/**
	 * Clear the dictionary (may be used when encoding multiple files).
	 */
	public void clearDictionary() {
		dictionary.clear();
	}

	/**
	 * Execute the compression.
	 */
	public void execute(File inputFile, File outputFile) throws IOException {
		execute(new BufferedReader(new FileReader(inputFile), INPUT_BUFFER_SIZE),
				new BufferedOutputStream(new FileOutputStream(inputFile),
						OUTPUT_BUFFER_SIZE));
	}

	/**
	 * Execute the compression.
	 */
	public void execute(BufferedReader reader, BufferedOutputStream outputStream)
			throws IOException {
		String line;
		int encoding = NO_VALUE + 1;

		// read values and fill the dictionary
		encodedColumn.clear();
		while ((line = reader.readLine()) != null) {
			int val = dictionary.getInt(line);
			if (val == NO_VALUE)
				dictionary.put(new String(line), val = encoding++);
			encodedColumn.add(val);
		}
		reader.close();

		// write the encoded file
		byte[] buffer = new byte[4];
		IntIterator i = encodedColumn.iterator();
		while (i.hasNext()) {
			int val = i.nextInt();
			ByteEncodingUtil.intToBytes(val, buffer);
			outputStream.write(buffer);
		}
		outputStream.close();
		reader.close();
	}

	/**
	 * Reads all column files from the input directory, compressed them, and
	 * writes them to the output directory.
	 */
	public static void compressAllFiles(File inputDirectory,
			File outputDirectory, int expectedNumberOfRows) throws IOException {

		outputDirectory.mkdirs();

		File[] columnFiles = inputDirectory
				.listFiles(FileSplitter.COLUMN_FILE_FILTER);
		DictionaryCompressor compressor = new DictionaryCompressor(
				expectedNumberOfRows);
		AsyncFileQueueInputStream queueInputStream = new AsyncFileQueueInputStream(
				Arrays.asList(columnFiles), 0x3000000, 10);
		File columnFile;
		try {
		while ((columnFile = queueInputStream.moveToNextFile()) != null) {
			compressor.clearDictionary();
			String originalFileName = columnFile.getName();
			String newFileName = originalFileName.replace(
					FileSplitter.COLUMN_FILE_EXTENSION,
					DictionaryCompressor.INT_COLUMN_FILE_EXTENSION);
			File outputFile = new File(outputDirectory, newFileName);
			System.out.println("Compressing " + outputFile.getName());
			BufferedReader reader = new BufferedReader(new InputStreamReader(
					queueInputStream));
			BufferedOutputStream outputStream = new BufferedOutputStream(
					new FileOutputStream(outputFile));
			compressor.execute(reader, outputStream);
		}
		queueInputStream.close();
		} finally {
			queueInputStream.stopReading();
		}
	}

	/**
	 * A loader for compressed columns.
	 */
	public static class ColumnLoader {

		private IntArrayList columnBuffer;

		public ColumnLoader(int estimatedRows) {
			this.columnBuffer = new IntArrayList(estimatedRows);
		}

		public synchronized int[] loadColumn(File inputFile) throws IOException {
			columnBuffer.clear();
			BufferedInputStream in = new BufferedInputStream(
					new FileInputStream(inputFile));
			byte[] buffer = new byte[4];
			while (in.read(buffer) == buffer.length) {
				columnBuffer.add(ByteEncodingUtil.byteToInt(buffer));
			}
			in.close();
			int[] column = columnBuffer.toIntArray();
			return column;
		}
	}

}
