package de.uni_potsdam.hpi.isg.utils.files;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileFilter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * Takes an input TSV file and generates multiple TSV files, each containing a
 * single column of the input file.
 */
public class FileSplitter {

	private static final int OUTPUT_BUFFER_SIZE = 1024*1024;

	public static final String COLUMN_FILE_EXTENSION = ".tsvcol";

	public static final FileFilter COLUMN_FILE_FILTER = new FileFilter() {

		@Override
		public boolean accept(File file) {
			return file.isFile()
					&& file.getName().endsWith(
							FileSplitter.COLUMN_FILE_EXTENSION);
		}
	};

	private List<BufferedWriter> writers = new ArrayList<BufferedWriter>();
	private File outputDir, inputFile;
	private int numCols;

	public FileSplitter(File inputFile, File dir, int numCols) {
		this.inputFile = inputFile;
		this.outputDir = dir;
		this.numCols = numCols;
	}

	public FileSplitter(File inputFile, File columnDir) {
		this.inputFile = inputFile;
		this.outputDir = columnDir;
		this.numCols = TsvReader.getNumCols(inputFile);
	}

	public void execute() throws IOException {
		// Prepare writers and reader.
		TsvReader reader = new TsvReader(inputFile, numCols);
		if (!outputDir.exists()) {
			outputDir.mkdirs();
		}
		for (int i = 0; i < numCols; i++) {
			File outputFile = new File(outputDir, createColumnFileName(i));
			if (!outputFile.exists()) {
				outputFile.createNewFile();
			}
			writers.add(new BufferedWriter(new FileWriter(outputFile, false), OUTPUT_BUFFER_SIZE));
		}

		// Read the records and write them to the output files.
		while (reader.moveToNextRow()) {
			String[] fields = reader.getCurrentRow();
			for (int col = 0; col < numCols; col++) {
				writers.get(col).write(fields[col]);
				writers.get(col).write("\n");
			}
		}

		// Close the writers.
		for (BufferedWriter writer : writers) {
			writer.close();
		}
	}

	private String createColumnFileName(int i) {
		String fileName = inputFile.getName().replace(TsvReader.TSV_FILE_EXTENSION, "");
		return String.format("%s-%03d%s", fileName, i,
				COLUMN_FILE_EXTENSION);
	}

	/**
	 * @param args
	 *            Input Folder, File Name
	 * @throws IOException
	 */
	public static void main(String[] args) throws IOException {
		File dir = new File(args[0]);
		File inputFile = new File(dir, args[1]);
		int numCols = Integer.parseInt(args[2]);

		if (!inputFile.exists()) {
			System.err.println("Invalid file: " + inputFile.getAbsolutePath());
			System.exit(1);
		}

		FileSplitter fileSplitter = new FileSplitter(inputFile, dir, numCols);
		fileSplitter.execute();
	}

}
