package de.uni_potsdam.hpi.isg.utils.files;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.List;

/**
 * Takes an input TSV file and outputs its rows as String arrays.
 */
public class TsvReader {

	private static final int INPUT_BUFFER_SIZE = 1024*1024*100;

	public static final String TSV_FILE_EXTENSION = ".tsv";
	
	private File inputFile;
	private int numCols;
	private String[] fields;
	private BufferedReader reader;
	private List<String> tempFieldList = new ArrayList<String>();

	public TsvReader(File inputFile, int numCols) throws IOException {
		this.inputFile = inputFile;
		this.numCols = numCols;
		open();
	}

	public TsvReader(File inputFile) throws IOException {
		this(inputFile, -1);
	}

	private void open() throws IOException {
		reader = new BufferedReader(new InputStreamReader(new FileInputStream(inputFile), Charset.forName("UTF-8")), INPUT_BUFFER_SIZE);
	}
	
	/**
	 * Tries to move the cursor to the next row of the input file.
	 * @return whether the cursor could be moved (i.e. there are rows left)
	 */
	public boolean moveToNextRow() {
		if (!isOpen()) {
			throw new IllegalStateException("The TSV file is not opened any more.");
		}
		
		String line = null;
		try {
			line = reader.readLine();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		if (line != null) {
			fields = split(line);
		} else {
			fields = null;
			close();
		}
		
		return fields != null;
	}
	
	private boolean isOpen() {
		return reader != null;
	}

	/**
	 * Splits the given line and writes the fields into <code>fields</code>.
	 */
	private String[] split(String line) {
		
		int startIndex = 0;
		int curIndex;
		tempFieldList.clear();
		for (curIndex = 0; curIndex < line.length(); curIndex++) {
			if (line.charAt(curIndex) == '\t') {
				tempFieldList.add(line.substring(startIndex, curIndex));
				startIndex = curIndex + 1;
			}
		}
		tempFieldList.add(line.substring(startIndex, curIndex));
		if ((numCols != -1)&&(numCols != tempFieldList.size())) {
			System.err.println("Warning: detected " + tempFieldList.size()
					+ " fields, expected " + numCols + ".");
		}
		if (fields == null || fields.length != tempFieldList.size()) {
			fields = new String[tempFieldList.size()];
		}
		return tempFieldList.toArray(fields);
	}

	public void close() {
		if (!isOpen())
			return;
		
		try {
			reader.close();
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			reader = null;
		}
	}

	public String[] getCurrentRow() {
		return fields;
	}

	/** Count the columns in the first row of the given file. 
	 * @throws IOException */
	public static int getNumCols(File inputFile) {
		TsvReader reader;
		try {
			reader = new TsvReader(inputFile);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
		if (!reader.moveToNextRow()) {
			return 0;
		}
		int numCols = reader.getCurrentRow().length;
		reader.close();
		return numCols;
	}

}
