package de.uni_potsdam.hpi.isg.utils.db.metamodel;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.net.URL;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import de.uni_potsdam.hpi.isg.utils.db.Database;
import de.uni_potsdam.hpi.isg.utils.db.PostgresqlDatabase;

public class PostgresSchemaExtractorTest {

	@SuppressWarnings("unused")
	private static Logger logger = LogManager
			.getLogger(PostgresSchemaExtractorTest.class);

	 @Before
	 public void setUpDatabase() throws SQLException, FileNotFoundException,
	 IOException {
	 PostgresqlDatabase db = (PostgresqlDatabase)
	 Database.newInstance(loadProperties());
	 db.create();
	 db.open();
	 URL scriptURL = this.getClass().getResource("/create_test_schema.sql");
	 try (BufferedReader reader = new BufferedReader(new FileReader(new
	 File(scriptURL.getFile())))) {
	 db.runScript(reader);
	 } finally {
	 db.close();
	 }
	 }

	@Test
	public void testExtractingSchema() throws FileNotFoundException,
			IOException, SQLException {
		final Properties properties = loadProperties();

		PostgresqlDatabase db = (PostgresqlDatabase) Database
				.newInstance(properties);
		db.open();
		PostgresqlSchemaExtractor extractor = new PostgresqlSchemaExtractor(db);
		RelationalSchema schema = extractor.extractSchema("public");
		db.close();

		assertSchemaContainsTable(schema, "album", "id", "name", "barcode");
		assertSchemaContainsTable(schema, "track", "album_id", "position",
				"name", "length");
		assertSchemaContainsUniqueConstraint(schema, "album", "id");
		assertSchemaContainsUniqueConstraint(schema, "album", "barcode");
		assertSchemaContainsUniqueConstraint(schema, "track", "album_id",
				"position");
		assertSchemaContainsNotNullConstraint(schema, "album", "id");
		assertSchemaContainsNotNullConstraint(schema, "album", "name");
		assertSchemaContainsNotNullConstraint(schema, "track", "album_id");
		assertSchemaContainsNotNullConstraint(schema, "track", "position");
		assertSchemaContainsNotNullConstraint(schema, "track", "name");
		assertSchemaContainsForeignKey(schema, "track",
				new String[] { "album_id" }, "album", new String[] { "id" });
	}

	private void assertSchemaContainsForeignKey(RelationalSchema schema,
			String sourceTableName, String[] sourceColumnNames,
			String targetTableName, String[] targetColumnNames) {
		Collection<Constraint> constraints = schema.getConstraints();
		int numMatchingConstraints = 0;
		for (Constraint constraint : constraints) {
			if (constraint.getType() == Constraint.Type.FOREIGN_KEY) {
				ForeignKeyConstraint fkConstraint = (ForeignKeyConstraint) constraint;
				boolean sourceMatches = matchTableAndColumns(sourceTableName,
						fkConstraint.getSourceTable(), sourceColumnNames,
						fkConstraint.getSourceColumns());
				boolean targetMatches = matchTableAndColumns(targetTableName,
						fkConstraint.getTargetTable(), targetColumnNames,
						fkConstraint.getTargetColumns());
				if (sourceMatches && targetMatches)
					numMatchingConstraints++;
			}
		}

		Assert.assertEquals(1, numMatchingConstraints);
	}

	private void assertSchemaContainsNotNullConstraint(RelationalSchema schema,
			String tableName, String columnName) {
		Collection<Constraint> constraints = schema.getConstraints();
		int numMatchingConstraints = 0;
		for (Constraint constraint : constraints) {
			if (constraint.getType() == Constraint.Type.NOT_NULL) {
				NotNullConstraint nnConstraint = (NotNullConstraint) constraint;
				if (nnConstraint.getTable().getName().equals(tableName)
						&& nnConstraint.getColumn().getName()
								.equals(columnName)) {
					numMatchingConstraints++;
				}
			}
		}

		Assert.assertEquals(1, numMatchingConstraints);
	}

	private void assertSchemaContainsUniqueConstraint(RelationalSchema schema,
			String tableName, String... columnNames) {

		Collection<Constraint> constraints = schema.getConstraints();
		int numMatchingConstraints = 0;
		for (Constraint constraint : constraints) {
			if (constraint.getType() == Constraint.Type.UNIQUE) {
				UniqueConstraint uniqueConstraint = (UniqueConstraint) constraint;
				Table table = uniqueConstraint.getTable();
				List<Column> columns = uniqueConstraint.getColumns();
				if (matchTableAndColumns(tableName, table, columnNames, columns))
					numMatchingConstraints++;
			}
		}

		Assert.assertEquals(1, numMatchingConstraints);
	}

	private boolean matchTableAndColumns(String tableName, Table table,
			String[] columnNames, List<Column> columns) {
		boolean tableNameMatches = table.getName().equals(tableName);
		List<String> expectedColumnNames = Arrays.asList(columnNames);
		List<String> actualColumnNames = new ArrayList<>();
		for (Column column : columns) {
			actualColumnNames.add(column.getName());
		}
		boolean columnsMatch = expectedColumnNames.equals(actualColumnNames);
		boolean match = tableNameMatches && columnsMatch;
		return match;
	}

	private void assertSchemaContainsTable(RelationalSchema schema,
			String tableName, String... columnNames) {

		Table table = schema.getTables().get(tableName);
		Assert.assertNotNull(table);

		Map<String, Column> columns = table.getColumns();
		Assert.assertEquals(columnNames.length, columns.size());

		for (String columnName : columnNames) {
			Assert.assertNotNull("Missing column: " + columnName,
					columns.get(columnName));
		}
	}

	private Properties loadProperties() {
		final URL propertiesURL = this.getClass().getResource(
				"/postgres.properties");
		final File propertiesFile = new File(propertiesURL.getFile());
		final Properties properties = new Properties();
		try (FileReader reader = new FileReader(propertiesFile)) {
			properties.load(reader);
		} catch (Exception e) {
			throw new RuntimeException("Test defect.", e);
		}
		return properties;
	}

	@After
	public void tearDownDatabase() throws SQLException {
		PostgresqlDatabase db = (PostgresqlDatabase) Database
				.newInstance(loadProperties());
		db.drop();
	}

}
